function nm_is_defined() {
    if ('undefined' === typeof NM) {
        console.log('NM is not defined');
        return false;
    }
    return true;
}

/**
 * Get the data from a form.
 *
 * @param {string|Element} formObject The selector for the form element such as '.myform'
 * or the form Element object
 * @param {boolean} returnAsString Whether to return the data as a url query param string. Default false.
 * @returns {Object|string}
 */
function Nmeri_getFormdata(formObject, returnAsString) {
    var form = 'string' === typeof formObject ? document.querySelector(formObject) : formObject;
    var obj = {};
    var formData = new FormData(form);
    formData.forEach(function(value, key) {
        if (-1 !== key.lastIndexOf('[]')) {
            if (obj[key]) {
                obj[key].push(value);
            } else {
                obj[key] = [value];
            }
        } else {
            obj[key] = value;
        }
    });

    var submitBtn = form.querySelector('[type="submit"]');
    if (submitBtn && submitBtn.name) {
        obj[submitBtn.name] = submitBtn.value;
    }

    return returnAsString ? new URLSearchParams(obj).toString() : obj;
}

function Nmeri_block(selector) {
    if (jQuery.blockUI) {
        var toBlock = Array.isArray(selector) ? selector : [selector];

        toBlock.forEach(function(item) {
            if (!(jQuery(selector).is('.nm.is-blocked') || jQuery(selector).parents('.nm.is-blocked').length)) {
                jQuery(item).addClass('is-nm-blocked').block({
                    message: null,
                    overlayCSS: {
                        backgroundColor: '#fff'
                    },
                    css: {
                        backgroundColor: 'transparent',
                        border: 'none'
                    }
                });
            }
        });
    }
}

function Nmeri_unblock(selector) {
    if (jQuery.blockUI) {
        var toBlock = Array.isArray(selector) ? selector : [selector];
        toBlock.forEach(function(item) {
            jQuery(item).removeClass('is-nm-blocked').unblock();
        });
    }
}

/**
 * Appends a dialog element to the document body to be shown later
 * @param {string} template The dialog template
 * @returns {string|null} The id of the dialog template that can be used to trigger
 * the show() method on it later, or null if the insert operation is simply a reload
 * of the template from the server.
 *
 * The id is not returned if the insert operation is a reload to prevent the template
 * from being shown by the showTemplate() function as this is the function in which the
 * insertTemplate() function is called.
 */
function Nmeri_insertTemplate(template) {
    var d = new DOMParser().parseFromString(template, 'text/html');
    var id = d.querySelector('body').firstChild.getAttribute('id');
    var isReload = false;

    // Remove previous identitcal component from dom
    var prevComp = document.getElementById(id);
    if (prevComp) {
        isReload = prevComp.dataset.isReload ? true : false;
        prevComp.remove();
    }

    document.body.appendChild(d.querySelector('body').children[0]);
    return {
        isReload: isReload,
        id: id
    };
}

/**
 * Show a dialog template
 * @param {string} template Template to show in a dialog
 * @param {object} args Arguments used to compose the template. "component" key is
 * required with a value.
 * @returns {undefined}
 */
function Nmeri_showTemplate(template, args) {
    if (!nm_is_defined()) {
        return;
    }

    // insert template into dom
    var inserted = Nmeri_insertTemplate(template);

    // Set component related target
    var dialog = document.getElementById(inserted.id);
    var comp = new NM[Nmeri_getDialogComponentName(dialog)](dialog);
    var relatedTarget;

    if (args && args.btn && args.btn.dataset.nmId) {
        relatedTarget = args.btn.dataset.nmId;
    }

    comp.relatedTarget = relatedTarget;

    // Close any currently open dialog
    var openDialog = document.querySelector('.nm.dialog.show');
    if (openDialog) {
        Nmeri_closeDialog(openDialog);
    }

    // show template
    if (!inserted.isReload) {
        comp.show(relatedTarget || null);
    }
}

function Nmeri_showToast(notice) {
    if (!nm_is_defined()) {
        return;
    }

    var toastDiv = jQuery('.nm-component.toast-container');
    if (toastDiv.length) {
        var notices = notice instanceof Array ? notice : [notice];

        notices.forEach(function(el) {
            var html = jQuery(el)[0];
            toastDiv.append(html); // append to start at bottom so that it would stack according to last
            (new NM.Toast(html)).show();
        });
    }
}

/**
 * Get the component used as a dialog
 * @param {Element} dialog The dialog element
 * @returns {String} The component name, Modal or Offcanvas
 */
function Nmeri_getDialogComponentName(dialog) {
    return dialog.classList.contains('modal') ? 'Modal' : 'Offcanvas';
}

/**
 * Close the disalog
 * @param {Element} dialog The dialog element
 * @param {boolean} force Whether to simply close the dialog only and perform no other action related
 * to closing. Default true.
 *
 * In closing the currently opened dialog, we can show another dialog
 * based on the data attributes supplied to the currently opened dialog.
 *
 * The data attributes must have values equal to the ids of elements in the dom.
 * These elements in most cases should be dialog elements.
 *
 * If the dialog has any of the following data attributes, the corresponding actions would
 * be performed:
 *
 * nm-redirect:
 * This opens a new dialog. First it checks if the dialog has already been previously opened and
 * the instance exists. If so, it searches the instance for the button clicked to open the dialog
 * (the relatedTarget property). If this is founds it triggers a click on the button again to reload the
 * dialog. If it doesn't find the button, it takes the value supplied to the data attribute as the id of a
 * clickable button to open the new dialog and triggers a click on it.
 *
 * nm-back:
 * This simply shows a dialog that has already been previously shown and it's instance exists in the dom,
 * by simply invoking the show() method on the instance again. In this case, the dialog is not refreshed.
 */
function Nmeri_closeDialog(dialog, force) {
    if (!nm_is_defined()) {
        return;
    }

    if (true === dialog) {
        Nmeri_closeOpenDialog();
        return;
    }

    var component = Nmeri_getDialogComponentName(dialog);

    if (!force) {
        NM[component].getInstance(document.getElementById(dialog.getAttribute('id'))).hide();
        return;
    }

    if (dialog.dataset.nmRedirect) {
        var target = document.querySelector(dialog.dataset.nmRedirect);
        var newDialog = NM[component].getInstance(target);
        if (newDialog && newDialog.relatedTarget) {
            document.querySelector('[data-nm-id="' + newDialog.relatedTarget + '"]').click();
        } else if (target) {
            target.click();
        }
    } else if (dialog.dataset.nmBack) {
        NM[component].getInstance(document.querySelector(dialog.dataset.nmBack)).show();
    } else if (dialog.dataset.nmClose) {
        NM[component].getInstance(document.getElementById(dialog.getAttribute('id'))).hide();
    }
}

/**
 * Close the currently opened dialog
 */
function Nmeri_closeOpenDialog() {
    var dialog = document.querySelector('.nm-component.dialog.show');
    if (dialog) {
        Nmeri_closeDialog(dialog);
    }
}

// Replace templates in the dom when they have been retrieved from the server
function Nmeri_replaceTemplates(templates) {
    for (var key in templates) {
        var collection = document.querySelectorAll(key);
        if (collection) {
            var template = templates[key];
            collection.forEach(function(coll) {
                var wrap = document.createElement('div');
                wrap.innerHTML = template;
                var newTemplate = wrap.children[0];
                coll.replaceWith(newTemplate);
                jQuery(document.body).trigger('nm_replaced_template', {
                    key: key,
                    template: newTemplate
                });
            });
        }
    }
}

/**
 * Retrieve dialog templates from the server to be loaded in the dom.
 *
 * This function is used to update dialogs that are already in the dom which cannot
 * be updated using data-nm-redirect or replaceTemplates() function as they are not
 * directly connected to the current dialog being shown.
 *
 * This function is called after a dialog operation has been successful (for example when
 * the form in the dialog is saved to the database). We loop through all the dialogs in the
 * dom to see if any of them should be reloaded. This is determined by a 'data-nm-reload'
 * attribute on the dialog containing the id of the currently shown dialog as it value.
 *
 * If we find such dialog, we simply reload it by triggering a click on its 'relatedTarget'
 * property which has to be set.
 *
 * @param {Element} dialog
 */
function Nmeri_reloadTemplates(dialog) {
    if (!nm_is_defined()) {
        return;
    }

    var collection = document.querySelectorAll('.dialog[data-nm-reload]');
    if (collection.length) {
        var component = Nmeri_getDialogComponentName(dialog);
        var dialogId = dialog.getAttribute('id');

        collection.forEach(function(coll) {
            var reloads = JSON.parse(coll.dataset.nmReload);
            if (reloads.indexOf('#' + dialogId) > -1) {
                var newDialog = NM[component].getInstance(coll);
                if (newDialog && newDialog.relatedTarget) {
                    coll.dataset.isReload = true;
                    document.querySelector('[data-nm-id="' + newDialog.relatedTarget + '"]').click();
                }
            }
        });
    }
}

/**
 * Initialize a specific bootstrap component
 *
 * How to use: for example, to initialize the tooltip component on the document:
 * - initComponent('tooltip', document);
 *
 * @param {string} component The name of the component e.g. tooltip, popover
 * @param {object} element The element to initialize the component on. Defaults to document
 */
function Nmeri_initComponent(component, element) {
    if ('undefined' !== typeof component && 'undefined' !== typeof NM) {
        var componentType = component.charAt(0).toUpperCase() + component.slice(1);
        var el = element instanceof Element ? element : document;
        var List = [].slice.call(el.querySelectorAll('.nm[data-bs-toggle="' + component + '"]'));
        var options = {};

        if ('tooltip' === component || 'popover' === component) {
            options['html'] = true;
            options['customClass'] = 'nm-component';
        }

        List.map(function(triggerEl) {
            return new NM[componentType](triggerEl, options);
        });
    }
}

/**
 * Go to a url
 * @param {string} url
 */
function Nmeri_redirect(url) {
    if (-1 === url.indexOf('https://') || -1 === url.indexOf('http://')) {
        window.location = url;
    } else {
        window.location = decodeURI(url);
    }
}

/**
 * Perform standard operations with the response from an ajax request
 */
function Nmeri_processResponse(response, args) {
    if (response.log) {
        console.log(response);
    }

    if (args && args.block) {
        Nmeri_unblock(args.block);
    }

    if ((args && args.close_dialog) || response.close_dialog) {
        Nmeri_closeDialog(args.close_dialog || response.close_dialog);
    }

    if (response.success && args && args.dialog) {
        Nmeri_reloadTemplates(args.dialog);
    }

    if (response.show_template) {
        Nmeri_showTemplate(response.show_template, args);
    }

    if (response.toast_notice) {
        Nmeri_showToast(response.toast_notice);
    }

    if (response.replace_templates) {
        Nmeri_replaceTemplates(response.replace_templates);
    }

    if (response.redirect) {
        Nmeri_redirect(response.redirect);
    }
}

/**
 * Common js code
 */

var nmgr = window.nmgr || {};

jQuery(function($) {

    /**
     * Define a common blockUI block setup for plugin
     *
     * @param {string | jQuery object} selector The element to be blocked
     */
    nmgr.block = function(selector) {
        if (!nmgr.is_blocked(selector)) {
            $(selector).addClass('nm is-blocked').block({
                message: null,
                overlayCSS: {
                    backgroundColor: '#fff'
                },
                css: {
                    backgroundColor: 'transparent',
                    border: 'none'
                }
            });
        }
    };

    /**
     * Define a common blockUI unblock setup for plugin
     *
     * @param {string | jQuery object} selector The element to be unblocked
     */
    nmgr.unblock = function(selector) {
        $(selector).removeClass('nm is-blocked').unblock();
    };

    /**
     * Check whether an element is currently blocked with blockUI
     *
     * @param {string | jQuery object} selector The element to be checked
     * @returns {jQuery.length|Boolean} True if the element is blocked, false if not
     */
    nmgr.is_blocked = function(selector) {
        return $(selector).is('.nm.is-blocked') || $(selector).parents('.nm.is-blocked').length;
    };

    /**
     * Scroll to an element
     *
     * @param {string | jQuery object} selector The element to scroll to
     */
    nmgr.scroll_to = function(selector) {
        if ($(selector).length) {
            if (nmgr.is_in_dialog(selector)) {
                $('#nmgr-modal').animate({
                    scrollTop: 0
                });
            } else {
                $('html, body').animate({
                    scrollTop: ($(selector).offset().top - 100)
                }, 1000);
            }
        }
    };

    /**
     * Go to a url
     * @param {string} url
     * @deprecated since 4.0.0
     */
    nmgr.go_to = function(url) {
        if ('function' === typeof Nmeri_redirect) {
            console.warn('nmgr.go_to() is deprecated since version 4.0.0. Use Nmeri_redirect()');
            Nmeri_redirect(url);
        } else {
            if (-1 === url.indexOf('https://') || -1 === url.indexOf('http://')) {
                window.location = url;
            } else {
                window.location = decodeURI(url);
            }
        }
    };

    /**
     * Remove woocommerce notices from the page
     * @returns {undefined}
     */
    nmgr.remove_notice = function(selector) {
        var notice_selectors = '.woocommerce-error, .woocommerce-message, .woocommerce-info';

        if ('undefined' !== typeof(selector)) {
            $(selector).find(notice_selectors).remove();
        } else {
            $(notice_selectors).remove();
        }
    };

    /**
     * Make a table responsive when it's width is above a certain size.
     *
     * The responsiveness in this case is for desktop display rather than mobile.
     * By default, the table should already be mobile responsive.
     *
     * @param {string} selector The table element to make responsive
     * @param {int} width Minimum width at which table would be responsive. Default 700px
     */
    nmgr.make_table_responsive = function(selector, width) {
        width = width || 600;
        selector = selector || '.nmgr-table';

        var table = $(selector);
        if (table.width() > width) {
            table.addClass('responsive');
        } else {
            table.removeClass('responsive');
        }
    };

    nmgr.tiptip = function() {
        $('.nmgr-tip').each(function() {
            var $element = $(this),
                options = {},
                title = $element.find('title');

            if ($element.attr('data-tiptip-position')) {
                options['defaultPosition'] = $element.attr('data-tiptip-position');
            }

            if ($element.is('svg')) {
                options['content'] = function() {
                    if (title.length) {
                        if (title.attr('data-title')) {
                            title.text('');
                            return title.attr('data-title');
                        }
                    }
                };
            }

            $element.tipTip(options);
        });
    };

    nmgr.tiptip();

    $(document.body).on('show.bs.modal', '.nmgr.dialog', function() {
        nmgr.tiptip();
        Nmeri_initComponent('tooltip', this);
    });


    // Check if we are in a modal window
    nmgr.is_in_modal = function(element) {
        console.log('nmgr.is_in_modal is deprecated. Use nmgr.is_in_dialog');
        return nmgr.is_in_dialog(element);
    };

    nmgr.is_in_dialog = function(element) {
        var dialog = '.nm-component.dialog.nmgr.show';
        if (element) {
            return document.querySelector(dialog) && $(element).closest(dialog).length > 0;
        } else {
            return document.querySelector(dialog);
        }
    };

    nmgr.init_datepicker = function(selector) {
        selector = selector || '.nmgr-use-datepicker';
        var datepickerClass = 'nmgr-datepicker';
        var datepickerOptions = nmgr_global_params.datepicker_options;

        if (datepickerOptions.styleDatepicker) {
            datepickerOptions = Object.assign(datepickerOptions, {
                beforeShow: function(input, inst) {
                    if (inst.input.hasClass('nmgr-use-datepicker')) {
                        inst.dpDiv.addClass(datepickerClass);
                    }
                },
                onClose: function(dateText, inst) {
                    if (inst.input.hasClass('nmgr-use-datepicker')) {
                        inst.dpDiv.removeClass(datepickerClass);
                    }
                }
            });
        }

        $(selector).each(function() {
            var data = this.dataset;
            for (var key in data) {
                if (key.includes('datepicker')) {
                    var dpkey = key.replace('datepicker', '');
                    var dpkey2 = dpkey.charAt(0).toLowerCase() + dpkey.slice(1);
                    datepickerOptions[dpkey2] = data[key];
                }
            }

            for (var key in datepickerOptions) {
                if ('string' === typeof(datepickerOptions[key]) && -1 < datepickerOptions[key].indexOf('function')) {
                    datepickerOptions[key] = Function('"use strict";return(' + datepickerOptions[key] + ')')();
                }
            }

            $(this).datepicker(datepickerOptions)
                .on("keyup", function() {
                    if (datepickerOptions.altField) {
                        var date;
                        var dateFormat = $(this).datepicker('option', 'dateFormat');

                        try {
                            date = $.datepicker.parseDate(dateFormat, this.value);
                        } catch (e) {}

                        if (!date) {
                            $(datepickerOptions.altField).val('');
                        }
                    }
                });

        });
    };

    // Make selected tables responsive
    if (nmgr_global_params.responsive_tables) {
        Object.values(nmgr_global_params.responsive_tables).forEach(function(table) {
            var responsiveWidth = table.width ? table.width : '';

            nmgr.make_table_responsive(table.selector, responsiveWidth);

            $(window).on('resize', function() {
                nmgr.make_table_responsive(table.selector, responsiveWidth);
            });

            if (table.events && 0 < table.events.length) {
                table.events.forEach(function(event) {
                    if ('object' === typeof event) {
                        var selector = event.selector;
                        var ev = event.event;
                    } else {
                        var selector = document.body;
                        var ev = event;
                    }
                    $(selector).on(ev, function() {
                        if ($(table.selector).is(':visible')) {
                            nmgr.make_table_responsive(table.selector, responsiveWidth);
                        }
                    });
                });
            }
        });
    }

    // Make selected tables datatable
    if (nmgr_global_params.datatables_tables) {
        Object.values(nmgr_global_params.datatables_tables).forEach(function(table) {
            /**
             *  After initialization always add width 100% to table to make it responsive
             *  This is particularly important if we are hiding some columns on the table as
             *  it would prevent the table from being responsive.
             *
             *  (This code must be before the actual initialization code for it to work)
             */
            $(document.body).on('init.dt', table.selector, function() {
                this.style.width = '100%';
            });

            // Initialize datatable on page load
            $(table.selector).DataTable(table.options);

            // Initialize datatable on specific events
            if (table.events && 0 < table.events.length) {
                table.events.forEach(function(event) {
                    if ('object' === typeof event) {
                        var selector = event.selector;
                        var ev = event.event;
                    } else {
                        var selector = document.body;
                        var ev = event;
                    }
                    $(selector).on(ev, function() {
                        if ($(table.selector).is(':visible')) {
                            $(table.selector).DataTable(table.options);
                        }
                    });
                });
            }

        });
    }

    nmgr.global_ajax_params = function() {
        var postdata = {
            action: 'nmgr_post_action',
            _wpnonce: nmgr_global_params.nonce,
            ajax_url: nmgr_global_params.ajax_url,
            nmgr_global: JSON.stringify(nmgr_global_params.global)
        };

        return postdata;
    }

    nmgr.post_action = function(args) {
        if (args.btn && args.btn.dataset && args.btn.dataset.notice && !window.confirm(args.btn.dataset.notice)) {
            return false;
        }

        if (args.block) {
            Nmeri_block(args.block);
        }

        if (!args.postdata) {
            return;
        }

        var postdata = nmgr.global_ajax_params();
        Object.assign(postdata, args.postdata);

        $.post(postdata.ajax_url, postdata, function(response) {
            Nmeri_processResponse(response, args);

            $(document.body).trigger('nmgr_post_action_response', {
                postdata: postdata,
                response: response
            });
        });
    }





});


jQuery(function($) {

    if (typeof nmgr_admin_params === 'undefined') {
        return false;
    }

    /**
     * Js events related to wishlist items template
     */


    nmgr.items = {
        container: '#nmgr-items',

        init: function() {
            $(document.body)
                .on('nmgr_post_action_response', this.post_action_response)
                .on('click', '#nmgr-items .nmgr-post-action', this.pre_post_action)
                .on('input', '.nmgr-item-part.nmgr_quantity input.quantity', this.save_item_quantity)
                .on('change', '.nmgr-product-search', this.new_add_items_row)
                .on('click', '.nmgr-add-items-dialog .nmgr-add', this.add_items)
                .on('removed_from_cart', nmgr.remove_notice)
                .on('click', '.nmgr_add_to_cart_button', this.add_single_product_to_cart)
                .on('click', '.nmgr-bulk-add-to-cart', this.add_multiple_products_to_cart)
                .on('change', '#nmgr-items .nmgr-select-all', this.toggle_select_items)
                .on('change', '#nmgr-items .nmgr-select', this.maybe_show_bulk_actions)
                .on('click', '#nmgr-items .nmgr-nav', this.paginate)
                .on('nmgr_add_to_cart_response', this.remove_loading_animation)
                .on('click', '.nmgr-items-display-modes .mode', this.save_display_mode_to_session)
                .on('input', '#nmgr-purchase-refund-form .nmgr-quantity', this.toggle_purchase_refund_options)
                .on('submit', '#nmgr-purchase-refund-form', this.submit_purchase_refund_form);

            var addItemsDialog = document.getElementById('nmgr-add-items-dialog');
            if (addItemsDialog) {
                addItemsDialog.addEventListener('show.bs.modal', this.init_add_items_form, false);
            }
        },

        reloaded: function(wishlist) {
            nmgr.tiptip();
            $(document.body).trigger('nmgr_items_reloaded', wishlist);
        },

        pre_post_action: function(e) {
            e.preventDefault();

            var postdata = Object.assign({}, this.dataset);
            postdata['dataset'] = $('#nmgr-items')[0].dataset;
            postdata['nmgr_global'] = $(nmgr.items.container).attr('data-nmgr_global');

            if ('bulk' === this.dataset.nmgr_post_mode) {
                postdata['wishlist_item_ids'] = nmgr.items.get_selected_item_ids();
                delete(postdata.wishlist_item_id);
            }

            var args = {
                btn: this,
                block: nmgr.items.container,
                postdata: postdata
            };
            nmgr.post_action(args);
        },

        post_action_response: function(e, args) {
            if (args.response.replace_templates && args.response.replace_templates.hasOwnProperty('#nmgr-items')) {
                nmgr.items.reloaded(args.response.wishlist || null);
            }

            if ('delete_item' === args.postdata.nmgr_post_action) {
                $(document.body).trigger('wc_fragment_refresh');
                $(document.body).trigger('nmgr_removed_from_wishlist', args.response);
            }
        },

        save_item_quantity: function() {
            var args = {
                postdata: {
                    nmgr_post_action: 'save_item_quantity',
                    wishlist_item_id: this.dataset.item_id,
                    quantity: this.value
                }
            };
            nmgr.post_action(args);

            var container = this.closest('.nmgr_quantity');
            Nmeri_block(container);
            setTimeout(function() {
                Nmeri_unblock(container);
            }, 500);
        },

        init_add_items_form: function() {
            // reset the modal content form
            nmgr.items.reset_add_items_form('.nmgr-add-items-dialog');

            // initialize select2 on the select field in the modal
            $(document.body).trigger('nmgr-select2-init');
        },

        reset_add_items_form: function(element) {
            if ('undefined' !== typeof element) {
                $(element).find('tr.added-row').remove();
                $(element).find('form').trigger('reset').find(':input').each(function() {
                    if ($(this).is('.nmgr-product-search')) {
                        $(this).val('');
                    }
                });
            }
        },

        new_add_items_row: function() {
            if (!$(this).closest('tr').is(':last-child')) {
                return;
            }

            var item_table = $(this).closest('table'),
                item_table_body = item_table.find('tbody'),
                index = item_table_body.find('tr').length,
                row = item_table_body.data('row').replace(/\[0\]/g, '[' + index + ']');

            item_table_body.append('<tr class="added-row">' + row + '</tr>');

            $(document.body).trigger('nmgr-select2-init');
        },

        add_items: function() {
            var add_items_data = [];

            $(this).closest('.nmgr-add-items-dialog').find('tbody tr').each(function() {
                var item = {};
                $($(this).find(':input').serializeArray()).each(function(index, obj) {
                    item[obj.name] = obj.value;
                });

                add_items_data.push(item);
            });

            var dialogEl = document.getElementById('nmgr-add-items-dialog');
            NM.Modal.getInstance(dialogEl).hide();

            var data = {
                nmgr_post_action: 'add_item',
                wishlist_id: $(nmgr.items.container).attr('data-wishlist-id'),
                nmgr_global: $(nmgr.items.container).attr('data-nmgr_global'),
                nmgr_add_items_data: add_items_data
            };

            // Check if items have changed, if so pass them through so we can save them before adding a new item.
            if ('true' === $('#nmgr-items button.save-action').attr('data-reload')) {
                data.items = $('.nmgr-items-view :input[name]').serialize();
            }

            var args = {
                block: nmgr.items.container,
                postdata: data
            };

            nmgr.post_action(args);
        },

        add_single_product_to_cart: function(e) {
            var $thisbutton = $(this),
                item = {},
                $form = $thisbutton.closest('.nmgr-add-to-cart-form');
            /**
             *  If we are not adding to cart via ajax, we are supposed to return here
             *  but we will only return if adding the wishlist item to the cart is not restricted.
             *  Otherwise we will continue into the ajax code below in order to retrieve the
             *  restricted notice set during the add-to-cart validation
             */
            if (!$thisbutton.is('.nmgr_ajax_add_to_cart')) {
                return true;
            }

            e.preventDefault();

            $($form.serializeArray()).each(function(index, obj) {
                item[obj.name] = obj.value;
            });

            $thisbutton.removeClass('added').addClass('loading');

            nmgr.items.add_to_cart([item]);
        },

        add_multiple_products_to_cart: function(e) {
            e.preventDefault();
            var items = [];
            var clickedButton = this;

            clickedButton.classList.add('loading');

            $(nmgr.items.container).find('.nmgr-select:checked').closest('.item').each(function() {
                var item = {},
                    form = $(this).find('.nmgr-add-to-cart-form');

                if (form.length) {
                    $(form.serializeArray()).each(function(index, obj) {
                        item[obj.name] = obj.value;
                    });
                    items.push(item);
                    //show loading animation
                    if (clickedButton.classList.contains('nmgr_ajax_add_to_cart')) {
                        $(this).find('.nmgr_add_to_cart_button').removeClass('added').addClass('loading');
                    }
                }
            });

            if (items.length) {
                if (clickedButton.classList.contains('nmgr_ajax_add_to_cart')) {
                    nmgr.items.add_to_cart(items);
                } else {
                    var stringed = JSON.stringify(items);
                    var form = document.createElement('form');
                    var input = document.createElement('input');

                    form.method = 'post';
                    input.type = 'hidden';
                    input.name = 'nmgr_add_to_cart_item_data_string';
                    input.value = stringed;
                    form.appendChild(input);
                    document.body.appendChild(form);
                    form.submit();
                }
            } else {
                clickedButton.classList.remove('loading');
            }
        },
        /**
         * Add single or multiple products to cart
         *
         * @param {array} items (product data to add to cart e.g. product_id, quantity, variation_id)
         */
        add_to_cart: function(items) {
            var data = {
                action: 'nmgr_add_to_cart',
                nmgr_add_to_cart_item_data: items,
                nmgr_global: $(nmgr.items.container).attr('data-nmgr_global')
            };

            $.post(nmgr_global_params.ajax_url, data, function(response) {
                if (!response.fragments) {
                    return;
                }

                if (response.fragments.success && response.fragments.redirect_url) {
                    window.location = response.fragments.redirect_url;
                    return;
                }

                $('.woocommerce-error, .woocommerce-message, .woocommerce-info').remove();

                $(document.body).trigger('nmgr_add_to_cart_response', response.fragments);

                if (response.fragments.notices) {
                    response.fragments.notices.forEach(function(item) {
                        Nmeri_showToast(item);
                    });
                }

                if (response.fragments.success) {
                    $(document.body).trigger('added_to_cart', [response.fragments, response.cart_hash]); // update cart fragments
                    $(document.body).trigger('nmgr_added_to_cart', [response.fragments.items_data ? response.fragments.items_data : '']);
                }
            });
        },

        /**
         * Remove the loading animation on the add to cart button on the items table after the ajax action
         * has been completed.
         *
         * It is necessary to do this here rather than in the add_to_cart function because the add_to_cart function
         * is used in various ways (to add single, multiple products, and products from the wishlist cart). For this reason
         * it is less effiicient to code for removing the icon animation in the add_to_cart function.
         */
        remove_loading_animation: function(e, fragments) {
            if ('undefined' !== typeof fragments && fragments.items_data) {
                fragments.items_data.forEach(function(item) {
                    $('[data-wishlist_item_id="' + item.wishlist_item_id + '"].nmgr_add_to_cart_button')
                        .removeClass('loading');
                });
            }

            var btn = document.querySelector('.nmgr-bulk-add-to-cart');
            if (btn) {
                btn.classList.remove('loading');
            }
        },

        toggle_select_items: function() {
            var items = document.querySelectorAll('.nmgr-select');
            if (0 < items.length) {
                for (var i = 0; i < items.length; i++) {
                    if (true === this.checked) {
                        items[i].checked = true;
                    } else {
                        items[i].checked = false;
                    }
                }

                nmgr.items.maybe_show_bulk_actions();
            }
        },
        maybe_show_bulk_actions: function() {
            var actions = document.querySelector('.nmgr-items-bulk-actions');
            var selectedItems = document.querySelectorAll('#nmgr-items .nmgr-select:checked');
            if (selectedItems.length && !actions.classList.contains('show')) {
                actions.classList.add('show');
            } else if (!selectedItems.length && actions.classList.contains('show')) {
                var selectAllBtn = document.querySelector('#nmgr-items .nmgr-select-all');
                if (selectAllBtn) {
                    selectAllBtn.checked = false;
                }
                actions.classList.remove('show');
            }
        },
        /**
         * Get the checkboxes of the items checked on the items table
         * @returns {NodeList}
         */
        get_selected_items: function() {
            return document.querySelectorAll('#nmgr-items .nmgr-select:checked');
        },
        /**
         * Get the ids of the items selected on the items table
         * @returns {Array}
         */
        get_selected_item_ids: function() {
            var item_ids = [],
                selected = this.get_selected_items();
            if (selected.length) {
                selected.forEach(function(item) {
                    item_ids.push(item.value);
                });
            }
            return item_ids;
        },

        save_display_mode_to_session: function(e) {
            e.preventDefault();

            var args = {
                block: nmgr.items.container,
                postdata: {
                    nmgr_post_action: 'save_display_mode_to_session',
                    dataset: $('#nmgr-items')[0].dataset,
                    mode: this.dataset.mode,
                    nmgr_global: $(nmgr.items.container).attr('data-nmgr_global')
                }
            };
            nmgr.post_action(args);
        },

        toggle_purchase_refund_options: function() {
            var $form = $(this).closest('form');
            var val = parseInt(this.value);
            var originalVal = parseInt(this.dataset.qty);
            if (val < originalVal) {
                $form.find('.nmgr-options').fadeOut();
            } else if (val > originalVal) {
                $form.find('.nmgr-options').fadeIn();
            }
        },

        submit_purchase_refund_form: function(e) {
            e.preventDefault();

            var args = {
                block: this,
                postdata: Object.assign(Nmeri_getFormdata(this), {
                    dataset: $('#nmgr-items')[0].dataset,
                    nmgr_post_action: 'save_purchase_refund_form'
                })
            };

            nmgr.post_action(args);
        },

        paginate: function(e) {
            e.preventDefault();

            var args = {
                block: nmgr.items.container,
                postdata: {
                    nmgr_post_action: 'paginate_items',
                    dataset: $('#nmgr-items')[0].dataset,
                    nav_action: this.dataset.nav_action,
                    nmgr_global: $(nmgr.items.container).attr('data-nmgr_global')
                }
            };
            nmgr.post_action(args);
        }

    };
    nmgr.items.init();

    /*
     * js events related to wishlist messages template
     */

    var nmgr_messages = {

        init: function() {
            $(document.body)
                .on('click', '.nmgr-message .delete-wishlist-message', this.delete);
        },

        delete: function(e) {
            e.preventDefault();

            var postdata = Object.assign({}, this.dataset);
            postdata['nmgr_global'] = $(this).closest('.nmgr-account-template').attr('data-nmgr_global');

            var args = {
                btn: this,
                block: $(this).closest('.nmgr-message'),
                postdata: postdata
            };

            nmgr.post_action(args);
        }
    };

    nmgr_messages.init();


    /**
     * Initializes all select2 elements used by plugin and retrieves their data
     */

    var nmgr_select2 = {
        init: function() {
            try {
                $(document.body)
                    .on('nmgr-select2-init', this.activate_select2)
                    .trigger('nmgr-select2-init');

                $('.nmgr-user-search').on('select2:select', this.update_account_shipping_address);

            } catch (e) {
                window.console.log(e);
            }
        },

        activate_select2: function() {
            nmgr_select2.select2_user_search();
            nmgr_select2.select2_product_search();
        },

        select2_product_search: function() {
            $(':input.nmgr-product-search').each(function() {
                var self = this;
                var select2_args = {
                    allowClear: this.dataset.allowClear ? true : false,
                    minimumInputLength: this.dataset.minimumInputLength ? this.dataset.minimumInputLength : '3',
                    escapeMarkup: function(m) {
                        return m;
                    },
                    ajax: {
                        url: nmgr_global_params.ajax_url,
                        dataType: 'json',
                        delay: 250,
                        data: function(params) {
                            return {
                                term: params.term,
                                action: 'nmgr_json_search_products',
                                security: nmgr_admin_params.search_products_nonce,
                                exclude: $(this).data('exclude'),
                                exclude_type: $(this).data('exclude_type'),
                                include: $(this).data('include'),
                                limit: $(this).data('limit'),
                                nmgr_global: $(nmgr.items.container).attr('data-nmgr_global'),
                                display_stock: $(this).data('display_stock'),
                            };
                        },
                        processResults: function(data) {
                            var terms = [];
                            if (data) {
                                $.each(data, function(id, text) {
                                    terms.push({
                                        id: id,
                                        text: text
                                    });
                                });
                            }
                            return {
                                results: terms
                            };
                        },
                        cache: true
                    }
                };

                if (self.dataset.textInputTooShort) {
                    select2_args.language = {};
                    select2_args.language['inputTooShort'] = function() {
                        return self.dataset.textInputTooShort;
                    };
                }

                $(this).select2(select2_args);

            });
        },

        select2_user_search: function() {
            $(':input.nmgr-user-search').each(function() {
                var select2_args = {
                    allowClear: $(this).data('allow-clear') ? true : false,
                    placeholder: $(this).data('placeholder'),
                    minimumInputLength: $(this).data('minimum_input_length') ? $(this).data('minimum_input_length') : '1',
                    escapeMarkup: function(m) {
                        return m;
                    },
                    ajax: {
                        url: nmgr_global_params.ajax_url,
                        dataType: 'json',
                        delay: 250,
                        data: function(params) {
                            return {
                                term: params.term,
                                action: 'nmgr_json_search_users',
                                security: nmgr_admin_params.search_users_nonce,
                                exclude: $(this).data('exclude')
                            };
                        },
                        processResults: function(data) {
                            var terms = [];
                            if (data) {
                                $.each(data, function(id, val) {
                                    terms.push({
                                        id: id,
                                        text: val.text,
                                        shipping_address: val.shipping_address
                                    });
                                });
                            }
                            return {
                                results: terms
                            };
                        },
                        cache: true
                    }
                };

                $(this).select2(select2_args);

            });
        },

        update_account_shipping_address: function(e) {
            var data = e.params.data;

            if (data.shipping_address) {
                $('.nmgr-copy-shipping-address').attr('data-nmgr-address', data.shipping_address);
            }
        }
    };

    nmgr_select2.init();
    var ship_to_account_address = {

        init: function() {
            $(document.body)
                .on('click', '.nmgr-copy-shipping-address', this.load_address);
        },

        load_address: function(e) {
            e.preventDefault();

            if ($('.nmgr-shipping-fields').length > 0) {
                var address = JSON.parse(e.target.dataset.nmgrAddress);

                for (var key in address) {
                    $('.nmgr-shipping-fields :input#shipping_' + key).val(address[key]).trigger('change');
                }
            }
        }

    };

    ship_to_account_address.init();


    nmgr.init_datepicker();

    /**
     * Adds image ui field to metabox
     */
    function render_image_field() {
        // Set all variables to be used in scope
        var frame,
            metaBox = $('#nm_gift_registry-bg-img.postbox'),
            addImgLink = metaBox.find('.upload-custom-img'),
            delImgLink = metaBox.find('.delete-custom-img'),
            descLink = metaBox.find('.nmgr-set-bg-img-desc'),
            imgContainer = metaBox.find('.custom-img-container'),
            imgIdInput = metaBox.find('.custom-img-id');

        // ADD IMAGE LINK
        metaBox.on('click', '.upload-custom-img, .custom-img-container', function(event) {

            event.preventDefault();

            // If the media frame already exists, reopen it.
            if (frame) {
                frame.open();
                return;
            }

            // Create a new media frame
            frame = wp.media({
                title: 'Select or Upload Image',
                button: {
                    text: 'Use this image'
                },
                multiple: false // Set to true to allow multiple files to be selected
            });

            // When an image is selected in the media frame...
            frame.on('select', function() {

                // Get media attachment details from the frame state
                var attachment = frame.state().get('selection').first().toJSON();

                if ('image' === attachment.type) {
                    // Send the attachment URL to our custom image input field.
                    imgContainer.removeClass('hidden').html('<img src="' + attachment.url + '" alt="" />');

                    // Send the attachment id to our hidden input
                    imgIdInput.val(attachment.id);

                    // Hide the add image link
                    addImgLink.addClass('hidden');

                    // Unhide the remove image link
                    delImgLink.removeClass('hidden');

                    // Unhide the description link
                    descLink.removeClass('hidden');
                }
            });

            // Finally, open the modal on click
            frame.open();
        });


        // DELETE IMAGE LINK
        delImgLink.on('click', function(event) {

            event.preventDefault();

            // Clear out the preview image
            imgContainer.html('').addClass('hidden');

            // Un-hide the add image link
            addImgLink.removeClass('hidden');

            // Hide the delete image link
            delImgLink.addClass('hidden');

            // hide the description link
            descLink.addClass('hidden');

            // Delete the image id from the hidden input
            imgIdInput.val('');

        });
    };
    render_image_field();

    /**
     * Use selectWoo to select the shipping country and state the woocommerce way in admin
     * This code is simply a modified version of the same code woocommerce uses to select shipping
     * country and state in the order screen. Modified from meta-boxes-order.js in woocommerce plugin.
     */
    var shipping_country_state = {
        states: null,

        init: function() {
            if (!(typeof nmgr_admin_params === 'undefined' || typeof nmgr_admin_params.countries === 'undefined')) {
                this.states = $.parseJSON(nmgr_admin_params.countries.replace(/&quot;/g, '"'));
            }

            $('.js_field-country').selectWoo().change(this.change_country);
            $('.js_field-country').trigger('change', [true]);
            $(document.body)
                .on('change', 'select.js_field-state', this.change_state)
                .on('country-change.woocommerce', this.hide_fields);
        },

        field_is_required: function(field, is_required) {
            if (is_required) {
                /**
                 * To prevent html5 woocommerce bug of
                 * 'An invalid form control with name='shipping_state' is not focusable.',
                 * we need to make sure that all fields that have the required attributes are also visible.
                 */
                field.find(':input').attr('required', '1').show();

                if (field.find('label .required').length === 0) {
                    field.find('label').append(
                        '&nbsp;<abbr class="required" title="' +
                        nmgr_admin_params.i18n_required_text +
                        '">*</abbr>'
                    );
                }
            } else {
                field.find('label .required').remove();
                field.find(':input').removeAttr('required');
            }
        },

        hide_fields: function(e, country, wrapper) {
            var thisform = wrapper,
                locale_json = nmgr_admin_params.locale.replace(/&quot;/g, '"'),
                locale = JSON.parse(locale_json),
                locale_fields = JSON.parse(nmgr_admin_params.locale_fields),
                thislocale;

            if (typeof locale[country] !== 'undefined') {
                thislocale = locale[country];
            } else {
                thislocale = locale['default'];
            }

            $.each(locale_fields, function(key, value) {
                var field = thisform.find(value),
                    fieldLocale = $.extend(true, {}, locale['default'][key], thislocale[key]);

                // Required.
                if (typeof fieldLocale.required !== 'undefined') {
                    shipping_country_state.field_is_required(field, fieldLocale.required);
                } else {
                    shipping_country_state.field_is_required(field, false);
                }

                // Hidden fields.
                if ('state' !== key) {
                    if (typeof fieldLocale.hidden !== 'undefined' && true === fieldLocale.hidden) {
                        field.hide().find(':input').val('');
                    } else {
                        field.show();
                    }
                }
            });

        },

        change_country: function(e, stickValue) {
            // Check for stickValue before using it
            if (typeof stickValue === 'undefined') {
                stickValue = false;
            }

            // Prevent if we don't have the metabox data
            if (shipping_country_state.states === null) {
                return;
            }

            var $this = $(this),
                country = $this.val(),
                $state = $this.parents('div.nmgr-shipping-fields').find(':input.js_field-state'),
                $parent = $state.parent(),
                input_name = $state.attr('name'),
                input_id = $state.attr('id'),
                value = $this.data('woocommerce.stickState-' + country) ? $this.data('woocommerce.stickState-' + country) : $state.val(),
                placeholder = $state.attr('placeholder'),
                $newstate;

            if (stickValue) {
                $this.data('woocommerce.stickState-' + country, value);
            }

            // Remove the previous DOM element
            $parent.show().find('.select2-container').remove();

            if (!$.isEmptyObject(shipping_country_state.states[country])) {
                var state = shipping_country_state.states[country],
                    $defaultOption = $('<option value=""></option>')
                    .text(nmgr_admin_params.i18n_select_state_text);

                $newstate = $('<select></select>')
                    .prop('id', input_id)
                    .prop('name', input_name)
                    .prop('placeholder', placeholder)
                    .addClass('js_field-state select short')
                    .append($defaultOption);

                $.each(state, function(index) {
                    var $option = $('<option></option>')
                        .prop('value', index)
                        .text(state[index]);
                    $newstate.append($option);
                });

                $newstate.val(value);

                $state.replaceWith($newstate);

                $newstate.show().selectWoo().hide().change();
            } else {
                $newstate = $('<input type="text" />')
                    .prop('id', input_id)
                    .prop('name', input_name)
                    .prop('placeholder', placeholder)
                    .addClass('js_field-state')
                    .val('');
                $state.replaceWith($newstate);
            }

            $(document.body).trigger('country-change.woocommerce', [country, $(this).closest('div')]);
        },

        change_state: function() {
            // Here we will find if state value on a select has changed and stick it to the country data
            var $this = $(this),
                state = $this.val(),
                $country = $this.parents('div.nmgr-shipping-fields').find(':input.js_field-country'),
                country = $country.val();

            $country.data('woocommerce.stickState-' + country, state);
        }
    };

    shipping_country_state.init();

});