function nm_is_defined() {
    if ('undefined' === typeof NM) {
        console.log('NM is not defined');
        return false;
    }
    return true;
}

/**
 * Get the data from a form.
 *
 * @param {string|Element} formObject The selector for the form element such as '.myform'
 * or the form Element object
 * @param {boolean} returnAsString Whether to return the data as a url query param string. Default false.
 * @returns {Object|string}
 */
function Nmeri_getFormdata(formObject, returnAsString) {
    var form = 'string' === typeof formObject ? document.querySelector(formObject) : formObject;
    var obj = {};
    var formData = new FormData(form);
    formData.forEach(function(value, key) {
        if (-1 !== key.lastIndexOf('[]')) {
            if (obj[key]) {
                obj[key].push(value);
            } else {
                obj[key] = [value];
            }
        } else {
            obj[key] = value;
        }
    });

    var submitBtn = form.querySelector('[type="submit"]');
    if (submitBtn && submitBtn.name) {
        obj[submitBtn.name] = submitBtn.value;
    }

    return returnAsString ? new URLSearchParams(obj).toString() : obj;
}

function Nmeri_block(selector) {
    if (jQuery.blockUI) {
        var toBlock = Array.isArray(selector) ? selector : [selector];

        toBlock.forEach(function(item) {
            if (!(jQuery(selector).is('.nm.is-blocked') || jQuery(selector).parents('.nm.is-blocked').length)) {
                jQuery(item).addClass('is-nm-blocked').block({
                    message: null,
                    overlayCSS: {
                        backgroundColor: '#fff'
                    },
                    css: {
                        backgroundColor: 'transparent',
                        border: 'none'
                    }
                });
            }
        });
    }
}

function Nmeri_unblock(selector) {
    if (jQuery.blockUI) {
        var toBlock = Array.isArray(selector) ? selector : [selector];
        toBlock.forEach(function(item) {
            jQuery(item).removeClass('is-nm-blocked').unblock();
        });
    }
}

/**
 * Appends a dialog element to the document body to be shown later
 * @param {string} template The dialog template
 * @returns {string|null} The id of the dialog template that can be used to trigger
 * the show() method on it later, or null if the insert operation is simply a reload
 * of the template from the server.
 *
 * The id is not returned if the insert operation is a reload to prevent the template
 * from being shown by the showTemplate() function as this is the function in which the
 * insertTemplate() function is called.
 */
function Nmeri_insertTemplate(template) {
    var d = new DOMParser().parseFromString(template, 'text/html');
    var id = d.querySelector('body').firstChild.getAttribute('id');
    var isReload = false;

    // Remove previous identitcal component from dom
    var prevComp = document.getElementById(id);
    if (prevComp) {
        isReload = prevComp.dataset.isReload ? true : false;
        prevComp.remove();
    }

    document.body.appendChild(d.querySelector('body').children[0]);
    return {
        isReload: isReload,
        id: id
    };
}

/**
 * Show a dialog template
 * @param {string} template Template to show in a dialog
 * @param {object} args Arguments used to compose the template. "component" key is
 * required with a value.
 * @returns {undefined}
 */
function Nmeri_showTemplate(template, args) {
    if (!nm_is_defined()) {
        return;
    }

    // insert template into dom
    var inserted = Nmeri_insertTemplate(template);

    // Set component related target
    var dialog = document.getElementById(inserted.id);
    var comp = new NM[Nmeri_getDialogComponentName(dialog)](dialog);
    var relatedTarget;

    if (args && args.btn && args.btn.dataset.nmId) {
        relatedTarget = args.btn.dataset.nmId;
    }

    comp.relatedTarget = relatedTarget;

    // Close any currently open dialog
    var openDialog = document.querySelector('.nm.dialog.show');
    if (openDialog) {
        Nmeri_closeDialog(openDialog);
    }

    // show template
    if (!inserted.isReload) {
        comp.show(relatedTarget || null);
    }
}

function Nmeri_showToast(notice) {
    if (!nm_is_defined()) {
        return;
    }

    var toastDiv = jQuery('.nm-component.toast-container');
    if (toastDiv.length) {
        var notices = notice instanceof Array ? notice : [notice];

        notices.forEach(function(el) {
            var html = jQuery(el)[0];
            toastDiv.append(html); // append to start at bottom so that it would stack according to last
            (new NM.Toast(html)).show();
        });
    }
}

/**
 * Get the component used as a dialog
 * @param {Element} dialog The dialog element
 * @returns {String} The component name, Modal or Offcanvas
 */
function Nmeri_getDialogComponentName(dialog) {
    return dialog.classList.contains('modal') ? 'Modal' : 'Offcanvas';
}

/**
 * Close the disalog
 * @param {Element} dialog The dialog element
 * @param {boolean} force Whether to simply close the dialog only and perform no other action related
 * to closing. Default true.
 *
 * In closing the currently opened dialog, we can show another dialog
 * based on the data attributes supplied to the currently opened dialog.
 *
 * The data attributes must have values equal to the ids of elements in the dom.
 * These elements in most cases should be dialog elements.
 *
 * If the dialog has any of the following data attributes, the corresponding actions would
 * be performed:
 *
 * nm-redirect:
 * This opens a new dialog. First it checks if the dialog has already been previously opened and
 * the instance exists. If so, it searches the instance for the button clicked to open the dialog
 * (the relatedTarget property). If this is founds it triggers a click on the button again to reload the
 * dialog. If it doesn't find the button, it takes the value supplied to the data attribute as the id of a
 * clickable button to open the new dialog and triggers a click on it.
 *
 * nm-back:
 * This simply shows a dialog that has already been previously shown and it's instance exists in the dom,
 * by simply invoking the show() method on the instance again. In this case, the dialog is not refreshed.
 */
function Nmeri_closeDialog(dialog, force) {
    if (!nm_is_defined()) {
        return;
    }

    if (true === dialog) {
        Nmeri_closeOpenDialog();
        return;
    }

    var component = Nmeri_getDialogComponentName(dialog);

    if (!force) {
        NM[component].getInstance(document.getElementById(dialog.getAttribute('id'))).hide();
        return;
    }

    if (dialog.dataset.nmRedirect) {
        var target = document.querySelector(dialog.dataset.nmRedirect);
        var newDialog = NM[component].getInstance(target);
        if (newDialog && newDialog.relatedTarget) {
            document.querySelector('[data-nm-id="' + newDialog.relatedTarget + '"]').click();
        } else if (target) {
            target.click();
        }
    } else if (dialog.dataset.nmBack) {
        NM[component].getInstance(document.querySelector(dialog.dataset.nmBack)).show();
    } else if (dialog.dataset.nmClose) {
        NM[component].getInstance(document.getElementById(dialog.getAttribute('id'))).hide();
    }
}

/**
 * Close the currently opened dialog
 */
function Nmeri_closeOpenDialog() {
    var dialog = document.querySelector('.nm-component.dialog.show');
    if (dialog) {
        Nmeri_closeDialog(dialog);
    }
}

// Replace templates in the dom when they have been retrieved from the server
function Nmeri_replaceTemplates(templates) {
    for (var key in templates) {
        var collection = document.querySelectorAll(key);
        if (collection) {
            var template = templates[key];
            collection.forEach(function(coll) {
                var wrap = document.createElement('div');
                wrap.innerHTML = template;
                var newTemplate = wrap.children[0];
                coll.replaceWith(newTemplate);
                jQuery(document.body).trigger('nm_replaced_template', {
                    key: key,
                    template: newTemplate
                });
            });
        }
    }
}

/**
 * Retrieve dialog templates from the server to be loaded in the dom.
 *
 * This function is used to update dialogs that are already in the dom which cannot
 * be updated using data-nm-redirect or replaceTemplates() function as they are not
 * directly connected to the current dialog being shown.
 *
 * This function is called after a dialog operation has been successful (for example when
 * the form in the dialog is saved to the database). We loop through all the dialogs in the
 * dom to see if any of them should be reloaded. This is determined by a 'data-nm-reload'
 * attribute on the dialog containing the id of the currently shown dialog as it value.
 *
 * If we find such dialog, we simply reload it by triggering a click on its 'relatedTarget'
 * property which has to be set.
 *
 * @param {Element} dialog
 */
function Nmeri_reloadTemplates(dialog) {
    if (!nm_is_defined()) {
        return;
    }

    var collection = document.querySelectorAll('.dialog[data-nm-reload]');
    if (collection.length) {
        var component = Nmeri_getDialogComponentName(dialog);
        var dialogId = dialog.getAttribute('id');

        collection.forEach(function(coll) {
            var reloads = JSON.parse(coll.dataset.nmReload);
            if (reloads.indexOf('#' + dialogId) > -1) {
                var newDialog = NM[component].getInstance(coll);
                if (newDialog && newDialog.relatedTarget) {
                    coll.dataset.isReload = true;
                    document.querySelector('[data-nm-id="' + newDialog.relatedTarget + '"]').click();
                }
            }
        });
    }
}

/**
 * Initialize a specific bootstrap component
 *
 * How to use: for example, to initialize the tooltip component on the document:
 * - initComponent('tooltip', document);
 *
 * @param {string} component The name of the component e.g. tooltip, popover
 * @param {object} element The element to initialize the component on. Defaults to document
 */
function Nmeri_initComponent(component, element) {
    if ('undefined' !== typeof component && 'undefined' !== typeof NM) {
        var componentType = component.charAt(0).toUpperCase() + component.slice(1);
        var el = element instanceof Element ? element : document;
        var List = [].slice.call(el.querySelectorAll('.nm[data-bs-toggle="' + component + '"]'));
        var options = {};

        if ('tooltip' === component || 'popover' === component) {
            options['html'] = true;
            options['customClass'] = 'nm-component';
        }

        List.map(function(triggerEl) {
            return new NM[componentType](triggerEl, options);
        });
    }
}

/**
 * Go to a url
 * @param {string} url
 */
function Nmeri_redirect(url) {
    if (-1 === url.indexOf('https://') || -1 === url.indexOf('http://')) {
        window.location = url;
    } else {
        window.location = decodeURI(url);
    }
}

/**
 * Perform standard operations with the response from an ajax request
 */
function Nmeri_processResponse(response, args) {
    if (response.log) {
        console.log(response);
    }

    if (args && args.block) {
        Nmeri_unblock(args.block);
    }

    if ((args && args.close_dialog) || response.close_dialog) {
        Nmeri_closeDialog(args.close_dialog || response.close_dialog);
    }

    if (response.success && args && args.dialog) {
        Nmeri_reloadTemplates(args.dialog);
    }

    if (response.show_template) {
        Nmeri_showTemplate(response.show_template, args);
    }

    if (response.toast_notice) {
        Nmeri_showToast(response.toast_notice);
    }

    if (response.replace_templates) {
        Nmeri_replaceTemplates(response.replace_templates);
    }

    if (response.redirect) {
        Nmeri_redirect(response.redirect);
    }
}

/**
 * Common js code
 */

var nmgr = window.nmgr || {};

jQuery(function($) {

    /**
     * Define a common blockUI block setup for plugin
     *
     * @param {string | jQuery object} selector The element to be blocked
     */
    nmgr.block = function(selector) {
        if (!nmgr.is_blocked(selector)) {
            $(selector).addClass('nm is-blocked').block({
                message: null,
                overlayCSS: {
                    backgroundColor: '#fff'
                },
                css: {
                    backgroundColor: 'transparent',
                    border: 'none'
                }
            });
        }
    };

    /**
     * Define a common blockUI unblock setup for plugin
     *
     * @param {string | jQuery object} selector The element to be unblocked
     */
    nmgr.unblock = function(selector) {
        $(selector).removeClass('nm is-blocked').unblock();
    };

    /**
     * Check whether an element is currently blocked with blockUI
     *
     * @param {string | jQuery object} selector The element to be checked
     * @returns {jQuery.length|Boolean} True if the element is blocked, false if not
     */
    nmgr.is_blocked = function(selector) {
        return $(selector).is('.nm.is-blocked') || $(selector).parents('.nm.is-blocked').length;
    };

    /**
     * Scroll to an element
     *
     * @param {string | jQuery object} selector The element to scroll to
     */
    nmgr.scroll_to = function(selector) {
        if ($(selector).length) {
            if (nmgr.is_in_dialog(selector)) {
                $('#nmgr-modal').animate({
                    scrollTop: 0
                });
            } else {
                $('html, body').animate({
                    scrollTop: ($(selector).offset().top - 100)
                }, 1000);
            }
        }
    };

    /**
     * Go to a url
     * @param {string} url
     * @deprecated since 4.0.0
     */
    nmgr.go_to = function(url) {
        if ('function' === typeof Nmeri_redirect) {
            console.warn('nmgr.go_to() is deprecated since version 4.0.0. Use Nmeri_redirect()');
            Nmeri_redirect(url);
        } else {
            if (-1 === url.indexOf('https://') || -1 === url.indexOf('http://')) {
                window.location = url;
            } else {
                window.location = decodeURI(url);
            }
        }
    };

    /**
     * Remove woocommerce notices from the page
     * @returns {undefined}
     */
    nmgr.remove_notice = function(selector) {
        var notice_selectors = '.woocommerce-error, .woocommerce-message, .woocommerce-info';

        if ('undefined' !== typeof(selector)) {
            $(selector).find(notice_selectors).remove();
        } else {
            $(notice_selectors).remove();
        }
    };

    /**
     * Make a table responsive when it's width is above a certain size.
     *
     * The responsiveness in this case is for desktop display rather than mobile.
     * By default, the table should already be mobile responsive.
     *
     * @param {string} selector The table element to make responsive
     * @param {int} width Minimum width at which table would be responsive. Default 700px
     */
    nmgr.make_table_responsive = function(selector, width) {
        width = width || 600;
        selector = selector || '.nmgr-table';

        var table = $(selector);
        if (table.width() > width) {
            table.addClass('responsive');
        } else {
            table.removeClass('responsive');
        }
    };

    nmgr.tiptip = function() {
        $('.nmgr-tip').each(function() {
            var $element = $(this),
                options = {},
                title = $element.find('title');

            if ($element.attr('data-tiptip-position')) {
                options['defaultPosition'] = $element.attr('data-tiptip-position');
            }

            if ($element.is('svg')) {
                options['content'] = function() {
                    if (title.length) {
                        if (title.attr('data-title')) {
                            title.text('');
                            return title.attr('data-title');
                        }
                    }
                };
            }

            $element.tipTip(options);
        });
    };

    nmgr.tiptip();

    $(document.body).on('show.bs.modal', '.nmgr.dialog', function() {
        nmgr.tiptip();
        Nmeri_initComponent('tooltip', this);
    });


    // Check if we are in a modal window
    nmgr.is_in_modal = function(element) {
        console.log('nmgr.is_in_modal is deprecated. Use nmgr.is_in_dialog');
        return nmgr.is_in_dialog(element);
    };

    nmgr.is_in_dialog = function(element) {
        var dialog = '.nm-component.dialog.nmgr.show';
        if (element) {
            return document.querySelector(dialog) && $(element).closest(dialog).length > 0;
        } else {
            return document.querySelector(dialog);
        }
    };

    nmgr.init_datepicker = function(selector) {
        selector = selector || '.nmgr-use-datepicker';
        var datepickerClass = 'nmgr-datepicker';
        var datepickerOptions = nmgr_global_params.datepicker_options;

        if (datepickerOptions.styleDatepicker) {
            datepickerOptions = Object.assign(datepickerOptions, {
                beforeShow: function(input, inst) {
                    if (inst.input.hasClass('nmgr-use-datepicker')) {
                        inst.dpDiv.addClass(datepickerClass);
                    }
                },
                onClose: function(dateText, inst) {
                    if (inst.input.hasClass('nmgr-use-datepicker')) {
                        inst.dpDiv.removeClass(datepickerClass);
                    }
                }
            });
        }

        $(selector).each(function() {
            var data = this.dataset;
            for (var key in data) {
                if (key.includes('datepicker')) {
                    var dpkey = key.replace('datepicker', '');
                    var dpkey2 = dpkey.charAt(0).toLowerCase() + dpkey.slice(1);
                    datepickerOptions[dpkey2] = data[key];
                }
            }

            for (var key in datepickerOptions) {
                if ('string' === typeof(datepickerOptions[key]) && -1 < datepickerOptions[key].indexOf('function')) {
                    datepickerOptions[key] = Function('"use strict";return(' + datepickerOptions[key] + ')')();
                }
            }

            $(this).datepicker(datepickerOptions)
                .on("keyup", function() {
                    if (datepickerOptions.altField) {
                        var date;
                        var dateFormat = $(this).datepicker('option', 'dateFormat');

                        try {
                            date = $.datepicker.parseDate(dateFormat, this.value);
                        } catch (e) {}

                        if (!date) {
                            $(datepickerOptions.altField).val('');
                        }
                    }
                });

        });
    };

    // Make selected tables responsive
    if (nmgr_global_params.responsive_tables) {
        Object.values(nmgr_global_params.responsive_tables).forEach(function(table) {
            var responsiveWidth = table.width ? table.width : '';

            nmgr.make_table_responsive(table.selector, responsiveWidth);

            $(window).on('resize', function() {
                nmgr.make_table_responsive(table.selector, responsiveWidth);
            });

            if (table.events && 0 < table.events.length) {
                table.events.forEach(function(event) {
                    if ('object' === typeof event) {
                        var selector = event.selector;
                        var ev = event.event;
                    } else {
                        var selector = document.body;
                        var ev = event;
                    }
                    $(selector).on(ev, function() {
                        if ($(table.selector).is(':visible')) {
                            nmgr.make_table_responsive(table.selector, responsiveWidth);
                        }
                    });
                });
            }
        });
    }

    // Make selected tables datatable
    if (nmgr_global_params.datatables_tables) {
        Object.values(nmgr_global_params.datatables_tables).forEach(function(table) {
            /**
             *  After initialization always add width 100% to table to make it responsive
             *  This is particularly important if we are hiding some columns on the table as
             *  it would prevent the table from being responsive.
             *
             *  (This code must be before the actual initialization code for it to work)
             */
            $(document.body).on('init.dt', table.selector, function() {
                this.style.width = '100%';
            });

            // Initialize datatable on page load
            $(table.selector).DataTable(table.options);

            // Initialize datatable on specific events
            if (table.events && 0 < table.events.length) {
                table.events.forEach(function(event) {
                    if ('object' === typeof event) {
                        var selector = event.selector;
                        var ev = event.event;
                    } else {
                        var selector = document.body;
                        var ev = event;
                    }
                    $(selector).on(ev, function() {
                        if ($(table.selector).is(':visible')) {
                            $(table.selector).DataTable(table.options);
                        }
                    });
                });
            }

        });
    }

    nmgr.global_ajax_params = function() {
        var postdata = {
            action: 'nmgr_post_action',
            _wpnonce: nmgr_global_params.nonce,
            ajax_url: nmgr_global_params.ajax_url,
            nmgr_global: JSON.stringify(nmgr_global_params.global)
        };

        return postdata;
    }

    nmgr.post_action = function(args) {
        if (args.btn && args.btn.dataset && args.btn.dataset.notice && !window.confirm(args.btn.dataset.notice)) {
            return false;
        }

        if (args.block) {
            Nmeri_block(args.block);
        }

        if (!args.postdata) {
            return;
        }

        var postdata = nmgr.global_ajax_params();
        Object.assign(postdata, args.postdata);

        $.post(postdata.ajax_url, postdata, function(response) {
            Nmeri_processResponse(response, args);

            $(document.body).trigger('nmgr_post_action_response', {
                postdata: postdata,
                response: response
            });
        });
    }





});


jQuery(function($) {

    if (typeof nmgr_frontend_params === 'undefined') {
        return false;
    }

    /**
     * Handles wishlist cart actions
     */

    var wishlist_cart = {
        init: function() {
            $(document.body)
                .on('nmgr_added_to_wishlist nmgr_items_reloaded', this.reload);
        },

        reload: function(e) {
            var carts = $('.nmgr-cart');

            if (carts.length > 0) {
                carts.each(function() {
                    wishlist_cart.get_cart_template(this);
                });
            }
        },

        get_cart_template: function(container) {
            var postdata = {
                nmgr_post_action: 'load_wishlist_cart',
                type: container.dataset.type,
                cart_id: container.getAttribute('id')
            };

            var data = {
                block: container,
                postdata: postdata
            };

            nmgr.post_action(data);
        }
    };

    wishlist_cart.init();

    /**
     * Handles adding a product to the wishlist
     * either normally or via ajax
     */

    jQuery(function($) {
        var selector = {
            wrapper: '.nmgr-add-to-wishlist-wrapper',
            form: '.nmgr-add-to-wishlist-form',
            add_to_wishlist_btn: '.nmgr-add-to-wishlist-button',
            dialog_submit_btn: '.nmgr-dialog-submit-button',
            dialog_form: '.nmgr-add-to-wishlist-content'
        };
        var add_to_wishlist = {
            init: function() {
                $(document.body)
                    .on('nmgr_post_action_response', this.post_action_response)
                    .on('click', selector.add_to_wishlist_btn, this.prepare_to_add)
                    .on('click', '.nm-dialog-button.nmgr_add_to_wishlist', this.add_to_wishlist_dialog_submit_btn_clicked)
                    .on('nmgr_removed_from_wishlist', this.removed_from_wishlist)
                    .on('click', '.nmgr-add-to-new-wishlist', this.add_to_new_wishlist)
                    .on('show.bs.modal', '.nmgr.dialog', this.set_selected_wishlist)
                    .on('change', 'select.list-of-wishlists', this.update_product_options_for_wishlist);
            },

            post_action_response: function(e, args) {
                var post_action = args.postdata.nmgr_post_action;

                if (['auto_create_wishlist', 'save_profile', 'save_shipping'].includes(post_action)) {
                    var $form = args.response.wishlist_type ?
                        $('.nmgr-add-to-wishlist-form.' + args.response.wishlist_type) :
                        $('');

                    // Remove the flag from creating a default wishlist for the user
                    if (args.response.success) {
                        $form.attr('data-create_wishlist', '');
                    }

                    if (args.response.wishlist) {
                        // update the wishlist inputs
                        var $wishlist_input = $form.find('input[name=nmgr_wid]');
                        $wishlist_input.val(args.response.wishlist.id);

                        if (!args.response.wishlist.needs_shipping_address) {
                            $wishlist_input.attr('data-shipping-address-required', 0);
                        }
                    }
                }

                /**
                 * atw_pid_click property is in the response object to trigger adding to wishlist
                 * using the simple mode
                 */
                if (parseInt(args.response.atw_pid_click || 0) && args.response.wishlist_type) {
                    Nmeri_closeOpenDialog();
                    $('.nmgr-add-to-wishlist-form-' + args.response.atw_pid_click + '.' + args.response.wishlist_type)
                        .find(selector.add_to_wishlist_btn)
                        .trigger('click');
                }

                /**
                 * atw_action property is in the response object to trigger adding to wishlist using the
                 * advanced mode
                 */
                if (args.response.atw_action && args.response.wishlist_type) {
                    var btn = $('.nmgr-add-to-wishlist-form-' +
                            args.response.atw_action.nmgr_pid +
                            '.' + args.response.wishlist_type)
                        .find(selector.add_to_wishlist_btn);
                    add_to_wishlist.ajax_action(btn, args.response.atw_action);
                    Nmeri_closeOpenDialog();
                }
            },

            /**
             * Events to perform after removing a product from the wishlist
             * Response object should contain 'items_data' key which is an array of objects containing for each item removed,
             * - wishlist_item_id
             * - wishlist_id
             * - product_id
             * - product_in_wishlist (This is an integer value, 0 or 1, representing whether the removed product is in
             * any of the user's wishlists)
             */
            removed_from_wishlist: function(e, response) {
                // Remove any 'view wishlists' button present in the page
                $(selector.wrapper).find('.nmgr-view-wishlist-button').remove();

                /**
                 * Remove the 'product-in-wishlist' icon from each add to wishlist form element of the product that has it.
                 *
                 * This doesn't work for grouped products at the moment becuase of how woocommerce handles grouped
                 * products. It is difficult to track a grouped product from its children. But it is only possible to track the
                 * children from the grouped product itself.
                 */
                if (response && response.items_data) {
                    response.items_data.forEach(function(item) {
                        var form = $('.nmgr-add-to-wishlist-form-' + item.product_id);
                        if (form.length > 0) {
                            if (!item.product_in_wishlist) {
                                // remove the in-wishlist icon on the form button
                                form.removeClass('product-in-wishlist');
                            }
                        }
                    });
                }
            },

            /**
             * Show notice on the page
             *
             * @param {string} html The notice html.
             * @param {jQuery oject} form The form element containing the add to wishlist button.
             */
            show_notice: function(html, form) {
                $target = $('.woocommerce-notices-wrapper:first') || $('.nmgr-add-to-wishlist-button').closest('.woocommerce');
                if ($target.length) {
                    $target.prepend(html);
                    $('html, body').animate({
                        scrollTop: ($target.offset().top - 100)
                    }, 1000);
                } else {
                    form.closest(selector.wrapper).after(html);
                }
            },

            /**
             * Check if a product should be added to a wishlist when the add to wishlist button is clicked on the page
             *
             * @param {object} e Event
             */
            prepare_to_add: function(e) {
                e.preventDefault();

                var $thisbutton = $(this);
                var form = this.closest('form');
                var formdata = Nmeri_getFormdata(form);
                var $form = $thisbutton.parent(selector.form);
                var postdata = Object.assign({
                    add_to_wishlist_formdata: $form.serialize(),
                    add_to_wishlist_mode: $form.attr('data-add_to_wishlist_mode')
                }, formdata);
                var params = {
                    clickedBtn: this,
                    postdata: postdata
                };

                // Deal with redirects immediately
                if ($form.is('.redirect')) {
                    $form.submit();
                    return;
                }

                // Remove any 'view wishlists' button present in the page
                $(selector.wrapper).find('.nmgr-view-wishlist-button').remove();

                // Deal with disabled status
                if ($thisbutton.is('.disabled')) {
                    // Deal with variation form
                    if ($thisbutton.is('.wc-variation-is-unavailable')) {
                        alert($form.attr('data-unavailable_text'));
                    } else if ($thisbutton.is('.wc-variation-selection-needed')) {
                        alert($form.attr('data-make_a_selection_text'));
                    }
                    return;
                }

                if ('auto' === $form.attr('data-create_wishlist')) {
                    // If we are supposed to create a new wishlist for the user, create it
                    add_to_wishlist.auto_create_wishlist(params);
                    return;
                } else if ('modal' === $form.attr('data-create_wishlist')) {
                    // If the user is supposed to create a new wishlist, let him do it
                    add_to_wishlist.show_create_new_wishlist_dialog(params);
                    return;
                } else if ('advanced' === params.postdata.add_to_wishlist_mode) {
                    add_to_wishlist.show_add_to_wishlist_dialog(params);
                    return;
                } else {
                    var $wishlist_input = $form.find('input[name=nmgr_wid]');
                    if ($wishlist_input.val() && 1 === parseInt($wishlist_input.attr('data-shipping-address-required'))) {
                        add_to_wishlist.dialog_set_shipping_address($wishlist_input.val(), params);
                        return;
                    } else if ($form.is('.product-type-grouped')) {
                        add_to_wishlist.show_add_to_wishlist_dialog(params);
                        return;
                    } else {
                        add_to_wishlist.ajax_action($(this), formdata);
                        return;
                    }
                }
            },

            auto_create_wishlist: function(params) {
                var postdata = params && params.postdata || {};
                postdata['nmgr_post_action'] = 'auto_create_wishlist';

                var args = {
                    block: params && params.clickedBtn,
                    postdata: postdata
                };

                nmgr.post_action(args);
            },

            add_to_new_wishlist: function(e) {
                e.preventDefault();

                var params = {
                    clickedBtn: this,
                    postdata: {
                        add_to_wishlist_formdata: new URLSearchParams(this.dataset).toString()
                    }
                };

                add_to_wishlist.show_create_new_wishlist_dialog(params);
            },

            /**
             * Create a new wishlist via the dialog
             * @param {object} params Params
             */
            show_create_new_wishlist_dialog: function(params) {
                var postdata = params && params.postdata || {};
                postdata['nmgr_post_action'] = 'show_create_new_wishlist_dialog';

                var args = {
                    block: params && params.clickedBtn,
                    postdata: postdata
                };

                nmgr.post_action(args);
            },

            /**
             * Display the add to wishlist dialog
             */
            show_add_to_wishlist_dialog: function(params) {
                var postdata = params && params.postdata || {};
                postdata['nmgr_post_action'] = 'show_add_to_wishlist_dialog';

                var args = {
                    block: params && params.clickedBtn,
                    postdata: postdata
                };

                nmgr.post_action(args);
            },

            /**
             * Add a product to a wishlist via ajax
             *
             * @param {jQuery object} $thisbutton The form submit button
             * @param {object} formdata The data from the add to wishlist form for the product
             */
            ajax_action: function($thisbutton, formdata) {
                var $form = $thisbutton.parent(selector.form);

                $thisbutton.removeClass('added');
                $thisbutton.addClass('loading');

                if ($form.attr('data-show_animation') > 0) {
                    $form.find('.nmgr-animate').removeClass('nmgr-animation-scaleshrink');
                }

                var data = Object.assign({
                    action: 'nmgr_add_to_wishlist',
                    _wpnonce: nmgr_frontend_params.nonce
                }, formdata);

                $(document.body).trigger('nmgr_adding_to_wishlist', [$thisbutton, $form, data]);

                $.post(nmgr_global_params.ajax_url, data, function(response) {
                    $('.woocommerce-error, .woocommerce-message, .woocommerce-info').remove();
                    $thisbutton.removeClass('loading');

                    if (!response) {
                        return;
                    }

                    if (response.error) {
                        add_to_wishlist.show_notice(response.notice, $form);
                        return;
                    } else if (response.success) {

                        $form.addClass('product-in-wishlist');
                        if ($form.attr('data-show_animation') > 0) {
                            $form.find('.nmgr-animate').addClass('nmgr-animation-scaleshrink');
                        }

                        if ($form.is('.archive') && $form.attr('data-show_notice_archive') > 0 ||
                            ($form.is('.single') && $form.attr('data-show_notice_single') > 0)) {
                            add_to_wishlist.show_notice(response.notice, $form);
                        }

                        if (response.notice &&
                            $(response.notice).find('.wc-forward').length &&
                            (($form.is('.archive') && $form.attr('data-show_notice_archive') < 1) ||
                                ($form.is('.single') && $form.attr('data-show_notice_single') < 1)) &&
                            $form.attr('data-show_view_wishlist_link') &&
                            !$form.parent(selector.wrapper).find('.nmgr-view-wishlist-button').length) {
                            var $wishlist_link = $('.wc-forward', $(response.notice))
                                .removeClass('button')
                                .addClass('nmgr-view-wishlist-button')[0]
                                .outerHTML;
                            $form.after($wishlist_link);
                        }
                    }

                    /**
                     * Trigger event after adding to wishlist
                     *
                     * @param {object} response The response from the add to wishlist action
                     *  - This is an object containing the error, success, notice and result values
                     *  - relating to the product added to the wishlist.
                     *  -- The 'result' key in the object contains the product_id, variation_id,
                     *  -- and wishlist_id of the product added to the wishlist
                     *
                     * @param {jQuery object} $thisbutton The button that was clicked to add to wishlist
                     * @param {jQuery object} $form The form parent element of the button
                     */
                    $(document.body).trigger('nmgr_added_to_wishlist', [response, $thisbutton, $form]);
                });
            },

            // Show the wishlist shipping address form in a modal
            dialog_set_shipping_address: function(wishlist_id, params) {
                var postdata = Object.assign(params.postdata, {
                    wishlist_id: wishlist_id,
                    nmgr_post_action: 'show_set_shipping_address_dialog'
                });
                var args = {
                    block: params && params.clickedBtn,
                    postdata: postdata
                };

                nmgr.post_action(args);
            },

            add_to_wishlist_dialog_submit_btn_clicked: function(e) {
                e.preventDefault();

                var dialog = this.closest('.dialog.nmgr');
                var dialog_form = dialog.querySelector('form');
                var formdata = Nmeri_getFormdata(dialog_form);
                var $dialog_form = $(dialog_form);
                var clickedBtn = this;

                // Get the wishlist select field
                var wishlist_select = dialog.querySelector('select[name=nmgr_wid]');

                // If the select wishlist requires a shipping address, show the shipping address form
                if (wishlist_select) {
                    if (1 === parseInt($(wishlist_select).find('option:selected').attr('data-shipping-address-required'))) {
                        var params = {
                            postdata: {
                                add_to_wishlist_formdata: new URLSearchParams(formdata).toString()
                            },
                            clickedBtn: this
                        };
                        add_to_wishlist.dialog_set_shipping_address(wishlist_select.value, params);
                        return;
                    }
                }

                // Make sure a grouped product is selected, for grouped products
                if ($dialog_form.hasClass('product-type-grouped')) {
                    var quantityInputs = $dialog_form.find('input.qty');
                    if (quantityInputs.length) {
                        var hasValue = false;
                        for (var i = 0; i < quantityInputs.length; i++) {
                            if ($(quantityInputs[i]).val() > 0) {
                                hasValue = true;
                            }
                        }

                        if (!hasValue && clickedBtn.dataset.select_quantity_text) {
                            alert([clickedBtn.dataset.select_quantity_text]);
                            return;
                        }
                    } else if ($dialog_form.find('input[name*=nmgr_qty]:not(".qty")').length &&
                        !$dialog_form.find('input[name*=nmgr_qty]:not(".qty"):checked').length &&
                        clickedBtn.dataset.select_product_text) {
                        alert([clickedBtn.dataset.select_product_text]);
                        return;
                    }
                }

                // At this point we're good to go. We have a selected wishlist and we need to submit the form.
                var $form = $('.nmgr-add-to-wishlist-form-' + formdata.nmgr_pid + '.' + formdata.nmgr_type);
                var $add_to_wishlist_button = $form.find(selector.add_to_wishlist_btn);
                if ($form.is('.nmgr-ajax-add-to-wishlist:not(".redirect, .disabled")') &&
                    $add_to_wishlist_button.is(':not(".disabled")')) {
                    add_to_wishlist.ajax_action($add_to_wishlist_button, formdata);
                    Nmeri_closeDialog(dialog);
                    return;
                }
            },

            /**
             * When a wishlist is selected to add a product to in the add to wishlist dialog,
             * update the options set for the product in the wishlist (that is if it is in the wishlist),
             * such as its favourite status so that the wishlist owner can know the options set for the
             * product in the currently selected wishlist
             */
            update_product_options_for_wishlist: function() {
                var wishlist_id = $(this).val();
                var data_name = 'data-in-wishlist-' + wishlist_id;
                var inputs = $(this.closest('.dialog')).find('[' + data_name + ']');

                if (inputs.length > 0) {
                    $(inputs).each(function() {
                        if ($(this).is(':checkbox')) {
                            var label = $('label[for="' + $(this).attr('id') + '"]');
                            if ($(this).attr(data_name) === $(this).val()) {
                                this.checked = true;
                                label.attr('title', label.attr('data-title-on'));
                            } else {
                                this.checked = false;
                                label.attr('title', label.attr('data-title-off'));
                            }
                        } else if ($(this).is('.nmgr-cart-qty-wrapper')) {
                            if ($(this).attr(data_name)) {
                                $(this).find('.nmgr-qty').text($(this).attr(data_name));
                                $(this).fadeIn();
                            } else {
                                $(this).fadeOut(300, function() {
                                    $(this).find('.nmgr-qty').text('');
                                });
                            }
                        }
                    });
                }
            },

            /**
             * If we have the list of wishlists select element in the modal,
             * trigger the change event on this element so that the currently selected wishlist
             * would be set up for event handlers attached to it to fire
             */
            set_selected_wishlist: function() {
                var list_of_wishlists = $(this).find('select.list-of-wishlists');
                if (list_of_wishlists.length > 0) {
                    list_of_wishlists.trigger('change');
                }
            }

        };

        /*
         * Add variable products to the wishlist
         *
         * (Used on single product page)
         * @todo convert to class
         */
        var add_to_wishlist_variable = {
            init: function() {
                $('.variations_form')
                    .on('found_variation.wc-variation-form', this.found_variation)
                    .on('show_variation', this.show_variation)
                    .on('hide_variation', this.hide_variation);
            },
            /**
             * Get the wishlist form attached to the variation form
             *
             * @param {jQuery object} obj Variation form
             * @returns {object}
             */
            get_wishlist_form: function(obj) {
                var product_id = $(obj).data('product_id');
                return $(obj).closest('.product').find('[data-nmgr_product_id=' + product_id + ']');
            },
            found_variation: function(e, variation) {
                var variationFormValues = $(this).serialize(),
                    $form = add_to_wishlist_variable.get_wishlist_form(this);
                $form.find('input[name="nmgr_vid"]').val(variation.variation_id).change();
                $form.find('input[name="nmgr_wc_form_values"]').val(variationFormValues).change();
                $form.find('input.qty').attr('min', variation.min_qty).attr('max', variation.max_qty);
            },
            show_variation: function(e, variation, purchasable) {
                var $form = add_to_wishlist_variable.get_wishlist_form(this),
                    $add_to_wishlist_btn = $form.find(selector.add_to_wishlist_btn);

                // Allow out of stock variable products to be added to wishlist (variation.is_in_stock)
                if (purchasable || !variation.is_in_stock) {
                    $form.removeClass('disabled');
                    $add_to_wishlist_btn.removeClass('disabled wc-variation-selection-needed');
                } else {
                    $form.addClass('disabled');
                    $add_to_wishlist_btn
                        .removeClass('wc-variation-selection-needed')
                        .addClass('disabled wc-variation-is-unavailable');
                }
            },
            hide_variation: function() {
                var $form = add_to_wishlist_variable.get_wishlist_form(this);
                $form.addClass('disabled');
                $form.find(selector.add_to_wishlist_btn).addClass('disabled wc-variation-selection-needed');
            }
        };

        add_to_wishlist.init();
        add_to_wishlist_variable.init();


    });

    nmgr.save_form = {
        init: function() {
            nmgr.init_datepicker();

            $(document.body)
                .on('show.bs.modal', '.nm-component.dialog', this.dialog_init)
                .on('submit', '.nmgr-ajax .nmgr-form', this.submit)
                .on('nmgr_post_action_response', this.post_action_response);
        },

        post_action_response: function(e, args) {
            var post_action = args.postdata.nmgr_post_action;

            if ('save_form' === post_action) {
                if (args.response.error_data) {
                    for (var key in args.response.error_data) {
                        var $container = $('#' + key).closest('.nmgr-validate');
                        $container.removeClass('nmgr-valid').addClass('nmgr-invalid');
                        var arr = args.response.error_data[key];
                        arr.forEach(function(item) {
                            $container.append('<div class="nmgr-error-text">' + item + '</div>');
                        });
                    }
                }

                $(document.body).trigger('nmgr_save_form', [args.response]);
            }
        },

        submit: function(e) {
            e.preventDefault();

            var $form = $(this);
            $form.find('.nmgr-validate').removeClass('nmgr-invalid').find('.nmgr-error-text').remove();

            var args = {
                block: this,
                postdata: {
                    data: $form.serialize(),
                    nmgr_post_action: 'save_form',
                    nmgr_global: $form.closest('.nmgr-account-template').attr('data-nmgr_global')
                }
            };

            nmgr.post_action(args);
        },

        // Actions to perform if the shipping address form is in the dialog
        dialog_init: function() {
            if (this.querySelector('#nmgr-shipping')) {
                /**
                 *  If we have the woocommerce shipping form, unbind the selectWoo function (wc_country_select_select2)
                 *  which prevents the modal from scrolling when a country is selected the second time.
                 *  The function is found in country-select.js. This is a bug in the woocommerce selectWoo plugin.
                 *  Because this function is not stored in a global variable, we can only unbind the event itself. This is currently
                 *  the easiest solution to the problem and we hope it doesn't affect other handlers bound to the event.
                 */
                $(document.body).off('country_to_state_changed');
            }
        }
    };

    nmgr.save_form.init();

    /*
     * js events related to wishlist images template
     */

    var nmgr_images = {
        container: '#nmgr-images',

        noticeSelector: '<div class="nmgr-image-notice woocommerce-error"></div>',

        noticeTimer: '',

        // Background image container selector
        bgContainer: '.nmgr-wrapper.nmgr-bg',

        // Featured image container selector
        featContainer: '.nmgr-wrapper.nmgr-thumbnail',

        // Delete image container selector
        delContainer: '.delete-image',

        init: function() {
            this.maybe_hide_delete_image_button();

            $(document.body)
                .on('change', '#nmgr-images .nmgr-image-input', this.validate)
                .on('click', '#nmgr-images .delete-image', this.delete)
                .on('nmgr_wishlist_created', function(e, wishlist) {
                    nmgr_images.reloaded(wishlist);
                })
                .on('nmgr_post_action_response', this.post_action_response);
        },

        post_action_response: function(e, args) {
            var post_action = args.postdata.nmgr_post_action;

            if (-1 !== ['delete_image'].indexOf(post_action)) {
                if (args.response.success) {
                    nmgr_images.reloaded(args.response.wishlist || null);
                }
            }
        },

        /**
         * Conditionally hide the delete image button for each image
         * based on whether an image already exists
         */
        maybe_hide_delete_image_button: function() {
            var wrapper = $('#nmgr-images .nmgr-wrapper');
            for (var i = 0; i < wrapper.length; i++) {
                if ($(wrapper[i]).attr('data-image_id') < 1) {
                    $(wrapper[i]).find('.delete-image').first().hide();
                } else {
                    $(wrapper[i]).find('.delete-image').first().show();
                }
            }
        },

        reloaded: function(wishlist) {
            $(document.body).trigger('nmgr_images_reloaded', wishlist);
            this.maybe_hide_delete_image_button();
        },

        delete: function(e) {
            e.preventDefault();

            var $form = $(this);
            var postdata = Object.assign({
                nmgr_post_action: 'delete_image',
                nmgr_global: $form.closest('.nmgr-account-template').attr('data-nmgr_global')
            }, this.dataset);
            var args = {
                block: this,
                btn: this,
                postdata: postdata
            };

            nmgr.post_action(args);
        },

        /**
         * Validate the selected file
         * only images accepted
         */
        validate: function(e) {
            var $element = $(e.target),
                wrapper = $element.closest('.nmgr-wrapper'),
                context = wrapper.attr('data-context'),
                file = $element[0].files[0],
                fileType = file.type;

            nmgr.block(wrapper);

            // reset file input field to enable consequent uploads of same file
            $element.val('');

            if (window.Blob) {
                if (!fileType.match('image.*')) {
                    alert(nmgr_frontend_params.images.i18n_invalid_type_text);
                    nmgr.unblock(wrapper);
                } else {
                    nmgr_images.upload(file, context);
                }
                return;
            }
            nmgr_images.upload(file, context);
        },

        /**
         * Upload the selected file to wordpress media
         *
         * @param object file Object
         * @param string context Whether the file should be uploaded as a featured or background image
         * (accepted values: featured, background)
         */
        upload: function(file, context) {
            var formData = new FormData(),
                postdata = nmgr.global_ajax_params(),
                $wrapper = $(nmgr_images.container);

            for (var item in postdata) {
                formData.append(item, postdata[item]);
            }

            formData.append('nmgr_post_action', 'save_image');
            formData.append('nmgr_global', $wrapper.attr('data-nmgr_global'));
            formData.append('nmgr-file', file);
            formData.append('context', context);
            formData.append('wishlist_id', $wrapper.attr('data-wishlist-id'));

            $.ajax({
                url: postdata.ajax_url,
                data: formData,
                processData: false,
                contentType: false,
                dataType: 'json',
                type: 'POST',
                success: function(response) {
                    Nmeri_processResponse(response);
                    if (response.success) {
                        nmgr_images.reloaded(response.wishlist || null);
                    }
                }
            });
        }

    };


    nmgr_images.init();

    /**
     * Js events related to wishlist items template
     */


    nmgr.items = {
        container: '#nmgr-items',

        init: function() {
            $(document.body)
                .on('nmgr_post_action_response', this.post_action_response)
                .on('click', '#nmgr-items .nmgr-post-action', this.pre_post_action)
                .on('input', '.nmgr-item-part.nmgr_quantity input.quantity', this.save_item_quantity)
                .on('change', '.nmgr-product-search', this.new_add_items_row)
                .on('click', '.nmgr-add-items-dialog .nmgr-add', this.add_items)
                .on('removed_from_cart', nmgr.remove_notice)
                .on('click', '.nmgr_add_to_cart_button', this.add_single_product_to_cart)
                .on('click', '.nmgr-bulk-add-to-cart', this.add_multiple_products_to_cart)
                .on('change', '#nmgr-items .nmgr-select-all', this.toggle_select_items)
                .on('change', '#nmgr-items .nmgr-select', this.maybe_show_bulk_actions)
                .on('click', '#nmgr-items .nmgr-nav', this.paginate)
                .on('nmgr_add_to_cart_response', this.remove_loading_animation)
                .on('click', '.nmgr-items-display-modes .mode', this.save_display_mode_to_session)
                .on('input', '#nmgr-purchase-refund-form .nmgr-quantity', this.toggle_purchase_refund_options)
                .on('submit', '#nmgr-purchase-refund-form', this.submit_purchase_refund_form);

            var addItemsDialog = document.getElementById('nmgr-add-items-dialog');
            if (addItemsDialog) {
                addItemsDialog.addEventListener('show.bs.modal', this.init_add_items_form, false);
            }
        },

        reloaded: function(wishlist) {
            nmgr.tiptip();
            $(document.body).trigger('nmgr_items_reloaded', wishlist);
        },

        pre_post_action: function(e) {
            e.preventDefault();

            var postdata = Object.assign({}, this.dataset);
            postdata['dataset'] = $('#nmgr-items')[0].dataset;
            postdata['nmgr_global'] = $(nmgr.items.container).attr('data-nmgr_global');

            if ('bulk' === this.dataset.nmgr_post_mode) {
                postdata['wishlist_item_ids'] = nmgr.items.get_selected_item_ids();
                delete(postdata.wishlist_item_id);
            }

            var args = {
                btn: this,
                block: nmgr.items.container,
                postdata: postdata
            };
            nmgr.post_action(args);
        },

        post_action_response: function(e, args) {
            if (args.response.replace_templates && args.response.replace_templates.hasOwnProperty('#nmgr-items')) {
                nmgr.items.reloaded(args.response.wishlist || null);
            }

            if ('delete_item' === args.postdata.nmgr_post_action) {
                $(document.body).trigger('wc_fragment_refresh');
                $(document.body).trigger('nmgr_removed_from_wishlist', args.response);
            }
        },

        save_item_quantity: function() {
            var args = {
                postdata: {
                    nmgr_post_action: 'save_item_quantity',
                    wishlist_item_id: this.dataset.item_id,
                    quantity: this.value
                }
            };
            nmgr.post_action(args);

            var container = this.closest('.nmgr_quantity');
            Nmeri_block(container);
            setTimeout(function() {
                Nmeri_unblock(container);
            }, 500);
        },

        init_add_items_form: function() {
            // reset the modal content form
            nmgr.items.reset_add_items_form('.nmgr-add-items-dialog');

            // initialize select2 on the select field in the modal
            $(document.body).trigger('nmgr-select2-init');
        },

        reset_add_items_form: function(element) {
            if ('undefined' !== typeof element) {
                $(element).find('tr.added-row').remove();
                $(element).find('form').trigger('reset').find(':input').each(function() {
                    if ($(this).is('.nmgr-product-search')) {
                        $(this).val('');
                    }
                });
            }
        },

        new_add_items_row: function() {
            if (!$(this).closest('tr').is(':last-child')) {
                return;
            }

            var item_table = $(this).closest('table'),
                item_table_body = item_table.find('tbody'),
                index = item_table_body.find('tr').length,
                row = item_table_body.data('row').replace(/\[0\]/g, '[' + index + ']');

            item_table_body.append('<tr class="added-row">' + row + '</tr>');

            $(document.body).trigger('nmgr-select2-init');
        },

        add_items: function() {
            var add_items_data = [];

            $(this).closest('.nmgr-add-items-dialog').find('tbody tr').each(function() {
                var item = {};
                $($(this).find(':input').serializeArray()).each(function(index, obj) {
                    item[obj.name] = obj.value;
                });

                add_items_data.push(item);
            });

            var dialogEl = document.getElementById('nmgr-add-items-dialog');
            NM.Modal.getInstance(dialogEl).hide();

            var data = {
                nmgr_post_action: 'add_item',
                wishlist_id: $(nmgr.items.container).attr('data-wishlist-id'),
                nmgr_global: $(nmgr.items.container).attr('data-nmgr_global'),
                nmgr_add_items_data: add_items_data
            };

            // Check if items have changed, if so pass them through so we can save them before adding a new item.
            if ('true' === $('#nmgr-items button.save-action').attr('data-reload')) {
                data.items = $('.nmgr-items-view :input[name]').serialize();
            }

            var args = {
                block: nmgr.items.container,
                postdata: data
            };

            nmgr.post_action(args);
        },

        add_single_product_to_cart: function(e) {
            var $thisbutton = $(this),
                item = {},
                $form = $thisbutton.closest('.nmgr-add-to-cart-form');
            /**
             *  If we are not adding to cart via ajax, we are supposed to return here
             *  but we will only return if adding the wishlist item to the cart is not restricted.
             *  Otherwise we will continue into the ajax code below in order to retrieve the
             *  restricted notice set during the add-to-cart validation
             */
            if (!$thisbutton.is('.nmgr_ajax_add_to_cart')) {
                return true;
            }

            e.preventDefault();

            $($form.serializeArray()).each(function(index, obj) {
                item[obj.name] = obj.value;
            });

            $thisbutton.removeClass('added').addClass('loading');

            nmgr.items.add_to_cart([item]);
        },

        add_multiple_products_to_cart: function(e) {
            e.preventDefault();
            var items = [];
            var clickedButton = this;

            clickedButton.classList.add('loading');

            $(nmgr.items.container).find('.nmgr-select:checked').closest('.item').each(function() {
                var item = {},
                    form = $(this).find('.nmgr-add-to-cart-form');

                if (form.length) {
                    $(form.serializeArray()).each(function(index, obj) {
                        item[obj.name] = obj.value;
                    });
                    items.push(item);
                    //show loading animation
                    if (clickedButton.classList.contains('nmgr_ajax_add_to_cart')) {
                        $(this).find('.nmgr_add_to_cart_button').removeClass('added').addClass('loading');
                    }
                }
            });

            if (items.length) {
                if (clickedButton.classList.contains('nmgr_ajax_add_to_cart')) {
                    nmgr.items.add_to_cart(items);
                } else {
                    var stringed = JSON.stringify(items);
                    var form = document.createElement('form');
                    var input = document.createElement('input');

                    form.method = 'post';
                    input.type = 'hidden';
                    input.name = 'nmgr_add_to_cart_item_data_string';
                    input.value = stringed;
                    form.appendChild(input);
                    document.body.appendChild(form);
                    form.submit();
                }
            } else {
                clickedButton.classList.remove('loading');
            }
        },
        /**
         * Add single or multiple products to cart
         *
         * @param {array} items (product data to add to cart e.g. product_id, quantity, variation_id)
         */
        add_to_cart: function(items) {
            var data = {
                action: 'nmgr_add_to_cart',
                nmgr_add_to_cart_item_data: items,
                nmgr_global: $(nmgr.items.container).attr('data-nmgr_global')
            };

            $.post(nmgr_global_params.ajax_url, data, function(response) {
                if (!response.fragments) {
                    return;
                }

                if (response.fragments.success && response.fragments.redirect_url) {
                    window.location = response.fragments.redirect_url;
                    return;
                }

                $('.woocommerce-error, .woocommerce-message, .woocommerce-info').remove();

                $(document.body).trigger('nmgr_add_to_cart_response', response.fragments);

                if (response.fragments.notices) {
                    response.fragments.notices.forEach(function(item) {
                        Nmeri_showToast(item);
                    });
                }

                if (response.fragments.success) {
                    $(document.body).trigger('added_to_cart', [response.fragments, response.cart_hash]); // update cart fragments
                    $(document.body).trigger('nmgr_added_to_cart', [response.fragments.items_data ? response.fragments.items_data : '']);
                }
            });
        },

        /**
         * Remove the loading animation on the add to cart button on the items table after the ajax action
         * has been completed.
         *
         * It is necessary to do this here rather than in the add_to_cart function because the add_to_cart function
         * is used in various ways (to add single, multiple products, and products from the wishlist cart). For this reason
         * it is less effiicient to code for removing the icon animation in the add_to_cart function.
         */
        remove_loading_animation: function(e, fragments) {
            if ('undefined' !== typeof fragments && fragments.items_data) {
                fragments.items_data.forEach(function(item) {
                    $('[data-wishlist_item_id="' + item.wishlist_item_id + '"].nmgr_add_to_cart_button')
                        .removeClass('loading');
                });
            }

            var btn = document.querySelector('.nmgr-bulk-add-to-cart');
            if (btn) {
                btn.classList.remove('loading');
            }
        },

        toggle_select_items: function() {
            var items = document.querySelectorAll('.nmgr-select');
            if (0 < items.length) {
                for (var i = 0; i < items.length; i++) {
                    if (true === this.checked) {
                        items[i].checked = true;
                    } else {
                        items[i].checked = false;
                    }
                }

                nmgr.items.maybe_show_bulk_actions();
            }
        },
        maybe_show_bulk_actions: function() {
            var actions = document.querySelector('.nmgr-items-bulk-actions');
            var selectedItems = document.querySelectorAll('#nmgr-items .nmgr-select:checked');
            if (selectedItems.length && !actions.classList.contains('show')) {
                actions.classList.add('show');
            } else if (!selectedItems.length && actions.classList.contains('show')) {
                var selectAllBtn = document.querySelector('#nmgr-items .nmgr-select-all');
                if (selectAllBtn) {
                    selectAllBtn.checked = false;
                }
                actions.classList.remove('show');
            }
        },
        /**
         * Get the checkboxes of the items checked on the items table
         * @returns {NodeList}
         */
        get_selected_items: function() {
            return document.querySelectorAll('#nmgr-items .nmgr-select:checked');
        },
        /**
         * Get the ids of the items selected on the items table
         * @returns {Array}
         */
        get_selected_item_ids: function() {
            var item_ids = [],
                selected = this.get_selected_items();
            if (selected.length) {
                selected.forEach(function(item) {
                    item_ids.push(item.value);
                });
            }
            return item_ids;
        },

        save_display_mode_to_session: function(e) {
            e.preventDefault();

            var args = {
                block: nmgr.items.container,
                postdata: {
                    nmgr_post_action: 'save_display_mode_to_session',
                    dataset: $('#nmgr-items')[0].dataset,
                    mode: this.dataset.mode,
                    nmgr_global: $(nmgr.items.container).attr('data-nmgr_global')
                }
            };
            nmgr.post_action(args);
        },

        toggle_purchase_refund_options: function() {
            var $form = $(this).closest('form');
            var val = parseInt(this.value);
            var originalVal = parseInt(this.dataset.qty);
            if (val < originalVal) {
                $form.find('.nmgr-options').fadeOut();
            } else if (val > originalVal) {
                $form.find('.nmgr-options').fadeIn();
            }
        },

        submit_purchase_refund_form: function(e) {
            e.preventDefault();

            var args = {
                block: this,
                postdata: Object.assign(Nmeri_getFormdata(this), {
                    dataset: $('#nmgr-items')[0].dataset,
                    nmgr_post_action: 'save_purchase_refund_form'
                })
            };

            nmgr.post_action(args);
        },

        paginate: function(e) {
            e.preventDefault();

            var args = {
                block: nmgr.items.container,
                postdata: {
                    nmgr_post_action: 'paginate_items',
                    dataset: $('#nmgr-items')[0].dataset,
                    nav_action: this.dataset.nav_action,
                    nmgr_global: $(nmgr.items.container).attr('data-nmgr_global')
                }
            };
            nmgr.post_action(args);
        }

    };
    nmgr.items.init();

    /*
     * js events related to wishlist messages template
     */

    var nmgr_messages = {

        init: function() {
            $(document.body)
                .on('click', '.nmgr-message .delete-wishlist-message', this.delete);
        },

        delete: function(e) {
            e.preventDefault();

            var postdata = Object.assign({}, this.dataset);
            postdata['nmgr_global'] = $(this).closest('.nmgr-account-template').attr('data-nmgr_global');

            var args = {
                btn: this,
                block: $(this).closest('.nmgr-message'),
                postdata: postdata
            };

            nmgr.post_action(args);
        }
    };

    nmgr_messages.init();


    var ship_to_account_address = {

        init: function() {
            $(document.body)
                .on('click', '.nmgr-copy-shipping-address', this.load_address);
        },

        load_address: function(e) {
            e.preventDefault();

            if ($('.nmgr-shipping-fields').length > 0) {
                var address = JSON.parse(e.target.dataset.nmgrAddress);

                for (var key in address) {
                    $('.nmgr-shipping-fields :input#shipping_' + key).val(address[key]).trigger('change');
                }
            }
        }

    };

    ship_to_account_address.init();

    $(document.body)
        .on('click', '.nmgr-copy', copy_link);

    function copy_link() {
        var $temp = $("<input>"),
            $self = $(this);
        $("body").append($temp);
        $temp.val($self.text()).select();
        document.execCommand("copy");
        $temp.remove();
        $self.addClass('nmgr-animation-scaleshrink');
        setTimeout(function() {
            $self.removeClass('nmgr-animation-scaleshrink');
        }, 700);
    }




});