<?php

/**
 * Sync
 */
defined( 'ABSPATH' ) || exit;

class NMGR_Account {

	/**
	 * The current wishlist being viewed
	 * @var NMGR_Wishlist
	 */
	protected $wishlist;

	/**
	 * The current section being processed
	 * @var string
	 */
	protected $section_key;
	protected $view;
	protected $type = 'gift-registry';
	protected $sections_data = [];
	protected $args = [];

	/**
	 * Inititialize the view for the wishlist items
	 *
	 * @param int|NMGR_Wishlist $wishlist Wishlist id or object
	 * @param array $args Arguments used to compose the view
	 */
	public function __construct( $wishlist = false ) {
		$this->set_wishlist( $wishlist );
		if ( $this->get_wishlist() ) {
			$this->set_type( $this->get_wishlist()->get_type() );
		}
	}

	public function set_section( $section_key ) {
		$this->section_key = $section_key;
		return $this;
	}

	/**
	 * @param int|NMGR_Wishlist $wishlist Wishlist id or object
	 */
	public function set_wishlist( $wishlist ) {
		if ( $wishlist ) {
			$this->wishlist = is_a( $wishlist, 'NMGR_Wishlist' ) ? $wishlist : nmgr_get_wishlist( $wishlist, true );
		}
	}

	public function set_type( $type ) {
		$this->type = $type;
	}

	/**
	 * @return NMGR_Wishlist
	 */
	public function get_wishlist() {
		return $this->wishlist;
	}

	public function get_type() {
		return $this->type;
	}

	public function is_gift_registry() {
		return 'gift-registry' === $this->get_type();
	}

	public function set_view( $view ) {
		$this->view = $view;
	}

	public function get_view() {
		return $this->view;
	}

	public function set_args( $args ) {
		$this->args = $args;
	}

	public function get_args() {
		return $this->args;
	}

	protected function get_sections() {
		$sections = [
			'profile' => [
				'title' => nmgr()->is_pro ?
				__( 'Profile', 'nm-gift-registry' ) :
				__( 'Profile', 'nm-gift-registry-lite' ),
				'priority' => 20,
				'content' => [ $this, 'get_profile_section' ],
				'show_for_user_only' => true,
			],
			'items' => [
				'title' => nmgr()->is_pro ?
				__( 'Items', 'nm-gift-registry' ) :
				__( 'Items', 'nm-gift-registry-lite' ),
				'priority' => 30,
				'content' => [ $this, 'get_items_section' ],
			],
			'shipping' => [
				'title' => nmgr()->is_pro ?
				__( 'Shipping', 'nm-gift-registry' ) :
				__( 'Shipping', 'nm-gift-registry-lite' ),
				'priority' => 50,
				'content' => [ $this, 'get_shipping_section' ],
				'show' => $this->is_gift_registry() && nmgr_get_option( 'enable_shipping' ),
				'show_for_user_only' => true,
				'replace_on_load' => [
					'items',
				],
			],
		];

		return $sections;
	}

	public function get_sections_data() {
		if ( !$this->sections_data ) {
			$data = apply_filters( 'nmgr_account_sections', $this->get_sections(), $this );
			uasort( $data, 'nmgr_sort_by_priority' );
			$this->sections_data = nmgr_get_elements_to_show( $data );
		}
		return $this->sections_data;
	}

	/**
	 * Get the data for a particular section
	 * @param string $section_key
	 * @return array
	 */
	public function get_section_data( $section_key = null ) {
		if ( !is_null( $section_key ) ) {
			_deprecated_argument( __METHOD__, '4.2.0', 'Use NMGR_Account::set section' );
			$this->section_key = $section_key;
		}
		return $this->get_sections_data()[ $this->section_key ] ?? [];
	}

	public function get_section_title( $section_key = null ) {
		if ( !is_null( $section_key ) ) {
			_deprecated_argument( __METHOD__, '4.2.0', 'Use NMGR_Account::set section' );
			$this->section_key = $section_key;
		}
		$data = $this->get_section_data();
		return $data ? apply_filters( "nmgr_{$this->section_key}_template_title", $data[ 'title' ] ?? '' ) : '';
	}

	protected function get_default_section_vars() {
		$vars = [
			'id' => $this->get_wishlist() ? $this->get_wishlist()->get_id() : 0,
			'wishlist' => $this->get_wishlist(),
			'title' => $this->get_section_title(),
			'type' => $this->get_type(),
			'section_name' => $this->section_key,
			'attributes' => $this->get_section_container_attributes(),
			'template_group' => 'account'
		];
		return $vars;
	}

	protected function get_template( $vars ) {
		return apply_filters(
			"nmgr_{$this->section_key}_template",
			nmgr_get_template( "account/{$this->section_key}.php", $vars ),
			$vars
		);
	}

	protected function get_profile_section( $vars ) {
		$vars[ 'wishlist' ] = !empty( $vars[ 'wishlist' ] ) ? $vars[ 'wishlist' ] : nmgr()->wishlist();
		$id = $vars[ 'wishlist' ]->get_id();
		$vars[ 'form' ] = new NMGR_Form( $id );
		if ( !$id && !empty( $vars[ 'type' ] ) ) {
			$vars[ 'form' ]->set_type( $vars[ 'type' ] );
		}
		return $this->get_template( $vars );
	}

	protected function get_items_section( $vars ) {
		$wishlist = $vars[ 'wishlist' ];
		$view = nmgr()->items_view( $wishlist );
		$class = array_filter( array(
			$wishlist && $wishlist->is_fulfilled() ? 'wishlist-fulfilled' : '',
			is_nmgr_wishlist() ? 'single' : '',
			is_nmgr_admin() ? 'admin' : '',
			is_nmgr_account_section() ? 'account' : '',
			$view->get_display()[ 'mode' ],
			) );

		if ( $wishlist ) {
			$total_items = $wishlist->get_items_count();
			$vars = $this->process_pagination( $vars, $total_items );
		}

		$vars2 = nmgr()->merge_args( $vars, array(
			'items' => $wishlist ? $wishlist->read_items( $vars[ 'pagination' ] ?? [] ) : '',
			'attributes' => [
				'class' => $class,
			],
			'view' => $view,
			) );

		return $this->get_template( $vars2 );
	}

	protected function get_shipping_section( $vars ) {
		if ( nmgr_get_type_option( $this->get_type(), 'enable_shipping' ) ) {
			$vars2 = array_merge( $vars, array(
				'form' => $vars[ 'wishlist' ] ? new \NMGR_Form( $vars[ 'wishlist' ]->get_id() ) : '',
				'customer' => $vars[ 'wishlist' ] ? new \WC_Customer( $vars[ 'wishlist' ]->get_user_id() ) : '',
				) );

			return $this->get_template( $vars2 );
		}
	}

	protected function get_section_content() {
		$wishlist_id = $this->get_wishlist() ? $this->get_wishlist()->get_id() : false;

		if ( !is_admin() && ('profile' !== $this->section_key) && !$wishlist_id ) {
			return;
		}

		$data = $this->get_section_data();

		if ( empty( $data ) ) {
			return;
		}

		$show_for_user_only = !empty( $data[ 'show_for_user_only' ] );

		if ( $show_for_user_only && (!is_nmgr_user( $this->get_type() ) ||
			($wishlist_id && !is_admin() && !nmgr_user_has_wishlist( $wishlist_id ))) ) {
			return;
		}

		$vars = $this->get_default_section_vars();

		$content = $data[ 'content' ] ?? '';

		if ( is_callable( $content ) ) {
			$content = call_user_func( $content, $vars, $this );
		}

		return $content;
	}

	/**
	 * Get an account section
	 * @deprecated since version 4.2.0
	 * @param string $section_key The section key, e.g. profile
	 * @return string Section template content
	 */
	public function get_section( $section_key ) {
		_deprecated_function( __METHOD__, '4.2.0', 'NMGR_Account::get_section_template' );
		$this->section_key = $section_key;
		return $this->get_section_content();
	}

	public function get_section_template() {
		return $this->get_section_content();
	}

	public function get_sections_by_ids( $section_keys ) {
		$keys = ( array ) $section_keys;
		$sections = [];
		foreach ( $keys as $section_key ) {
			$this->set_section( $section_key );
			$sections[ '#nmgr-' . $section_key ] = $this->get_section_template();
		}
		return $sections;
	}

	public function get_sections_to_replace() {
		$section_keys = $this->get_other_sections_to_replace();
		$section_keys[] = $this->section_key;
		return $this->get_sections_by_ids( $section_keys );
	}

	public function get_other_sections_to_replace() {
		return $this->get_section_data()[ 'replace_on_load' ] ?? [];
	}

	public function get_section_container_attributes() {
		if ( isset( $GLOBALS[ 'nmgr' ] ) ) {
			$nmgr_global = $GLOBALS[ 'nmgr' ];

			if ( isset( $nmgr_global->wishlist ) ) {
				unset( $nmgr_global->wishlist );
			}
		}

		$section_data = $this->get_section_data();

		$id = $section_data[ 'id' ] ?? ('nmgr-' . $this->section_key);

		$attributes = array(
			'id' => $id,
			'class' => [
				'woocommerce',
				'nmgr-account-template',
				'nmgr-ajax',
			],
			'data-nonce' => wp_create_nonce( 'nmgr_manage_wishlist' ),
			// @deprecated 'data-wishlist-id'
			'data-wishlist-id' => !empty( $this->get_wishlist() ) ? absint( $this->get_wishlist()->get_id() ) : 0,
			'data-wishlist_id' => !empty( $this->get_wishlist() ) ? absint( $this->get_wishlist()->get_id() ) : 0,
			'data-nmgr_global' => isset( $nmgr_global ) ? htmlspecialchars( json_encode( $nmgr_global ) ) : '',
		);

		return apply_filters( 'nmgr_account_section_attributes', $attributes, $this->section_key );
	}

	protected function get_items_per_page() {
		return apply_filters( 'nmgr_items_per_page', 12, $this->section_key );
	}

	protected function get_page() {
		return !empty( $this->get_args()[ 'page' ] ) ? $this->get_args()[ 'page' ] : 1;
	}

	protected function process_pagination( $vars, $total_items ) {
		return $vars;
	}

}
