<?php

/**
 * Sync
 */
defined( 'ABSPATH' ) || exit;

class NMGR_Ajax {

	public function run() {
		$this->ajax_events();
		add_action( 'admin_init', array( $this, 'set_global_variable' ), -1 );
		add_action( 'wp_handle_upload_prefilter', array( $this, 'check_file_type' ) );
		add_filter( 'nmgr_get_wishlist_data', array( $this, 'get_wishlist_data' ), 10, 2 );
	}

	/**
	 * Make the plugin global variable available for all ajax requests
	 *
	 * json_decode is used here for situations where JSON.stringify is used in
	 * the ajax script to post the global nmgr variable  (like when using the formData object)
	 */
	public function set_global_variable() {
		if ( wp_doing_ajax() && isset( $_REQUEST[ 'nmgr_global' ] ) ) { // phpcs:ignore WordPress.Security.NonceVerification
			$nmgr = $_REQUEST[ 'nmgr_global' ]; // phpcs:ignore

			if ( is_string( $nmgr ) ) {
				$GLOBALS[ 'nmgr' ] = json_decode( stripslashes( $nmgr ) ); // phpcs:ignore
			} else {
				$GLOBALS[ 'nmgr' ] = ( object ) wc_clean( $nmgr );
			}
		}
	}

	public function mimes() {
		return array(
			'jpg|jpeg|jpe' => 'image/jpeg',
			'gif' => 'image/gif',
			'png' => 'image/png',
			'bmp' => 'image/bmp',
		);
	}

	/**
	 * Allow only image files to be accepted as uploads
	 *
	 * Initial validation for uploaded images is done in the javascript file before the ajax request fires.
	 * Subsequent validation is done in the 'wp_ajax_nmgr_save_image' function using the overrides attribute
	 * of the 'media_handle_upload' function
	 * This validation is the third and is done using the 'wp_handle_uploads_prefilter' in the same 'media_handle_upload' function
	 *
	 * The accepted mime times for the media_handle_upload function is already overridden to accept only image mimes.
	 * This allows only images to be uploaded to the uploads folder.
	 * However in a situation where 'ALLOW_UNFILTERED_UPLOADS' is defined as true, any file type would still be accepted.
	 * This is where this function comes in. It specifically prevents non-image files from being uploaded
	 * regardless of the status of ALLOW_UNFILTERED_UPLOADS
	 * It is simply a modified copy of the same 'wp_check_filetype_and_ext' code snippet found in the 'media_handle_upload' function
	 *
	 * This function is used instead of that snippet because the snippet uses current_user_can( 'unfiltered_upload' )
	 * to allow unfiltered uploads if defined
	 *
	 * @param type $file Data for a single file in $_FILES
	 * @return array $file
	 */
	public function check_file_type( $file ) {
		if ( isset( $_FILES[ 'nmgr-file' ] ) && $file === $_FILES[ 'nmgr-file' ] ) {
			$wp_filetype = wp_check_filetype_and_ext( $file[ 'tmp_name' ], $file[ 'name' ], $this->mimes() );

			if ( empty( $wp_filetype[ 'type' ] ) || empty( $wp_filetype[ 'ext' ] ) ) {
				$file[ 'error' ] = __( 'Please upload only image files.', 'nm-gift-registry' );
			}
		}
		return $file;
	}

	/**
	 * Add extra properties to wishlist data for ajax requests
	 *
	 * @param array $data Wishlist data
	 * @param NMGR_Wishlist $wishlist
	 * @return array
	 */
	public function get_wishlist_data( $data, $wishlist ) {
		if ( !wp_doing_ajax() ) {
			return $data;
		}

		$data = array_merge( $data, array(
			'has_shipping_address' => $wishlist->has_shipping_address(),
			'needs_shipping_address' => $wishlist->needs_shipping_address(),
			) );

		return $data;
	}

	public function ajax_events() {
		$ajax_events = array(
			'add_to_wishlist',
			'json_search_products',
			'json_search_users',
			'add_to_cart',
			'get_cart_fragments',
			'post_action',
		);

		foreach ( $ajax_events as $event ) {
			if ( method_exists( $this, $event ) ) {
				add_action( 'wp_ajax_nmgr_' . $event, array( $this, $event ) );
				add_action( 'wp_ajax_nopriv_nmgr_' . $event, array( $this, $event ) );
			}
		}

		/**
		 * Remove woocommerce's add to cart action when running our own so that the add to cart action
		 * runs only once for us
		 */
		if ( wp_doing_ajax() && isset( $_REQUEST[ 'action' ] ) && ('nmgr_add_to_cart' === $_REQUEST[ 'action' ]) ) { // phpcs:ignore WordPress.Security.NonceVerification
			remove_action( 'wp_loaded', array( 'WC_Form_Handler', 'add_to_cart_action' ), 20 );
		}
	}

	public function add_to_wishlist() {
		check_ajax_referer( 'nmgr-frontend' );

		$result = nmgr()->wordpress()->add_to_wishlist_action();

		// We are expecting notices from the add to wishlist action, so get them
		$error = wc_notice_count( 'error' ) ? true : false;
		$success = wc_notice_count( 'success' ) ? true : false;

		// This clears the notices after printing so we get it last
		$notice = wc_print_notices( true );

		wp_send_json( array(
			'notice' => $notice,
			'error' => $error,
			'success' => $success,
			'result' => $result,
		) );
	}

	public function add_to_cart() {
		if ( !isset( $_POST[ 'nmgr_add_to_cart_item_data' ] ) || empty( $_POST[ 'nmgr_add_to_cart_item_data' ] ) ) {
			wp_die( -1 );
		}

		$data = $_POST[ 'nmgr_add_to_cart_item_data' ];

		$items_data = nmgr_add_to_cart( $data );

		$success = wc_notice_count( 'success' ) ? true : false;
		$redirect_url = $success ? apply_filters( 'woocommerce_add_to_cart_redirect', false, false ) : false;
		if ( $success && !$redirect_url && 'yes' === get_option( 'woocommerce_cart_redirect_after_add' ) ) {
			$redirect_url = wc_get_cart_url();
		}

		// We are expecting notices from the add to cart action, so get them
		$custom_data = array(
			'success' => $success,
			'notices' => $success && $redirect_url ? '' : nmgr_get_wc_toast_notices(),
			'items_data' => $items_data,
			'redirect_url' => esc_url( apply_filters( 'nmgr_ajax_add_to_cart_redirect_url', $redirect_url, $success, $items_data ) ),
		);

		/**
		 * Due to the way woocommerce adds products to cart via ajax, the refreshed cart fragments must be returned to
		 * the script directly as a result of the add to cart action in order to properly refresh the cart fragments.
		 * This poses a problem if we want to return additional results from the add-to-cart action as it is in this case.
		 * The solution is to add the additional results to the cart fragments array using the filter 'woocommerce_add_to_cart_fragments'
		 * so that our results can be returned to the script with the cart fragments array.
		 *
		 * This hack is necessary because Calling 'WC_AJAX::get_refreshed_fragments()' at the end of the add to cart action
		 * kills the script (using wp_die()) and so we have no other way of returning our custom results to the script except by
		 * hooking into the fragments array before the script is killed.
		 */
		if ( class_exists( 'wc_ajax' ) && method_exists( 'wc_ajax', 'get_refreshed_fragments' ) ) {
			add_filter( 'woocommerce_add_to_cart_fragments', function( $fragments ) use ( $custom_data ) {
				return array_merge( $fragments, $custom_data );
			} );

			\WC_AJAX::get_refreshed_fragments();
		}
	}

	public function get_cart_fragments() {
		if ( class_exists( 'wc_ajax' ) && method_exists( 'wc_ajax', 'get_refreshed_fragments' ) ) {
			\WC_AJAX::get_refreshed_fragments();
		}
	}

	public function send_ajax_response( $data = array() ) {
		_deprecated_function( __METHOD__, '4.0.0' );

		if ( !wp_doing_ajax() ) {
			return;
		}

		$data = ( array ) $data;

		$error = isset( $data[ 'error' ] ) ? $data[ 'error' ] : (wc_notice_count( 'error' ) ? true : false);
		$success = isset( $data[ 'success' ] ) ? $data[ 'success' ] : (wc_notice_count( 'success' ) ? true : false);

		if ( isset( $data[ 'error' ] ) ) {
			unset( $data[ 'error' ] );
		}

		if ( isset( $data[ 'success' ] ) ) {
			unset( $data[ 'success' ] );
		}

		// This clears the notices after printing so we get it last
		$data[ 'notice' ] = isset( $data[ 'notice' ] ) ? $data[ 'notice' ] : wc_print_notices( true );

		wp_send_json( array(
			'error' => $error,
			'success' => $success,
			'data' => $data
		) );
	}

	/**
	 * Save a wishlist profile
	 *
	 * This function is used for both admin and frontend (ajax requests)
	 */
	protected function save_form() {
		$form_data = [];
		parse_str( $_POST[ 'data' ] ?? '', $form_data );
		$wishlist_id = $form_data[ 'nmgr_wishlist_id' ] ?? 0;

		if ( $wishlist_id ) {
			nmgr_check_wishlist_permission( $wishlist_id );
		}

		$form = new \NMGR_Form( $wishlist_id );
		$form->set_data( $form_data )->validate();

		if ( $form->has_errors() ) {
			wp_send_json( [
				'error_data' => $form->get_fields_error_messages(),
				'toast_notice' => nmgr_get_toast_notice( nmgr_get_error_text(), 'error' ),
			] );
		}

		$id = $form->save();
		$wishlist = nmgr_get_wishlist( $id );

		$response_data = array(
			'action' => $form_data[ 'nmgr_save' ] ?? null,
			'success' => true,
			'wishlist_type' => $wishlist->get_type(),
			'wishlist' => $wishlist->get_data(),
			'created' => ( bool ) !$wishlist_id
		);

		$add_to_wishlist_formdata = $form_data[ 'add_to_wishlist_formdata' ] ?? null;

		if ( $add_to_wishlist_formdata ) {
			$data = [];
			parse_str( $add_to_wishlist_formdata, $data );

			if ( $wishlist->needs_shipping_address() ) {
				$response_data[ 'show_template' ] = $this->get_set_shipping_address_dialog(
					$id, $add_to_wishlist_formdata
				);
			} elseif ( 'advanced' === ($data[ 'nmgr_atw_mode' ]) ) {
				/**
				 * 'nmgr_wid' key is only present if we are creating a new wishlist while saving
				 * the shipping address so we use it to determine whether to show the add to wishlist
				 * dialog or to simply submit the add to wishlist formdata
				 */
				if ( empty( $data[ 'nmgr_wid' ] ) ) {
					$response_data[ 'show_template' ] = $this->get_add_to_wishlist_dialog( $data );
				} else {
					$response_data[ 'atw_action' ] = $data;
				}
			} else {
				/**
				 * Here we assume the add to wishlist mode is simple, so send the product id to the js script
				 * as a flag for us to add to the wishlist via the simple action
				 */
				$response_data[ 'atw_pid_click' ] = ($data[ 'nmgr_pid' ] ?? 0);
			}
		} elseif ( $response_data[ 'created' ] ) {
			if ( $wishlist->needs_shipping_address() ) {
				$response_data[ 'redirect' ] = add_query_arg(
					'nmgr_redirect',
					1,
					trailingslashit( $wishlist->get_permalink() ) . 'shipping'
				);
			} else {
				$response_data[ 'redirect' ] = $wishlist->get_permalink();
			}
		} elseif ( !empty( $form_data[ '_wp_http_referer' ] ) &&
			false !== strpos( $form_data[ '_wp_http_referer' ], 'nmgr_redirect' ) ) {
			$response_data[ 'redirect' ] = $wishlist->get_permalink();
		}

		if ( !$add_to_wishlist_formdata ) {
			$redirect = $response_data[ 'redirect' ] ?? null;
			$response_data[ 'redirect' ] = apply_filters( 'nmgr_redirect_after_save', $redirect, $response_data );
		}

		if ( $response_data[ 'success' ] && empty( $response_data[ 'redirect' ] ) ) {
			$response_data[ 'toast_notice' ] = nmgr_get_toast_notice( nmgr_get_success_text() );
		}

		wp_send_json( $response_data );
	}

	protected function add_item() {
		$wishlist_id = absint( wp_unslash( $_POST[ 'wishlist_id' ] ) );
		nmgr_check_wishlist_permission( $wishlist_id );

		$wishlist = nmgr_get_wishlist( $wishlist_id, true );
		$items_to_add = isset( $_POST[ 'nmgr_add_items_data' ] ) ?
			array_filter( wc_clean( wp_unslash( ( array ) $_POST[ 'nmgr_add_items_data' ] ) ) ) :
			array();

		// Add items to wishlist.
		foreach ( $items_to_add as $item ) {
			if ( !isset( $item[ 'product_id' ] ) || empty( $item[ 'product_id' ] ) ) {
				continue;
			}
			$product_id = absint( $item[ 'product_id' ] );
			$qty = wc_stock_amount( isset( $item[ 'product_qty' ] ) && $item[ 'product_qty' ] ? $item[ 'product_qty' ] : 1 );
			$product = wc_get_product( $product_id );

			if ( $product ) {
				$favourite = isset( $item[ 'product_fav' ] ) ? absint( $item[ 'product_fav' ] ) : 0;
				$wishlist->add_item( $product, $qty, $favourite );
			}
		}

		$acc = nmgr()->account( $wishlist );
		wp_send_json( array(
			'success' => true,
			'toast_notice' => nmgr_get_toast_notice( nmgr_get_success_text() ),
			'wishlist' => $wishlist->get_data(),
			'replace_templates' => $acc->get_sections_by_ids( 'items' ),
		) );
	}

	protected function save_item_quantity() {
		$posted_ids = nmgr_get_posted_wishlist_and_item_ids();
		nmgr_check_wishlist_permission( $posted_ids[ 'wishlist_id' ] ?? false  );

		$item = nmgr_get_wishlist_item( $posted_ids[ 'wishlist_item_id' ] ?? 0 );

		if ( $item && isset( $_POST[ 'quantity' ] ) ) {
			$item->set_quantity( ( int ) $_POST[ 'quantity' ] );
			$item->save();
		}
	}

	/**
	 * Search for products and echo json.
	 */
	public function json_search_products() {
		check_ajax_referer( 'nmgr-search-products', 'security' );

		$term = '';
		$include_variations = ( bool ) apply_filters( 'nmgr_search_products_include_variations', true );

		if ( empty( $term ) && isset( $_GET[ 'term' ] ) ) {
			$term = ( string ) wc_clean( wp_unslash( $_GET[ 'term' ] ) ); // phpcs:ignore WordPress.Security.ValidatedSanitizedInput.InputNotSanitized
		}

		if ( empty( $term ) ) {
			wp_die();
		}

		if ( !empty( $_GET[ 'limit' ] ) ) {
			$limit = absint( $_GET[ 'limit' ] );
		} else {
			$limit = absint( apply_filters( 'nmgr_json_search_limit', 30 ) );
		}

		$include_ids = !empty( $_GET[ 'include' ] ) ? array_map( 'absint', ( array ) wp_unslash( $_GET[ 'include' ] ) ) : array();
		$exclude_ids = !empty( $_GET[ 'exclude' ] ) ? array_map( 'absint', ( array ) wp_unslash( $_GET[ 'exclude' ] ) ) : array();

		$exclude_types = array();
		if ( !empty( $_GET[ 'exclude_type' ] ) ) {
			// Support both comma-delimited and array format inputs.
			$exclude_types = wp_unslash( $_GET[ 'exclude_type' ] );
			if ( !is_array( $exclude_types ) ) {
				$exclude_types = explode( ',', $exclude_types );
			}

			// Sanitize the excluded types against valid product types.
			foreach ( $exclude_types as &$exclude_type ) {
				$exclude_type = strtolower( trim( $exclude_type ) );
			}
			$exclude_types = array_intersect(
				array_merge( array( 'variation' ), array_keys( wc_get_product_types() ) ),
				$exclude_types
			);
		}

		$data_store = \WC_Data_Store::load( 'product' );
		$ids = $data_store->search_products( $term, '', $include_variations, false, $limit, $include_ids, $exclude_ids );

		$product_objects = array_filter( array_map( 'wc_get_product', $ids ), 'wc_products_array_filter_readable' );
		$products = array();

		foreach ( $product_objects as $product_object ) {
			if ( in_array( $product_object->get_type(), $exclude_types, true ) ) {
				continue;
			}

			$formatted_name = $product_object->get_formatted_name();
			$managing_stock = $product_object->managing_stock();

			if ( $managing_stock && !empty( $_GET[ 'display_stock' ] ) ) {
				$stock_amount = $product_object->get_stock_quantity();
				$formatted_name .= ' &ndash; ' . sprintf(
						/* translators: %d: stock quantity */
						nmgr()->is_pro ? __( 'Stock: %d', 'nm-gift-registry' ) : __( 'Stock: %d', 'nm-gift-registry-lite' ),
						wc_format_stock_quantity_for_display( $stock_amount, $product_object )
				);
			}

			$products[ $product_object->get_id() ] = rawurldecode( $formatted_name );
		}

		wp_send_json( $products );
	}

	/**
	 * Search for users
	 */
	public function json_search_users() {
		ob_start();

		check_ajax_referer( 'nmgr-search-users', 'security' );

		if ( !current_user_can( 'edit_nm_gift_registries' ) ) {
			wp_die( -1 );
		}

		$term = isset( $_GET[ 'term' ] ) ? ( string ) wc_clean( wp_unslash( $_GET[ 'term' ] ) ) : ''; // phpcs:ignore WordPress.Security.ValidatedSanitizedInput.InputNotSanitized
		$limit = 0;

		if ( empty( $term ) ) {
			wp_die();
		}

		$ids = array();
		// Search by ID.
		if ( is_numeric( $term ) ) {
			$customer = new \WC_Customer( ( int ) $term );

			// Customer exists.
			if ( 0 !== $customer->get_id() ) {
				$ids = array( $customer->get_id() );
			}
		}

		/**
		 * If the customer doesn't exist by searching by id, search for numeric username,
		 * this prevents performance issues with ID lookups.
		 */
		if ( empty( $ids ) ) {
			$data_store = \WC_Data_Store::load( 'customer' );

			/**
			 * If search is smaller than 3 characters, limit result set to avoid
			 * too many rows being returned.
			 */
			if ( 3 > strlen( $term ) ) {
				$limit = 20;
			}
			$ids = $data_store->search_customers( $term, $limit );
		}

		$found_customers = array();

		if ( !empty( $_GET[ 'exclude' ] ) ) {
			$ids = array_diff( $ids, array_map( 'absint', ( array ) wp_unslash( $_GET[ 'exclude' ] ) ) );
		}

		foreach ( $ids as $id ) {
			$customer = new \WC_Customer( $id );
			$name = $customer->get_first_name() . ' ' . $customer->get_last_name();

			$found_customers[ $id ] = [
				'text' => sprintf(
					/* translators: $1: customer name, $2: customer email */
					nmgr()->is_pro ? esc_html__( '%1$s (%2$s)', 'nm-gift-registry' ) : esc_html__( '%1$s (%2$s)', 'nm-gift-registry-lite' ),
					trim( $name ) ? $name : $customer->get_display_name(),
					$customer->get_email()
				),
				'shipping_address' => json_encode( $customer->get_shipping() ),
			];
		}

		wp_send_json( $found_customers );
	}

	protected function show_create_new_wishlist_dialog() {
		add_filter( 'nmgr_fields', [ $this, 'add_add_to_wishlist_formdata_to_profile_fields' ] );

		$formdata = [];
		parse_str( $_POST[ 'add_to_wishlist_formdata' ], $formdata );

		$type = $formdata[ 'nmgr_type' ];

		$css = '<style>.dialog.show.nmgr .nmgr-account-template .save-action { display:none;}</style>';
		$modal = nmgr_get_modal();
		$modal->set_title( sprintf(
				/* translators: %s: wishlist type title */
				nmgr()->is_pro ? __( 'Create new %s', 'nm-gift-registry' ) : __( 'Create new %s', 'nm-gift-registry-lite' ),
				nmgr_get_type_title( '', false, $type )
		) );
		$modal->set_id( 'nmgr-create-new-wishlist-dialog' );
		$modal->set_content( $css . nmgr_get_profile_template( [ 'type' => $type ] ) );
		$modal->make_large()->make_vertically_centered();
		$modal->set_footer(
			$modal->get_save_button( [
				'attributes' => [
					'type' => 'submit',
					'class' => 'nmgr_save_profile',
					'form' => 'nmgr-profile-form'
				]
			] )
		);

		wp_send_json( [ 'show_template' => $modal->get() ] );
	}

	public function add_add_to_wishlist_formdata_to_profile_fields( $fields ) {
		if ( !empty( $_POST[ 'add_to_wishlist_formdata' ] ) ) {
			$add_to_wishlist_formdata = sanitize_text_field( wp_unslash( $_POST[ 'add_to_wishlist_formdata' ] ) );
			$fields[ 'add_to_wishlist_formdata' ] = $this->get_add_to_wishlist_formdata_field( 'profile', $add_to_wishlist_formdata );
		}
		return $fields;
	}

	protected function get_add_to_wishlist_formdata_field( $fieldset, $add_to_wishlist_formdata ) {
		return array(
			'type' => 'nmgr-hidden',
			'fieldset' => $fieldset,
			'prefix' => false,
			'value' => $add_to_wishlist_formdata,
		);
	}

	protected function auto_create_wishlist() {
		$add_to_wishlist_formdata = sanitize_text_field( wp_unslash( $_POST[ 'add_to_wishlist_formdata' ] ?? '' ) );
		$data = [];
		parse_str( $add_to_wishlist_formdata, $data );

		$type = $data[ 'nmgr_type' ];

		$wishlist_id = wp_insert_post(
			array(
				'post_title' => 'Auto Draft',
				'post_type' => 'nm_gift_registry',
				'post_status' => 'auto-draft',
			)
		);

		$title = str_replace(
			array( '{wishlist_type_title}', '{site_title}', '{wishlist_id}' ),
			array( nmgr_get_type_title( 'c', false, $type ), get_bloginfo( 'name' ), $wishlist_id ),
			nmgr_get_type_option( $type, 'default_wishlist_title' )
		);

		$wishlist = nmgr_get_wishlist( $wishlist_id );
		$wishlist->set_title( $title );
		$wishlist->set_status( 'publish' );
		$wishlist->set_user_id( esc_html( nmgr_get_current_user_id() ) );
		$wishlist->set_type( $type );

		if ( 'gift-registry' === $type && is_user_logged_in() ) {
			$user = wp_get_current_user();
			$wishlist->set_email( $user->user_email );
			$wishlist->set_first_name( $user->first_name );
			$wishlist->set_last_name( $user->last_name );
		}

		$id = $wishlist->save();

		$response = array(
			'success' => true,
			'wishlist_type' => $type,
			'wishlist' => $wishlist->get_data()
		);

		if ( $id ) {
			if ( $wishlist->needs_shipping_address() ) {
				wp_send_json( array_merge( $response, [
					'show_template' => $this->get_set_shipping_address_dialog( $id, $add_to_wishlist_formdata )
				] ) );
			} else {
				if ( 'advanced' === ($data[ 'nmgr_atw_mode' ] ?? '') ) {
					wp_send_json( array_merge( $response, [
						'show_template' => $this->get_add_to_wishlist_dialog( $data )
					] ) );
				} else {
					/**
					 * Here we assume the add to wishlist mode is simple, so send the product id to the js script
					 * as a flag for us to add to the wishlist via the simple action
					 */
					wp_send_json(
						array_merge( $response, [
						'atw_pid_click' => ($data[ 'nmgr_pid' ] ?? 0)
						] )
					);
				}
			}
		}
	}

	protected function get_add_to_wishlist_dialog( $formdata ) {
		$args = $this->get_add_to_wishlist_dialog_args( $formdata );
		$form = $this->get_add_to_wishlist_dialog_form( $formdata );

		if ( $form ) {
			$type = $formdata[ 'nmgr_type' ];
			$type_title = nmgr_get_type_title( '', false, $type );
			$modal = nmgr_get_modal();
			$modal->set_id( 'nmgr-add-to-wishlist-dialog' );
			$modal->set_content( $form );
			$modal->make_large()->make_vertically_centered();

			if ( !empty( $args[ 'wishlists' ] ) ) {
				$modal->set_footer(
					$modal->get_save_button( [
						'text' => str_replace(
							'{wishlist_type_title}',
							$type_title,
							nmgr_get_type_option( $type, 'add_to_wishlist_button_text' )
						),
						'attributes' => [
							'data-action' => 'add_to_wishlist',
							'class' => 'nmgr_add_to_wishlist',
							'data-select_product_text' => sprintf(
								/* translators: %s: wishlist type title */
								nmgr()->is_pro ? esc_attr__( 'Please select an item to add to your %s.', 'nm-gift-registry' ) : esc_attr__( 'Please select an item to add to your %s.', 'nm-gift-registry-lite' ),
								$type_title
							),
							'data-select_quantity_text' => sprintf(
								/* translators: %s: wishlist type title */
								nmgr()->is_pro ? esc_attr__( 'Please choose the quantity of items you wish to add to your %s.', 'nm-gift-registry' ) : esc_attr__( 'Please choose the quantity of items you wish to add to your %s.', 'nm-gift-registry-lite' ),
								$type_title
							),
						],
					] )
				);
			}

			return $modal->get();
		}
	}

	protected function get_add_to_wishlist_dialog_form( $formdata ) {
		$args = $this->get_add_to_wishlist_dialog_args( $formdata );
		if ( !empty( $args ) ) {
			return nmgr_get_template( 'add-to-wishlist/dialog-content.php',
				apply_filters( 'nmgr_add_to_wishlist_dialog_content_args', $args )
			);
		}
	}

	protected function get_add_to_wishlist_dialog_args( $formdata ) {
		$product = !empty( $formdata[ 'nmgr_pid' ] ) ? wc_get_product( $formdata[ 'nmgr_pid' ] ) : false;
		$type = $formdata[ 'nmgr_type' ];
		$wishlist_id = $formdata[ 'nmgr_wid' ] ?? 0;
		$args = [];

		if ( $product ) {
			$wishlists = nmgr_get_user_wishlists( '', $type );

			foreach ( $wishlists as $key => $wish ) {
				if ( method_exists( $wish, 'is_archived' ) && $wish->is_archived() ) {
					unset( $wishlists[ $key ] );
				}
			}

			if ( !$wishlist_id && !empty( $wishlists ) ) {
				$first = reset( $wishlists );
				$wishlist_id = $first->get_id();
			}

			$formdata[ 'nmgr_wc_form_values' ] = $_POST[ 'nmgr_wc_form_values' ] ?? '';

			$args = array(
				'type' => $type,
				'product' => $product,
				'wishlists' => $wishlists,
				'selected_wishlist_id' => $wishlist_id,
				'grouped_products' => array(),
				'formdata' => $formdata,
			);

			if ( $product->is_type( 'grouped' ) ) {
				$args[ 'grouped_products' ] = array_filter(
					array_map( 'wc_get_product', $product->get_children() ),
					'wc_products_array_filter_visible_grouped' );
			}
		}
		return $args;
	}

	protected function delete_item() {
		$wid = nmgr_get_posted_wishlist_and_item_ids();
		$wishlist_id = $wid[ 'wishlist_id' ];
		$wishlist_item_ids = $wid[ 'wishlist_item_ids' ];

		nmgr_check_wishlist_permission( $wishlist_id );

		$items_data = array();
		$wishlist = nmgr_get_wishlist( $wishlist_id );
		if ( $wishlist && !empty( $wishlist_item_ids ) ) {
			$error = false;
			$success = false;
			foreach ( $wishlist_item_ids as $item_id ) {
				$item = $wishlist->get_item( $item_id );

				if ( !$item ||
					(!is_nmgr_admin() && $wishlist->is_type( 'gift-registry' ) &&
					((method_exists( $item, 'is_archived' ) && $item->is_archived()) ||
					$item->is_purchased() ||
					$item->is_fulfilled())
					)
				) {
					$error = true;
					continue;
				}

				$product = $item->get_product();

				$item_data = array(
					'wishlist_item_id' => $item_id,
					'wishlist_id' => $wishlist_id,
					'product_id' => $product ? $product->get_id() : '',
					'variation_id' => $product ? $product->get_id() : '',
				);

				$wishlist->delete_item( $item_id );

				$item_data[ 'product_in_wishlist' ] = $product ?
					( int ) nmgr_user_has_product_in_wishlist( $product ) : '';
				$items_data[] = $item_data;

				$success = true;
			}

			$acc = nmgr()->account( $wishlist );
			$acc->set_section( 'items' );
			$acc->set_args( [ 'page' => $this->get_current_page() ] );
			$items_section = $acc->get_sections_by_ids( 'items' );
			$acc->set_args( [] );
			$sections = array_merge( $items_section, $acc->get_other_sections_to_replace() );
			$toast_notice = [];
			if ( $error ) {
				$toast_notice[] = nmgr_get_error_toast_notice();
			}

			if ( $success ) {
				$toast_notice[] = nmgr_get_success_toast_notice();
			}

			wp_send_json( array(
				'replace_templates' => $sections,
				'toast_notice' => $toast_notice,
				'wishlist' => $wishlist->get_data(),
				'items_data' => $items_data
			) );
		}
	}

	protected function show_purchase_refund_item_dialog() {
		$wid = nmgr_get_posted_wishlist_and_item_ids();
		$wishlist_id = $wid[ 'wishlist_id' ];
		$wishlist_item_id = $wid[ 'wishlist_item_id' ];

		nmgr_check_wishlist_permission( $wishlist_id );

		$modal = nmgr_get_modal();
		$modal->set_title( nmgr()->is_pro ?
				__( 'Update purchased quantity', 'nm-gift-registry' ) :
				__( 'Update purchased quantity', 'nm-gift-registry-lite' )
		);
		$modal->set_id( 'nmgr-purchase-refund-dialog' );
		$modal->set_content( nmgr_get_purchase_refund_template( $wishlist_item_id ) );
		$modal->set_footer(
			$modal->get_save_button( [
				'attributes' => [
					'type' => 'submit',
					'class' => [ 'button-primary', 'nmgr_save_purchase_refund_form' ],
					'form' => 'nmgr-purchase-refund-form'
				]
			] )
		);

		wp_send_json( [ 'show_template' => $modal->get() ] );
	}

	protected function save_purchase_refund_form() {
		$wid = nmgr_get_posted_wishlist_and_item_ids();
		$wishlist_id = $wid[ 'wishlist_id' ];
		$wishlist_item_id = $wid[ 'wishlist_item_id' ];

		nmgr_check_wishlist_permission( $wishlist_id );

		$item = nmgr_get_wishlist_item( $wishlist_item_id );

		if ( ( int ) $_POST[ 'quantity' ] !== $item->get_purchased_quantity() ) {
			$args = [
				'quantity' => ( int ) $_POST[ 'quantity' ],
				'create_order' => !empty( $_POST[ 'create_order' ] ),
				'apply_price' => !empty( $_POST[ 'apply_price' ] ),
			];
			nmgr_update_item_purchased_quantity( $wishlist_item_id, $args );

			$acc = nmgr()->account( $wishlist_id );
			$acc->set_section( 'items' );
			$acc->set_args( [ 'page' => $this->get_current_page() ] );
			$items_section = $acc->get_sections_by_ids( 'items' );
			$acc->set_args( [] );
			$sections = array_merge( $items_section, $acc->get_other_sections_to_replace() );

			wp_send_json( [
				'toast_notice' => nmgr_get_toast_notice( nmgr_get_success_text() ),
				'replace_templates' => $sections,
				'close_dialog' => true,
			] );
		}
	}

	protected function show_add_to_wishlist_dialog() {
		$formdata = [];
		parse_str( $_POST[ 'add_to_wishlist_formdata' ], $formdata );
		wp_send_json( [ 'show_template' => $this->get_add_to_wishlist_dialog( $formdata ) ] );
	}

	/**
	 * Get the dialog for setting a wishlist shipping address
	 * @param int $wishlist_id The wishlist id
	 * @param string $add_to_wishlist_formdata The data from the add to wishlist form that should be submitted
	 * after the wishlist shipping address is successfully set.
	 * @return string
	 */
	protected function get_set_shipping_address_dialog( $wishlist_id, $add_to_wishlist_formdata ) {
		nmgr_check_wishlist_permission( $wishlist_id );

		add_filter( 'nmgr_fields', function( $fields ) use( $add_to_wishlist_formdata ) {
			$fields[ 'add_to_wishlist_formdata' ] = $this->get_add_to_wishlist_formdata_field( 'shipping', $add_to_wishlist_formdata );
			return $fields;
		} );

		// hide default submit button for shipping form so that we can set our own
		$css = '<style>.dialog.show.nmgr .nmgr-account-template .save-action { display:none;}</style>';
		$modal = nmgr_get_modal();
		$modal->set_title( nmgr()->is_pro ?
				__( 'Set shipping address', 'nm-gift-registry' ) :
				__( 'Set shipping address', 'nm-gift-registry-lite' )  );
		$modal->set_id( 'nmgr-set-shipping-address-dialog' );
		$modal->set_content( $css . nmgr_get_shipping_template( $wishlist_id ) );
		$modal->make_large()->make_vertically_centered();
		$modal->set_footer(
			$modal->get_save_button( [
				'attributes' => [
					'type' => 'submit',
					'class' => 'nmgr_save_shipping_address',
					'form' => 'nmgr-shipping-form'
				]
			] )
		);

		if ( $add_to_wishlist_formdata ) {
			$modal->set_container_attributes( [ 'data-add_to_wishlist_formdata' => $add_to_wishlist_formdata ] );
		}
		return $modal->get();
	}

	protected function show_set_shipping_address_dialog() {
		$wishlist_id = ( int ) wp_unslash( $_POST[ 'wishlist_id' ] ?? 0 );
		$add_to_wishlist_formdata = !empty( $_POST[ 'add_to_wishlist_formdata' ] ) ?
			sanitize_text_field( wp_unslash( $_POST[ 'add_to_wishlist_formdata' ] ) ) : '';

		wp_send_json( [
			'show_template' => $this->get_set_shipping_address_dialog( $wishlist_id, $add_to_wishlist_formdata )
		] );
	}

	protected function load_wishlist_cart() {
		if ( !empty( $_POST[ 'cart_id' ] ) ) {
			$args = [ 'type' => $_POST[ 'type' ] ?? '' ];
			wp_send_json( [
				'replace_templates' => [
					'#' . sanitize_text_field( $_POST[ 'cart_id' ] ) => nmgr_get_cart_template( $args )
				]
			] );
		}
	}

	protected function get_current_page() {
		return ( int ) ($_POST[ 'dataset' ][ 'page' ] ?? 1);
	}

	public function post_action() {
		check_ajax_referer( 'nmgr' );

		$action = $_POST[ 'nmgr_post_action' ] ?? '';

		switch ( $action ) {
			case 'save_form':
				$this->save_form();
				break;
			case 'save_item_quantity':
				$this->save_item_quantity();
				break;
			case 'add_item':
				$this->add_item();
				break;
			case 'delete_item':
				$this->delete_item();
				break;
			case 'purchase_refund_item':
				$this->show_purchase_refund_item_dialog();
				break;
			case 'save_purchase_refund_form':
				$this->save_purchase_refund_form();
				break;
			case 'auto_create_wishlist':
				$this->auto_create_wishlist();
				break;
			case 'show_create_new_wishlist_dialog':
				$this->show_create_new_wishlist_dialog();
				break;
			case 'show_add_to_wishlist_dialog':
				$this->show_add_to_wishlist_dialog();
				break;
			case 'show_set_shipping_address_dialog':
				$this->show_set_shipping_address_dialog();
				break;
			case 'load_wishlist_cart':
				$this->load_wishlist_cart();
				break;
			default:
				$args = array_merge( [
					'post_action' => $action,
					], nmgr_get_posted_wishlist_and_item_ids() );
				do_action( 'nmgr_post_action', $args );
				break;
		}

		wp_die();
	}

}
