<?php

defined( 'ABSPATH' ) || exit;

/*
 * Handles CRUD operations for a wishlist item
 */

class NMGR_Database_Wishlist_Item {

	public function create( $item ) {
		global $wpdb;

		$wpdb->insert(
			$wpdb->prefix . 'nmgr_wishlist_items', array(
			'product_id' => $item->get_product_id(),
			'wishlist_id' => $item->get_wishlist_id(),
			'date_created' => current_time( 'mysql', 1 ),
			'date_modified' => current_time( 'mysql', 1 ),
			)
		);
		$item->set_id( $wpdb->insert_id );
		$this->update_meta_data( $item, true );
		$item->apply_changes();

		do_action_deprecated( 'nmgr_new_wishlist_item', array( $item->get_id(), $item, $item->get_wishlist_id() ), '2.1.0', 'nmgr_wishlist_item_created' );
		do_action( 'nmgr_wishlist_item_created', $item, $item->get_wishlist() );
	}

	public function read( $item ) {
		global $wpdb;

		$item->set_defaults();

		// Get from cache if available.
		$data = 0 < $item->get_id() ? wp_cache_get( 'nmgr-item-' . $item->get_id() ) : false;

		if ( false === $data ) {
			$data = $wpdb->get_row( $wpdb->prepare( "SELECT * FROM {$wpdb->prefix}nmgr_wishlist_items WHERE wishlist_item_id = %d LIMIT 1;", $item->get_id() ) );

			if ( !$data ) {
				/* translators: %s: wishlist type title */
				throw new Exception( sprintf(
						/* translators: %s: wishlist type title */
						nmgr()->is_pro ? __( 'Invalid %s item.', 'nm-gift-registry' ) : __( 'Invalid %s item.', 'nm-gift-registry-lite' ),
						nmgr_get_type_title()
					) );
			}

			if ( 0 < $item->get_id() ) {
				wp_cache_set( 'nmgr-item-' . $item->get_id(), $data );
			}
		}

		$item->set_props(
			array(
				'wishlist_id' => $data->wishlist_id,
				'product_id' => $data->product_id,
				'date_created' => 0 < $data->date_created ? $data->date_created : null,
				'date_modified' => 0 < $data->date_modified ? $data->date_modified : null,
			)
		);

		$this->read_meta_data( $item );
		$item->set_object_read( true );
	}

	public function update( $item ) {
		global $wpdb;

		/**
		 * We want to update the nmgr_wishlist_items table if any of the item's properties has changed
		 * This is mainly so that we can update the 'date_modified' value
		 */
		if ( array_intersect( array_keys( $item->get_data() ), array_keys( $item->get_changes() ) ) ) {
			$wpdb->update(
				$wpdb->prefix . 'nmgr_wishlist_items', array(
				'product_id' => $item->get_product_id(),
				'wishlist_id' => $item->get_wishlist_id(),
				'date_modified' => current_time( 'mysql', 1 ),
				), array( 'wishlist_item_id' => $item->get_id() )
			);

			$this->clear_cache( $item->get_id() );
		}

		$this->update_meta_data( $item );
		$item->apply_changes();

		do_action_deprecated( 'nmgr_update_wishlist_item', array( $item->get_id(), $item, $item->get_wishlist_id() ), '2.1.0', 'nmgr_wishlist_item_updated' );
		do_action( 'nmgr_wishlist_item_updated', $item, $item->get_wishlist() );
	}

	public function delete( $item ) {
		global $wpdb;

		$id = 0;
		if ( is_numeric( $item ) ) {
			$id = $item;
		} elseif ( is_object( $item ) && $item->get_id() ) {
			$id = $item->get_id();
		}

		if ( $id ) {
			$wishlist_id = $item->get_wishlist_id();

			do_action( 'nmgr_before_delete_wishlist_item', $item->get_id() );

			$wpdb->delete( $wpdb->prefix . 'nmgr_wishlist_items', array( 'wishlist_item_id' => $id ) );
			$wpdb->delete( $wpdb->prefix . 'nmgr_wishlist_itemmeta', array( 'wishlist_item_id' => $id ) );

			do_action_deprecated( 'nmgr_delete_wishlist_item', array( $item->get_id(), $item->get_wishlist_id() ), '2.1.0', 'nmgr_wishlist_item_deleted' );
			do_action( 'nmgr_wishlist_item_deleted', $id, $wishlist_id );

			$this->clear_cache( $id );
		}
	}

	public function read_meta_data( $item ) {
		$props_to_meta_keys = $item->get_meta_keys( $item->get_meta_data() );
		foreach ( $props_to_meta_keys as $prop => $meta_key ) {
			if ( metadata_exists( $item->meta_type, $item->get_id(), $meta_key ) ) {
				$val = get_metadata( $item->meta_type, $item->get_id(), $meta_key, true );
			} else {
				$val = $item->get_default_data()[ $prop ] ?? '';
			}
			$item->set_prop( $prop, $val );
		}
	}

	public function update_meta_data( $item, $force = false ) {
		$props_to_meta_keys = $item->get_props_to_update( $force );

		if ( !empty( $props_to_meta_keys ) ) {
			$old_values = get_metadata( $item->meta_type, $item->get_id() );

			foreach ( $props_to_meta_keys as $prop => $meta_key ) {
				if ( is_callable( array( $item, "get_$prop" ) ) ) {
					$new_value = $item->{"get_$prop"}();
				} else {
					$new_value = $item->get_prop( $prop );
				}

				$new_value = is_string( $new_value ) ? wp_slash( $new_value ) : $new_value;
				$old_value = get_metadata( $item->meta_type, $item->get_id(), $meta_key, true );
				update_metadata( $item->meta_type, $item->get_id(), $meta_key, $new_value );
				do_action( 'nmgr_updated_meta_key', $meta_key, $new_value, $old_value, $item );
			}

			$new_values = get_metadata( $item->meta_type, $item->get_id() );
			do_action( 'nmgr_updated_metadata', $new_values, $old_values, $item );
		}
	}

	public function clear_cache( $item_id ) {
		wp_cache_delete( 'nmgr-item-' . $item_id );
	}

}
