<?php

/**
 * Sync
 */
defined( 'ABSPATH' ) || exit;

/*
 * Handles CRUD operations for a wishlist
 */

class NMGR_Database_Wishlist {

	/**
	 * Create a new wishlist in the database.
	 *
	 * @param NMGR_Wishlist $wishlist Wishlist object.
	 */
	public function create( $wishlist ) {
		$id = wp_insert_post( array_intersect_key( $wishlist->get_data(), $wishlist->get_core_data() ) );

		if ( $id && !is_wp_error( $id ) ) {
			$wishlist->set_id( $id );
			$this->update_tax_data( $wishlist );
			$this->update_meta_data( $wishlist, true );
			$wishlist->apply_changes();

			do_action( 'nmgr_created_wishlist', $id );
		}
	}

	/**
	 * Read a wishlist from the database
	 *
	 * @param NMGR_Wishlist $wishlist Wishlist  object.
	 * @throws Exception for an invalid wishlist.
	 */
	public function read( $wishlist ) {
		$wishlist->set_defaults();
		$post = get_post( $wishlist->get_id() );

		if ( !$wishlist->get_id() || !$post || 'nm_gift_registry' !== $post->post_type ) {
			throw new Exception( sprintf(
					/* translators: %s: wishlist type title */
					nmgr()->is_pro ? __( 'Invalid %s.', 'nm-gift-registry' ) : __( 'Invalid %s.', 'nm-gift-registry-lite' ),
					nmgr_get_type_title()
				) );
		}

		foreach ( array_keys( $wishlist->get_data() ) as $key ) {
			if ( property_exists( $post, $key ) ) {
				$wishlist->set_prop( $key, $post->$key );
			}
		}

		$terms = wp_get_object_terms( $wishlist->get_id(), 'nm_gift_registry_type', [ 'fields' => 'slugs' ] );
		if ( !empty( $terms ) ) {
			$terms_array = ( array ) $terms;
			$type = reset( $terms_array );
			$wishlist->set_type( $type );
		}

		$this->read_meta_data( $wishlist );

		do_action( 'nmgr_read_wishlist', $wishlist );

		$wishlist->set_object_read( true );
	}

	/**
	 * Update a wishlist item in the database.
	 *
	 * @param NMGR_Wishlist $wishlist Wishlist object.
	 */
	public function update( $wishlist ) {
		$core_data = array_intersect_key( $wishlist->get_changes(), $wishlist->get_core_data() );

		if ( !empty( $core_data ) ) {
			/**
			 * Use $wpdb->update to update directly if doing the save_post action (such as in admin screen)
			 * to prevent infinite loops
			 */
			if ( doing_action( 'save_post_nm_gift_registry' ) ) {
				$GLOBALS[ 'wpdb' ]->update( $GLOBALS[ 'wpdb' ]->posts, $core_data, array( 'ID' => $wishlist->get_id() ) );
				clean_post_cache( $wishlist->get_id() );
			} else {
				wp_update_post( array_merge( array( 'ID' => $wishlist->get_id() ), $core_data ) );
			}
		}

		$this->update_tax_data( $wishlist );
		$this->update_meta_data( $wishlist );
		$wishlist->apply_changes();

		do_action( 'nmgr_updated_wishlist', $wishlist->get_id() );
	}

	/**
	 * Remove a wishlist from the database.
	 *
	 * @paramNMGR_Wishlist $wishlist Wishlist object.
	 * @param bool $force_delete Whether to permanently delete the wishlist. Default false
	 */
	public function delete( $wishlist, $force_delete = false ) {
		$id = $wishlist->get_id();

		if ( !$id ) {
			return;
		}

		if ( $force_delete ) {
			wp_delete_post( $id );
			$wishlist->set_id( 0 );
		} else {
			wp_trash_post( $id );
			$wishlist->set_status( 'trash' );
		}
	}

	/**
	 * Read a wishlist's meta data
	 *
	 * @param NMGR_Wishlist $wishlist Wishlist object
	 */
	public function read_meta_data( $wishlist ) {
		$data = array();
		$props_to_meta_keys = $wishlist->get_meta_keys( $wishlist->get_meta_data() );

		foreach ( $props_to_meta_keys as $prop => $meta_key ) {
			if ( metadata_exists( 'post', $wishlist->get_id(), $meta_key ) ) {
				$data[ $prop ] = get_post_meta( $wishlist->get_id(), $meta_key, true );
			} else {
				$data[ $prop ] = $wishlist->get_default_data()[ $prop ] ?? '';
			}
		}

		$wishlist->set_props( $data );
	}

	/**
	 * Saves an item's meta data to the database
	 *
	 * @param NMGR_Wishlist $item Wishlist object
	 * @param bool $force	Whether to update all meta or only changed ones. Used during create
	 */
	public function update_meta_data( $item, $force = false ) {
		$props_to_meta_keys = $item->get_props_to_update( $force );

		if ( !empty( $props_to_meta_keys ) ) {
			$old_values = get_metadata( $item->meta_type, $item->get_id() );

			foreach ( $props_to_meta_keys as $prop => $meta_key ) {
				if ( is_callable( array( $item, "get_$prop" ) ) ) {
					$new_value = $item->{"get_$prop"}();
				} else {
					$new_value = $item->get_prop( $prop );
				}

				$new_value = is_string( $new_value ) ? wp_slash( $new_value ) : $new_value;
				$old_value = get_metadata( $item->meta_type, $item->get_id(), $meta_key, true );
				do_action( 'nmgr_updated_meta_key', $meta_key, $new_value, $old_value, $item );
				update_metadata( $item->meta_type, $item->get_id(), $meta_key, $new_value );
			}

			$new_values = get_metadata( $item->meta_type, $item->get_id() );
			do_action( 'nmgr_updated_metadata', $new_values, $old_values, $item );
		}
	}

	public function update_tax_data( $wishlist ) {
		$extra_data = $wishlist->get_extra_data();
		if ( !empty( $extra_data[ 'tax_input' ] ) ) {
			foreach ( $extra_data[ 'tax_input' ] as $taxonomy => $tags ) {
				wp_set_post_terms( $wishlist->get_id(), $tags, $taxonomy );
			}
		}
	}

	/**
	 * Read wishlist items from the database for this wishlist.
	 * @deprecated since version 4.2.0
	 * @param  NMGR_Wishlist $wishlist Wishlist object.
	 * @return array
	 */
	public function read_items( $wishlist ) {
		_deprecated_function( __METHOD__, '4.2.0', 'NMGR_Wishlist::read_items' );
		return $wishlist->read_items();
	}

	/**
	 * Remove all items from the wishlist.
	 *
	 * @deprecated since version 4.2.0
	 * @param  NMGR_Wishlist $wishlist Wishlist object.
	 */
	public function delete_items( $wishlist ) {
		_deprecated_function( __METHOD__, '4.2.0', 'NMGR_Wishlist::delete_items' );
		$wishlist->delete_items();
	}

	/**
	 * Delete all images from the wishlist
	 *
	 * Saves updated wishlist data without images data to database after deleting.
	 *
	 * @deprecated since version 3.0.0
	 * @param NMGR_Wishlist $wishlist
	 * @param int|array $image_id Image id(s) to delete. If none is supplied, deletes all wishlist images
	 */
	public function delete_images( $wishlist, $image_id = [] ) {
		_deprecated_function( __FUNCTION__, '3.0.0', '$Wishlist->delete_images()' );
		if ( method_exists( $wishlist, 'delete_images' ) ) {
			$wishlist->delete_images( $image_id );
		}
	}

	/**
	 * @deprecated since version 4.2.0
	 */
	public function clear_items_cache( $wishlist ) {
		_deprecated_function( __METHOD__, '4.2.0', 'NMGR_Wishlist::clear_items_cache' );
		$wishlist->clear_items_cache();
	}

}
