<?php

/**
 * Sync
 */
defined( 'ABSPATH' ) || exit;

use NMGR\Int\Nmerimedia_Strings,
		NMGR\Setup\Wizard,
		NmeriMedia\V2\Utils;

class NMGR_Installer extends Utils {

	use \NmeriMedia\V2\Traits\PluginProps,
		 \NmeriMedia\V2\Traits\Constructor;

	/**
	 * @deprecated since version 4.0.0
	 * @var \NmeriMedia\V2\Settings
	 */
	public $settings;
	private $gift_registry_settings;
	private $wishlist_settings;
	private $default_field_values = [];
	private $default_pro_fields_values = [];
	public $docs_url;
	public $product_url;
	public $review_url;
	public $support_url;

	/**
	 * Notices to be shown
	 *
	 * @var array
	 */
	private $notices = array();

	public function load() {
		spl_autoload_register( array( $this, 'autoload' ) );

		$this->set_extra_props();

		register_deactivation_hook( $this->file, array( $this, 'deactivate' ) );
		register_uninstall_hook( $this->file, array( static::class, 'uninstall' ) );
		add_action( 'plugins_loaded', array( $this, 'maybe_install_and_run' ) );
	}

	private function set_extra_props() {
		$this->is_pro = false === strpos( $this->slug, 'lite' );
		$this->is_licensed = false === strpos( $this->slug, 'lite' );

		$plugin_headers = array(
			'docs_url' => 'Nmerimedia Docs URI',
			'product_url' => 'Nmerimedia Product URI',
			'review_url' => 'Nmerimedia Review URI',
			'support_url' => 'Nmerimedia Support URI',
		);

		$filedata = get_file_data( $this->file, $plugin_headers );

		foreach ( $filedata as $key => $val ) {
			if ( property_exists( $this, $key ) ) {
				$this->{$key} = $val;
			}
		}
	}

	public function post_thumbnail_size() {
		return apply_filters( 'nmgr_medium_size', 190 );
	}

	public function theme_path() {
		return apply_filters( 'nmgr_theme_path', trailingslashit( $this->slug ) );
	}

	public function template_path() {
		return plugin_dir_path( $this->file ) . 'templates/';
	}

	public static function get_plugin_data() {
		_deprecated_function( __METHOD__, '3.0.0', 'nmgr' );
		return static::get_instance();
	}

	public function autoload( $class ) {
		if ( class_exists( $class ) || false === stripos( $class, 'nmgr' ) ) {
			return;
		}

		$file = 'class-' . str_replace( '_', '-', strtolower( $class ) ) . '.php';
		$classes_path = realpath( dirname( $this->file ) ) . DIRECTORY_SEPARATOR . 'includes' . DIRECTORY_SEPARATOR;
		$filepath = $classes_path . $file;

		if ( file_exists( $filepath ) ) {
			include_once $filepath;
		}
	}

	public function deactivate() {
		flush_rewrite_rules();
	}

	/**
	 * @return \NmeriMedia\V2\Settings
	 */
	public function get_settings() {
		_deprecated_function( __METHOD__, '4.0.0', 'gift_registry_settings()' );
		return $this->gift_registry_settings();
	}

	/**
	 *
	 * @return NMGR_Gift_Registry_Settings
	 */
	public function gift_registry_settings() {
		if ( !$this->gift_registry_settings ) {
			$this->gift_registry_settings = new NMGR_Gift_Registry_Settings( [ $this, 'get_instance' ] );
		}

		return $this->gift_registry_settings;
	}

	/**
	 *
	 * @return NMGR_Wishlist_Settings
	 */
	public function wishlist_settings() {
		if ( !$this->wishlist_settings ) {
			$this->wishlist_settings = new NMGR_Wishlist_Settings( [ $this, 'get_instance' ] );
		}

		return $this->wishlist_settings;
	}

	public function settings() {
		_deprecated_function( __METHOD__, '4.0.0', 'gift_registry_settings()' );
		return $this->get_settings();
	}

	public function get_default_field_values() {
		return $this->default_field_values;
	}

	public function get_default_pro_fields_values() {
		return $this->default_pro_fields_values;
	}

	public function set_default_field_values() {
		$this->default_field_values = array_merge(
			$this->wishlist_settings()->get_default_field_values(),
			$this->gift_registry_settings()->get_default_field_values()
		);

		$this->default_pro_fields_values = array_merge(
			$this->wishlist_settings()->get_default_pro_fields_values(),
			$this->gift_registry_settings()->get_default_pro_fields_values()
		);
	}

	public function maybe_install_and_run() {
		if ( $this->maybe_deactivate_plugin() ) {
			add_action( 'admin_init', array( $this, 'deactivate_plugin' ) );
			add_action( 'admin_notices', array( $this, 'show_deactivation_notice' ) );
			return;
		}

		/**
		 * Unregister the autoload function registered with lite versions prior to 3.0.0
		 * This is necessary in other to prevent fatal error when activating lite version
		 * prior to 3.0.0 with pro version from 3.0.0 already activated.
		 * @since version 3.0.0
		 * @todo remove in version 4.0.0
		 */
		if ( $this->is_pro && method_exists( 'NMGR_Lite_Install', 'autoload' ) &&
			(new ReflectionMethod( 'NMGR_Lite_Install', 'autoload' ) )->isStatic() ) {
			spl_autoload_unregister( [ 'NMGR_Lite_Install', 'autoload' ] );
		}

		// At this point we can install and run our plugin

		include_once $this->path . 'includes/nmgr-functions.php';

		add_filter( 'nmerimedia_strings_class', [ $this, 'strings_class' ] );

		// Run after woocommerce_init which runs at priority 1
		add_action( 'init', [ $this, 'set_default_field_values' ] );

		// Installation action
		if ( version_compare( get_option( $this->plugin_prefix . '_version' ), $this->version, '<' ) ) {
			add_action( 'init', array( $this, 'install_actions' ) );
		}

		// Run plugin
		$this->run();
	}

	public function strings_class() {
		return Nmerimedia_Strings::class;
	}

	public function install_actions() {
		Wizard::install_actions();

		/**
		 * Cleanup: Remove old cron hooks
		 * @todo Remove code in later version
		 */
		if ( version_compare( nmgr()->version, '4.2.0', '<' ) ) {
			wp_clear_scheduled_hook( 'nmgr_delete_guest_wishlists' );
			wp_clear_scheduled_hook( 'nmgr_set_expired_wishlists' );
			wp_clear_scheduled_hook( 'nmgr_gift-registry_SetExpiredWishlists' );
		}

		$this->add_default_settings();
		$this->create_tables();
		static::add_capabilities();

		// register taxonomy (for inserting terms and flushing rewrite rules)
		$this->wordpress()->register_taxonomy();

		// Insert terms
		$this->insert_terms();

		// register custom post type (for flushing rewrite rules)
		$this->wordpress()->register_post_types();

		// register custom wishlist single and archive pages (for flushing rewrite rules)
		$this->wordpress()->add_rewrite_rules();

		// Update version
		update_option( $this->plugin_prefix . '_version', $this->version );

		/**
		 * Delete deprecated options
		 * @since 4.0.0
		 * @todo Remove in a later version
		 */
		delete_option( 'nmgr_show_current_version_notices' );

		// flush rewrite rules
		flush_rewrite_rules();

		// Runs on register activation hook (on activation) or init (on update)
		do_action( $this->plugin_prefix . '_installed' );
	}

	public function run() {
		$this->gift_registry_settings()->init();
		$this->wishlist_settings()->init();

		Wizard::run();
		NMGR_Scripts::run();
		$this->ajax()->run();
		$this->admin_post()->run();
		NMGR_Admin::run();
		$this->templates()->run();
		NMGR_Form::run();
		NMGR_Events::run();
		$this->order()->run();
		$this->wordpress()->run();

		if ( method_exists( nmgr(), 'mailer' ) ) {
			nmgr()->mailer()->run();
		}

		if ( $this->is_licensed && class_exists( '\\NMGR\\Sub\\Update' ) ) {
			new \NMGR\Sub\Update();
		}

		new NMGR_Widget_Cart();
		new NMGR_Widget_Search();

		add_action( 'init', array( $this, 'register_meta_tables' ), 0 );
		add_action( 'switch_blog', array( $this, 'register_meta_tables' ), 0 );
		add_filter( 'plugin_action_links_' . $this->basename, array( $this, 'plugin_action_links' ) );
		add_filter( 'plugin_row_meta', array( $this, 'plugin_row_meta' ), 10, 2 );

		do_action( $this->plugin_prefix . '_plugin_loaded' );
	}

	/**
	 * Set wishlist types
	 * Added in version 4.0.0
	 */
	public function insert_terms() {
		$terms = [ 'wishlist', 'gift registry' ];
		foreach ( $terms as $term ) {
			if ( empty( term_exists( $term, 'nm_gift_registry_type' ) ) ) {
				wp_insert_term( $term, 'nm_gift_registry_type' );
			}
		}
	}

	public function add_default_settings() {
		$defaults = $this->get_default_field_values();
		$existing_settings = get_option( 'nmgr_settings' );

		if ( $existing_settings ) {
			$defaults = array_merge( $defaults, $existing_settings );

			if ( version_compare( get_option( $this->plugin_prefix . '_version' ), '4.0.0', '<' ) ) {
				/**
				 * Option key 'add_to_wishlist_single' is deprecated.
				 * If it's value is already set, set it for the option key 'add_to_wishlist_button_position_single'
				 * remove the deprecated option key in a later version
				 * @since 4.0.0
				 */
				if ( array_key_exists( 'add_to_wishlist_single', $defaults ) && !$defaults[ 'add_to_wishlist_single' ] ) {
					$defaults[ 'add_to_wishlist_button_position_single' ] = '';
				}

				/**
				 * Option key 'add_to_wishlist_archive' is deprecated.
				 * If it's value is already set, set it for the option key 'add_to_wishlist_button_position_archive'
				 * remove the deprecated option key in a later version
				 * @since 4.0.0
				 */
				if ( array_key_exists( 'add_to_wishlist_archive', $defaults ) && !$defaults[ 'add_to_wishlist_archive' ] ) {
					$defaults[ 'add_to_wishlist_button_position_archive' ] = '';
				}
			}

			delete_option( 'nmgr_settings' );
		}

		add_option( 'nmgr_settings', $defaults );

		if ( $this->is_pro ) {
			add_option( 'nmgr_exclude_from_search', array() );
		}
	}

	public function create_tables() {
		global $wpdb;

		$collate = '';

		if ( $wpdb->has_cap( 'collation' ) ) {
			$collate = $wpdb->get_charset_collate();
		}

		$tables = "
		  CREATE TABLE {$wpdb->prefix}nmgr_wishlist_items (
				wishlist_item_id BIGINT UNSIGNED NOT NULL auto_increment,
				product_id BIGINT UNSIGNED NOT NULL,
				wishlist_id BIGINT UNSIGNED NOT NULL,
				date_created datetime NOT NULL DEFAULT '0000-00-00 00:00:00',
				date_modified datetime NOT NULL DEFAULT '0000-00-00 00:00:00',
				PRIMARY KEY  (wishlist_item_id),
				KEY wishlist_id (wishlist_id)
			 ) $collate;
		  CREATE TABLE {$wpdb->prefix}nmgr_wishlist_itemmeta (
			 meta_id BIGINT UNSIGNED NOT NULL auto_increment,
			 wishlist_item_id BIGINT UNSIGNED NOT NULL,
			 meta_key varchar(255) default NULL,
			 meta_value longtext NULL,
			 PRIMARY KEY  (meta_id),
			 KEY wishlist_item_id (wishlist_item_id),
			 KEY meta_key (meta_key(32))
		  ) $collate;
		  ";

		// update schema with dbdelta
		require_once ABSPATH . 'wp-admin/includes/upgrade.php';

		if ( $wpdb->get_var( "SHOW TABLES LIKE '{$wpdb->prefix}nmgr_wishlist_items';" ) ) {
			$items_table = "{$wpdb->prefix}nmgr_wishlist_items";

			/**
			 * Drop the name column from the wishlist items table
			 * @since 2.4
			 */
			if ( $wpdb->query( "SHOW columns from $items_table LIKE 'name'" ) ) {
				$wpdb->query( "ALTER TABLE $items_table DROP `name`" );
			}

			/**
			 * Add the product_id column to the wishlist items table and populate it
			 * @since 2.4
			 */
			if ( !$wpdb->query( "SHOW columns from $items_table LIKE 'product_id'" ) ) {
				$wpdb->query( "ALTER TABLE $items_table ADD `product_id` BIGINT UNSIGNED NOT NULL AFTER `wishlist_item_id`" );

				$wpdb->query( "UPDATE $items_table AS a INNER JOIN {$wpdb->prefix}nmgr_wishlist_itemmeta AS b ON a.wishlist_item_id = b.wishlist_item_id SET a.product_id = b.meta_value WHERE b.meta_key = '_product_id'" );
			}
		}

		dbDelta( $tables );
	}

	/**
	 * Get the roles that have been assigned a specific plugin capability
	 *
	 * Default roles are admiinstrator and shop_manager
	 *
	 * @param string $capability_type The plugin capability
	 * @return array The roles that have this capability
	 */
	public static function get_roles( $capability_type ) {
		return apply_filters( "nmgr_{$capability_type}_roles", array(
			'administrator',
			'shop_manager'
			) );
	}

	public static function get_capabilities() {
		$capabilities = array();
		$post_type = 'nm_gift_registry';
		$post_type_plural = 'nm_gift_registries';

		// Permission for managing plugin settings
		$capabilities[ 'manage_settings' ] = array(
			"manage_{$post_type}_settings"
		);

		// Permission for managing gift registry post type CRUD operations
		$capabilities[ "manage_CRUD" ] = array(
			"edit_{$post_type}",
			"read_{$post_type}",
			"delete_{$post_type}",
			"edit_{$post_type_plural}",
			"edit_others_{$post_type_plural}",
			"publish_{$post_type_plural}",
			"read_private_{$post_type_plural}",
			"delete_{$post_type_plural}",
			"delete_private_{$post_type_plural}",
			"delete_published_{$post_type_plural}",
			"delete_others_{$post_type_plural}",
			"edit_private_{$post_type_plural}",
			"edit_published_{$post_type_plural}",
		);

		return $capabilities;
	}

	public static function add_capabilities() {
		global $wp_roles;

		if ( !class_exists( 'WP_Roles' ) ) {
			return;
		}

		if ( !isset( $wp_roles ) ) {
			$wp_roles = new WP_Roles(); // phpcs:ignore
		}

		$capabilities = static::get_capabilities();

		foreach ( $capabilities as $type => $permissions ) {
			$roles = static::get_roles( $type );

			if ( !empty( $roles ) ) {
				foreach ( $permissions as $permission ) {
					foreach ( $roles as $role ) {
						$wp_roles->add_cap( $role, $permission );
					}
				}
			}
		}
	}

	public static function remove_capabilities() {
		global $wp_roles;

		if ( !class_exists( 'WP_Roles' ) ) {
			return;
		}

		if ( !isset( $wp_roles ) ) {
			$wp_roles = new WP_Roles(); // phpcs:ignore
		}

		$capabilities = static::get_capabilities();

		foreach ( $capabilities as $type => $permissions ) {
			$roles = static::get_roles( $type );

			if ( !empty( $roles ) ) {
				foreach ( $permissions as $permission ) {
					foreach ( $roles as $role ) {
						$wp_roles->remove_cap( $role, $permission );
					}
				}
			}
		}
	}

	/**
	 * Register our custom meta tables with wordpress so that we can use the meta api
	 * @global object $wpdb
	 */
	public function register_meta_tables() {
		global $wpdb;

		$wpdb->wishlist_itemmeta = $wpdb->prefix . 'nmgr_wishlist_itemmeta';
		$wpdb->tables[] = 'nmgr_wishlist_itemmeta';
	}

	public function plugin_action_links( $links ) {
		$gr = $this->gift_registry_settings();
		$wl = $this->wishlist_settings();

		return array_merge( $links, array(
			'<a href="' . $gr->get_page_url() . '">' . $gr->get_heading() . '</a>',
			'<a href="' . $wl->get_page_url() . '">' . $wl->get_heading() . '</a>',
			) );
	}

	public function plugin_row_meta( $links, $file ) {
		if ( $file == $this->basename ) {
			$docs = $this->is_pro ?
				__( 'Docs', 'nm-gift-registry' ) :
				__( 'Docs', 'nm-gift-registry-lite' );

			$links[] = '<a href="' . esc_url( $this->docs_url ) . '">' . $docs . '</a>';

			if ( !$this->is_pro ) {
				$links[] = '<a href="https://nmerimedia.com/product/nm-gift-registry" style="color:#b71401;"><strong>' . __( 'Get PRO', 'nm-gift-registry-lite' ) . '</strong></a>';
			}
		}
		return $links;
	}

	public function maybe_deactivate_plugin() {
		if ( !$this->is_pro && class_exists( '\\NMGR\\Sub\\Mailer' ) ) {
			$plugin_name = '<strong>' . $this->name . '</strong>';
			$this->notices[] = [
				'message' => $plugin_name . ' - ' . __( 'The lite version of the plugin cannot be activated while the full version is active.', 'nm-gift-registry-lite' ),
				'type' => 'info',
			];
		}

		if ( !class_exists( 'Woocommerce' ) ) {
			$this->notices[] = [ 'message' => sprintf(
					/* translators: %s: plugin name */
					$this->is_pro ? __( 'You need the <strong>WooCommerce plugin</strong> to be installed and activated for %s to work.', 'nm-gift-registry' ) : __( 'You need the <strong>WooCommerce plugin</strong> to be installed and activated for %s to work.', 'nm-gift-registry-lite' ),
					'<strong>' . $this->name . '</strong>'
				),
			];
		} elseif ( version_compare( WC_VERSION, $this->requires_wc, '<' ) ) {
			$this->notices[] = [ 'message' => sprintf(
					/* translators: %1$s: plugin name, %2$s: required woocommerce version */
					$this->is_pro ? __( '%1$s needs <strong>WooCommerce %2$s or higher</strong> to work. Please update WooCommerce.', 'nm-gift-registry' ) : __( '%1$s needs <strong>WooCommerce %2$s or higher</strong> to work. Please update WooCommerce.', 'nm-gift-registry-lite' ),
					'<strong>' . $this->name . '</strong>',
					$this->requires_wc
				), ];
		}

		if ( $this->is_pro && function_exists( 'nmgrcf' ) && property_exists( $this, 'requires_nmgrcf' ) &&
			version_compare( nmgrcf()->version, $this->requires_nmgrcf, '<' ) ) {
			$msg = '<p><strong>' . $this->name . '</strong> deactivated.</p>';
			$this->notices[] = [ 'message' => $msg . sprintf(
					/* translators: 1,2: plugin name, 3: required version */
					__( '%1$s needs %2$s %3$s or higher to work. Please update %2$s or deactivate the current version.', 'nm-gift-registry' ),
					'<strong>' . $this->name . '</strong>',
					'<strong>' . nmgrcf()->name . '</strong>',
					'<u>' . $this->requires_nmgrcf . '</u>'
				), ];
		}

		return !empty( $this->notices ) ? true : false;
	}

	public function deactivate_plugin() {
		deactivate_plugins( $this->basename );
		if ( isset( $_GET[ 'activate' ] ) ) { // phpcs:ignore WordPress.Security.NonceVerification
			unset( $_GET[ 'activate' ] ); // phpcs:ignore WordPress.Security.NonceVerification
		}
	}

	public function show_deactivation_notice() {
		foreach ( $this->notices as $notice ) {
			printf( '<div class="notice notice-%1$s"><p>%2$s</p></div>',
				$notice[ 'type' ] ?? 'error',
				wp_kses_post( $notice[ 'message' ] )
			);
		}
	}

	public static function uninstall() {
		if ( static::get_instance()->is_pro ) {
			static::uninstall_pro();
		} else {
			static::uninstall_lite();
		}
	}

	private static function uninstall_pro() {
		global $wpdb;

		if ( !apply_filters( 'nmgr_delete_data_on_uninstall', false ) ) {
			return;
		}

		// delete images
		$post_ids = $wpdb->get_col( "SELECT ID FROM {$wpdb->posts} WHERE post_type = 'nm_gift_registry'" );
		foreach ( $post_ids as $post_id ) {
			$attachment_ids = $wpdb->get_col( "SELECT ID FROM {$wpdb->posts} WHERE post_parent = $post_id AND post_type = 'attachment'" );
			foreach ( $attachment_ids as $id ) {
				wp_delete_attachment( $id, true );
			}
		}

		// delete comments and commentmeta
		$wpdb->query( "DELETE FROM {$wpdb->comments} WHERE comment_type IN ( 'nmgiftregistry_msg' );" );
		$wpdb->query( "DELETE meta FROM {$wpdb->commentmeta} meta LEFT JOIN {$wpdb->comments} comments ON comments.comment_ID = meta.comment_id WHERE comments.comment_ID IS NULL;" );

		if ( !get_option( 'nmgrlite_version' ) ) {
			static::uninstall_actions();
		}

		$wpdb->query( "DELETE FROM $wpdb->options WHERE option_name = 'nmgr_version';" );
		$wpdb->query( "DELETE FROM $wpdb->options WHERE option_name = 'nmgr_exclude_from_search';" );
		$wpdb->query( "DELETE FROM $wpdb->options WHERE option_name LIKE '\_transient\_nmgr\_%';" );

		wp_cache_flush();
	}

	private static function uninstall_lite() {
		global $wpdb;

		if ( !apply_filters( 'nmgr_delete_data_on_uninstall', false ) ) {
			return;
		}

		if ( !get_option( 'nmgr_version' ) ) {
			static::uninstall_actions();
		}

		$wpdb->query( "DELETE FROM $wpdb->options WHERE option_name LIKE 'nmgrlite\_%';" );
	}

	private static function uninstall_actions() {
		global $wpdb;

		// remove capabilities
		static::remove_capabilities();

		// delete tables
		$wpdb->query( "DROP TABLE IF EXISTS {$wpdb->prefix}nmgr_wishlist_items" );
		$wpdb->query( "DROP TABLE IF EXISTS {$wpdb->prefix}nmgr_wishlist_itemmeta" );

		// delete posts and postmeta
		$wpdb->query( "DELETE FROM {$wpdb->posts} WHERE post_type IN ( 'nm_gift_registry');" );
		$wpdb->query( "DELETE meta FROM {$wpdb->postmeta} meta LEFT JOIN {$wpdb->posts} posts ON posts.ID = meta.post_id WHERE posts.ID IS NULL;" );

		// delete user meta
		$wpdb->query( "DELETE FROM $wpdb->usermeta WHERE meta_key LIKE 'nmgr\_%';" );

		// delete options
		$wpdb->query( "DELETE FROM $wpdb->options WHERE option_name LIKE '%%nmgr\_%%';" );

		// delete taxonomy
		$wpdb->delete( $wpdb->term_taxonomy, array( 'taxonomy' => 'nm_gift_registry_type' ) );

		// Delete orphan relationships.
		$wpdb->query( "DELETE tr FROM {$wpdb->term_relationships} tr LEFT JOIN {$wpdb->posts} posts ON posts.ID = tr.object_id WHERE posts.ID IS NULL;" );

		// Delete orphan terms.
		$wpdb->query( "DELETE t FROM {$wpdb->terms} t LEFT JOIN {$wpdb->term_taxonomy} tt ON t.term_id = tt.term_id WHERE tt.term_id IS NULL;" );
	}

	/**
	 * @return \NMGR_Sub
	 */
	public function sub() {
		return new NMGR_Sub();
	}

	/**
	 * @return \NMGR_Admin_Post
	 */
	public function admin_post() {
		return new NMGR_Admin_Post();
	}

	/**
	 * @return \NMGR_Ajax
	 */
	public function ajax() {
		return new NMGR_Ajax();
	}

	public function items_view( $wishlist = null ) {
		return new NMGR_Items_View( $wishlist );
	}

	/**
	 * @return \NMGR_Templates
	 */
	public function templates() {
		return new NMGR_Templates();
	}

	/**
	 * @return \NMGR_Wordpress
	 */
	public function wordpress() {
		return new NMGR_Wordpress();
	}

	/**
	 * @return \NMGR_Order
	 */
	public function order() {
		return new NMGR_Order();
	}

	/**
	 * @param type $item
	 * @return \NMGR_Wishlist_Item
	 */
	public function wishlist_item( $item = 0 ) {
		$class = apply_filters( 'nmgr_wishlist_item_class', NMGR_Wishlist_Item::class );
		return new $class( $item );
	}

	/**
	 * @param type $wishlist
	 * @return \NMGR_Wishlist
	 */
	public function wishlist( $wishlist = 0 ) {
		$class = apply_filters( 'nmgr_wishlist_class', NMGR_Wishlist::class );
		return new $class( $wishlist );
	}

	/**
	 * @param type $wishlist
	 * @return \NMGR_Account
	 */
	public function account( $wishlist = false ) {
		return new NMGR_Account( $wishlist );
	}

}
