<?php

/**
 * Sync
 */
defined( 'ABSPATH' ) || exit;

/**
 * Output the view related to the wishlist item
 */
class NMGR_Items_View {

	/**
	 * Arguments used to compose the view
	 * @var array
	 */
	protected $args = [];

	/**
	 * The current wishlist item in the view
	 * @var \NMGR_Wishlist_Item|NMGR\Sub\Wishlist_Item
	 */
	public $item;

	/**
	 * The current wishlist being viewed
	 * @var NMGR_Wishlist
	 */
	public $wishlist;

	/**
	 * The display properties
	 * E.g. display mode (list or grid), display row/columns
	 * @var array
	 */
	public $display = [];
	public $is_gift_registry = true;

	/**
	 * Inititialize the view for the wishlist items
	 *
	 * @param int|NMGR_Wishlist $wishlist Wishlist id or object
	 */
	public function __construct( $wishlist = null ) {
		$this->set_wishlist( $wishlist );

		if ( $this->get_wishlist() ) {
			$this->is_gift_registry = $this->get_wishlist()->is_type( 'gift-registry' );
		}
	}

	public function __get( $param ) {
		if ( 'parts' === $param ) {
			_deprecated_function( __CLASS__ . '->parts', '3.0.0', __CLASS__ . '->get_parts_data()' );
			return array_keys( $this->get_parts_data() );
		}
	}

	/**
	 * @param int|NMGR_Wishlist_Item|\NMGR\Sub\Wishlist_Item $item Item id or object
	 */
	public function set_item( $item ) {
		$this->item = is_a( $item, 'NMGR_Wishlist_Item' ) ? $item : nmgr_get_wishlist_item( $item );
	}

	/**
	 * @return NMGR_Wishlist_Item|\NMGR\Sub\Wishlist_Item
	 */
	public function get_item() {
		return $this->item;
	}

	/**
	 * @param int|NMGR_Wishlist $wishlist Wishlist id or object
	 */
	public function set_wishlist( $wishlist ) {
		$this->wishlist = is_a( $wishlist, 'NMGR_Wishlist' ) ? $wishlist : nmgr_get_wishlist( $wishlist );
	}

	/**
	 * @return NMGR_Wishlist
	 */
	public function get_wishlist() {
		return $this->wishlist;
	}

	/**
	 * @deprecated since 4.0.0
	 * @todo Remove in later version
	 */
	public function get_args() {
		return apply_filters( 'nmgr_items_view_args', $this->args, $this );
	}

	/**
	 * @deprecated since 4.0.0
	 * @todo Remove in later version
	 */
	public function set_args( $args ) {
		$this->args = $args;
	}

	public function is_gift_registry() {
		return $this->is_gift_registry;
	}

	/**
	 * Get the display properties for the view.
	 * @return array
	 */
	public function get_display() {
		if ( method_exists( $this, 'get_display_pro' ) ) {
			$display = $this->get_display_pro();
		} else {
			$display = [
				'mode' => 'list',
				'columns' => '',
				'row-gap' => '',
				'column-gap' => '',
				'toggle_mode' => false,
			];
		}

		if ( !is_admin() ) {
			$display = apply_filters( 'nmgr_items_display', $display, $this );
		}
		return $display;
	}

	public function get_type() {
		return $this->is_gift_registry() ? 'gift-registry' : 'wishlist';
	}

	/**
	 * Get the parts of an item that should be displayed in the items template
	 * @param boolean $filtered Whether the returned results should be filtered. Default true.
	 * @return array
	 */
	public function get_parts() {
		_deprecated_function( __METHOD__, '4.2.0' );
		return array_keys( $this->get_parts_data() );
	}

	public function get_parts_data( $all = false ) {
		$item = $this->get_item();
		$permitted = !$this->is_public();
		$is_gift_registry = $this->is_gift_registry();

		$quantity_label = nmgr()->is_pro ?
			__( 'Desired quantity', 'nm-gift-registry' ) :
			__( 'Desired Quantity', 'nm-gift-registry-lite' );

		$purchased_quantity_label = nmgr()->is_pro ?
			__( 'Purchased quantity', 'nm-gift-registry' ) :
			__( 'Purchased quantity', 'nm-gift-registry-lite' );

		$favourite_label = __( 'Favourite', 'nm-gift-registry' );

		$data = [
			'checkbox' => [
				'label' => nmgr()->is_pro ?
				__( 'Select item', 'nm-gift-registry' ) :
				__( 'Select item', 'nm-gift-registry-lite' ),
				'table_header_content' => '',
				'show_in_settings' => true,
				'content' => [ $this, 'get_item_checkbox' ],
				'priority' => 10,
				'content_container_attributes' => [
					'class' => [
						'nmgr-select',
						'list' == $this->get_display()[ 'mode' ] ? 'left-align' : '',
					],
				],
			],
			'thumbnail' => [
				'label' => nmgr()->is_pro ?
				__( 'Thumbnail', 'nm-gift-registry' ) :
				__( 'Thumbnail', 'nm-gift-registry-lite' ),
				'table_header_content' => '',
				'show_in_settings' => true,
				'content' => [ $this, 'get_item_thumbnail' ],
				'priority' => 20,
				'content_container_attributes' => [
					'class' => 'thumb',
				],
			],
			'title' => [
				'label' => nmgr()->is_pro ?
				__( 'Product', 'nm-gift-registry' ) :
				__( 'Product', 'nm-gift-registry-lite' ),
				'show_in_settings' => true,
				'content' => [ $this, 'get_item_title' ],
				'priority' => 30,
				'content_container_attributes' => [
					'class' => 'name', //keep legacy class
					'data-sort' => $item && $item->get_product() ? $item->get_product()->get_name() : '',
					'data-title' => nmgr()->is_pro ?
					__( 'Product', 'nm-gift-registry' ) :
					__( 'Product', 'nm-gift-registry-lite' ),
				],
			],
			'cost' => [
				'label' => nmgr()->is_pro ?
				__( 'Cost', 'nm-gift-registry' ) :
				__( 'Cost', 'nm-gift-registry-lite' ),
				'show_in_settings' => true,
				'content' => [ $this, 'get_item_cost' ],
				'priority' => 40,
				'content_container_attributes' => [
					'data-sort' => $item && $item->get_product() ? $item->get_product()->get_price() : '',
					'data-title' => nmgr()->is_pro ?
					__( 'Cost', 'nm-gift-registry' ) :
					__( 'Cost', 'nm-gift-registry-lite' ),
				],
			],
			'quantity' => [
				'label' => $quantity_label,
				'table_header_content' => '<span>' . nmgr_get_svg( array(
					'icon' => 'cart-empty',
					'size' => 1,
					'fill' => 'currentColor',
					'class' => 'nmgr-tip',
					'title' => $quantity_label
				) ) . '</span>',
				'show_in_settings' => $is_gift_registry,
				'show' => $is_gift_registry,
				'content' => [ $this, 'get_item_quantity' ],
				'priority' => 50,
				'content_container_attributes' => [
					'data-sort' => $item ? (
					($this->is_public() && $is_gift_registry &&
					nmgr_get_type_option( $this->get_type(), 'display_item_purchased_quantity' )) ?
					$item->get_unpurchased_quantity() : $item->get_quantity()
					) : '',
					'data-title' => nmgr()->is_pro ?
					__( 'Quantity', 'nm-gift-registry' ) :
					__( 'Quantity', 'nm-gift-registry-lite' ),
				],
			],
			'purchased_quantity' => [
				'label' => $purchased_quantity_label,
				'table_header_content' => '<span>' . nmgr_get_svg( array(
					'icon' => 'cart-full',
					'size' => 1,
					'fill' => 'currentColor',
					'class' => 'nmgr-tip',
					'title' => $purchased_quantity_label
				) ) . '</span>',
				'show_in_settings' => $is_gift_registry,
				'show' => $is_gift_registry && $permitted,
				'priority' => 60,
				'content' => [ $this, 'get_item_purchased_quantity' ],
				'content_container_attributes' => [
					'class' => 'purchased-quantity', // keep legacy class
					'data-sort' => $item ? $item->get_purchased_quantity() : '',
					'data-title' => nmgr()->is_pro ?
					__( 'Purchased quantity', 'nm-gift-registry' ) :
					__( 'Purchased quantity', 'nm-gift-registry-lite' ),
				],
			],
			'favourite' => [
				'label' => $favourite_label,
				'table_header_content' => '<span>' . nmgr_get_svg( array(
					'icon' => 'star-empty',
					'size' => 1,
					'fill' => 'currentColor',
					'class' => 'nmgr-tip nmgr-cursor-help',
					'title' => $favourite_label
				) ) . '</span>',
				'show_in_settings' => $is_gift_registry,
				'show' => $is_gift_registry,
				'priority' => 70,
				'content' => [ $this, 'get_item_favourite' ],
				'content_container_attributes' => [
					'data-sort' => $item && method_exists( $item, 'get_favourite' ) ? $item->get_favourite() : 0,
					'data-title' => $favourite_label,
				],
			],
			'total_cost' => [
				'label' => nmgr()->is_pro ?
				__( 'Total', 'nm-gift-registry' ) :
				__( 'Total', 'nm-gift-registry-lite' ),
				'show_in_settings' => $is_gift_registry,
				'show' => $is_gift_registry && $permitted,
				'priority' => 80,
				'content' => [ $this, 'get_item_total_cost' ],
				'content_container_attributes' => [
					'class' => 'total-cost', // keep legacy class
					'data-sort' => $item ? $item->get_total() : '',
					'data-title' => nmgr()->is_pro ?
					__( 'Total', 'nm-gift-registry' ) :
					__( 'Total', 'nm-gift-registry-lite' ),
				],
			],
			'add_to_cart_button' => [
				'label' => nmgr()->is_pro ?
				__( 'Add to cart', 'nm-gift-registry' ) :
				__( 'Add to cart', 'nm-gift-registry-lite' ),
				'table_header_content' => '',
				'show_in_settings' => true,
				'show' => !is_nmgr_admin() &&
				$this->get_wishlist() &&
				!$this->get_wishlist()->needs_shipping_address(),
				'priority' => 90,
				'content' => [ $this, 'get_item_add_to_cart_button' ],
				'content_container_attributes' => [
					'class' => [
						'actions', // keep legacy class
						'add_to_cart', // keep legacy class
					],
				],
			],
			'action_buttons' => [
				'label' => nmgr()->is_pro ?
				__( 'Actions', 'nm-gift-registry' ) :
				__( 'Actions', 'nm-gift-registry-lite' ),
				'table_header_content' => '<div>' . nmgr_get_svg( array(
					'icon' => 'gear',
					'size' => 1,
					'fill' => 'currentColor'
				) ) . '</div>',
				'priority' => 100,
				'show_in_settings' => true,
				'show' => $permitted,
				'content' => [ $this, 'get_item_action_buttons' ],
				'content_container_attributes' => [
					'class' => [
						'actions', // keep legacy class
						'edit-delete', // keep legacy class
					],
				],
			],
		];

		if ( !nmgr()->is_pro ) {
			unset( $data[ 'checkbox' ] );
			unset( $data[ 'favourite' ] );
		}

		$showing_data = apply_filters( 'nmgr_items_view_parts_data', $data, $this );

		if ( !$all ) {
			$showing_data = nmgr_get_elements_to_show( $showing_data );

			if ( $is_gift_registry && (is_nmgr_admin() || is_nmgr_wishlist()) ) {
				foreach ( $showing_data as $key => $part ) {
					if ( true === ( bool ) ($part[ 'show_in_settings' ] ?? false) &&
						!nmgr_get_type_option( $this->get_type(), "display_item_$key" ) ) {
						unset( $showing_data[ $key ] );
					}
				}
			}
		}

		uasort( $showing_data, 'nmgr_sort_by_priority' );
		return $showing_data;
	}

	/**
	 * Get the attributes for the view container ('list' or 'grid')
	 * @param boolean $formatted Whether to format the attributes
	 * @return string
	 */
	public function get_items_container_attributes( $formatted = true ) {
		$attributes = [
			'class' => [ 'nmgr-items-view', $this->get_display()[ 'mode' ] ]
		];

		if ( 'list' == $this->get_display()[ 'mode' ] ) {
			$attributes[ 'class' ][] = 'nmgr-table';
			$attributes[ 'class' ][] = 'nmgr-items-table';
		}

		$attr = apply_filters( 'nmgr_items_view_attributes', $attributes, $this );

		return $formatted ? nmgr_format_attributes( $attr ) : $attr;
	}

	/**
	 * Whether the item should be hidden from view
	 * (All items in the wishlist are shown by default)
	 * @return boolean
	 */
	public function item_is_disabled() {
		/**
		 * Don't show the item if we are on the frontend single wishlist page
		 * and it is fulfilled and should be hidden
		 */
		$disabled = false;
		if ( $this->is_gift_registry() && $this->is_public() ) {
			$disabled = ($this->get_item()->is_fulfilled() &&
				nmgr_get_type_option( $this->get_type(), 'hide_fulfilled_items' )) ||
				(method_exists( $this->get_item(), 'is_archived' ) &&
				$this->get_item()->is_archived() &&
				!$this->get_item()->is_fulfilled());
		}
		return apply_filters( 'nmgr_item_view_is_disabled', $disabled, $this );
	}

	public function get_item_statuses() {
		$statuses = [];
		$product = $this->get_item()->get_product();

		$statuses[ 'item-purchased' ] = [
			'label' => nmgr()->is_pro ?
			__( 'Purchased', 'nm-gift-registry' ) :
			__( 'Purchased', 'nm-gift-registry-lite' ),
			'show_notice' => true,
			'show' => $this->get_item()->is_purchased(),
		];

		$statuses[ 'item-fulfilled' ] = [
			'label' => nmgr()->is_pro ?
			__( 'Fulfilled', 'nm-gift-registry' ) :
			__( 'Fulfilled', 'nm-gift-registry-lite' ),
			'blur' => true,
			'show_notice' => true,
			'show' => $this->get_item()->is_fulfilled(),
			'show_in_add_to_cart_column' => true,
			'add_to_cart_column_attributes' => [
				'class' => 'nmgr-tip',
				'title' => apply_filters( 'nmgr_item_fulfilled_notice',
					sprintf(
						/* translators: %s : wishlist type title */
						nmgr()->is_pro ? __( 'This item has been bought for the %s owner.', 'nm-gift-registry' ) : __( 'This item has been bought for the %s owner.', 'nm-gift-registry-lite' ),
						nmgr_get_type_title( '', 0, $this->get_type() )
					)
				),
			],
		];

		$statuses[ 'item-archived' ] = [
			'label' => nmgr()->is_pro ?
			__( 'Archived', 'nm-gift-registry' ) :
			__( 'Archived', 'nm-gift-registry-lite' ),
			'blur' => true,
			'show' => method_exists( $this->get_item(), 'is_archived' ) && $this->get_item()->is_archived(),
			'show_notice' => true,
			'show_in_add_to_cart_column' => true,
		];

		$statuses[ 'out-of-stock' ] = [
			'label' => nmgr()->is_pro ?
			__( 'Out of stock', 'nm-gift-registry' ) :
			__( 'Out of stock', 'nm-gift-registry-lite' ),
			'blur' => true,
			'show' => !$product->is_in_stock(),
			'show_in_add_to_cart_column' => true,
		];

		$statuses[ 'not-purchasable' ] = [
			'label' => nmgr()->is_pro ?
			__( 'Not purchasable', 'nm-gift-registry' ) :
			__( 'Not purchasable', 'nm-gift-registry-lite' ),
			'blur' => !$product->is_type( 'external' ),
			'show_notice' => true,
			'show' => !$product->is_purchasable(),
			'show_in_add_to_cart_column' => true,
		];

		$fstatuses = apply_filters( 'nmgr_item_view_statuses', $statuses, $this );

		return nmgr_get_elements_to_show( $fstatuses );
	}

	public function get_item_statuses_for_disabling_add_to_cart() {
		return array_filter( $this->get_item_statuses(), function ( $status ) {
			return true === ($status[ 'show_in_add_to_cart_column' ] ?? null);
		} );
	}

	/**
	 * Get the attributes used for the item container html
	 * @param boolean $formatted Whether to format the attributes for html output
	 * @return array
	 */
	public function get_item_container_attributes( $formatted = true ) {
		$product = $this->get_item()->get_product();
		$item_statuses = $this->get_item_statuses();
		$classes = array_merge( array( 'item', $product->get_type() ), array_keys( $item_statuses ) );

		$attr = array(
			'class' => array_filter( $classes ),
			'data-product_title' => sanitize_text_field( $this->get_item()->get_product()->get_title() ),
			'data-wishlist_item_id' => absint( $this->get_item()->get_id() ),
			'data-wishlist_id' => absint( $this->get_wishlist()->get_id() ),
		);

		$attributes = apply_filters( 'nmgr_item_view_attributes', $attr, $this );
		return $formatted ? nmgr_format_attributes( $attributes ) : $attributes;
	}

	/**
	 * Replace the default html tag used for the template with the current tag
	 * @param string $template The html template
	 * @return string Template html
	 */
	protected function replace_tag( $template ) {
		_deprecated_function( __METHOD__, '3.0.0' );
		$tag = 'list' === $this->get_display()[ 'mode' ] ? 'td' : 'div';
		return str_replace( array( '<td', '</td' ), array( "<$tag", "</$tag" ), $template );
	}

	/**
	 * Get a template part for displaying an item in the items template
	 *
	 * @param string $part
	 * @return string Template html
	 */
	public function get_part( $part ) {
		_deprecated_function( __METHOD__, '3.0.0', __CLASS__ . '->get_item_part_content' );
		$html = is_callable( [ $this, "get_item_$part" ] ) ? $this->{"get_item_$part"}() : '';
		return $this->replace_tag(
				apply_filters_deprecated(
					'nmgr_item_view_part',
					[ $html, $part, $this ],
					'3.0.0',
					'nmgr_items_view_parts_data'
				)
		);
	}

	public function get_part_attributes( $part ) {
		$attributes = [
			'class' => [
				'nmgr-item-part',
				'nmgr_' . $part
			],
		];

		if ( 'action_buttons' !== $part ) {
			foreach ( $this->get_item_statuses() as $status ) {
				if ( true === ($status[ 'blur' ] ?? null) ) {
					$attributes[ 'class' ][] = 'nmgr-blur';
					break;
				}
			}
		}

		return $attributes;
	}

	/**
	 * Get the product image template
	 * @return string Template html
	 */
	public function get_item_thumbnail() {
		$product = $this->get_item()->get_product();
		$size = 'grid' == $this->get_display()[ 'mode' ] ? 'nmgr_medium' : 'nmgr_thumbnail';
		$image_size = apply_filters( 'nmgr_item_view_image_size', $size, $this );
		$thumbnail = $product ? $product->get_image( $image_size, array(
				'title' => $product->get_name(),
				'class' => 'nmgr-tip',
				'alt' => $product->get_name()
			) ) : '';

		return nmgr_get_template(
			'account/items/item-thumbnail.php',
			array(
				'thumbnail' => $thumbnail,
				'item' => $this->get_item(),
				'view' => $this,
			)
		);
	}

	/**
	 * Get the template for the item's title
	 * @return string Template html
	 */
	public function get_item_title() {
		$product = $this->get_item()->get_product();
		$product_link = is_nmgr_admin() ?
			get_edit_post_link( $product->get_id() ) :
			$product->get_permalink( $this->get_item()->get_data() );

		if ( !is_nmgr_admin() && $this->is_gift_registry() &&
			empty( $this->get_item_statuses_for_disabling_add_to_cart() ) ) {
			$product_link = add_query_arg( 'nmgr_item_id', $this->get_item()->get_id(), $product_link );
		}

		return nmgr_get_template( 'account/items/item-title.php', array(
			'item' => $this->get_item(),
			'product' => $product,
			'product_link' => $product_link,
			'view' => $this,
			) );
	}

	/**
	 * Get the template for the item's cost
	 * @return string Template html
	 */
	public function get_item_cost() {
		return nmgr_get_template( 'account/items/item-cost.php', array(
			'item' => $this->get_item(),
			'product' => $this->get_item()->get_product(),
			'view' => $this,
			) );
	}

	/**
	 * Get the template for the item's quantity
	 * @return string Template html
	 */
	public function get_item_quantity() {
		$product = $this->get_item()->get_product();
		$quantity = ($this->is_gift_registry() && $this->is_public() &&
			nmgr_get_type_option( $this->get_type(), 'display_item_purchased_quantity' )) ?
			$this->get_item()->get_unpurchased_quantity() :
			$this->get_item()->get_quantity();

		return nmgr_get_template( 'account/items/item-quantity.php', array(
			'item' => $this->get_item(),
			'product' => $product,
			'quantity' => $quantity,
			'view' => $this,
			) );
	}

	/**
	 * Get the template for the item's purchased quantity
	 * @return string Template html
	 */
	public function get_item_purchased_quantity() {
		return nmgr_get_template( 'account/items/item-purchased-quantity.php', array(
			'item' => $this->get_item(),
			'product' => $this->get_item()->get_product(),
			'view' => $this,
			) );
	}

	/**
	 * Get the template for the item's total cost
	 * @return string Template html
	 */
	public function get_item_total_cost() {
		return nmgr_get_template( 'account/items/item-total_cost.php', array(
			'item' => $this->get_item(),
			'view' => $this,
			) );
	}

	/**
	 * Get the template for the item's add to cart button
	 * @return string Template html
	 */
	public function get_item_add_to_cart_button() {
		$product = $this->get_item()->get_product();
		$type = $this->get_type();

		// Get the maximum quantity of the item that can be added to the cart
		$max_purchase_qty = $product->get_max_purchase_quantity();

		if ( 'gift-registry' === $type ) {
			if ( nmgr_get_type_option( $type, 'display_item_quantity' ) &&
				!nmgr_get_type_option( $type, 'display_item_purchased_quantity' ) ) {
				$desired_qty = $this->get_item()->get_quantity();
			} elseif ( nmgr_get_type_option( $type, 'display_item_quantity' ) &&
				nmgr_get_type_option( $type, 'display_item_purchased_quantity' ) ) {
				$desired_qty = $this->get_item()->get_unpurchased_quantity();
			} else {
				$desired_qty = $max_purchase_qty;
			}

			if ( $max_purchase_qty > 0 ) {
				if ( $max_purchase_qty > $desired_qty ) {
					$max_qty = $desired_qty;
				} else {
					$max_qty = $max_purchase_qty;
				}
			} else {
				$max_qty = $desired_qty;
			}
		} else {
			$max_qty = $max_purchase_qty;
		}

		$button_classes = [
			'nmgr_add_to_cart_button',
			'button',
			'alt'
		];

		if ( nmgr()->is_pro && method_exists( $this, 'get_pro_add_to_cart_button_classes' ) ) {
			$button_classes = array_merge( $button_classes, $this->get_pro_add_to_cart_button_classes( $type ) );
		} else {
			$button_classes[] = 'nmgr_ajax_add_to_cart';
		}

		return nmgr_get_template( 'account/items/item-actions-add_to_cart.php', array(
			'item' => $this->get_item(),
			'wishlist' => $this->get_wishlist(),
			'product' => $product,
			'max_qty' => $max_qty,
			'button_classes' => $button_classes,
			'view' => $this,
			) );
	}

	/**
	 * Get all the actions that should be performed on the specific item in the view
	 * @param boolean $all Whether to get all the actions without any conditions of display. Default false.
	 * @return array
	 */
	public function get_item_actions( $all = false ) {
		$item = $this->get_item();
		$item_id = $item ? $item->get_id() : 0;
		$is_gift_registry = $this->is_gift_registry();
		$item_has_archive_prop = $item && method_exists( $item, 'is_archived' );
		$item_is_archived = $item_has_archive_prop && $item->is_archived();
		$show_favourite = $is_gift_registry && nmgr()->is_pro &&
			nmgr_get_type_option( $this->get_type(), 'display_item_favourite' );
		$is_admin = is_nmgr_admin();

		$actions = [
			'delete' => [
				'text' => nmgr()->is_pro ?
				__( 'Delete', 'nm-gift-registry' ) :
				__( 'Delete', 'nm-gift-registry-lite' ),
				'priority' => 20,
				'attributes' => [
					'class' => [
						'delete-wishlist-item',
						'nmgr-post-action',
					],
					'href' => '#',
					'data-notice' => $item ? nmgr_get_delete_item_notice( $item ) : '',
					'data-nmgr_post_action' => 'delete_item',
					'data-wishlist_item_id' => $item_id,
				],
				'show' => $is_admin || (!$is_admin && $item && (!$item_is_archived && !$item->is_fulfilled() &&
				!$item->is_purchased() )),
				'show_in_bulk_actions' => true,
			],
			'purchase_refund' => [
				'text' => nmgr()->is_pro ?
				__( 'Update purchased quantity', 'nm-gift-registry' ) :
				__( 'Update purchased quantity', 'nm-gift-registry-lite' ),
				'priority' => 25,
				'attributes' => [
					'class' => [
						'purchase-refund-wishlist-item',
						'nmgr-post-action',
					],
					'href' => '#',
					'data-nmgr_post_action' => 'purchase_refund_item',
					'data-wishlist_item_id' => $item_id,
				],
				'show' => $is_gift_registry && $is_admin && !$item_is_archived,
			],
			'favourite' => [
				'text' => __( 'Toggle favourite status', 'nm-gift-registry' ),
				'priority' => 30,
				'attributes' => [
					'class' => [
						'favourite-wishlist-item',
						'nmgr-post-action',
					],
					'href' => '#',
					'data-nmgr_post_action' => 'toggle_item_favourite',
					'data-wishlist_item_id' => $item_id,
				],
				'show' => $is_gift_registry && $show_favourite && !$item_is_archived,
				'show_in_bulk_actions' => $is_gift_registry && $show_favourite,
			],
			'archive' => [
				'text' => __( 'Toggle archive status', 'nm-gift-registry' ),
				'priority' => 40,
				'attributes' => [
					'class' => [
						'archive-wishlist-item',
						'nmgr-post-action',
					],
					'data-archive_action' => 'archive',
					'data-nmgr_post_action' => 'toggle_item_archive',
					'data-wishlist_item_id' => $item_id,
					'href' => '#'
				],
				'show' => $is_gift_registry && $is_admin && $item_has_archive_prop,
				'show_in_bulk_actions' => $is_gift_registry && $is_admin,
			],
		];

		if ( $is_gift_registry &&
			is_callable( [ $this->get_wishlist(), 'is_archived' ] ) &&
			$this->get_wishlist()->is_archived() ) {
			$actions = [];
		} else {
			$actions = apply_filters( 'nmgr_item_actions', $actions, $this );

			if ( !$all ) {
				foreach ( $actions as $key => $args ) {
					if ( isset( $args[ 'show' ] ) && !$args[ 'show' ] ) {
						unset( $actions[ $key ] );
					} elseif ( !isset( $args[ 'show' ] ) && $is_gift_registry &&
						$item_is_archived && !($args[ 'show_if_archived' ] ?? false) ) {
						unset( $actions[ $key ] );
					}
				}
			}
		}

		uasort( $actions, 'nmgr_sort_by_priority' );
		return $actions;
	}

	public function get_item_dropdown() {
		$dropdown = nmgr_get_dropdown();

		foreach ( $this->get_item_actions() as $action_key => $action_args ) {
			if ( is_array( $action_args ) ) {
				$dropdown->set_menu_item(
					($action_args[ 'text' ] ?? '' ),
					($action_args[ 'attributes' ] ?? [] ),
					$action_key,
					($action_args[ 'tag' ] ?? 'a' )
				);
			}
		}

		$show_notices_statuses = array_filter( $this->get_item_statuses(), function ( $status ) {
			return true === ($status[ 'show_notice' ] ?? null);
		} );

		if ( !empty( $show_notices_statuses ) ) {
			if ( $dropdown->has_items() ) {
				$dropdown->set_menu_divider( [], 'notice_divider' );
			}

			$statuses_text = nmgr()->is_pro ?
				__( 'Item statuses', 'nm-gift-registry' ) :
				__( 'Item statuses', 'nm-gift-registry-lite' );

			$dropdown->set_menu_header( $statuses_text, [], 'item_status_header_text' );

			foreach ( $show_notices_statuses as $status_key => $status ) {
				$dropdown->set_menu_text(
					($status[ 'label' ] ?? null ),
					[ 'class' => 'nmgr-view-item-notice' ],
					$status_key,
					'span'
				);
			}
		}

		return $dropdown->has_items() ? $dropdown->get() : '';
	}

	/**
	 * Get the template for the item's edit and delete buttons
	 * @return string Template html
	 */
	public function get_item_action_buttons() {
		return nmgr_get_template( 'account/items/item-actions-edit-delete.php', array(
			'dropdown' => $this->get_item_dropdown(),
			'item' => $this->get_item(),
			'items_args' => $this->get_args(),
			'view' => $this,
			) );
	}

	public function get_table_header( $part ) {
		$part_data = $this->get_parts_data()[ $part ];
		$header = $part_data[ 'table_header_content' ] ?? ($part_data[ 'label' ] ?? '');

		if ( has_action( "nmgr_items_table_header_$part" ) ) {
			ob_start();
			do_action_deprecated(
				"nmgr_items_table_header_$part",
				[ $this ],
				'3.0.0',
				'nmgr_items_view_parts_data'
			);
			$header = ob_get_clean();
		}

		$attributes = [ 'class' => [ "item_$part" ] ];

		if ( !empty( $part_data[ 'table_header_attributes' ] ?? false  ) ) {
			$attributes = nmgr()->merge_args( $attributes, ( array ) $part_data[ 'table_header_attributes' ] );
		}

		$content = apply_filters_deprecated(
			'nmgr_items_table_header_content',
			[ $header, $part, $this ],
			'3.0.0',
			'nmgr_items_view_parts_data'
		);

		return sprintf(
			'<th %1$s>%2$s</th>',
			nmgr_format_attributes( $attributes ),
			$content
		);
	}

	public function get_item_part_content( $part ) {
		$part_data = $this->get_parts_data()[ $part ];
		$content = $part_data[ 'content' ] ?? '';

		if ( is_callable( $content ) ) {
			$content = call_user_func( $content, $this );
		}

		// Strip opening and closing td tags for backwards compatibility
		$raw = preg_replace( '/<td[^>]*?>(.*?)<\/td[^>]*>$/si', '$1', trim( $content ) );

		$attributes = $this->get_part_attributes( $part );

		if ( !empty( $part_data[ 'content_container_attributes' ] ?? false  ) ) {
			$attributes = nmgr()->merge_args( $attributes, ( array ) $part_data[ 'content_container_attributes' ] );
		}

		return $this->wrap_with_container( $raw, $attributes );
	}

	protected function wrap_with_container( $content, $attributes = '' ) {
		$tag = 'list' === $this->get_display()[ 'mode' ] ? 'td' : 'div';
		return sprintf(
			'<%1$s %2$s>%3$s</%4$s>',
			$tag,
			nmgr_format_attributes( $attributes ),
			$content,
			$tag
		);
	}

	/**
	 * Check if we are on the wishlist page for public view.
	 * This is the view that is not for the wishlist owner who has the wishlist.
	 * @return bool
	 */
	public function is_public() {
		return false === (is_nmgr_admin() || nmgr_user_has_wishlist( $this->get_wishlist() ));
	}

}
