<?php

/**
 * Sync
 */
defined( 'ABSPATH' ) || exit;

class NMGR_Wishlist_Item extends NMGR_Data {

	/**
	 * Wishlist item data stored in nmgr_wishlist_items table
	 *
	 * @var array
	 */
	protected $core_data = array(
		// General wishlist id containing many items
		'wishlist_id' => 0,
		// product id
		'product_id' => 0,
		// Date the item was added to the wishlist
		'date_created' => '',
		// Date the item was last updated
		'date_modified' => '',
	);

	/**
	 * Wishlist item meta data stored in nmgr_wishlist_itemmeta table
	 *
	 * Internal meta keys for wishlist item
	 *
	 * @var array
	 */
	protected $meta_data = array(
		'product_id' => 0,
		'variation_id' => 0,
		'variation' => array(),
		/**
		 * Quantity of the item in the wishlist
		 */
		'quantity' => 1,
		/**
		 * Quantity of the item purchased from the wishlist
		 */
		'purchased_quantity' => 0,
		/**
		 * The unique id of the item
		 * This id is generated based on the item properties such as product id, variation id and wishlist id
		 * as such it provides unique identification for the item in a wishlist and among all wishlists
		 */
		'unique_id' => '',
		/**
		 * Quantity reference of item for all orders
		 *
		 * This is an array of the ordered, refunded and original purchased quantity (if set) for the item,
		 * stored in the database with the order_id as array key
		 *
		 * This reference is used to calculate/update the purchased quantity of the item
		 *
		 * An item would have quantity references for all the orders which have certain quantities of the item.
		 * Each of these orders is accessed in the quantity reference array using the order id as array key.
		 */
		'quantity_reference' => array(),
		/**
		 * The method for purchasing a particular quantity of the item.
		 * this should typically be as a woocommerce order.
		 */
		'purchase_log' => [],
	);

	/**
	 * Meta type.
	 *
	 * @see https://developer.wordpress.org/reference/functions/add_metadata/.
	 * @var string
	 */
	public $meta_type = 'wishlist_item';

	/**
	 * Name of the object type
	 *
	 * @var string
	 */
	protected $object_type = 'wishlist_item';

	/**
	 * Constructor.
	 *
	 * @param int|object $item ID to load from the DB, or NMGR_Wishlist_Item object.
	 */
	public function __construct( $item = 0 ) {
		// $item would be an object with 'wishlist_item_id' if read from database
		if ( is_object( $item ) && !empty( $item->wishlist_item_id ) ) {
			$this->set_id( absint( $item->wishlist_item_id ) );
		}

		if ( !empty( $this->pro_meta_data ) ) {
			$this->meta_data = array_merge( $this->pro_meta_data, $this->meta_data );
		}

		parent::__construct( $item );

		$this->db = new NMGR_Database_Wishlist_Item();

		if ( $this->get_id() > 0 ) {
			$this->db->read( $this );
		}
	}

	/*
	  |--------------------------------------------------------------------------
	  | Getters
	  |--------------------------------------------------------------------------
	 */

	/**
	 * Get the name of the wishlist item
	 *
	 * @deprecated since version 2.4
	 * @return string
	 */
	public function get_name() {
		_deprecated_function( __FUNCTION__, '2.4', 'NMGR_Wishlist_Item->get_product()->get_name()' );
		return $this->get_product()->get_name();
	}

	/**
	 * Get the id of the wishlist the item belongs to
	 *
	 * @return int
	 */
	public function get_wishlist_id() {
		return absint( $this->get_prop( 'wishlist_id' ) );
	}

	/**
	 * Get the date the item was added to the wishlist
	 *
	 * @return Timestamp
	 */
	public function get_date_created() {
		return $this->get_prop( 'date_created' );
	}

	/**
	 * Get the date the item was last updated in the wishlist
	 *
	 * @return Timestamp
	 */
	public function get_date_modified() {
		return $this->get_prop( 'date_modified' );
	}

	/**
	 * Get the id of the product this item represents
	 *
	 * @return int
	 */
	public function get_product_id() {
		return absint( $this->get_prop( 'product_id' ) );
	}

	/**
	 * Get the id of the product variation this item represents
	 *
	 * @return int
	 */
	public function get_variation_id() {
		return absint( $this->get_prop( 'variation_id' ) );
	}

	/**
	 * Get the variation of the product this item represents
	 *
	 * @return array
	 */
	public function get_variation() {
		return array_filter( ( array ) $this->get_prop( 'variation' ) );
	}

	/**
	 * Get the quantity of this item in the wishlist
	 *
	 * @return int
	 */
	public function get_quantity() {
		return wc_stock_amount( $this->get_prop( 'quantity' ) );
	}

	/**
	 * Get the purchased quantity of this item in the wishlist
	 *
	 * @return int
	 */
	public function get_purchased_quantity() {
		return ( int ) apply_filters( 'nmgr_item_purchased_quantity', $this->get_prop( 'purchased_quantity' ), $this );
	}

	/**
	 * Get the unpurchased quantity of the item
	 * This only works if the quantity and purchased quantity columns are visible on the items table
	 *
	 * @return int
	 */
	public function get_unpurchased_quantity() {
		$quantity = nmgr_get_type_option( $this->get_wishlist()->get_type(), 'display_item_quantity' ) ?
			$this->get_quantity() : 0;
		return max( $quantity - $this->get_purchased_quantity(), 0 );
	}

	/**
	 * Get the unique id of this item
	 *
	 * @return string
	 */
	public function get_unique_id() {
		return $this->get_prop( 'unique_id' );
	}

	/**
	 * Get the quantity reference of this item
	 *
	 * @return array
	 */
	public function get_quantity_reference() {
		return ( array ) $this->get_prop( 'quantity_reference' );
	}

	/**
	 * Get the product this item represents
	 *
	 * @return WC_Product
	 */
	public function get_product() {
		if ( $this->get_variation_id() ) {
			$product = wc_get_product( $this->get_variation_id() );
		} else {
			$product = wc_get_product( $this->get_product_id() );
		}
		return $product;
	}

	/**
	 * Get the wishlist this item belongs to
	 *
	 * @return NMGR_Wishlist
	 */
	public function get_wishlist() {
		return nmgr_get_wishlist( $this->get_wishlist_id() );
	}

	/**
	 * Get the total cost of the wishlist item (cost of product x qty)
	 *
	 * @param bool $currency_symbol Whether to return the value formatted with the currency symbol
	 * @return string
	 */
	public function get_total( $currency_symbol = false ) {
		$product = $this->get_product();
		if ( $product ) {
			$total = ( float ) apply_filters( 'nmgr_item_total', wc_get_price_to_display( $product, array( 'qty' => $this->get_quantity() ) ), $this );
			return $currency_symbol ? wc_price( $total, array( 'currency' => get_woocommerce_currency() ) ) : $total;
		}
		return 0;
	}

	/**
	 * Get variations that are not shown in the item title
	 * This is because variation titles display the attributes
	 *
	 * @return array Array of variation name, value pairs
	 */
	public function get_variations_for_display() {
		if ( empty( $this->get_variation() ) ) {
			return;
		}

		$variations = array();

		foreach ( $this->get_variation() as $name => $value ) {
			$taxonomy = wc_attribute_taxonomy_name( str_replace( 'attribute_pa_', '', urldecode( $name ) ) );

			if ( taxonomy_exists( $taxonomy ) ) {
				// If this is a term slug, get the term's nice name.
				$term = get_term_by( 'slug', $value, $taxonomy );
				if ( !is_wp_error( $term ) && $term && $term->name ) {
					$value = $term->name;
				}
				$label = wc_attribute_label( $taxonomy );
			} else {
				// If this is a custom option slug, get the options name.
				$label = wc_attribute_label( str_replace( 'attribute_', '', $name ), $this->get_product() );
			}

			// Check the nicename against the title.
			if ( '' === $value || wc_is_attribute_in_product_name( $value, $this->get_product()->get_name() ) ) {
				continue;
			}

			$variations[] = array(
				'key' => $label,
				'value' => $value,
			);
		}

		return $variations;
	}

	/**
	 * Get the method used to purchase quantities of the item
	 * @return array
	 */
	public function get_purchase_log() {
		return $this->get_prop( 'purchase_log' );
	}

	/**
	 * Get the default purchase log properties
	 * @return array
	 */
	public function get_purchase_log_props() {
		return apply_filters(
			'nmgr_item_purchase_log_props',
			[
				'id' => nmgr()->is_pro ?
				__( 'ID', 'nm-gift-registry' ) :
				__( 'ID', 'nm-gift-registry-lite' ),
				'type' => nmgr()->is_pro ?
				__( 'Type', 'nm-gift-registry' ) :
				__( 'Type', 'nm-gift-registry-lite' ),
				'quantity' => nmgr()->is_pro ?
				__( 'Quantity', 'nm-gift-registry' ) :
				__( 'Quantity', 'nm-gift-registry-lite' ),
				'user_id' => nmgr()->is_pro ?
				__( 'User ID', 'nm-gift-registry' ) :
				__( 'User ID', 'nm-gift-registry-lite' ),
				'route' => nmgr()->is_pro ?
				__( 'Route', 'nm-gift-registry' ) :
				__( 'Route', 'nm-gift-registry-lite' ),
				'date' => nmgr()->is_pro ?
				__( 'Date', 'nm-gift-registry' ) :
				__( 'Date', 'nm-gift-registry-lite' ),
			],
			$this
		);
	}

	/**
	 * Add a purchase method for an item when it's purchased quantity has been updated
	 * to log details about the particular update
	 *
	 * @param int $quantity The quantity of the item reflecting the current purchase.
	 * This number should be negative if the purchase is a refund and positive if it
	 * is indeed a purchase.
	 * @param type $method_type The type of purchase. Default values are 'order' and 'manual'
	 * to determine whether the purchased quantity was updated via an order or manually via
	 * the wishlist items table
	 * @param array $args Extra data to add to the purchase method log. For example 'order_id'.
	 */
	public function add_purchase_log( $quantity, $method_type = 'order', $args = [] ) {
		$original = $this->get_purchase_log();
		$id = $this->get_wishlist_id() . '_' . $this->get_id() . '_' . time();
		$default = [
			'id' => $id,
			'type' => $method_type,
			'quantity' => ( int ) $quantity,
			'user_id' => get_current_user_id(),
			'route' => is_nmgr_admin() ? 'admin' : 'frontend',
			'date' => current_time( 'mysql' ),
		];

		$fargs = array_merge( $default, $args );
		$original[ $id ] = $fargs;
		$this->set_purchase_log( $original );
	}

	public function set_purchase_log( $value ) {
		$this->set_prop( 'purchase_log', $value );
	}

	/**
	 * Get the amount purchased for the item from orders
	 *
	 * This is taken from the actual price of the item in each order
	 * typical made from the checkout page.
	 * @return int|float
	 */
	public function get_purchased_amount() {
		$order_item_ids = $this->get_order_item_ids();
		$tax_display = 'incl' === get_option( 'woocommerce_tax_display_shop' );
		$inc_tax = apply_filters( 'nmgr_item_purchased_amount_from-order_inc_tax', $tax_display, $this );
		$order_amounts = 0;

		foreach ( $order_item_ids as $order_item_id ) {
			$order_item = \WC_Order_Factory::get_order_item( $order_item_id );
			if ( $order_item ) {
				$order = $order_item->get_order();
				$order_amounts += $order->get_line_total( $order_item, $inc_tax );
			}
		}

		/**
		 * We're already rounding the value of the item in the order using $order->get_line_total
		 * However we round here again to fix a bug that makes some results return unrounded when
		 * this function is called via select2 ajax
		 */
		$amt = round( $order_amounts, wc_get_price_decimals() );
		return ( float ) apply_filters( 'nmgr_item_purchased_amount', $amt, $this );
	}

	/**
	 * Get the total amount purchased for the item.
	 * This may include amount purchased from orders and other methods such as crowdfunds.
	 * @return float
	 */
	public function get_total_purchased_amount() {
		return ( float ) apply_filters( 'nmgr_item_total_purchased_amount', $this->get_purchased_amount(), $this );
	}

	/**
	 * Get the amount left to be purchased for the item
	 * @return int|float
	 */
	public function get_unpurchased_amount() {
		return $this->get_total() - $this->get_purchased_amount();
	}

	/**
	 * Get the total amount left to be purchased for the item
	 * This may include amount unpurchased from orders and other methods such as crowdfunds.
	 * @return int|float
	 */
	public function get_total_unpurchased_amount() {
		return $this->get_total() - $this->get_total_purchased_amount();
	}

	/**
	 * Get the data that would be used to identify the item in the cart (as cart item data)
	 * or in the order (as order item metadata)
	 *
	 * Wishlist information gotten:
	 * - wishlist_id
	 * - wishlist_item_id
	 * - product_id
	 * - variation_id
	 * - type
	 *
	 * @return array
	 */
	public function get_cart_order_data() {
		return array(
			'wishlist_id' => ( int ) $this->get_wishlist_id(),
			'wishlist_item_id' => ( int ) $this->get_id(),
			'product_id' => $this->get_product_id(),
			'variation_id' => $this->get_variation_id(),
			'type' => 'wishlist_item',
		);
	}

	/**
	 * Get the ids of all the order items that have this item's product
	 */
	public function get_order_item_ids() {
		global $wpdb;

		$order_item_ids = wp_cache_get( 'nmgr_item_order_item_ids_' . $this->get_id() );

		if ( false === $order_item_ids ) {
			$order_item_ids = [];

			$results = $wpdb->get_col( $wpdb->prepare( "
		SELECT DISTINCT order_itemmeta.order_item_id FROM {$wpdb->prefix}woocommerce_order_itemmeta AS order_itemmeta
		INNER JOIN {$wpdb->prefix}wc_order_product_lookup AS product_lookup
			ON product_lookup.order_item_id = order_itemmeta.order_item_id
		INNER JOIN {$wpdb->prefix}nmgr_wishlist_items AS wishlist_items
			ON (product_lookup.product_id = wishlist_items.product_id)
		WHERE wishlist_items.wishlist_item_id = %d
		AND order_itemmeta.meta_key = 'nm_gift_registry'
		",
					$this->get_id()
				) );

			if ( !empty( $results ) ) {
				foreach ( $results as $result_id ) {
					$item = \WC_Order_Factory::get_order_item( $result_id );
					if ( $item ) {
						$meta = $item->get_meta( 'nm_gift_registry' );
						if ( $this->get_id() === ( int ) ($meta[ 'wishlist_item_id' ] ?? null) ) {
							$order_item_ids[] = $result_id;
						}
					}
				}

				wp_cache_set( 'nmgr_item_order_item_ids_' . $this->get_id(), $order_item_ids );
			}
		}

		return $order_item_ids;
	}

	/*
	  |--------------------------------------------------------------------------
	  | Setters
	  |--------------------------------------------------------------------------
	 */

	/**
	 * Set wishlist ID.
	 *
	 * @param int $value Wishlist ID.
	 */
	public function set_wishlist_id( $value ) {
		$this->set_prop( 'wishlist_id', absint( $value ) );
	}

	/**
	 * Set item name.
	 *
	 * @param string $value Item name.
	 * @deprecated since version 2.4
	 */
	public function set_name( $value ) {
		_deprecated_function( __FUNCTION__, '2.4' );
	}

	/**
	 * Set item desired quantity
	 *
	 * @param int $value Desired quantity
	 */
	public function set_quantity( $value ) {
		$this->set_prop( 'quantity', wc_stock_amount( $value ) );
	}

	/**
	 * Set item purchased quantity
	 *
	 * @param int $value purchased quantity
	 */
	public function set_purchased_quantity( $value ) {
		$this->set_prop( 'purchased_quantity', absint( $value ) );
	}

	/**
	 * Set item product id
	 *
	 * @param int $value Product id.
	 */
	public function set_product_id( $value ) {
		$this->set_prop( 'product_id', absint( $value ) );
	}

	/**
	 * Set item variation id
	 * @param int $value Product Id/Variation id
	 */
	public function set_variation_id( $value ) {
		$this->set_prop( 'variation_id', absint( $value ) );
	}

	/**
	 * Set the item variation
	 * @param array Product variation
	 */
	public function set_variation( $value ) {
		$this->set_prop( 'variation', $value );
	}

	/**
	 * Set all product details for item at once based on the product the item represents
	 *
	 * This sets the product id, variation id and variation
	 *
	 * @param WC_Product $product Product the item represents
	 */
	public function set_product( $product ) {
		if ( $product->is_type( 'variation' ) ) {
			$this->set_product_id( $product->get_parent_id() );
			$this->set_variation_id( $product->get_id() );
			$this->set_variation( is_callable( array( $product, 'get_variation_attributes' ) ) ? $product->get_variation_attributes() : array()  );
		} else {
			$this->set_product_id( $product->get_id() );
		}
	}

	/**
	 * Set the unique id for the item
	 * @param string $value unique id
	 */
	public function set_unique_id( $value ) {
		$this->set_prop( 'unique_id', $value );
	}

	/**
	 * Set the quantity reference for the item
	 * @param array $value Quantity reference
	 */
	public function set_quantity_reference( $value ) {
		$this->set_prop( 'quantity_reference', $value );
	}

	/**
	 * Add the data for this wishlist item as order item meta
	 * Make sure to save the order item ($order_item) afterwards as this
	 * function doesn't save it.
	 * @param WC_Order_Item $order_item
	 */
	public function add_order_item_meta( $order_item ) {
		$order_item->add_meta_data( 'nm_gift_registry', $this->get_cart_order_data() );
		$order_item->add_meta_data( 'nmgr_wishlist_title', $this->get_wishlist()->get_title() );
	}

	/*
	  |--------------------------------------------------------------------------
	  | Conditionals
	  |--------------------------------------------------------------------------
	 */

	/**
	 * Get whether any quantity of this item has been purchased
	 *
	 * This is only possible if the 'purchased quantity' column is visible on the items table
	 * as it is the column used to determine that item purchased would be accounted for
	 *
	 * @return boolean True or false
	 */
	public function is_purchased() {
		$purchased = false;
		if ( nmgr_get_type_option( $this->get_wishlist()->get_type(), 'display_item_purchased_quantity' ) ) {
			$purchased = ( bool ) $this->get_purchased_quantity();
		}
		return ( bool ) apply_filters( 'nmgr_item_is_purchased', $purchased, $this );
	}

	/**
	 * Get whether the desired quantity of this item has been completely purchased
	 *
	 * This is typically only possible if the 'quantity' and 'purchased_quantity' columns are visible on the items table
	 *
	 * @return boolean
	 */
	public function is_fulfilled() {
		$fulfilled = false;
		if ( nmgr_get_type_option( $this->get_wishlist()->get_type(), 'display_item_quantity' ) &&
			nmgr_get_type_option( $this->get_wishlist()->get_type(), 'display_item_purchased_quantity' ) ) {
			$fulfilled = ( bool ) 0 >= $this->get_unpurchased_quantity();
		}
		return ( bool ) apply_filters( 'nmgr_item_is_fulfilled', $fulfilled, $this );
	}

	public function clear_cache() {
		$this->db->clear_cache( $this->get_id() );
	}

	public function read_meta_data() {
		$props_to_meta_keys = $this->get_meta_keys( $this->get_meta_data() );
		foreach ( $props_to_meta_keys as $prop => $meta_key ) {
			if ( metadata_exists( $this->meta_type, $this->get_id(), $meta_key ) ) {
				$val = get_metadata( $this->meta_type, $this->get_id(), $meta_key, true );
			} else {
				$val = $this->get_default_data()[ $prop ] ?? '';
			}
			$this->set_prop( $prop, $val );
		}
	}

}
