<?php

/**
 * Sync
 */
defined( 'ABSPATH' ) || exit;

class NMGR_Wishlist extends NMGR_Data {

	/**
	 * Wishlist data stored in wp posts table
	 *
	 * @var array
	 */
	protected $core_data = array(
		'post_title' => '',
		'post_status' => 'publish',
		'post_type' => 'nm_gift_registry',
		'post_excerpt' => '',
		'post_name' => '',
		'post_date' => '',
		'post_author' => 0,
	);

	/**
	 * Wishlist meta data stored in post meta table
	 *
	 * Internal meta keys for wishlist
	 *
	 * @var array
	 */
	protected $meta_data = array(
		'_first_name' => '',
		'_last_name' => '',
		'_partner_first_name' => '',
		'_partner_last_name' => '',
		'_email' => '',
		'_event_date' => '',
		'shipping_first_name' => '',
		'shipping_last_name' => '',
		'shipping_company' => '',
		'shipping_address_1' => '',
		'shipping_address_2' => '',
		'shipping_city' => '',
		'shipping_postcode' => '',
		'shipping_country' => '',
		'shipping_state' => '',
		'_date_fulfilled' => null,
		'_nmgr_user_id' => 0,
		'_nmgr_guest' => 0,
		'_nmgr_expired' => 0,
	);

	/**
	 * Properties that have sub properties
	 *
	 * @var array
	 */
	protected $parent_props = array(
		'shipping'
	);
	public $items = array();
	protected $object_type = 'wishlist';
	public $meta_type = 'post';
	protected $type = null;
	protected $items_cache_keys = [];

	/**
	 * Get the wishlist if ID is passed, otherwise the wishlist is new and empty.
	 *
	 * @param  int|object|NMGR_Wishlist $wishlist Wishlist to read.
	 */
	public function __construct( $wishlist = 0 ) {
		if ( !empty( $wishlist->ID ) ) {
			$this->set_id( absint( $wishlist->ID ) );
		}

		if ( !empty( $this->pro_core_data ) ) {
			$this->core_data = array_merge( $this->pro_core_data, $this->core_data );
		}

		if ( !empty( $this->pro_meta_data ) ) {
			$this->meta_data = array_merge( $this->pro_meta_data, $this->meta_data );
		}

		parent::__construct( $wishlist );

		$this->db = new NMGR_Database_Wishlist();

		if ( $this->get_id() > 0 ) {
			$this->db->read( $this );
		}
	}

	/*
	  |--------------------------------------------------------------------------
	  | Getters
	  |--------------------------------------------------------------------------
	 */

	/**
	 * Get all data for this wishlist including wishlist items
	 *
	 * @param bool $items Whether to get the wishlist items with the data. Default false
	 * @return array Wishlist Data
	 */
	public function get_data( $items = false ) {
		$data = parent::get_data();

		if ( $items ) {
			$items_data = array_map( function( $obj ) {
				$d = nmgr_get_wishlist_item( $obj );
				return $d ? $d->get_data() : array();
			}, $this->get_items() );

			$data = array_merge( $data, array( 'items' => $items_data ) );
		}

		return apply_filters( 'nmgr_get_wishlist_data', $data, $this );
	}

	/**
	 * Get the title of the wishlist
	 *
	 * @return string
	 */
	public function get_title() {
		return $this->get_prop( 'post_title' );
	}

	/**
	 * Get the post status of the wishlist (e.g. publish, draft)
	 *
	 * @return string
	 */
	public function get_status() {
		return $this->get_prop( 'post_status' );
	}

	/**
	 * Get the first name of the wishlist owner
	 *
	 * @return string
	 */
	public function get_first_name() {
		return $this->get_prop( '_first_name' );
	}

	/**
	 * Get the last name of the wishlist owner
	 *
	 * @return string
	 */
	public function get_last_name() {
		return $this->get_prop( '_last_name' );
	}

	/**
	 * Get the first name and last name of the wishlist owner
	 *
	 * @return string
	 */
	public function get_full_name() {
		return trim( sprintf( '%1$s %2$s', $this->get_first_name(), $this->get_last_name() ) );
	}

	/**
	 * Get the first name of the wishlist owner's partner
	 *
	 * @return string
	 */
	public function get_partner_first_name() {
		return $this->get_prop( '_partner_first_name' );
	}

	/**
	 * Get the last name of the wishlist owner's partner
	 *
	 * @return string
	 */
	public function get_partner_last_name() {
		return $this->get_prop( '_partner_last_name' );
	}

	/**
	 * Get the first name and last name of the wishlist owner's partner
	 *
	 * @return string
	 */
	public function get_partner_full_name() {
		return trim( sprintf( '%1$s %2$s', $this->get_partner_first_name(), $this->get_partner_last_name() ) );
	}

	/**
	 * Get the display name for the wishlist
	 * This is the combination of the names of the wishlist owner and wishlist owner's partner if available
	 *
	 * @return string
	 */
	public function get_display_name() {
		$display_name = '';
		if ( $this->get_full_name() && $this->get_partner_full_name() ) {
			$display_name = "{$this->get_full_name()} &amp; {$this->get_partner_full_name()}";
		} elseif ( $this->get_full_name() ) {
			$display_name = $this->get_full_name();
		}
		return $display_name;
	}

	/**
	 * Get the registered email for the wishlist
	 *
	 * @return string
	 */
	public function get_email() {
		return $this->get_prop( '_email' );
	}

	/**
	 * Get the date for the wishlist event
	 *
	 * @return string
	 */
	public function get_event_date() {
		return $this->get_prop( '_event_date' );
	}

	/**
	 * Get the wishlist description
	 *
	 * @return string
	 */
	public function get_description() {
		return $this->get_prop( 'post_excerpt' );
	}

	/**
	 * Get all shipping fields
	 *
	 * @return array
	 */
	public function get_shipping() {
		$shipping_keys = array();
		$shipping = array();

		foreach ( array_keys( $this->get_meta_data() ) as $key ) {
			if ( false !== strpos( $key, 'shipping_' ) ) {
				$shipping_keys[] = $key;
			}
		}

		foreach ( $shipping_keys as $key ) {
			$u_key = str_replace( 'shipping_', '', $key );
			if ( is_callable( array( $this, "get_$key" ) ) ) {
				$shipping[ $u_key ] = $this->{"get_$key"}();
			} else {
				$shipping[ $u_key ] = $this->get_prop( $key );
			}
		}

		return $shipping;
	}

	/**
	 * Get shipping first name.
	 *
	 * @return string
	 */
	public function get_shipping_first_name() {
		return $this->get_prop( 'shipping_first_name' );
	}

	/**
	 * Get shipping_last_name.
	 *
	 * @return string
	 */
	public function get_shipping_last_name() {
		return $this->get_prop( 'shipping_last_name' );
	}

	/**
	 * Get shipping company.
	 *
	 * @return string
	 */
	public function get_shipping_company() {
		return $this->get_prop( 'shipping_company' );
	}

	/**
	 * Get shipping address line 1
	 *
	 * @return string
	 */
	public function get_shipping_address() {
		return $this->get_prop( 'shipping_address_1' );
	}

	/**
	 * Get shipping address line 1.
	 *
	 * @return string
	 */
	public function get_shipping_address_1() {
		return $this->get_prop( 'shipping_address_1' );
	}

	/**
	 * Get shipping address line 2.
	 *
	 * @return string
	 */
	public function get_shipping_address_2() {
		return $this->get_prop( 'shipping_address_2' );
	}

	/**
	 * Get shipping city.
	 *
	 * @return string
	 */
	public function get_shipping_city() {
		return $this->get_prop( 'shipping_city' );
	}

	/**
	 * Get shipping state.
	 *
	 * @return string
	 */
	public function get_shipping_state() {
		return $this->get_prop( 'shipping_state' );
	}

	/**
	 * Get shipping postcode.
	 *
	 * @return string
	 */
	public function get_shipping_postcode() {
		return $this->get_prop( 'shipping_postcode' );
	}

	/**
	 * Get shipping country.
	 *
	 * @return string
	 */
	public function get_shipping_country() {
		return $this->get_prop( 'shipping_country' );
	}

	/**
	 * Get the date the wishlist was fulfilled
	 *
	 * This is the date all items in the wishlist were marked as purchased
	 *
	 * @return DateTime object
	 */
	public function get_date_fulfilled() {
		return $this->get_prop( '_date_fulfilled' );
	}

	/**
	 * Get the total price of all items in the wishlist
	 *
	 * @param boolean $currency_symbol Whether to prefix the returned amount with the currency symbol
	 * @return string|int|float
	 */
	public function get_total( $currency_symbol = false ) {
		$total = 0;
		foreach ( $this->get_items() as $item ) {
			$total += $item->get_total();
		}
		return $currency_symbol ? wc_price( $total, array( 'currency' => get_woocommerce_currency() ) ) : $total;
	}

	/**
	 * Get the amount purchased for all items from orders
	 * @return float|int
	 */
	public function get_purchased_amount() {
		$total = 0;
		foreach ( $this->get_items() as $item ) {
			$total += $item->get_purchased_amount();
		}
		return $total;
	}

	/**
	 * Get the amount left to be purchased for all items from orders
	 * @return int|float
	 */
	public function get_unpurchased_amount() {
		return $this->get_total() - $this->get_purchased_amount();
	}

	/**
	 * Get the total amount purchased from the wishlist
	 * @return int|float
	 */
	public function get_total_purchased_amount() {
		$total = 0;
		foreach ( $this->get_items() as $item ) {
			$total += $item->get_total_purchased_amount();
		}
		return ( float ) apply_filters( 'nmgr_wishlist_total_purchased_amount', $total, $this );
	}

	/**
	 * Get the total amount left to be purchased for the wishlist
	 * @return int|float
	 */
	public function get_total_unpurchased_amount() {
		return $this->get_total() - $this->get_total_purchased_amount();
	}

	/**
	 * Get the permalink for the wishlist
	 *
	 * @return string
	 */
	public function get_permalink() {
		return apply_filters( 'nmgr_wishlist_permalink', get_permalink( $this->get_id() ), $this );
	}

	/**
	 * Get the user id of the user associated with the wishlist
	 *
	 * @return int
	 */
	public function get_user_id() {
		return $this->get_prop( '_nmgr_user_id' );
	}

	/**
	 * Get the user associated with the wishlist
	 *
	 * @return WP_User|false
	 */
	public function get_user() {
		return is_numeric( $this->get_user_id() ) ? get_user_by( 'id', $this->get_user_id() ) : false;
	}

	/**
	 * Get the customer associated with the wishlist
	 *
	 * This should be the same as the user associated with the wishlist
	 * but simply retrieved as a WC_Customer object
	 */
	public function get_customer() {
		return new \WC_Customer( $this->get_user_id() );
	}

	/**
	 * Get the slug of the wishlist
	 *
	 * @return string
	 */
	public function get_slug() {
		return $this->get_prop( 'post_name' );
	}

	/**
	 * Get the date the wishlist was created
	 *
	 * @return string
	 */
	public function get_date_created() {
		return $this->get_prop( 'post_date' );
	}

	/*
	  |--------------------------------------------------------------------------
	  | Setters
	  |--------------------------------------------------------------------------
	 */

	/**
	 * Set the title of the wishlist
	 */
	public function set_title( $value ) {
		$this->set_prop( 'post_title', $value );
	}

	/**
	 * Set the post status of the wishlist
	 */
	public function set_status( $value ) {
		$this->set_prop( 'post_status', $value );
	}

	/**
	 * Set the first name of the wishlist owner
	 */
	public function set_first_name( $value ) {
		$this->set_prop( '_first_name', $value );
	}

	/**
	 * Set the last name of the wishlist owner
	 */
	public function set_last_name( $value ) {
		$this->set_prop( '_last_name', $value );
	}

	/**
	 * Set the first name of the wishlist owner's partner
	 */
	public function set_partner_first_name( $value ) {
		$this->set_prop( '_partner_first_name', $value );
	}

	/**
	 * Set the last name of the wishlist owner's partner
	 */
	public function set_partner_last_name( $value ) {
		$this->set_prop( '_partner_last_name', $value );
	}

	/**
	 * Set the registered email for the wishlist
	 */
	public function set_email( $value ) {
		$this->set_prop( '_email', $value );
	}

	/**
	 * Set the date for the wishlist event
	 */
	public function set_event_date( $value ) {
		$this->set_prop( '_event_date', $value );
	}

	/**
	 * Set the wishlist description
	 */
	public function set_description( $value ) {
		$this->set_prop( 'post_excerpt', $value );
	}

	/**
	 * Set shipping first name.
	 *
	 * @param string $value Shipping first name.
	 */
	public function set_shipping_first_name( $value ) {
		$this->set_prop( 'shipping_first_name', $value );
	}

	/**
	 * Set shipping last name.
	 *
	 * @param string $value Shipping last name.
	 */
	public function set_shipping_last_name( $value ) {
		$this->set_prop( 'shipping_last_name', $value );
	}

	/**
	 * Set shipping company.
	 *
	 * @param string $value Shipping company.
	 */
	public function set_shipping_company( $value ) {
		$this->set_prop( 'shipping_company', $value );
	}

	/**
	 * Set shipping address line 1.
	 *
	 * @param string $value Shipping address line 1.
	 */
	public function set_shipping_address_1( $value ) {
		$this->set_prop( 'shipping_address_1', $value );
	}

	/**
	 * Set shipping address line 2.
	 *
	 * @param string $value Shipping address line 2.
	 */
	public function set_shipping_address_2( $value ) {
		$this->set_prop( 'shipping_address_2', $value );
	}

	/**
	 * Set shipping city.
	 *
	 * @param string $value Shipping city.
	 */
	public function set_shipping_city( $value ) {
		$this->set_prop( 'shipping_city', $value );
	}

	/**
	 * Set shipping state.
	 *
	 * @param string $value Shipping state.
	 */
	public function set_shipping_state( $value ) {
		$this->set_prop( 'shipping_state', $value );
	}

	/**
	 * Set shipping postcode.
	 *
	 * @param string $value Shipping postcode.
	 */
	public function set_shipping_postcode( $value ) {
		$this->set_prop( 'shipping_postcode', $value );
	}

	/**
	 * Set shipping country.
	 *
	 * @param string $value Shipping country.
	 */
	public function set_shipping_country( $value ) {
		$this->set_prop( 'shipping_country', $value );
	}

	/**
	 * Set user id.
	 *
	 * @param int $value User ID.
	 */
	public function set_user_id( $value ) {
		$this->set_prop( 'post_author', (is_numeric( $value ) ? ( int ) $value : 0 ) );
		$this->set_prop( '_nmgr_user_id', $value );
	}

	/**
	 * Set the date the wishlist was fulfilled
	 *
	 * This is the date all items in the wishlist were marked as purchased
	 *
	 * @param string|integer|null $date UTC timestamp
	 */
	public function set_date_fulfilled( $date = null ) {
		$this->set_prop( '_date_fulfilled', $date );
	}

	/**
	 * Set the wishlist type
	 * @param string $type Taxonomy term slug
	 */
	public function set_type( $type ) {
		$this->type = $type;
		$this->set_prop( 'tax_input', [
			'nm_gift_registry_type' => nmgr_get_term_id_by_slug( $type )
		] );
	}

	/*
	  |--------------------------------------------------------------------------
	  | Wishlist Items
	  |--------------------------------------------------------------------------
	 */

	/**
	 * Remove all items  from this wishlist
	 *
	 * @return void
	 */
	public function delete_items() {
		global $wpdb;

		$wpdb->query( $wpdb->prepare( "DELETE FROM itemmeta USING {$wpdb->prefix}nmgr_wishlist_itemmeta itemmeta INNER JOIN {$wpdb->prefix}nmgr_wishlist_items items WHERE itemmeta.wishlist_item_id = items.wishlist_item_id and items.wishlist_id = %d", $this->get_id() ) );
		$wpdb->query( $wpdb->prepare( "DELETE FROM {$wpdb->prefix}nmgr_wishlist_items WHERE wishlist_id = %d", $this->get_id() ) );

		$this->clear_items_cache();
	}

	/**
	 * Get all items in this wishlist
	 *
	 * @return NMGR_Wishlist_Item[]|\NMGR\Sub\Wishlist_Item[]
	 */
	public function get_items() {
		if ( !$this->items ) {
			$this->items = apply_filters( 'nmgr_items', $this->read_items(), $this );
		}
		return $this->items;
	}

	/**
	 * Get a single wishlist item from the wishlist
	 *
	 * @param  int  $id wishlist_item_id
	 * @return NMGR_Wishlist_Item|\NMGR\Sub\Wishlist_Item|false
	 */
	public function get_item( $id ) {
		if ( isset( $this->items[ $id ] ) ) {
			return $this->items[ $id ];
		}

		$args = [
			'where' => 'AND wishlist_item_id = ' . ( int ) $id,
		];

		$val = $this->read_items( $args );
		return !empty( $val ) ? reset( $val ) : false;
	}

	public function get_item_by_unique_id( $unique_id ) {
		return $this->get_item_by_meta( '_unique_id', $unique_id );
	}

	/**
	 * Bulk updates wishlist items from the wishlist items table
	 * (used for updating item properties such as quantity, purchased_quantity and favourite)
	 *
	 * @deprecated since version 4.1.0
	 * @param array $posted_data Posted data containing wishlist items properties to save.
	 */
	public function update_items( $posted_data ) {
		_deprecated_function( __METHOD__, '4.1.0' );
		if ( isset( $posted_data[ 'wishlist_item_id' ] ) ) {

			foreach ( $posted_data[ 'wishlist_item_id' ] as $item_id ) {
				$item = $this->get_item( $item_id );

				if ( !$item ) {
					continue;
				}

				// This array holds the props we are updating for each wishlist item
				$data_keys = array(
					'wishlist_item_qty' => $item->get_quantity(),
				);

				$item_data = array();

				foreach ( $data_keys as $key => $default ) {
					$item_data[ $key ] = isset( $posted_data[ $key ][ $item_id ] ) ? wc_check_invalid_utf8( wp_unslash( $posted_data[ $key ][ $item_id ] ) ) : $default;
				}

				if ( '0' === $item_data[ 'wishlist_item_qty' ] ) {
					$item->delete();
					continue;
				}

				$props = array(
					'quantity' => $item_data[ 'wishlist_item_qty' ],
				);

				$item->set_props( $props );
				$item->save();
			}

			$this->clear_items_cache();
		}
	}

	/**
	 * Remove an item from the wishlist.
	 *
	 * @param int $item_id Item ID to delete.
	 * @return false|void
	 */
	public function delete_item( $item_id ) {
		$item = $this->get_item( $item_id );
		$item->delete();
		$this->clear_items_cache();
	}

	/**
	 * Gets the count of items in this wishlist
	 *
	 * @return int
	 */
	public function get_item_count() {
		_deprecated_function( __METHOD__, '4.2.0', __CLASS__ . '::get_item_quantity_count' );
		return $this->get_items_quantity_count();
	}

	/**
	 * Get the total quantities of all items in the wishlist
	 * @return int
	 */
	public function get_items_quantity_count() {
		global $wpdb;

		$cache_key = 'nmgr_items_quantity_count_' . $this->get_id();
		$val = wp_cache_get( $cache_key );

		if ( false === $val ) {
			$val = $wpdb->get_var( $wpdb->prepare( "
			SELECT SUM(itemmeta.meta_value)
			FROM {$wpdb->prefix}nmgr_wishlist_itemmeta AS itemmeta
				LEFT JOIN {$wpdb->prefix}nmgr_wishlist_items AS items
					ON itemmeta.wishlist_item_id = items.wishlist_item_id
				WHERE wishlist_id = %d
				AND itemmeta.meta_key = '_quantity'
				",
					$this->get_id()
				) );

			wp_cache_set( $cache_key, $val );
			$this->add_items_cache_key( $cache_key );
		}

		return ( int ) $val;
	}

	public function get_item_purchased_count() {
		_deprecated_function( __METHOD__, '4.2.0', __CLASS__ . '::get_items_purchased_quantity_count' );
		return $this->get_items_purchased_quantity_count();
	}

	/**
	 * Gets the count of purchased wishlist items
	 * @return int
	 */
	public function get_items_purchased_quantity_count() {
		global $wpdb;

		$cache_key = 'nmgr_items_purchased_quantity_count_' . $this->get_id();
		$val = wp_cache_get( $cache_key );

		if ( false === $val ) {
			$val = $wpdb->get_var( $wpdb->prepare( "
			SELECT SUM(itemmeta.meta_value)
			FROM {$wpdb->prefix}nmgr_wishlist_itemmeta AS itemmeta
				LEFT JOIN {$wpdb->prefix}nmgr_wishlist_items AS items
					ON itemmeta.wishlist_item_id = items.wishlist_item_id
				WHERE wishlist_id = %d
				AND itemmeta.meta_key = '_purchased_quantity'
				",
					$this->get_id()
				) );

			wp_cache_set( $cache_key, $val );
			$this->add_items_cache_key( $cache_key );
		}

		return apply_filters( 'nmgr_wishlist_item_purchased_count', $val, $this );
	}

	/**
	 * Add an item (product) to this wishlist and save the item in the database
	 *
	 * @param  int|WC_Product $product Product id or object.
	 * @param  int $qty Quantity to add.
	 * @param int $favourite Whether the product is marked as favourite in the wishlist. Values are 1 or 0.
	 * @param array $variation Product variations if the product is a variation.
	 * @param array $item_data Extra data associated with the item to be added.
	 *
	 * @return int
	 */
	public function add_item( $product, $qty = 1, $favourite = null, $variation = array(), $item_data = array() ) {
		$product = wc_get_product( $product );

		if ( !$product ) {
			return 0;
		}

		$product_id = $product->is_type( 'variation' ) ? $product->get_parent_id() : $product->get_id();
		$variation_id = $product->is_type( 'variation' ) ? $product->get_id() : 0;

		// Generate a unique id to identify item in wishlist based on product ID, variation ID, and variation data
		$unique_id = $this->generate_unique_id( $product_id, $variation_id, $variation, $item_data );

		$args = array(
			'wishlist_id' => $this->get_id(),
			'name' => $product->get_name(),
			'product_id' => $product_id,
			'variation_id' => $variation_id,
			'variation' => $variation,
			'quantity' => $qty,
			'favourite' => $favourite,
			'unique_id' => $unique_id,
		);

		$item = nmgr()->wishlist_item();

		$item_with_unique_id = $this->get_item_by_unique_id( $unique_id );
		if ( $item_with_unique_id ) {
			$item = $item_with_unique_id;

			// if the wishlist already has the item, we can only update the quantity
			$args[ 'quantity' ] = $item->get_quantity() + $qty;
		}

		$item->set_props( $args );
		$item->save();
		$this->clear_items_cache();

		return $item->get_id();
	}

	/**
	 * Checks if an item is already in the wishlist
	 *
	 * @param string $unique_id The unique id of the item in the wishlist
	 * @return boolean
	 */
	public function has_item( $unique_id ) {
		return ( bool ) $this->get_item_by_unique_id( $unique_id );
	}

	/**
	 * Checks if the wishlist has items
	 *
	 * @return boolean
	 */
	public function has_items() {
		return ( bool ) $this->get_items_count();
	}

	/**
	 * Generate a unique id for the wishlist item being added
	 *
	 * @param int $product_id ID of the product the key is being generated for
	 * @param int $variation_id Variation id of the product the key is being generated for
	 * @param array $variation Variation data for the wishlist item
	 * @param array $item_data Extra data passed to denote the uniqueness of the item in the
	 * wishlist
	 */
	public function generate_unique_id( $product_id, $variation_id = 0, $variation = array(), $item_data = array() ) {
		$id_parts = array( $product_id );

		if ( $variation_id && 0 !== $variation_id ) {
			$id_parts[] = $variation_id;
		}

		if ( is_array( $variation ) && !empty( $variation ) ) {
			$variation_key = '';
			foreach ( $variation as $key => $value ) {
				$variation_key .= trim( $key ) . trim( $value );
			}
			$id_parts[] = $variation_key;
		}

		if ( is_array( $item_data ) && !empty( $item_data ) ) {
			$item_data_key = '';
			foreach ( $item_data as $key => $value ) {
				if ( is_array( $value ) || is_object( $value ) ) {
					$value = http_build_query( $value );
				}
				$item_data_key .= trim( $key ) . trim( $value );
			}
			$id_parts[] = $item_data_key;
		}

		return md5( implode( '_', $id_parts ) );
	}

	/**
	 * Get all items in an order for this wishlist
	 *
	 * @param type $order_id Order id
	 */
	public function get_items_in_order( $order_id ) {
		if ( is_numeric( $order_id ) ) {
			$order = wc_get_order( $order_id );
		} elseif ( $order_id instanceof WC_Order ) {
			$order = $order_id;
		}

		if ( !$order ) {
			return;
		}

		$items = $order->get_items();
		$items_in_order = array();

		foreach ( $items as $item_id => $item ) {
			$meta = $item->get_meta( 'nm_gift_registry' );
			// We want to make sure the item has not been completely refunded
			if ( $meta && ( int ) $item->get_quantity() > absint( $order->get_qty_refunded_for_item( $item->get_id() ) ) ) {
				if ( $meta[ 'wishlist_id' ] === $this->get_id() ) {
					$items_in_order[ $item_id ] = array(
						'name' => $item->get_name(),
						'quantity' => $item->get_quantity(),
						'variation_id' => $item->get_variation_id(),
						'total' => $item->get_total(),
					);
				}
			}
		}

		return apply_filters( 'nmgr_wishlist_get_items_in_order', $items_in_order, $order, $this );
	}

	/**
	 * Get the wishlist item representing a product, if the product is in the wishlist
	 *
	 * @param int|WC_Product $product_id The product id or object
	 * @return NMGR_Wishlist_Item|\NMGR\Sub\Wishlist_Item|false
	 */
	public function get_item_by_product( $product_id ) {
		$product = wc_get_product( $product_id );
		if ( $product ) {
			return $this->get_item_by_meta( '_product_id', $product->get_id() );
		}
	}

	public function get_item_by_meta( $meta_key, $meta_value ) {
		global $wpdb;

		$args = [
			'select' => 'items.wishlist_item_id',
			'where' => $wpdb->prepare( "AND itemmeta.meta_key = %s AND itemmeta.meta_value = %s", $meta_key, $meta_value ),
			'with_meta' => true,
			'limit' => 1
		];

		$val = $this->read_items( $args );
		return !empty( $val ) ? reset( $val ) : false;
	}

	/*
	  |--------------------------------------------------------------------------
	  | Conditionals
	  |--------------------------------------------------------------------------
	 */

	/**
	 * Whether the wishlist has a shipping address
	 *
	 * The wishlist has a shipping address if all the required fields are filled
	 * or if the country and address 1 fields are filled.
	 *
	 * @return boolean
	 */
	public function has_shipping_address() {
		$address = $this->get_shipping();

		if ( !isset( $address[ 'country' ] ) || !$address[ 'country' ] ) {
			return false;
		}

		$fields = is_a( wc()->countries, 'WC_Countries' ) ? wc()->countries->get_address_fields( $address[ 'country' ], 'shipping_' ) : array();
		$required = array_keys( array_filter( $fields, function( $field ) {
				return isset( $field[ 'required' ] ) && $field[ 'required' ];
			} ) );

		foreach ( $required as $field ) {
			$unprefixed = str_replace( 'shipping_', '', $field );
			if ( !isset( $address[ $unprefixed ] ) ||
				(isset( $address[ $unprefixed ] ) && !$address[ $unprefixed ] ) ) {
				return false;
			}
		}

		return (isset( $address[ 'address_1' ] ) && $address[ 'address_1' ]) ||
			(isset( $address[ 'address_2' ] ) && $address[ 'address_2' ]);
	}

	/**
	 * Whether the wishlist needs its shipping address to be filled
	 *
	 * The wishlist needs a shipping address if the shipping address is required in the
	 * plugin setting and it's shipping address is not completely filled.
	 *
	 * @return boolean
	 */
	public function needs_shipping_address() {
		$val = nmgr_get_type_option( $this->get_type(), 'shipping_address_required' ) &&
			!$this->has_shipping_address();
		return apply_filters( 'nmgr_wishlist_needs_shipping_address', $val, $this );
	}

	/**
	 * Whether the wishlist ships to the customer's account shipping address
	 *
	 * @return boolean
	 */
	public function is_shipping_to_account_address() {
		_deprecated_function( __METHOD__, '4.0.0' );

		if ( $this->is_guest() ) {
			return false;
		}

		$bool = ( bool ) $this->get_prop( 'ship_to_account_address' );

		if ( $bool ) {
			$customer = $this->get_customer();
			if ( $customer &&
				((method_exists( $customer, 'has_shipping_address' ) && $customer->has_shipping_address()) ||
				$customer->get_shipping_address())
			) {
				foreach ( $customer->get_shipping() as $k => $v ) {
					update_post_meta( $this->get_id(), '_shipping_' . $k, $v );
				}
			}
			delete_post_meta( $this->get_id(), '_ship_to_account_address' );
		}

		return false;
	}

	/**
	 * Checks if the wishlist has a product
	 *
	 * Wishlist items can be single products, variable products or items of group products.
	 *
	 * This function allows us to check if the product in question  (if it is a grouped product)
	 * has an item in the wishlist, or whether the product (if it is a simple or variable product) is an item in the wishlist.
	 * It basically queries the presence of the product in the wishlist in a general way.
	 *
	 * In this sense it differs from the 'has_item' function which checks for a particular item's
	 * unique id in the wishlist and cannot tell whether the item is part of a variable or grouped product
	 *
	 * @param WC_Product $product
	 * @return boolean true|false
	 */
	public function has_product( $product ) {
		global $wpdb;

		if ( !$product ) {
			return false;
		}

		if ( $product->is_type( 'grouped' ) && $product->has_child() ) {
			$product_item_ids = $product->get_children();
		} else {
			$product_item_ids = ( array ) $product->get_id();
		}

		$product_str = implode( '_', $product_item_ids );
		$cache_key = 'nmgr_items_has_product_' . $this->get_id() . '_' . $product_str;

		$val = wp_cache_get( $cache_key );

		if ( false === $val ) {
			$val = $wpdb->get_var( $wpdb->prepare( "
			SELECT COUNT(*)
			FROM {$wpdb->prefix}nmgr_wishlist_items AS items
				LEFT JOIN {$wpdb->prefix}nmgr_wishlist_itemmeta AS itemmeta
					ON items.wishlist_item_id = itemmeta.wishlist_item_id
				WHERE wishlist_id = %d
				AND itemmeta.meta_key = '_product_id'
				AND itemmeta.meta_value IN ('" . implode( "','", array_map( 'intval', $product_item_ids ) ) . "')
				LIMIT 1
				",
					$this->get_id()
				) );

			wp_cache_set( $cache_key, $val );
			$this->add_items_cache_key( $cache_key );
		}

		return ( bool ) $val;
	}

	/**
	 * Check if all the items in the wishlist have been fully purchased
	 *
	 * @return boolean
	 */
	public function is_fulfilled() {
		if ( $this->has_items() &&
			nmgr_get_type_option( $this->get_type(), 'display_item_quantity' ) &&
			nmgr_get_type_option( $this->get_type(), 'display_item_purchased_quantity' )
		) {
			$items = $this->get_items();

			foreach ( $items as $item ) {
				if ( !$item->is_fulfilled() ) {
					return false;
				}
			}
			return true;
		}
		return false;
	}

	/**
	 * Check whether the wishlist is active.
	 * An active wishlist is a wishlist that has any of the registered post statuses and is not trashed.
	 *
	 * @return boolean
	 */
	public function is_active() {
		return $this->get_id() && in_array( $this->get_status(), nmgr_get_post_statuses() );
	}

	/**
	 * Check if the wishlist belongs to a guest
	 *
	 * @since 2.0.0
	 * @return boolean
	 */
	public function is_guest() {
		return $this->get_user_id() === get_post_meta( $this->get_id(), '_nmgr_guest', true );
	}

	/**
	 * Check if the wishlist has the term
	 * @param string $term_slug Default values are gift-registry, wishlist
	 * @return boolean
	 */
	public function is_type( $term_slug ) {
		return $term_slug === $this->get_type();
	}

	/**
	 * Check if the wishlist is expired.
	 * This happens if the event date is set and it is passed the current date
	 *
	 * We get this value directly from php using the event date rather than from the _nmgr_expired
	 * meta_key which is set in the database during a cron job and may not be available at the
	 * time of calling this function
	 *
	 * @return boolean
	 */
	public function is_expired() {
		return $this->get_event_date() && $this->get_expiry_days() < 0;
	}

	public function get_type() {
		return $this->type;
	}

	/**
	 * Get the days of expiry related to the event date
	 * @return boolean|int False if wishlist has no event date. Positive number if wishlist expiry
	 * is in the future, negative number if wishlist has expired, Zero if wishlist expiry is today.
	 */
	public function get_expiry_days() {
		if ( $this->get_event_date() ) {
			$event_date = nmgr_get_datetime( $this->get_event_date() );
			if ( $event_date ) {
				$diff = date_diff( new DateTime( current_time( 'Y-m-d' ) ), new DateTime( $event_date->format( 'Y-m-d' ) ) );
				return ( int ) $diff->format( "%R%a" );
			}
		}
		return false;
	}

	public function get_expiry() {
		return absint( $this->get_prop( '_nmgr_expired' ) );
	}

	public function set_expiry( $value ) {
		$this->set_prop( '_nmgr_expired', absint( $value ) );
	}

	public function clear_items_cache() {
		$this->items = [];
		foreach ( $this->get_items_cache_keys() as $cache_key ) {
			wp_cache_delete( $cache_key );
		}
	}

	/**
	 * Get the number of items in the wishlist
	 * @global type $wpdb
	 * @return type
	 */
	public function get_items_count() {
		global $wpdb;

		$cache_key = 'nmgr_items_count_' . $this->get_id();
		$count = wp_cache_get( $cache_key );

		if ( false === $count ) {
			$count = $wpdb->get_var( $wpdb->prepare( "SELECT COUNT(*) FROM {$wpdb->prefix}nmgr_wishlist_items WHERE wishlist_id = %d", $this->get_id() ) );

			wp_cache_set( $cache_key, $count );
			$this->add_items_cache_key( $cache_key );
		}

		return ( int ) $count;
	}

	/**
	 * @return NMGR_Wishlist_Item[]|\NMGR\Sub\Wishlist_Item[]
	 */
	public function read_items( $args = [] ) {
		global $wpdb;

		if ( !$this->get_id() ) {
			return [];
		}

		$defaults = [
			'limit' => null,
			'offset' => 0,
			'order' => 'DESC',
			'orderby' => 'date_created',
			'where' => null,
			'page' => null, // 'page' and 'offset' should not be used together.
			'select' => '*',
			'with_meta' => false,
			'return' => 'items', // other values are query, raw_results
		];

		$p_args = wp_parse_args( $args, $defaults );
		$limit = max( 0, ( int ) $p_args[ 'limit' ] );

		if ( $p_args[ 'page' ] ) {
			$p_args[ 'offset' ] = max( 0, (( int ) $p_args[ 'page' ] - 1) * $limit );
		}

		$join_sql = '';
		if ( $p_args[ 'with_meta' ] ) {
			$join_sql = "LEFT JOIN {$wpdb->prefix}nmgr_wishlist_itemmeta AS itemmeta
					ON items.wishlist_item_id = itemmeta.wishlist_item_id";
		}

		$select_sql = $p_args[ 'select' ];
		$where_sql = "AND wishlist_id = {$this->get_id()} " . ($p_args[ 'where' ] ?? '');
		$limit_sql = $limit ? $wpdb->prepare( "LIMIT %d", $limit ) : '';
		$offset_sql = $limit ? 'OFFSET ' . $p_args[ 'offset' ] : '';
		$order_sql = esc_sql( "ORDER BY {$p_args[ 'orderby' ]} {$p_args[ 'order' ]}" );
		$sql = "SELECT $select_sql "
			. "FROM {$wpdb->prefix}nmgr_wishlist_items AS items "
			. "$join_sql "
			. "WHERE 1=1 $where_sql "
			. "$order_sql "
			. "$limit_sql "
			. "$offset_sql";

		if ( 'query' === $p_args[ 'return' ] ) {
			return $sql;
		}

		$cache_key = 'nmgr_items_' . str_replace( ' ', '_', trim( $sql ) );
		$items_data = wp_cache_get( $cache_key );

		if ( empty( $items_data ) ) {
			// Items from database are returned as objects
			$items_data = $wpdb->get_results( $sql );
			wp_cache_set( $cache_key, $items_data );
			$this->add_items_cache_key( $cache_key );
		}

		if ( 'raw_results' === $p_args[ 'return' ] ) {
			return $items_data;
		}

		$item_ids_to_class_objs = [];

		foreach ( $items_data as $item_data ) {
			$item = nmgr()->wishlist_item();
			$item->set_id( $item_data->wishlist_item_id );
			foreach ( array_keys( $item->get_data() ) as $key ) {
				if ( property_exists( $item_data, $key ) ) {
					$item->set_prop( $key, $item_data->$key );
				}
			}
			$item->read_meta_data();
			$item->set_object_read();

			$item_ids_to_class_objs[ $item->get_id() ] = $item;
		}

		return $item_ids_to_class_objs;
	}

	public function get_items_cache_keys() {
		return $this->items_cache_keys;
	}

	public function add_items_cache_key( $cache_key ) {
		$this->items_cache_keys[] = $cache_key;
	}

}
