<?php

use \NmeriMedia\V2\Bootstrap\Modal;
use \NmeriMedia\V2\Bootstrap\Dropdown;
use \NmeriMedia\V2\Bootstrap\Toast;
use NMGR\Int\Strings;

/**
 * @sync
 */
defined( 'ABSPATH' ) || exit;

if ( file_exists( __DIR__ . '/Sub/nmgr-functions-pro.php' ) ) {
	require_once 'Sub/nmgr-functions-pro.php';
}

/**
 * Returns Plugin main properties
 *
 * @return object
 */
function nmgr() {
	if ( function_exists( 'nm_gift_registry' ) ) {
		return nm_gift_registry();
	} elseif ( function_exists( 'nm_gift_registry_lite' ) ) {
		return nm_gift_registry_lite();
	}
}

/**
 * Checks if the post is for an nm gift registry post type
 *
 * @param int | WP_Post | null $post The post id or object
 * @return boolean True | False
 */
function is_nmgr_post( $post = null ) {
	return 'nm_gift_registry' === get_post_type( $post );
}

/**
 * Check if we are viewing a single wishlist page on the frontend
 *
 * @return boolean
 */
function is_nmgr_wishlist() {
	return is_nmgr_wishlist_page( 'single' );
}

/**
 * Check if we are on a single wishlist edit page or the all wishlists page in the admin area
 *
 * @return boolean
 */
function is_nmgr_admin() {
	global $current_screen;

	$bool = false;

	if ( wp_doing_ajax() && isset( $GLOBALS[ 'nmgr' ]->is_admin ) ) {
		$bool = filter_var( $GLOBALS[ 'nmgr' ]->is_admin, FILTER_VALIDATE_BOOLEAN );
	} elseif ( is_admin() && !wp_doing_ajax() ) {
		if ( !isset( $current_screen ) ) {
			$bool = false;
		} else {
			$bool = isset( $current_screen->post_type ) && 'nm_gift_registry' === $current_screen->post_type;
		}
	}

	return apply_filters( 'is_nmgr_admin', $bool );
}

/**
 * Check if we are on the woocommerce account page endpoint for NM Gift Registry
 *
 * @return boolean
 */
function is_nmgr_account() {
	_deprecated_function( __FUNCTION__, '4.0.0', 'is_nmgr_account_section' );
	return is_nmgr_account_section();
}

/**
 * Check if we are on the wishlists search page
 *
 * @param WP_Query $query The query object. Uses global $wp_query by default.
 * @return boolean
 */
function is_nmgr_search( $query = '' ) {
	global $wp_query;

	$the_query = is_a( $query, 'WP_Query' ) ? $query : $wp_query;

	$bool = (is_search() &&
		isset( $the_query->query_vars[ 'post_type' ] ) &&
		'nm_gift_registry' === $the_query->query_vars[ 'post_type' ]) ||
		isset( $the_query->query_vars[ 'nmgr_s' ] ) || isset( $_GET[ 'nmgr_s' ] ) ||
		nmgr_post_content_has_shortcode( '[nmgr_search]' ) ||
		nmgr_post_content_has_shortcode( '[nmgr_search_results]' );

	return apply_filters( 'is_nmgr_search', $bool );
}

/**
 * Check if we are in an NM Gift Registry account tab
 *
 * @return boolean
 */
function is_nmgr_account_tab() {
	_deprecated_function( __FUNCTION__, '4.0.0' );
	return apply_filters( 'is_nmgr_account_tab', false );
}

/**
 * Check if we are in an NM Gift Registry modal window
 *
 * @return boolean
 */
function is_nmgr_modal() {
	return apply_filters( 'is_nmgr_modal', (nmgr_get_global_var()->is_modal ?? false ) );
}

/**
 * Check if we are currently in an NM Gift Registry account section
 *
 * Registered account sections are:
 * - profile
 * - items
 * - images
 * - shipping
 * - messages
 * - settings
 *
 * @return boolean
 */
function is_nmgr_account_section( $section = '' ) {
	$bool = is_nmgr_wishlist_page( 'account_section' );
	if ( $section ) {
		$bool = $bool && $section === get_query_var( 'nmgr_action' );
	}

	return apply_filters( 'is_nmgr_account_section', $bool );
}

/**
 * Check if we are on a page which uses nmgr templates
 *
 * @return boolean
 */
function is_nmgr() {
	return apply_filters( 'is_nmgr',
		is_nmgr_wishlist() ||
		is_nmgr_admin() ||
		is_nmgr_search() ||
		is_nmgr_account_section()
	);
}

/**
 * Check whether the gift-registry or wishlist module is enabled
 *
 * @param int $user_id User id. Optional. Defaults to current logged in user.
 * @return boolean
 */
function is_nmgr_enabled( $type = 'gift-registry' ) {
	if ( !in_array( $type, [ 'gift-registry', 'wishlist' ] ) ) {
		_deprecated_argument( __FUNCTION__, '4.0.0' );
		$type = 'gift-registry';
	}
	return apply_filters( 'is_nmgr_enabled', (( bool ) nmgr_get_type_option( $type, 'enable' ) ), $type );
}

/**
 * Determine whether we are in the shop loop
 *
 * This is based on whether any of the registered action hooks for displaying product
 * content within loops in content-product.php is being fired.
 *
 * This function is typically preferred to checking for woocommerce archive
 * pages with is_shop() or is_product_taxonomy() because it covers the shop loop
 * in every location including places which may not be typical archive locations
 * such as the 'Related Products' section on single product pages.
 *
 * @return boolean
 */
function is_nmgr_shop_loop() {
	$actions = apply_filters( 'nmgr_shop_loop_actions', array(
		'woocommerce_before_shop_loop_item',
		'woocommerce_before_shop_loop_item_title',
		'woocommerce_shop_loop_item_title',
		'woocommerce_after_shop_loop_item_title',
		'woocommerce_after_shop_loop_item'
		) );

	foreach ( $actions as $action ) {
		if ( doing_action( $action ) ) {
			return true;
		}
	}
	return false;
}

/**
 * Get the search form for finding a wishlist on the frontend
 */
function nmgr_get_search_form( $args = [] ) {
	$defaults = array(
		'show_results' => false,
		'form_action' => home_url(),
	);

	return nmgr_get_search_template( wp_parse_args( $args, $defaults ) );
}

/**
 * Paged navigation for search results
 */
function nmgr_paging_nav() {
	$args = apply_filters( 'nmgr_paging_nav_args', array(
		'next_text' => nmgr()->is_pro ?
		__( 'Next', 'nm-gift-registry' ) :
		__( 'Next', 'nm-gift-registry-lite' ),
		'prev_text' => nmgr()->is_pro ?
		__( 'Previous', 'nm-gift-registry' ) :
		__( 'Previous', 'nm-gift-registry-lite' ),
		) );
	$pagination = apply_filters( 'nmgr_paging_nav', get_the_posts_pagination( $args ) );
	echo $pagination;
}

/**
 * Get the permalink for the wishlist account page
 * or for managing a specific wishlist on the account page
 *
 * @param mixed $id_or_slug The post id, post slug or NMGR_Wishlist object for the wishlist. Optional.
 * @return string
 */
function nmgr_get_account_url( $id_or_slug = '' ) {
	_deprecated_function( __FUNCTION__, '4.0.0' );

	if ( !$id_or_slug ) {
		$account_url = nmgr_get_url( 'gift-registry', 'home' );
	} else {
		if ( is_a( $id_or_slug, 'NMGR_Wishlist' ) ) {
			$slug = $id_or_slug->get_slug();
		} elseif ( is_numeric( $id_or_slug ) ) {
			$post = get_post( $id_or_slug );
			$slug = $post ? $post->post_name : '';
		} else {
			$slug = ( string ) $id_or_slug;
		}
		$account_url = nmgr_get_url( 'gift-registry', $slug );
	}

	return apply_filters( 'nmgr_account_url', $account_url );
}

/**
 * Get an option value for the plugin
 *
 * If an option key is not provided, values for all plugin keys are returned
 * @return mixed
 */
function nmgr_get_option( $option_key = '', $default_value = null, $type = '' ) {
	$settings = nmgr()->is_pro ? [] : nmgr()->get_default_pro_fields_values();
	$options = array_diff_key( get_option( 'nmgr_settings', array() ), $settings );

	if ( $option_key ) {
		$option_key = 'wishlist' === $type ? 'wishlist_' . $option_key : $option_key;
		$value = $default_value;

		if ( array_key_exists( $option_key, $options ) ) {
			$value = $options[ $option_key ];
		}

		return apply_filters( 'nmgr_get_option', $value, [
			'option_key' => $option_key,
			'default_value' => $default_value,
			'type' => $type,
			'options' => $options
			] );
	} elseif ( $type ) {
		if ( 'wishlist' === $type ) {
			foreach ( $options as $key => $value ) {
				if ( 0 !== strpos( $key, 'wishlist_' ) ) {
					unset( $options[ $key ] );
				}
			}
		} elseif ( 'gift-registry' === $type ) {
			foreach ( $options as $key => $value ) {
				if ( 0 === strpos( $key, 'wishlist_' ) ) {
					unset( $options[ $key ] );
				}
			}
		}
	}

	return $options;
}

function nmgr_get_type_option( $type, $option_key = '', $default_value = null ) {
	return nmgr_get_option( $option_key, $default_value, $type );
}

/**
 * Verify the standard form nonce supplied in NMGR_Form
 *
 * @param array $request Array to check in for existing nonce key or $_REQUEST if not supplied
 * @return false|int False if the nonce is invalid, 1 if the nonce is valid and generated between
 *                   0-12 hours ago, 2 if the nonce is valid and generated between 12-24 hours ago.
 */
function nmgr_verify_form_nonce( $request = '' ) {
	_deprecated_function( __FUNCTION__, '4.0.0' );
	return NMGR_Form::verify_nonce( $request );
}

/**
 * Verify the wishlist id posted during a request and make sure that it is the
 * same wishlist id sent to the page
 *
 * @deprecated 2.0.0
 * @param array $posted_data The posted data containing the 'wishlist id' and 'nonce' keys. Optional. Default is $_POST.
 * @return int|null Verified wishlist id or null if the wishlist id supplied is invalid.
 */
function nmgr_get_verified_wishlist_id( $posted_data = '' ) {
	_deprecated_function( __FUNCTION__, '2.0.0', 'nmgr_verify_request' );

	$wishlist_id = isset( $_REQUEST[ 'wishlist_id' ] ) ? ( int ) $_REQUEST[ 'wishlist_id' ] : 0;
	if ( 0 === $wishlist_id && nmgr_user_has_wishlist( $wishlist_id ) ) {
		return $wishlist_id;
	}
	return null;
}

/**
 * Get a template file from the templates path
 *
 * This function searches the templates path in the theme folder before defaulting to
 * the templates path in the plugin folder if it doesn't find the file.
 * This way, It allows plugin templates to be overridden by copying them to the theme folder
 * similar to the way woocommerce works.
 *
 * The default expected theme template path where overridden templates reside is: yourtheme/plugin-slug'
 * where 'yourtheme' is the name of your theme and 'plugin-slug' is nm-gift-registry-lite for the lite version or
 * nm-gift-registry for the full version of the plugin.
 *
 * @param string $name Name of template file to get (prefixed with subfolder if it exists in a subfolder of the template path).
 * @param array $args Variables to send to the template file.
 *
 * @return string Template html
 */
function nmgr_get_template( $name, $args = array() ) {
	global $nmgr_template;
	$nmgr_template = $name;

	$defaults = array(
		'template_name' => $name,
		'args' => $args,
		'template_path' => nmgr()->theme_path(),
		'default_path' => nmgr()->template_path(),
	);

	$fargs = apply_filters( 'nmgr_get_template_args', $defaults );

	$template = wc_get_template_html(
		$fargs[ 'template_name' ],
		$fargs[ 'args' ],
		$fargs[ 'template_path' ],
		$fargs[ 'default_path' ]
	);

	return $template;
}

/**
 * Output a template file
 *
 * @param type $name Name of template file to get (prefixed with subfolder if it exists in a subfolder of the template path).
 * @param type $args Variables to send to the template file.
 */
function nmgr_template( $name, $args = array() ) {
	echo nmgr_get_template( $name, $args ); // phpcs:ignore WordPress.Security.EscapeOutput.OutputNotEscaped
}

/**
 * Get the full anchor tag for a user's wishlist page on the frontend or the edit page in the admin area if in admin
 *
 * @param obj $wishlist Wishlist object
 * @param array $args Parameters to be used in the anchor tag.
 * Accepted parameters
 *  title - title attriubte of the link element.
 * content - content to display as the link text.
 * @return string anchor html tag
 */
function nmgr_get_wishlist_link( $wishlist, $args = array() ) {
	if ( $wishlist ) {
		$url = !is_nmgr_admin() ? $wishlist->get_permalink() : get_edit_post_link( $wishlist->get_id() );
		$title = isset( $args[ 'title' ] ) ? esc_attr( $args[ 'title' ] ) : '';
		$content = nmgr_kses_post( isset( $args[ 'content' ] ) ? $args[ 'content' ] : $wishlist->get_title() );
		return sprintf( '<a href="%s" title="%s">%s</a>', $url, $title, $content );
	}
}

/**
 * Get the current wishlist id based on the global context, current page, or query
 *
 * @return int wishlist id
 */
function nmgr_get_current_wishlist_id() {
	global $nmgr;

	$the_post = get_queried_object_id();

	if ( is_nmgr_wishlist() || is_nmgr_wishlist_page( 'account_section' ) ) {
		$nmgr_w = sanitize_text_field( wp_unslash( get_query_var( 'nmgr_w' ) ) );
		$type = nmgr_get_current_type();

		if ( wp_doing_ajax() && !empty( $nmgr->wishlist_id ) ) {
			$the_post = $nmgr->wishlist_id;
		} elseif ( $nmgr_w ) {
			$the_post = is_numeric( $nmgr_w ) ? absint( $nmgr_w ) : get_page_by_path( $nmgr_w, '', 'nm_gift_registry' );
		} elseif ( !nmgr_get_type_option( $type, 'allow_multiple_wishlists' ) ) {
			$the_post = nmgr_get_user_default_wishlist_id( '', $type );
		}
	}

	if ( $the_post && is_numeric( $the_post ) ) {
		$the_post = get_post( $the_post );
	}

	return ( isset( $the_post->ID ) && is_nmgr_post( $the_post->ID ) ) ? ( int ) $the_post->ID : 0;
}

/**
 * Get all the users that have active wishlists
 *
 * Active wishlists are wishlists that are not in trash.
 *
 * @global wpdb $wpdb
 * @param string category The category of wishlists to get.
 * @return array User ids
 */
function nmgr_get_users( $category = 'gift-registry' ) {
	global $wpdb;

	if ( !in_array( $category, [ 'wishlist', 'gift-registry' ] ) ) {
		_deprecated_argument( __FUNCTION__, '4.0.0', 'Use "gift-registry" or "wishlist".' );
		return [];
	}

	$ids = wp_cache_get( 'nmgr_user_ids' );

	if ( false === $ids ) {
		$ids = $wpdb->get_col( $wpdb->prepare( "
				SELECT DISTINCT post_author FROM {$wpdb->posts} posts
		INNER JOIN {$wpdb->term_relationships} term_relationships
		ON posts.ID = term_relationships.object_id
		INNER JOIN {$wpdb->term_taxonomy} term_taxonomy
		ON term_taxonomy.term_taxonomy_id = term_relationships.term_taxonomy_id
		INNER JOIN {$wpdb->terms} terms
		ON terms.term_id = term_taxonomy.term_taxonomy_id
					WHERE post_author != 0
					AND post_type = 'nm_gift_registry'
					AND post_status != 'trash'
					AND term_taxonomy.taxonomy = 'nm_gift_registry_type'
		AND terms.slug = %s
				 ",
				$category
			) );

		wp_cache_set( 'nmgr_user_ids', $ids );
	}
	return $ids;
}

/**
 * Get all the wishlist ids for a user
 *
 * @param int|string $user_id The user id (optional).
 * Defaults to current logged in user id or guest user id cookie value.
 * @param string $type The wishlist type taxonomy term slug (gift-registry or wishlist)
 * @return array
 */
function nmgr_get_user_wishlist_ids( $user_id = '', $type = 'gift-registry' ) {
	$userid = $user_id ? $user_id : nmgr_get_current_user_id();

	if ( !$userid ) {
		return array();
	}

	$args = [
		'posts_per_page' => -1,
		'post_type' => 'nm_gift_registry',
		'post_status' => nmgr_get_post_statuses(),
		'fields' => 'ids',
		'meta_query' => [
			[
				'key' => '_nmgr_user_id',
				'value' => $userid,
				'compare' => '=',
			]
		],
		'tax_query' => [
			[
				'taxonomy' => 'nm_gift_registry_type',
				'field' => 'slug',
				'terms' => $type,
				'operator' => 'IN',
			]
		],
	];

	return get_posts( $args );
}

/**
 * Get all the wishlists for a user
 *
 * This function only retrieves active wishlists (wishlists with valid plugin statuses - @see nmgr_get_post_statuses())
 * as these are the statuses used for wishlists on the frontend.
 *
 * @param int|string $user_id The user id (optional).
 * Defaults to current logged in user id or guest user id cookie value.
 * @param string $type The wishlist type taxonomy term slug (gift-registry or wishlist)
 * @return array
 */
function nmgr_get_user_wishlists( $user_id = '', $type = 'gift-registry' ) {
	$posts = nmgr_get_user_wishlist_ids( $user_id, $type );

	foreach ( $posts as $key => $id ) {
		if ( !in_array( get_post_status( $id ), nmgr_get_post_statuses() ) ) {
			unset( $posts[ $key ] );
			continue;
		}
		$posts[ $id ] = $posts[ $key ];
		unset( $posts[ $key ] );
	}

	return !empty( $posts ) ? array_filter( array_map( 'nmgr_get_wishlist', $posts ) ) : array();
}

/**
 * Get the count of all a user's wishlists or the current user if no user id is supplied
 *
 * This function only retrieves wishlists with valid plugin statuses (@see nmgr_get_statuses())
 * as these are the statuses used for wishlists on the frontend
 *
 * @param int $user_id The user id (optional)
 * @param string $type The wishlist type taxonomy term slug (gift-registry or wishlist)
 * @return mixed int | NULL
 */
function nmgr_get_user_wishlists_count( $user_id = '', $type = 'gift-registry' ) {
	$post_ids = nmgr_get_user_wishlist_ids( $user_id, $type );

	foreach ( $post_ids as $key => $id ) {
		if ( !is_nmgr_post( $id ) || !in_array( get_post_status( $id ), nmgr_get_post_statuses() ) ) {
			unset( $post_ids[ $key ] );
		}
	}

	return count( $post_ids );
}

/**
 * Compose an svg icon using dynamic arguments provided
 *
 * @param string|array $args The name of the svg icon to get or array of icon parameters.
 *
 * If $args is an array, registered array keys needed to compose the svg are:
 * - icon - [required] string The svg icon name. This should correspond to the name of an svg file in the assets/svg directory or the last part of the id of a symbol element in the sprite file - assets/svg/sprite.svg.
 * - size - [optional] integer|string The svg icon width and height. (uses em unit if no unit is specified). Default 1em (16px).
 * - class - [optional] array Classes to add to the svg icon.
 * - sprite - [optional] boolean Whether to use the loaded svg sprite file (Default),
 *            or to load the single svg icon from another path.
 * - role - [optional] string Svg role attribute. Default img.
 * - id - [optional|required] string Svg id attribute. Expected if "sprite" is true.
 *        If not set, the 'icon' parameter is set as the id by default
 * - style - [optional] string Svg Inline style.
 * - title - [optional] string Svg title attribute.
 * - fill - [optional] string Svg fill attribute.
 * - path - [optional] Full path to the single svg icon file.
 *   (Usually used if not using the loaded svg sprite file).
 *
 * @return string Svg HTML element
 */
function nmgr_get_svg( $args ) {
	$args[ 'id' ] = $args[ 'id' ] ?? ('nmgr-icon-' . ($args[ 'icon' ] ?? ''));
	$args[ 'class' ] = array_merge( ( array ) ($args[ 'class' ] ?? []), [ 'nmgr-icon' ] );
	$args[ 'path' ] = $args[ 'path' ] ?? nmgr()->path . 'assets/svg/';

	return apply_filters( 'nmgr_svg', nmgr()->get_svg( $args ), $args );
}

/**
 * Get a single svg icon file unmodified from the icon directory
 *
 * @param string $icon_name The name of the icon e.g. user.
 * @param string $path The full path to the icon file.
 *
 * @return string Icon html
 */
function nmgr_get_svg_file( $icon_name, $path = '' ) {
	_deprecated_function( __FUNCTION__, '3.0.0', 'nmgr()->get_svg_file' );
	return nmgr()->get_svg_file( $icon_name, $path );
}

/**
 * Get a wishlist by a specified id,
 * or by the current wishlist id in the global object if no id is specified.
 *
 * @param mixed $wishlist_id The wishlist id used to retrieve the wishlist. If zero is supplied, the
 * wishlist id is taken from the current wishlist id in the global context. If the value supplied
 * equates to false, the function assumes no wishlist id is specified and returns false.
 * @param bool $active Whether the wishlist must be active. Default false. (An active wishlist has its
 * post status in the registered post statuses for wishlists. Active wishlists appear on the frontend).
 * @return NMGR_Wishlist|false
 */
function nmgr_get_wishlist( $wishlist_id = 0, $active = false ) {
	if ( false !== $wishlist_id ) {
		$wishlist_id = $wishlist_id ? $wishlist_id : nmgr_get_current_wishlist_id();
	}

	if ( !$wishlist_id ) {
		return false;
	}

	/**
	 * try catch statement is used because the wishlist db class throws an exception
	 * if the wishlist cannot be read from database
	 */
	try {
		$wishlist = apply_filters( 'nmgr_get_wishlist', null, $wishlist_id, $active );

		if ( $wishlist ) {
			return $wishlist;
		} else {
			$wishlist = nmgr()->wishlist( $wishlist_id );
			return $active ? ($wishlist->is_active() ? $wishlist : false) : $wishlist;
		}
	} catch ( Exception $e ) {
		return false;
	}
}

/**
 * Get specific information concerning the main wishlist account page
 *
 * For example, get the name or slug of the account page, as set in the admin settings screen
 *
 * @param string $param The information to get e.g. 'name', 'slug'
 * @return string Wishlist account page information
 */
function nmgr_get_account_details( $param ) {
	$options = nmgr_get_option();
	switch ( $param ) {
		case 'name':
			return isset( $options[ 'my_account_name' ] ) ? esc_html( $options[ 'my_account_name' ] ) : false;
		case 'slug':
			$slug = isset( $options[ 'my_account_name' ] ) ? $options[ 'my_account_name' ] : null;
			return $slug ? remove_accents( strtolower( str_replace( ' ', '-', esc_html( $slug ) ) ) ) : false;
	}
}

/**
 * Add to wishlist messages.
 *
 * @param NMGR_Wishlist $wishlist Wishlist Object
 * @param int|array $products Array of product id to quantity wishlist or single product ID.
 * @param bool      $show_qty Should qty's be shown?
 * @param bool      $return   Return message rather than add it.
 *
 * @return mixed
 */
function nmgr_add_to_wishlist_notice( $wishlist, $products, $show_qty = false, $return = false ) {
	$titles = array();
	$count = 0;
	$type = $wishlist->get_type();

	if ( !is_array( $products ) ) {
		$products = array( $products => 1 );
		$show_qty = false;
	}

	if ( !$show_qty ) {
		$products = array_fill_keys( array_keys( $products ), 1 );
	}

	foreach ( $products as $product_id => $qty ) {
		$titles[] = apply_filters( 'nmgr_add_to_wishlist_qty_html', ( $qty > 1 ? absint( $qty ) . ' &times; ' : '' ), $product_id ) .
			apply_filters( 'nmgr_add_to_wishlist_item_name_in_quotes',
				sprintf(
					/* translators: %s: item name */
					nmgr()->is_pro ? _x( '&ldquo;%s&rdquo;', 'Item name in quotes', 'nm-gift-registry' ) : _x( '&ldquo;%s&rdquo;', 'Item name in quotes', 'nm-gift-registry-lite' ),
					strip_tags( get_the_title( $product_id ) )
				),
				$product_id
		);
		$count += $qty;
	}

	$titles = array_filter( $titles );
	$added_text = sprintf(
		/* translators: 1: item names, 2: wishlist type title */
		nmgr()->is_pro ? _n( '%1$s has been added to your %2$s.', '%1$s have been added to your %2$s.', $count, 'nm-gift-registry' ) : _n( '%1$s has been added to your %2$s.', '%1$s have been added to your %2$s.', $count, 'nm-gift-registry-lite' ),
		wc_format_list_of_items( $titles ),
		esc_html( nmgr_get_type_title( '', false, $type ) )
	);

	// Get success messages.
	$message = sprintf( '<a href="%s" tabindex="1" class="button wc-forward">%s</a> %s',
		esc_url( $wishlist->get_permalink() ),
		sprintf(
			/* translators: %s: wishlist type title */
			nmgr()->is_pro ? __( 'View %s', 'nm-gift-registry' ) : __( 'View %s', 'nm-gift-registry-lite' ),
			nmgr_get_type_title( '', false, $type )
		),
		esc_html( $added_text )
	);

	$message = apply_filters( 'nmgr_add_to_wishlist_notice', $message, $wishlist, $products, $show_qty );

	if ( $return ) {
		return $message;
	} else {
		wc_add_notice( $message );
	}
}

/**
 * Get all wishlists which are in the cart
 * @return array Array of unique wishlist ids
 */
function nmgr_get_wishlists_in_cart( $type = 'gift-registry' ) {
	$wishlists = array();
	if ( is_a( wc()->cart, 'WC_Cart' ) && !WC()->cart->is_empty() ) {
		$cart = WC()->cart->get_cart();

		foreach ( $cart as $cart_item ) {
			$nmgr_data = nmgr_get_cart_item_data( $cart_item, 'wishlist_item' );
			if ( $nmgr_data ) {
				$wishlist_id = $nmgr_data[ 'wishlist_id' ];
				$wishlist = nmgr_get_wishlist( $wishlist_id, true );
				if ( $wishlist && $wishlist->is_type( $type ) ) {
					$wishlists[] = ( int ) $wishlist_id;
				}
			}
		}

		$wishlists = apply_filters( 'nmgr_get_wishlists_in_cart', $wishlists, $cart, $type );
	}
	return array_unique( $wishlists );
}

/**
 * Get the title attribute to show when a wishlist has a product
 *
 * This depends on the kind of product - simple, variable, grouped
 * as well as the number of wishlists the user has
 */
function nmgr_get_product_in_wishlist_title_attribute( $product, $type = 'gift-registry' ) {
	if ( !$product ) {
		return;
	}

	$count = nmgr_get_user_wishlists_count( '', $type );
	$product_type = $product->get_type();
	$wishlist_type_title = ($count > 1 ? nmgr_get_type_title( '', 1, $type ) : nmgr_get_type_title( '', false, $type ) );

	switch ( $product_type ) {
		case 'variable':
			return sprintf(
				nmgr()->is_pro ?
				/* translators: %s: wishlist type title */
				_n(
					'This product has variations in your %s',
					'This product has variations in one or more of your %s',
					$count,
					'nm-gift-registry'
				) :
				/* translators: %s: wishlist type title */
				_n(
					'This product has variations in your %s',
					'This product has variations in one or more of your %s',
					$count,
					'nm-gift-registry-lite'
				),
				$wishlist_type_title
			);
		case 'grouped':
			return sprintf(
				nmgr()->is_pro ?
				/* translators: %s: wishlist type title */
				_n(
					'This product has child products in your %s',
					'This product has child products in one or more of your %s',
					$count,
					'nm-gift-registry'
				) :
				/* translators: %s: wishlist type title */
				_n(
					'This product has child products in your %s',
					'This product has child products in one or more of your %s',
					$count,
					'nm-gift-registry-lite'
				),
				$wishlist_type_title
			);
		default:
			return sprintf(
				nmgr()->is_pro ?
				/* translators: %s: wishlist type title */
				_n(
					'This product is in your %s',
					'This product is in one or more of your %s',
					$count,
					'nm-gift-registry'
				) :
				/* translators: %s: wishlist type title */
				_n(
					'This product is in your %s',
					'This product is in one or more of your %s',
					$count,
					'nm-gift-registry-lite'
				),
				$wishlist_type_title
			);
	}
}

/**
 * Check whether the current user has a product in any of his wishlists
 *
 * @param WC_Product $product
 * @return boolean True or false
 */
function nmgr_user_has_product_in_wishlist( $product, $type = 'gift-registry' ) {
	$wishlists = nmgr_get_user_wishlists( '', $type );
	if ( $wishlists ) {
		foreach ( $wishlists as $wishlist ) {
			if ( $wishlist && $wishlist->has_product( $product ) ) {
				return true;
			}
		}
	}
	return false;
}

/**
 * Query string keys for adding an item to the wishlist
 *
 * This function is used to prevent hardcoding the query keys in various files.
 * In case they change later, they would only change here
 *
 * @param string $name query key to get for value
 * @return string Query key
 */
function nmgr_query_key( $name = '' ) {
	$query_keys = array(
		'product_id' => 'nmgr_pid',
		'wishlist_id' => 'nmgr_wid',
		'quantity' => 'nmgr_qty',
		'favourite' => 'nmgr_fav',
		'variation_id' => 'nmgr_vid',
		'wishlist' => 'nmgr_w', // accepts wishlist id or slug.
	);

	return $name ? (isset( $query_keys[ $name ] ) ? $query_keys[ $name ] : null) : $query_keys;
}

/**
 * Get the post statuses used by the plugin
 *
 * These are the statuses the plugin uses for active wishlists on the frontend.
 * All other wordpress post statuses are currently ignored.
 *
 * Default post statuses:
 * - publish
 * - private
 *
 * @return array
 */
function nmgr_get_post_statuses() {
	return apply_filters( 'nmgr_post_statuses', array( 'publish', 'private' ) );
}

/**
 * Get the standard date format used by the plugin to display dates sitewide
 *
 * This function simply allows the date format to be filtered so that a different
 * date format can be used to display dates sitewide.
 *
 * @return string
 */
function nmgr_date_format() {
	_deprecated_function( __FUNCTION__, '2.4.2' );
	return apply_filters_deprecated( 'nmgr_date_format', [ get_option( 'date_format' ) ], '2.4.2' );
}

/**
 * Get the registered php date format for the plugin.
 *
 * This date format is used by default to format all dates displayed by the plugin sitewide
 * and also to validate dates submitted through jquery-datepicker
 *
 * @return string
 */
function nmgr_php_date_format() {
	_deprecated_function( __FUNCTION__, '2.4.2' );
	return get_option( 'date_format' );
}

/**
 * Get a localized date based on date format
 *
 * @param string $date Date to format
 * @param string $format Date format to use. Default is default wordpress date format
 * @return string
 */
function nmgr_format_date( $date, $format = '' ) {
	$datetime = nmgr_get_datetime( $date );
	if ( $date && $datetime ) {
		$date_format = $format ? $format : get_option( 'date_format' );
		$function = function_exists( 'wp_date' ) ? 'wp_date' : 'date_i18n';
		return call_user_func( $function, $date_format, $datetime->getTimestamp() );
	}
	return $date;
}

/**
 * Get variation attributes posted by a form
 *
 * @param int $variation_id Id of the variation product
 * @param array $post posted data contaIning variations
 * @return array
 */
function nmgr_get_posted_variations( $variation_id, $post = '' ) {
	$variations = array();
	$posted_data = array();

	if ( !$variation_id ) {
		return $variations;
	}

	if ( empty( $post ) ) {
		$posted_data = $_REQUEST; // phpcs:ignore WordPress.Security.NonceVerification
	} elseif ( is_string( $post ) ) {
		parse_str( $post, $posted_data );
	} elseif ( is_array( $post ) ) {
		$posted_data = $post;
	}

	$product = wc_get_product( $variation_id );
	$product_parent_id = $product->get_parent_id();
	$product_parent = wc_get_product( $product_parent_id );

	if ( !$product_parent ) {
		return $variations;
	}

	foreach ( $posted_data as $key => $value ) {
		if ( false === strpos( $key, 'attribute_' ) ) {
			unset( $posted_data[ $key ] );
		}
	}

	foreach ( $product_parent->get_attributes() as $attribute ) {
		if ( !$attribute[ 'is_variation' ] ) {
			continue;
		}
		$attribute_key = 'attribute_' . sanitize_title( $attribute[ 'name' ] );

		if ( isset( $posted_data[ $attribute_key ] ) ) {
			if ( $attribute[ 'is_taxonomy' ] ) {
				$value = sanitize_title( wp_unslash( $posted_data[ $attribute_key ] ) );
			} else {
				$value = html_entity_decode( wc_clean( wp_unslash( $posted_data[ $attribute_key ] ) ), ENT_QUOTES, get_bloginfo( 'charset' ) );
			}
			$variations[ $attribute_key ] = $value;
		}
	}

	return $variations;
}

/**
 * Add the plugin prefix to the specified fields keys
 * except fields that have $args['prefix'] set to false
 *
 * @param array $fields Form fields to add prefix to
 * @return array Prefixed form fields
 */
function nmgr_add_prefix( $fields ) {
	$prefixed = array();
	foreach ( $fields as $name => $args ) {
		if ( (isset( $args[ 'prefix' ] ) && !$args[ 'prefix' ]) || false !== strpos( $name, 'nmgr_' ) ) {
			$prefixed[ $name ] = $args;
			continue;
		}
		$prefixed[ 'nmgr_' . $name ] = $args;
	}
	return $prefixed;
}

/**
 * Remove plugin prefix from supplied fields
 *
 * The function removes the prefix from field keys and values if they are strings or arrays of strings
 *
 * @param string | array $data Fields to remove prefix from
 * @return array Fields with prefix removed from keys and values
 */
function nmgr_remove_prefix( $data ) {
	$data_array = ( array ) $data;
	$new_data = array();

	foreach ( $data_array as $key => $value ) {
		$key = str_replace( 'nmgr_', '', $key );
		$value = is_string( $value ) ? str_replace( 'nmgr_', '', $value ) : $value;

		if ( is_array( $value ) ) {
			$value = array_map( function( $val ) {
				return is_string( $val ) ? str_replace( 'nmgr_', '', $val ) : $val;
			}, $value );
		}
		$new_data[ $key ] = $value;
	}
	return $new_data;
}

/**
 * Include the svg sprite file in a page
 */
function nmgr_include_sprite_file() {
	$sprite_file = nmgr()->path . 'assets/svg/sprite.svg';
	if ( file_exists( $sprite_file ) ) {
		include_once $sprite_file;
	}
}

/**
 * Checks whether the post content contains any of the specified shortcodes
 *
 * This function is a simply modification of woocommerce's wc_post_content_has_shortcode
 * function to allow checking for multiple shortcodes at once
 *
 * @param string|array $tags Shortcode tag(s) to check for
 * @see wc_post_content_has_shortcode()
 * @return boolean
 */
function nmgr_post_content_has_shortcodes( $tags ) {
	_deprecated_function( __FUNCTION__, '4.0.0', 'nmgr_post_content_has_shortcode' );

	foreach ( ( array ) $tags as $tag ) {
		if ( wc_post_content_has_shortcode( $tag ) ) {
			return true;
		}
	}
	return false;
}

if ( !function_exists( 'nmgr_get_account_tabs' ) ) {

	/**
	 * Tabs used on account page
	 *
	 * @return array
	 */
	function nmgr_get_account_tabs( $wishlist = false ) {
		_deprecated_function( __FUNCTION__, '4.0.0' );
		$account = nmgr()->account( $wishlist );
		$sections_data = $account->get_sections_data();

		foreach ( $sections_data as $key => $data ) {
			$sections_data[ $key ][ 'tab_id' ] = "nmgr-tab-{$key}";
			$sections_data[ $key ][ 'tab_content_id' ] = "tab-{$key}";
		}

		return $sections_data;
	}

}

if ( !function_exists( 'nmgr_get_no_wishlist_placeholder' ) ) {

	/**
	 * Placeholder content displayed on wishlist account tabs when no wishlist exists
	 *
	 * @param string $tab The name of the tab.
	 * @param boolean $echo Whether to echo the content. Default false.
	 */
	function nmgr_get_no_wishlist_placeholder( $tab = '', $echo = false ) {
		_deprecated_function( __FUNCTION__, '2.2.1', 'nmgr_get_default_account_section_content' );

		$content = nmgr_get_default_account_section_content( $tab );

		if ( $echo ) {
			echo $content;
		} else {
			return $content;
		}
	}

}

/**
 * Get the default content that should be shown on an account section.
 * This content is typically shown when there is no wishlist or the account section has no content.
 *
 * @param string $section The name of the account section or tab
 * @param mixed $wishlist The wishlist object, if a wishlist already exists.
 * @return html
 */
function nmgr_get_default_account_section_content( $section = '', $wishlist = '' ) {
	switch ( $section ) {
		case 'items':
			$icon = 'cart-empty';
			break;
		case 'images':
			$icon = 'camera';
			break;
		case 'messages':
			$icon = 'bubble';
			break;
		case 'orders':
			$icon = 'cart-full';
			break;
		case 'shipping':
			$icon = 'box-open';
			break;
		case 'settings':
			$icon = 'gear';
			break;

		default:
			$icon = 'heart';
			break;
	}

	$svg = '';
	$svg_args = '';

	$icon = apply_filters( 'nmgr_default_account_section_svg_icon', $icon, $section, $wishlist );

	if ( $icon ) {
		$default_svg_args = array(
			'icon' => $icon,
			'size' => nmgr()->post_thumbnail_size() / 16, // convert px to em
			'fill' => '#f8f8f8',
		);

		$svg_args = apply_filters( 'nmgr_default_account_section_svg_args', $default_svg_args, $section, $wishlist );
	}

	if ( $svg_args ) {
		$svg = sprintf( '<div class="nmgr-no-wishlist-placeholder-svg nmgr-text-center">%s</div>',
			nmgr_get_svg( $svg_args )
		);
	}

	$template = nmgr_get_template( 'account/call-to-action-no-wishlist.php', array( 'section' => $section, 'wishlist' => $wishlist ) );

	$call_to_action = apply_filters( 'nmgr_default_account_section_call_to_action', $template, $section, $wishlist );

	$content = apply_filters( 'nmgr_default_account_section_content', ($svg . $call_to_action ), $section, $wishlist );

	return $content;
}

if ( !function_exists( 'nmgr_get_account_template' ) ) {

	/**
	 * Template for displaying a user's wishlist account information
	 *
	 * @param int|string|NMGR_Wishlist|array $atts Attributes needed to compose the template.
	 * Currently accepted $atts attributes if array:
	 * - id [int|string|NMGR_Wishlist] optional. The id, slug or object of the wishlist to display account information for.
	 * Default to current wishlist id in global context.
	 *
	 * @param boolean $echo Whether to echo the template. Default false.
	 * @return string Template html
	 */
	function nmgr_get_account_template( $atts = '', $echo = false ) {
		_deprecated_function( __FUNCTION__, '4.0.0', 'nmgr_get_wishlist_template()' );
		ob_start();
		function_exists( 'wc_print_notice' ) ? wc_print_notice( nmgr_get_relocation_notice(), 'notice' ) : '';
		$template = ob_get_clean();

		if ( $echo ) {
			echo $template; // phpcs:ignore WordPress.Security.EscapeOutput.OutputNotEscaped
		} else {
			return $template;
		}
	}

}

if ( !function_exists( 'nmgr_get_account_wishlist_template' ) ) {

	/**
	 * Template for managing account information for a single wishlist
	 *
	 * @param int|NMGR_Wishlist|array $atts Attributes needed to compose the template.
	 * Currently accepted $atts attributes if array:
	 * - id [int|NMGR_Wishlist] Wishlist id or instance of NMGR_Wishlist.
	 *   Default none - id is taken from the global context if present @see nmgr_get_current_wishlist_id().
	 *
	 * @param boolean $echo Whether to echo the template. Default false.
	 *
	 * @return string Template html
	 */
	function nmgr_get_account_wishlist_template( $atts = '', $echo = false ) {
		_deprecated_function( __FUNCTION__, '4.0.0' );
	}

}

function nmgr_get_account_section( $section, $att = false, $echo = false ) {
	$id = 0;
	if ( is_array( $att ) && !empty( $att[ 'id' ] ) ) {
		$id = $att[ 'id' ];
		unset( $att[ 'id' ] );
	} elseif ( is_numeric( $att ) ) {
		$id = $att;
		$att = [];
	} elseif ( is_a( $att, NMGR_Wishlist::class ) ) {
		$id = $att->get_id();
		$att = [];
	}

	$acc = nmgr()->account( $id )->set_section( $section );

	if ( !empty( $att ) ) {
		$acc->set_args( $att );
	}

	$template = $acc->get_section_template();

	if ( $echo ) {
		echo $template; // phpcs:ignore WordPress.Security.EscapeOutput.OutputNotEscaped
	} else {
		return $template;
	}
}

if ( !function_exists( 'nmgr_get_overview_template' ) ) {

	function nmgr_get_overview_template( $atts = '', $echo = false ) {
		_deprecated_function( __FUNCTION__, '4.0.0' );
		return '';
	}

}

if ( !function_exists( 'nmgr_get_profile_template' ) ) {

	function nmgr_get_profile_template( $atts = '', $echo = false ) {
		$account = nmgr()->account( $atts[ 'id' ] ?? false );
		if ( !empty( $atts[ 'type' ] ) ) {
			$account->set_type( $atts[ 'type' ] );
		}
		$template = $account->set_section( 'profile' )->get_section_template();

		if ( $echo ) {
			echo $template; // phpcs:ignore WordPress.Security.EscapeOutput.OutputNotEscaped
		} else {
			return $template;
		}
	}

}

if ( !function_exists( 'nmgr_get_items_template' ) ) {


	function nmgr_get_items_template( $id, $echo = false ) {
		return nmgr_get_account_section( 'items', $id, $echo );
	}

}

if ( !function_exists( 'nmgr_get_shipping_template' ) ) {

	function nmgr_get_shipping_template( $id, $echo = false ) {
		return nmgr_get_account_section( 'shipping', $id, $echo );
	}

}

/**
 * Template for displaying a single wishlist
 *
 * @param int|NMGR_Wishlist|array $atts Attributes needed to compose the template.
 * Currently accepted $atts attributes if array:
 * - id [int|NMGR_Wishlist] Wishlist id or instance of NMGR_Wishlist.
 *   Default none - id is taken from the global context if present @see nmgr_get_current_wishlist_id().
 *
 * @param boolean $echo Whether to echo the template. Default false.
 *
 * @return string Template html
 */
function nmgr_get_wishlist_template( $atts = '', $echo = false ) {
	$vars = nmgr()->merge_args(
		array(
			'id' => is_array( $atts ) ? 0 : $atts,
		),
		array_filter( ( array ) $atts )
	);


	$id = $vars[ 'id' ];

	if ( !empty( $id ) && is_numeric( $id ) ) {
		$id = absint( $id );
	} elseif ( is_a( $id, 'NMGR_Wishlist' ) ) {
		$id = $id->get_id();
	} elseif ( empty( $id ) ) {
		$id = ( int ) nmgr_get_current_wishlist_id();
	}

	$template_class = new NMGR_Wishlist_Template( $id );

	if ( $echo ) {
		echo $template_class->get();
	} else {
		return $template_class->get();
	}
}

if ( !function_exists( 'nmgr_get_share_template' ) ) {

	function nmgr_get_share_template( $id, $echo = false ) {
		$wishlist = nmgr_get_wishlist( $id, true );
		$type = $wishlist ? $wishlist->get_type() : 'gift-registry';

		if ( !$wishlist ||
			!is_nmgr_enabled( $type ) ||
			('publish' !== $wishlist->get_status()) ) {
			return;
		}

		// We should show the sharing links if at least one share option is enabled
		$options = nmgr_get_type_option( $type );
		$to_share = false;

		foreach ( $options as $key => $value ) {
			if ( false !== strpos( $key, 'share_on' ) && !empty( $value ) ) {
				$to_share = true;
				break;
			}
		}

		if ( !$to_share ) {
			return;
		}

		$vars = array(
			'wishlist' => $wishlist,
		);

		$template = apply_filters( 'nmgr_share_template', nmgr_get_template( 'account/sharing.php', $vars ), $vars );

		if ( $echo ) {
			echo $template; // phpcs:ignore WordPress.Security.EscapeOutput.OutputNotEscaped
		} else {
			return $template;
		}
	}

}

/**
 * Add a product to a wishlist
 *
 * @param NMGR_Wishlist $wishlist The wishlist object
 * @param WC_Product $product The product object
 * @param int $quantity Quantity to be added
 * @param int $favourite Product favourite in wishlist
 * @param array $variations Product variations
 * @param array $item_data Extra data passed to the wishlist item
 *
 * @return array|false Array of product id to quantity added to wishlist.
 * False if product is not added to the wishlist
 * @throws Exception
 */
function nmgr_add_to_wishlist( $wishlist, $product, $quantity, $favourite = null, $variations = array(), $item_data = array() ) {
	if ( $quantity < 1 || !$product || 'trash' === $product->get_status() ) {
		return false;
	}

	$add_to_wishlist_params = array(
		'wishlist' => $wishlist,
		'product' => $product,
		'quantity' => $quantity,
		'favourite' => $favourite,
		'variations' => $variations,
		'item_data' => $item_data,
	);

	$product_id = $product->is_type( 'variation' ) ? $product->get_parent_id() : $product->get_id();
	$variation_id = $product->is_type( 'variation' ) ? $product->get_id() : 0;
	$extra_item_data = apply_filters( 'nmgr_add_to_wishlist_item_data', $item_data, $add_to_wishlist_params );
	$add_to_wishlist_params[ 'item_data' ] = $extra_item_data;
	$unique_id = $wishlist->generate_unique_id( $product_id, $variation_id, $variations, $add_to_wishlist_params[ 'item_data' ] );
	$add_to_wishlist_params[ 'unique_id' ] = $unique_id;
	$type = $wishlist->get_type();

	if ( !apply_filters( 'nmgr_maybe_add_to_wishlist', true, $add_to_wishlist_params ) ) {
		return false;
	}

	$item_id = $wishlist->add_item( $product, $quantity, $favourite, $variations, $add_to_wishlist_params[ 'item_data' ] );

	if ( $item_id ) {
		$add_to_wishlist_params[ 'item_id' ] = $item_id;
		do_action( 'nmgr_added_to_wishlist', $add_to_wishlist_params );
		return array( $product->get_id() => $quantity );
	} else {
		throw new Exception( sprintf(
				/* translators: 1: product name, 2: wishlist type title */
				nmgr()->is_pro ? __( '%1$s could not be added to your %2$s', 'nm-gift-registry' ) : __( '%1$s could not be added to your %2$s', 'nm-gift-registry-lite' ),
				$product->get_name(),
				nmgr_get_type_title( '', false, $type )
			) );
	}
}

/**
 * Get the button for adding a product to the wishlist
 *
 * @param int|WC_Product $atts Attributes needed to compose the button
 * Currently accepted $atts attributes if array:
 * - id [int|WC_Product] Product id or instance of WC_Product.
 *   Default none - id is taken from the global product variable if present.
 *
 * @param $echo boolean Whether to echo the template. Default false.
 * @return string Button html
 */
function nmgr_get_add_to_wishlist_button( $atts = false, $echo = false ) {
	global $product;

	if ( !is_array( $atts ) && !empty( $atts ) ) {
		$product = wc_get_product( $atts );
	} elseif ( is_array( $atts ) && isset( $atts[ 'id' ] ) ) {
		$product = wc_get_product( $atts[ 'id' ] );
	} else {
		$product = wc_get_product( $product );
	}

	$type = !empty( $atts[ 'type' ] ) ? $atts[ 'type' ] : 'gift-registry';

	if ( !is_nmgr_enabled( $type ) || !$product ) {
		return;
	}

	/**
	 * Whether to show the button conditionally
	 * (This can be modified by a filter - see 'nmgr_show_add_to_wishlist_button')
	 */
	$show_button = true;

	// Don't show the button for grouped products if all have no options
	if ( $product->is_type( 'grouped' ) ) {
		$grouped_products = array_filter(
			array_map( 'wc_get_product', $product->get_children() ),
			'wc_products_array_filter_visible_grouped' );

		if ( !$grouped_products ) {
			return;
		}

		$maybe_show = count( $grouped_products );

		foreach ( $grouped_products as $grouped_product ) {
			if ( $grouped_product->has_options() ) {
				--$maybe_show;
			}
		}

		$show_button = ( bool ) $maybe_show;
	}

	// Should we show the add to wishlist button
	if ( !apply_filters( 'nmgr_show_add_to_wishlist_button', $show_button, $product, $type ) ) {
		return;
	}

	/**
	 * At this point, we are showing the wishlist button, so run through all the use cases
	 */
	// Get the user's wishlists
	$wishlists = (nmgr()->is_pro && nmgr_get_type_option( $type, 'allow_multiple_wishlists' )) ?
		nmgr_get_user_wishlists( '', $type ) :
		array_filter( array( nmgr_get_wishlist( nmgr_get_user_default_wishlist_id( '', $type ), true ) ) );

	// How should we create a wishlist for users without any wishlist when the add to wishlist button is clicked?
	$create_wishlist = false;
	if ( empty( $wishlists ) ) {
		if ( 'gift-registry' === $type ) {
			if ( apply_filters( 'nmgr_redirect_to_create_wishlist', false ) ) {
				$create_wishlist = 'redirect';
			} elseif ( nmgr_get_option( 'default_wishlist_title' ) ) {
				$create_wishlist = 'auto';
			} else {
				$create_wishlist = 'modal';
			}
		} else {
			$create_wishlist = 'auto';
		}
	} else {
		foreach ( $wishlists as $key => $wish ) {
			if ( method_exists( $wish, 'is_archived' ) && $wish->is_archived() ) {
				unset( $wishlists[ $key ] );
			}
		}
	}

	// Set class for the add to wishlist button position on single and archive pages
	$wrapper_class = array();
	$thumbnail_positions = array( 'thumbnail_top_left', 'thumbnail_top_right', 'thumbnail_bottom_left', 'thumbnail_bottom_right' );
	$archive_position = nmgr_get_type_option( $type, 'add_to_wishlist_button_position_archive' );
	$single_position = nmgr_get_type_option( $type, 'add_to_wishlist_button_position_single' );
	$button_location = is_product() ? $single_position : (is_nmgr_shop_loop() ? $archive_position : null);

	if ( in_array( $button_location, $thumbnail_positions ) ) {
		$wrapper_class[] = 'on-thumbnail';

		switch ( $button_location ) {
			case 'thumbnail_top_left':
				$wrapper_class[] = 'nmgr-left';
				$wrapper_class[] = 'nmgr-top';
				break;
			case 'thumbnail_top_right':
				$wrapper_class[] = 'nmgr-right';
				$wrapper_class[] = 'nmgr-top';
				break;
			case 'thumbnail_bottom_right':
				$wrapper_class[] = 'nmgr-right';
				$wrapper_class[] = 'nmgr-bottom';
				break;
			case 'thumbnail_bottom_left':
				$wrapper_class[] = 'nmgr-left';
				$wrapper_class[] = 'nmgr-bottom';
				break;
		}
	}

	$umode = 'simple';
	if ( nmgr()->is_pro &&
		('gift-registry' === $type || nmgr_get_type_option( $type, 'allow_multiple_wishlists' ) ) ) {
		$umode = 'advanced';
	}
	$mode = apply_filters( 'nmgr_add_to_wishlist_mode', $umode, [ 'type' => $type ] );

	$button_text = str_replace(
		'{wishlist_type_title}',
		nmgr_get_type_title( '', false, $type ),
		nmgr_get_type_option( $type, 'add_to_wishlist_button_text' )
	);

	// Set default arguments for the button
	$defaults = array(
		'type' => $type,
		'wishlists' => $wishlists,
		'wishlist' => reset( $wishlists ),
		'mode' => $mode,
		'wrapper_class' => $wrapper_class,
		'form_class' => array(
			'nmgr-add-to-wishlist-form',
			"nmgr-add-to-wishlist-form-{$product->get_id()}",
			$type,
			nmgr_user_has_product_in_wishlist( $product, $type ) ? 'product-in-wishlist' : '',
			'product-type-' . $product->get_type(),
			is_nmgr_shop_loop() ? 'archive' : 'single',
			'nmgr-ajax-add-to-wishlist',
		),
		'form_attributes' => array(
			'data-nmgr_product_id' => $product->get_id(),
			'data-create_wishlist' => $create_wishlist,
			'data-show_notice_archive' => 0,
			'data-show_notice_single' => 0,
			'data-show_view_wishlist_link' => 1,
			'data-show_animation' => 1,
			'data-add_to_wishlist_mode' => $mode,
			'data-wishlist_type' => $type,
		),
		'button_class' => array(
			!is_nmgr_shop_loop() ? 'alt' : '',
			'nmgr-add-to-wishlist-button',
		),
		'button_attributes' => array(
			'aria-label' => $button_text,
			'rel' => 'nofollow',
			'role' => 'button',
			'tabindex' => 0,
		),
		'button_text' => $button_text,
		'button_icon' => nmgr_get_svg( array(
			'icon' => 'heart',
			'size' => .75,
			'fill' => '#ccc',
			'class' => 'nmgr-animate in-wishlist-icon in-wishlist',
			'style' => 'margin-left:0.1875em;',
			'title' => nmgr_get_product_in_wishlist_title_attribute( $product, $type ),
		) ),
	);

	if ( $product->is_type( 'variable' ) ) {
		$defaults[ 'form_attributes' ][ 'data-make_a_selection_text' ] = sprintf(
			/* translators: %s wishlist type title */
			nmgr()->is_pro ? esc_attr__( 'Please select some product options before adding this product to your %s.', 'nm-gift-registry' ) : esc_attr__( 'Please select some product options before adding this product to your %s.', 'nm-gift-registry-lite' ),
			nmgr_get_type_title( '', false, $type )
		);
		$defaults[ 'form_attributes' ][ 'data-unavailable_text' ] = nmgr()->is_pro ?
			esc_attr__( 'Sorry, this product is unavailable. Please choose a different combination.', 'nm-gift-registry' ) :
			esc_attr__( 'Sorry, this product is unavailable. Please choose a different combination.', 'nm-gift-registry-lite' );
	}

	// Should we redirect when the add to wishlist button is clicked?
	$redirect = false;

	// if we don't have a valid user but we are showing the wishlist button, redirect to login page with notice
	if ( !is_nmgr_user( $type ) ) {
		$redirect = true;
		$defaults[ 'permalink' ] = wc_get_page_permalink( 'myaccount' );
		$defaults[ 'permalink_args' ] = array(
			'nmgr-notice' => 'require-login',
			'nmgr-redirect' => $_SERVER[ 'REQUEST_URI' ],
			'nmgr-type' => $type,
		);
	} else {
		/**
		 * if we have a valid user who has no wishlists to add products to and the admin wants to
		 * redirect him to create a wishlist instead of using the modal, setup the redirect
		 */
		if ( 'redirect' === $create_wishlist ) {
			$redirect = true;
			$defaults[ 'permalink' ] = nmgr_get_url( $type, 'new' );
			$defaults[ 'permalink_args' ] = array(
				'nmgr-notice' => 'create-wishlist',
				'nmgr-type' => $type,
			);
		} elseif ( is_nmgr_shop_loop() && $product->is_type( array( 'variable' ) ) ) {
			/**
			 * The user is logged in and if he has wishlists to add products to
			 * For variable and grouped products on archive pages redirect to actual product page with notice
			 */
			$redirect = true;
			$defaults[ 'permalink' ] = $product->get_permalink();
			$defaults[ 'permalink_args' ] = array(
				'nmgr-notice' => 'select-product',
				'nmgr-pt' => $product->get_type(),
				'nmgr-type' => $type,
			);
		}
	}

	if ( is_nmgr_shop_loop() && $product->is_type( array( 'variable' ) ) ) {
		$defaults[ 'button_text' ] .= ' *';
	}

	$defaults[ 'form_class' ][] = $redirect ? 'redirect' : '';
	$defaults[ 'button_text' ] = apply_filters( 'nmgr_add_to_wishlist_button_text', $defaults[ 'button_text' ], $type );
	$args = wp_parse_args( apply_filters( 'nmgr_add_to_wishlist_button_args', $defaults, $product, $type ), $defaults );

	$args[ 'wrapper_class' ] = isset( $args[ 'wrapper_class' ] ) ? implode( ' ', array_filter( ( array ) $args[ 'wrapper_class' ] ) ) : '';
	$args[ 'form_class' ] = isset( $args[ 'form_class' ] ) ? implode( ' ', array_filter( ( array ) $args[ 'form_class' ] ) ) : '';
	$args[ 'form_attributes' ] = isset( $args[ 'form_attributes' ] ) ? wc_implode_html_attributes( ( array ) $args[ 'form_attributes' ] ) : '';
	$args[ 'button_class' ] = isset( $args[ 'button_class' ] ) ? implode( ' ', array_filter( ( array ) $args[ 'button_class' ] ) ) : '';

	if ( isset( $args[ 'button_attributes' ][ 'aria-label' ] ) ) {
		$args[ 'button_attributes' ][ 'aria-label' ] = wp_strip_all_tags( $args[ 'button_attributes' ][ 'aria-label' ] );
	}

	$button_params = array(
		'class' => esc_attr( $args[ 'button_class' ] ),
		'attributes' => wc_implode_html_attributes( ( array ) $args[ 'button_attributes' ] ),
		'text' => isset( $args[ 'button_text' ] ) ? esc_html( $args[ 'button_text' ] ) : '',
		'icon' => isset( $args[ 'button_icon' ] ) ? $args[ 'button_icon' ] : null
	);

	// Get the add-to-wishlist button type
	$button_type = nmgr_get_type_option( $type, 'add_to_wishlist_button_type', 'button' );
	$button_html = '';

	switch ( $button_type ) {
		case 'icon-heart':
			$not_in_wishlist_svg_args = array_merge( array(
				'icon' => 'heart-empty',
				'size' => 2,
				'class' => 'not-in-wishlist ' . $button_params[ 'class' ],
				'fill' => 'currentColor',
				'title' => $button_text,
				), $args[ 'button_attributes' ] );

			$in_wishlist_svg_args = array_merge( array(
				'icon' => 'heart',
				'size' => 2,
				'class' => 'nmgr-animate in-wishlist ' . $button_params[ 'class' ],
				'fill' => 'currentColor',
				'title' => nmgr_get_product_in_wishlist_title_attribute( $product, $type ),
				), $args[ 'button_attributes' ] );

			$button_html = nmgr_get_svg( $not_in_wishlist_svg_args ) . nmgr_get_svg( $in_wishlist_svg_args );
			break;

		case 'custom':
			if ( nmgr()->is_pro && function_exists( 'nmgr_get_custom_add_to_wishlist_button' ) ) {
				$button_html = nmgr_get_custom_add_to_wishlist_button( $type );
			}
			break;
	}

	if ( !$button_html ) {
		if ( is_nmgr_shop_loop() ) {
			$button_html = sprintf( "<a href='#' class='button %s' %s>%s</a>",
				$button_params[ 'class' ],
				$button_params[ 'attributes' ],
				$button_params[ 'text' ] . $button_params[ 'icon' ]
			);
		} else {
			$button_html = sprintf( "<button type='submit' class='button %s' %s>%s</button>",
				$button_params[ 'class' ],
				$button_params[ 'attributes' ],
				$button_params[ 'text' ] . $button_params[ 'icon' ]
			);
		}
	}

	/**
	 * Filter the add to wishlist button
	 *
	 * @param string $button_template The button html
	 * @param array $button_params Array keys of parameters used to compose the button html
	 * 					and their values. Default parameters: id, class, attributes, text, icon
	 * @param array $args Arguments used to compose the overall button template
	 */
	$args[ 'button' ] = apply_filters( 'nmgr_add_to_wishlist_button', $button_html, $button_params, $args, $type );

	$template = !empty( $args[ 'button' ] ) ? nmgr_get_template( 'add-to-wishlist/form.php', $args ) : '';

	if ( $echo ) {
		echo $template; // phpcs:ignore WordPress.Security.EscapeOutput.OutputNotEscaped
	} else {
		return $template;
	}
}

/**
 * Displays the button for adding a product to the wishlist
 *
 * @param int|WC_Product $atts Attributes needed to compose the button
 * Currently accepted $atts attributes if array:
 * - id [int|WC_Product] Product id or instance of WC_Product. Default none - id is taken from the global product variable context if present.
 * @return string button html
 */
function nmgr_add_to_wishlist_button( $atts = false ) {
	nmgr_get_add_to_wishlist_button( $atts, true );
}

function nmgr_add_to_wishlist_button_2() {
	nmgr_get_add_to_wishlist_button( [ 'type' => 'wishlist' ], true );
}

/**
 * Get the possible titles that can be used for the wishlist type
 *
 * @return array
 */
function nmgr_get_type_titles() {
	$titles = array(
		'gift_registry' => array(
			'singular' => nmgr()->is_pro ?
			__( 'gift registry', 'nm-gift-registry' ) :
			__( 'gift registry', 'nm-gift-registry-lite' ),
			'plural' => nmgr()->is_pro ?
			__( 'gift registries', 'nm-gift-registry' ) :
			__( 'gift registries', 'nm-gift-registry-lite' )
		),
		'wishlist' => array(
			'singular' => nmgr()->is_pro ?
			__( 'wishlist', 'nm-gift-registry' ) :
			__( 'wishlist', 'nm-gift-registry-lite' ),
			'plural' => nmgr()->is_pro ?
			__( 'wishlists', 'nm-gift-registry' ) :
			__( 'wishlists', 'nm-gift-registry-lite' )
		),
		'gift_list' => array(
			'singular' => nmgr()->is_pro ?
			__( 'gift list', 'nm-gift-registry' ) :
			__( 'gift list', 'nm-gift-registry-lite' ),
			'plural' => nmgr()->is_pro ?
			__( 'gift lists', 'nm-gift-registry' ) :
			__( 'gift lists', 'nm-gift-registry-lite' )
		),
		'list' => array(
			'singular' => nmgr()->is_pro ?
			__( 'list', 'nm-gift-registry' ) :
			__( 'list', 'nm-gift-registry-lite' ),
			'plural' => nmgr()->is_pro ?
			__( 'lists', 'nm-gift-registry' ) :
			__( 'lists', 'nm-gift-registry-lite' )
		)
	);

	return apply_filters( 'nmgr_type_titles', $titles );
}

/**
 * Get the current title used for the wishlist type
 *
 * @param string $formatting How to format the title (default is lowercase).
 * Possible values are
 * - 'c': Capitalize
 * - 'u': Uppercase
 * - 'cf': Capitalize first word
 *
 * @param boolean $pluralize Whether to get the plural form of the title. Default false
 * @param string $type The wishlist type (gift-registry or wishlist). Default is gift-registry
 *
 * @return string
 */
function nmgr_get_type_title( $formatting = '', $pluralize = false, $type = '' ) {
	$type_title = '';
	$type_titles = nmgr_get_type_titles();
	$wishlist_type = $type ? $type : (nmgr_get_current_type() ?? 'gift-registry');

	if ( 'wishlist' === $wishlist_type ) {
		$type_title = $pluralize ?
			($type_titles[ 'wishlist' ][ 'plural' ] ?? '') :
			($type_titles[ 'wishlist' ][ 'singular' ] ?? '');
	} else {
		$key = nmgr_get_option( 'type_title', 'gift_registry' );

		if ( isset( $type_titles[ $key ] ) ) {
			$singular = isset( $type_titles[ $key ][ 'singular' ] ) ? $type_titles[ $key ][ 'singular' ] : '';
			$plural = isset( $type_titles[ $key ][ 'plural' ] ) ? $type_titles[ $key ][ 'plural' ] : '';
			$type_title = !$pluralize ? $singular : $plural;
		}

		// Default list title
		if ( !$type_title ) {
			if ( $pluralize ) {
				$type_title = nmgr()->is_pro ?
					__( 'gift registries', 'nm-gift-registry' ) :
					__( 'gift registries', 'nm-gift-registry-lite' );
			} else {
				$type_title = nmgr()->is_pro ?
					__( 'gift registry', 'nm-gift-registry' ) :
					__( 'gift registry', 'nm-gift-registry-lite' );
			}
		}
	}

	switch ( $formatting ) {
		case 'cf': // capitalize first word
			$type_title = ucfirst( $type_title );
			break;

		case 'c': // capitalize
			$type_title = ucwords( $type_title );
			break;

		case 'u': // uppercase
			$type_title = strtoupper( $type_title );
			break;

		default: //lowercase
			$type_title = strtolower( $type_title );
			break;
	}

	return apply_filters( 'nmgr_type_title', $type_title, $formatting, $pluralize, $type_titles, $wishlist_type );
}

/**
 * Returns the admin url base page for the plugin
 */
function nmgr_get_admin_url() {
	return admin_url( 'edit.php?post_type=nm_gift_registry' );
}

/**
 * Return a standard plugin tooltip notification
 *
 * @param string $title The notification message to be in the tooltip
 * @return string
 */
function nmgr_get_help_tip( $title ) {
	return '<span class="nmgr-tip nmgr-help-tip nmgr-cursor-help" title="' . $title . '"> &#9432;</span>';
}

/**
 * Order statuses used by the plugin to determine if a payment is cancelled
 *
 * These are the same statuses used by woocommerce to
 * determine whether to increase stock levels
 *
 * Default:
 * - cancelled
 * - pending
 *
 * @return array
 */
function nmgr_get_payment_cancelled_order_statuses() {
	return apply_filters( 'nmgr_payment_cancelled_order_statuses', array( 'cancelled', 'pending' ) );
}

/**
 * Get the url where items can be added to the wishlist
 *
 * This should typically be the shop page url
 *
 * @return string
 */
function nmgr_get_add_items_url() {
	return apply_filters( 'nmgr_add_items_url', wc_get_page_permalink( 'shop' ) );
}

/**
 * Svg tags allowed by the plugin
 *
 * @return array
 */
function nmgr_allowed_svg_tags() {
	return array(
		'svg' => array(
			'id' => true,
			'role' => true,
			'width' => true,
			'height' => true,
			'class' => true,
			'style' => true,
			'fill' => true,
			'xmlns' => true,
			'viewbox' => true,
			'aria-hidden' => true,
			'focusable' => true,
			'data-notice' => true, // may be deprecated soon. Used temporarily.
		),
		'use' => array(
			'xlink:href' => true
		),
		'title' => array(
			'data-title' => true
		),
		'path' => array(
			'fill' => true,
			'fill-rule' => true,
			'd' => true,
			'transform' => true,
		),
		'polygon' => array(
			'fill' => true,
			'fill-rule' => true,
			'points' => true,
			'transform' => true,
			'focusable' => true,
		),
	);
}

/**
 * Sanitize content to allow for svg tags used by the plugin
 *
 * @param string $data Content to sanitize
 * @return string Sanitized content
 */
function nmgr_kses_svg( $data ) {
	return wp_kses( $data, nmgr_allowed_svg_tags() );
}

/**
 * Sanitize content to allow for HTML tags used by WordPress in post content and svg tags used by the plugin
 *
 * This function is simply used to allow the plugin svg tags to be used alongside
 * WordPress allowed HTML tags in post content.
 *
 * @param string $data Content to sanitize
 * @return string Sanitized content
 */
function nmgr_kses_post( $data ) {
	return wp_kses( $data, array_merge( wp_kses_allowed_html( 'post' ), nmgr_allowed_svg_tags() ) );
}

/**
 * Check if a wishlist belongs to a particular user
 *
 * @param int|NMGR_Wishlist $wishlist_id The wishlist id or object.
 * @param int|string $user_id The user id. Optional. Defaults to current logged in user or guest.
 * @return boolean True if user has the wishlist. False if not.
 */
function nmgr_user_has_wishlist( $wishlist_id, $user_id = '' ) {
	global $wpdb;

	$id = 0;
	if ( is_numeric( $wishlist_id ) && 0 < $wishlist_id ) {
		$id = $wishlist_id;
	} elseif ( is_a( $wishlist_id, 'NMGR_Wishlist' ) ) {
		$id = $wishlist_id->get_id();
	}

	if ( $id ) {
		$userid = $user_id ? $user_id : nmgr_get_current_user_id();
		if ( $userid ) {
			$cache_key = 'nmgr_user_wishlist_id_' . $userid;
			$post_id = wp_cache_get( $cache_key );

			if ( !$post_id ) {
				$post_id = $wpdb->get_var( $wpdb->prepare(
						"SELECT COUNT(*) FROM $wpdb->postmeta WHERE meta_key = '_nmgr_user_id' AND post_id = %s AND meta_value = %s",
						$id,
						$userid
					) );
				wp_cache_set( $cache_key, $post_id );
			}
			return ( bool ) $post_id;
		}
	}
	return false;
}

/**
 * Check whether the cart has items belonging to a particular wishlist
 *
 * If no wishlist id is supplied, the function checks for the first wishlist that may have an item in the cart.
 *
 * Alias of nmgr_get_wishlist_in_cart()
 *
 * @param int $wishlist_id The wishlist id to check for.
 * @return boolean
 */
function nmgr_cart_has_wishlist( $wishlist_id = '' ) {
	return ( bool ) nmgr_get_wishlist_in_cart( $wishlist_id );
}

/**
 * Check whether the cart has items belonging to a particular wishlist
 *
 * If no wishlist id is supplied, the function checks for the first wishlist that may have an item in the cart.
 *
 * @param int $wishlist_id The wishlist id to check for.
 * @return int The wishlist id if the cart has the wishlist or 0 if the cart doesn't.
 */
function nmgr_get_wishlist_in_cart( $wishlist_id = '' ) {
	if ( is_a( wc()->cart, 'WC_Cart' ) && !WC()->cart->is_empty() ) {
		$id = 0;
		$cart = WC()->cart->get_cart();

		foreach ( $cart as $cart_item ) {
			$nmgr_data = nmgr_get_cart_item_data( $cart_item, 'wishlist_item' );
			if ( $nmgr_data ) {
				if ( $wishlist_id &&
					(absint( $wishlist_id ) === absint( $nmgr_data[ 'wishlist_id' ] )) &&
					nmgr_get_wishlist( $wishlist_id, true ) ) {
					$id = $wishlist_id;
					break;
				} elseif ( !$wishlist_id && nmgr_get_wishlist( $nmgr_data[ 'wishlist_id' ], true ) ) {
					$id = $nmgr_data[ 'wishlist_id' ];
					break;
				}
			}
		}
		return ( int ) apply_filters( 'nmgr_get_wishlist_in_cart', $id, $cart );
	}
	return 0;
}

if ( !function_exists( 'nmgr_get_search_results_template' ) ) {

	/**
	 * Get the template for outputting wishlist search results
	 *
	 * Alias of nmgr_get_archive_template()
	 *
	 * This function or the shortcode attached to it should be used after wp_loaded hook
	 * as that is when the wp_query global exists.
	 *
	 * @param array $atts Attributes needed to compose the template.
	 * @param bool $echo Whether to echo or return the template.
	 */
	function nmgr_get_search_results_template( $atts = array(), $echo = false ) {
		$defaults = [
			'show_form' => false,
		];

		return nmgr_get_search_template( wp_parse_args( $atts, $defaults ), $echo );
	}

}

/**
 * Check if it is an admin request
 *
 * @return boolean
 */
function is_nmgr_admin_request() {
	return nmgr()->is_admin_request();
}

function nmgr_get_search_template_args( $atts = [] ) {
	if ( isset( $atts[ 'placeholder' ] ) ) {
		_deprecated_argument( __FUNCTION__, '2.5.1', nmgr()->is_pro ?
				__( 'Use "input_placeholder" instead of "placeholder"', 'nm-gift-registry' ) :
				__( 'Use "input_placeholder" instead of "placeholder"', 'nm-gift-registry-lite' )
		);
		$atts[ 'input_placeholder' ] = $atts[ 'placeholder' ];
	}

	$args_unfiltered = nmgr()->merge_args(
		array(
			'show_form' => true,
			'show_results' => true,
			'form_action' => '',
			'show_results_if_no_search_query' => true,
			'input_required' => false,
			'input_placeholder' => sprintf(
				/* translators: %s: wishlist type title */
				nmgr()->is_pro ? __( 'Search %s&hellip;', 'nm-gift-registry' ) : __( 'Search %s&hellip;', 'nm-gift-registry-lite' ),
				nmgr_get_type_title( '', true )
			),
			'submit_button_text' => apply_filters( 'nmgr_search_submit_button_text',
				nmgr()->is_pro ?
				_x( 'Search', 'submit button', 'nm-gift-registry' ) :
				_x( 'Search', 'submit button', 'nm-gift-registry-lite' )
			),
			'hidden_fields' => array() // for now hidden fields can only be added via code and not shortcode
		),
		( array ) $atts
	);

	$using_wp_search = untrailingslashit( home_url() ) ===
		untrailingslashit( esc_url( $args_unfiltered[ 'form_action' ] ) );

	if ( $using_wp_search ) {
		$args_unfiltered[ 'hidden_fields' ][ 'post_type' ] = 'nm_gift_registry';
	}

	$args = wp_parse_args( filter_var_array( $args_unfiltered, array(
		'show_form' => FILTER_VALIDATE_BOOLEAN,
		'show_results' => FILTER_VALIDATE_BOOLEAN,
		'form_action' => FILTER_SANITIZE_URL,
		'show_results_if_no_search_query' => FILTER_VALIDATE_BOOLEAN,
		'input_required' => FILTER_VALIDATE_BOOLEAN,
		'input_placeholder' => FILTER_DEFAULT,
		'submit_button_text' => FILTER_DEFAULT,
		) ), $args_unfiltered );

	return apply_filters( 'nmgr_search_template_args', $args );
}

/**
 * Get the wishlist search template
 * This includes the search form and search results, depending on the arguments provided
 *
 * @param type $atts Attributes needed to compose the template
 * @param boolean $echo Whether to echo the template. Default false.
 * @return string Template html
 */
function nmgr_get_search_template( $atts = '', $echo = false ) {
	$args = nmgr_get_search_template_args( $atts );

	$template = '';

	if ( !$args[ 'show_results_if_no_search_query' ] && !nmgr_get_search_query() ) {
		$args[ 'show_results' ] = false;
	}

	// Whether we are using the default wordpress search that uses 's' query var on the home_url permalink
	$using_wp_search = untrailingslashit( home_url() ) === untrailingslashit( esc_url( $args[ 'form_action' ] ) );

	if ( $args[ 'show_form' ] ) {
		$vars = array(
			'form_action' => $args[ 'form_action' ],
			'input_name' => $using_wp_search ? 's' : 'nmgr_s',
			'input_value' => is_nmgr_search() ? ($using_wp_search ? get_query_var( 's' ) : get_query_var( 'nmgr_s' )) : '',
			'using_wp_search' => $using_wp_search,
			'input_placeholder' => $args[ 'input_placeholder' ],
			'input_required' => $args[ 'input_required' ],
			'hidden_fields' => array_filter( ( array ) $args[ 'hidden_fields' ] ),
			'submit_button_text' => $args[ 'submit_button_text' ],
		);

		$template .= apply_filters( 'nmgr_search_form', nmgr_get_template( 'form-search-wishlist.php', $vars ), $vars );
	}

	if ( $args[ 'show_results' ] ) {
		$template .= nmgr_get_archive_template( $args );
	}

	if ( $echo ) {
		echo $template; // phpcs:ignore WordPress.Security.EscapeOutput.OutputNotEscaped
	} else {
		return $template;
	}
}

/**
 * Check if the current user is a logged in user or a guest.
 *
 * Guests exist when non-logged-in users have been permitted to create and manage wishlists.
 *
 * @return boolean
 */
function is_nmgr_user( $type = 'gift-registry' ) {
	return apply_filters( 'is_nmgr_user', (is_user_logged_in() || is_nmgr_guest( $type ) ) );
}

/**
 * Check if the current user is a guest
 *
 * Guests exist when non-logged-in users have been permitted to create and manage wishlists.
 *
 * @return boolean
 */
function is_nmgr_guest( $type = 'gift-registry' ) {
	$val = ( bool ) !is_user_logged_in() && nmgr_get_type_option( $type, 'allow_guest_wishlists' );
	return apply_filters( 'is_nmgr_guest', $val, $type );
}

/**
 * Get the user id for the current logged in user or guest
 * (Note that this function returns the string value of the user id)
 *
 * @return string
 */
function nmgr_get_current_user_id() {
	if ( get_current_user_id() ) {
		/**
		 * Always return logged in user id as a string so that it can be compatible with the guest
		 * user id cookie value type and be tested with ===
		 */
		return ( string ) get_current_user_id();
	}

	if ( is_nmgr_guest( 'gift-registry' ) || is_nmgr_guest( 'wishlist' ) ) {
		return nmgr_get_user_id_cookie();
	}
}

/**
 * Get the user id stored in a cookie.
 * (This is typically used for guests)
 *
 * @return string
 */
function nmgr_get_user_id_cookie() {
	return isset( $_COOKIE[ 'nmgr_user_id' ] ) ? sanitize_key( wp_unslash( $_COOKIE[ 'nmgr_user_id' ] ) ) : 0;
}

/**
 * Generate a user id for wishlist users.
 * This is typically done for guests who are not logged in and so have no user id
 *
 * @return string
 */
function nmgr_generate_user_id() {
	require_once ABSPATH . 'wp-includes/class-phpass.php';
	$hasher = new PasswordHash( 8, false );
	return md5( $hasher->get_random_bytes( 32 ) );
}

/**
 * Set a cookie - wrapper for setcookie using WP constants.
 *
 * @param string $name Cookie name
 * @param string $value Cookie value
 * @param integer $expire Cookie expiry
 */
function nmgr_setcookie( $name, $value, $expire = 0 ) {
	if ( !headers_sent() ) {
		setcookie( $name, $value, $expire, COOKIEPATH ? COOKIEPATH : '/', COOKIE_DOMAIN, false, false );
	}
}

if ( !function_exists( 'nmgr_get_cart_template' ) ) {

	/**
	 * Template for displaying wishlists in cart fashion
	 *
	 * By default wishlists are displayed as a dropdown but this can be changed
	 * by supplying relevant arguments
	 *
	 * @param mixed $atts Attributes needed to compose the template.
	 *
	 * @param boolean $echo Whether to echo the template. Default false.
	 *
	 * @return string Template html
	 */
	function nmgr_get_cart_template( $atts = '', $echo = false ) {
		$args = array(
			'type' => 'gift-registry',
		);

		$vars = shortcode_atts(
			$args,
			$atts,
			'nmgr_cart'
		);

		$type = $vars[ 'type' ];

		if ( !is_nmgr_enabled( $type ) ) {
			return;
		}

		$vars[ 'cart_qty' ] = 0;
		$vars[ 'url' ] = nmgr_get_url( $type, 'home' );

		foreach ( nmgr_get_user_wishlist_ids( '', $type ) as $wishlist_id ) {
			$wishlist = nmgr()->wishlist();
			$wishlist->set_id( $wishlist_id );
			$vars[ 'cart_qty' ] = $vars[ 'cart_qty' ] + ( int ) $wishlist->get_items_quantity_count();
		}

		if ( !is_nmgr_user( $type ) ) {
			$vars[ 'url' ] = add_query_arg( array(
				'nmgr-notice' => 'login-to-access',
				'nmgr-redirect' => $_SERVER[ 'REQUEST_URI' ],
				'nmgr-type' => $type,
				), wc_get_page_permalink( 'myaccount' ) );
		}

		$template = apply_filters( 'nmgr_cart_template', nmgr_get_template( 'cart.php', $vars ), $vars );

		if ( $echo ) {
			echo $template;
		} else {
			return $template;
		}
	}

}

function nmgr_get_dialog_template( $args ) {
	$defaults = array(
		'id' => '',
		'show_header' => true,
		'show_header_close_button' => true,
		'show_body_close_button' => false,
		'title' => '',
		'content' => '',
		'footer' => '',
	);

	$vars = apply_filters( 'nmgr_dialog_template_args', wp_parse_args( $args, $defaults ) );
	$vars[ 'args' ] = $vars;

	return nmgr_get_template( 'dialog-template.php', $vars );
}

/**
 * Check if the current user has permissions to manage a wishlist
 *
 * By default the administrator and shop_manager roles have permision to manage all wishlists.
 *
 * @param int|NMGR_Wishlist $wishlist_id Wishlist id or object
 * @return boolean
 */
function nmgr_user_can_manage_wishlist( $wishlist_id = 0 ) {
	return $wishlist_id ?
		(current_user_can( 'manage_nm_gift_registry_settings' ) || nmgr_user_has_wishlist( $wishlist_id )) :
		false;
}

/**
 * Get the DateTime object representation of a date
 *
 * @param string $date
 * @return DateTime
 */
function nmgr_get_datetime( $date ) {
	// Y-m-d is 2021-10-04 This is the default date format used for the jquery ui datepicker.
	$date_format = apply_filters( 'nmgr_validate_date_format', 'Y-m-d' );

	$datetime = DateTime::createFromFormat( $date_format, $date, wp_timezone() );

	if ( !$datetime ) {
		try {
			$datetime = new DateTime( $date, wp_timezone() );
		} catch ( Exception $ex ) {
			$datetime = false;
		}
	}
	return $datetime;
}

function nmgr_get_dialog_submit_button( $args = array() ) {
	$class = isset( $args[ 'class' ] ) ? esc_attr( implode( ' ', ( array ) $args[ 'class' ] ) ) : '';
	$text = isset( $args[ 'text' ] ) ?
		nmgr_kses_post( $args[ 'text' ] ) :
		(nmgr()->is_pro ?
		__( 'Done', 'nm-gift-registry' ) :
		__( 'Done', 'nm-gift-registry-lite' ));
	$attributes = array();

	if ( isset( $args[ 'attributes' ] ) && is_array( $args[ 'attributes' ] ) ) {
		foreach ( $args[ 'attributes' ] as $attribute => $attribute_value ) {
			$attributes[] = esc_attr( $attribute ) . '="' . esc_attr( $attribute_value ) . '"';
		}
	}

	return sprintf( '<button class="nmgr-dialog-submit-button %1$s" %2$s>%3$s</button>',
		$class,
		implode( ' ', $attributes ),
		$text
	);
}

/**
 * Get a wishlist item by a specified id,
 *
 * @param int $item_id The wishlist item id used to retrieve the wishlist item
 * @return mixed NMGR_Wishlist_Item | false
 */
function nmgr_get_wishlist_item( $item_id ) {
	if ( !$item_id ) {
		return false;
	}

	$item = apply_filters( 'nmgr_get_wishlist_item', null, $item_id );
	if ( $item ) {
		return $item;
	}

	/**
	 * try catch statement is used because the wishlist item db class throws an exception
	 * if the wishlist item cannot be read from database
	 */
	try {
		$item = nmgr()->wishlist_item( $item_id );
		return $item;
	} catch ( Exception $e ) {
		return false;
	}
}

function nmgr_get_delete_item_notice( $item ) {
	return apply_filters( 'nmgr_delete_item_notice', sprintf(
			/* translators: %s: wishlist type title */
			nmgr()->is_pro ? __( 'Are you sure you want to remove the %s item?', 'nm-gift-registry' ) : __( 'Are you sure you want to remove the %s item?', 'nm-gift-registry-lite' ),
			nmgr_get_type_title( '', false, $item->get_wishlist()->get_type() )
		), $item );
}

/**
 * Get the default wishlist id for a user
 *
 * This is the wishlist id that is set for the user when he's allowed to have only one wishlist.
 * It should typically be the users latest wishlist which is in an active state.
 *
 * This function is different from 'nmgr_get_current_wishlist_id' because that function gets
 * the current wishlist id being viewed or operated on depending on the context. This function
 * does not depend on any context.
 *
 * @param int $user_id The user id. Defaults to the current wishlist user id if no user id is supplied.
 * @return int
 */
function nmgr_get_user_default_wishlist_id( $user_id = 0, $type = 'gift-registry' ) {
	$wishlist_ids = nmgr_get_user_wishlist_ids( $user_id, $type );

	if ( !empty( $wishlist_ids ) ) {
		foreach ( $wishlist_ids as $wishlist_id ) {
			if ( in_array( get_post_status( $wishlist_id ), nmgr_get_post_statuses() ) ) {
				return ( int ) $wishlist_id;
			}
		}
	}
	return 0;
}

/**
 * Display posts found using the standard wishlist template.
 *
 * This function is meant to provide a consistent template for displaying wishlist posts on any nm_gift_registry
 * archive page such as search, categories, tags, post_type_archive e.t.c. It also works with a custom $wp_query object.
 *
 * @param WP_Query $query Custom query object if provided. Uses global $wp_query by default.
 * @param array $action_args The arguments supplied to the action hooks used in the function.
 *
 * @deprecated since version 2.5
 */
function nmgr_archive_content( $query = null, $action_args = array() ) {
	_deprecated_function( __FUNCTION__, '2.5', 'nmgr_archive_loop' );
	nmgr_archive_loop( $query, $action_args );
}

/**
 * Display posts found using the standard wishlist template.
 *
 * This function is meant to provide a consistent template for displaying wishlist posts on any nm_gift_registry
 * archive page such as search, categories, tags, post_type_archive e.t.c. It also works with a custom $wp_query object.
 *
 * @param WP_Query $query Custom query object if provided. Uses global $wp_query by default.
 * @param array $action_args The arguments supplied to the action hooks used in the function.
 */
function nmgr_archive_loop( $query = null, $action_args = array() ) {
	global $wp_query;

	$custom_query = is_a( $query, 'WP_Query' ) && !$query->is_main_query();
	$the_query = $query ? $query : $wp_query;
	$is_nmgr_archive = is_nmgr_archive();
	$type = $action_args[ 'type' ] ?? null;

	if ( 'nm_gift_registry' !== $the_query->get( 'post_type' ) ) {
		return;
	}

	if ( $type && !nmgr_get_type_option( $type, 'enable' ) ) {
		return;
	}

	add_filter( 'is_nmgr_archive', '__return_true' );

	do_action( 'nmgr_archive_header', $the_query, $action_args );

	if ( $the_query->have_posts() ) :

		do_action( 'nmgr_before_archive_loop', $the_query, $action_args );

		while ( $the_query->have_posts() ) :

			$the_query->the_post();

			nmgr_template( 'content-archive-nm_gift_registry.php' );

		endwhile;

		do_action( 'nmgr_after_archive_loop', $the_query, $action_args );

		/**
		 * Set the custom query to the global $wp_query object so that we can use
		 * wordpress default pagination
		 */
		if ( $custom_query ) {
			$GLOBALS[ 'wp_query' ] = $query;
		}

		nmgr_paging_nav();

		/**
		 * reset the global $wp_query object after pagination.
		 */
		if ( $custom_query ) :
			wp_reset_query();
		endif;

	else:
		do_action_deprecated( 'nmgr_no_search_results', array( $action_args ), '2.2.0', 'nmgr_archive_no_results_found' );
		do_action( 'nmgr_archive_no_results_found', $the_query, $action_args );
	endif;

	add_filter( 'is_nmgr_archive', ($is_nmgr_archive ? '__return_true' : '__return_false' ) );
}

/**
 * Check if we are on an nm_gift_registry archive page
 * e.g. search results, categories, tags and post type archives.
 *
 * This function doesn't return true for function nmgr_archive_content when used in
 * pages via shortcodes because it is not strictly a wordpress archive page.
 *
 * @return boolean
 */
function is_nmgr_archive() {
	return apply_filters( 'is_nmgr_archive',
		/**
		 * 'is_search' checks to see if we are using wordpress' default search key 's'
		 * 'is_nmgr_search' check to see that we are searching for wishlists
		 * Using these two functions is necessary to make sure we are searching for wishlists
		 * using wordpress' default search and not the custom wishlist search key 'nmgr_s' used
		 * in the search shortcode which makes it not to be a search archive
		 */
		(is_search() && is_nmgr_search()) ||
		is_post_type_archive( 'nm_gift_registry' ) ||
		(is_tax() &&
		isset( get_queried_object()->taxonomy ) &&
		in_array( get_queried_object()->taxonomy, get_object_taxonomies( 'nm_gift_registry' ) )) ||
		is_nmgr_wishlist_page( 'archive' )
	);
}

/**
 * Check if we are on the custom page used for wishlist archives
 * @deprecated since version 4.0.0
 * @todo Add deprecated function notice.
 * @return boolean
 */
function is_nmgr_archive_page() {
	return is_nmgr_wishlist_page( 'archive' );
}

/**
 * Check if we are on the custom page used for displaying wishlists
 * @return boolean
 */
function is_nmgr_wishlist_page( $query = '' ) {
	$gr_page_id = ( int ) nmgr_get_option( 'page_id' );
	$wh_page_id = ( int ) nmgr_get_option( 'wishlist_page_id' );
	$queried_id = ( int ) get_queried_object_id();
	$is_custom_page = $queried_id && ($queried_id === $gr_page_id || $queried_id === $wh_page_id );
	$is_post_type_archive = is_post_type_archive( 'nm_gift_registry' );
	$is_singular = is_singular( 'nm_gift_registry' );
	$doing_ajax = wp_doing_ajax();
	$base_actions = nmgr_get_base_actions();
	$has_wishlist_shortcode = nmgr_post_content_has_shortcode( '[nmgr_wishlist]', $queried_id );
	$has_archive_shortcode = nmgr_post_content_has_shortcode( '[nmgr_archive]', $queried_id );

	$is_page = ($is_custom_page && ($has_wishlist_shortcode || $has_archive_shortcode)) ||
		$is_post_type_archive ||
		$is_singular ||
		($doing_ajax && !empty( $GLOBALS[ 'nmgr' ]->is_wishlist_page ));

	if ( $is_page && $query ) {
		$type = nmgr_get_current_type();
		$nmgr_w = get_query_var( 'nmgr_w' );
		$nmgr_action = get_query_var( 'nmgr_action' );
		$ajax_query = $doing_ajax && !empty( $GLOBALS[ 'nmgr' ]->{'is_wishlist_page_' . $query} );
		$allow_multi = nmgr()->is_pro && nmgr_get_type_option( $type, 'allow_multiple_wishlists' );
		$base = !$nmgr_w && !$nmgr_action && $has_wishlist_shortcode;

		switch ( $query ) {
			case 'base':
				$is_page = $base;
				break;
			case 'new':
				$is_page = ($query === $nmgr_w) || $ajax_query;
				break;
			case 'home':
				$is_page = ($query === $nmgr_w) || ($base && $allow_multi ) || $ajax_query;
				break;
			case 'archive':
				$is_page = (!$nmgr_w && !$nmgr_action && ($has_archive_shortcode || $is_post_type_archive)) || $ajax_query;
				break;
			case 'single':
				$single = ($nmgr_w && !in_array( $nmgr_w, $base_actions ) && !$nmgr_action) ||
					($base && !$allow_multi ) ||
					($is_singular && !$nmgr_action) ||
					$ajax_query;
				$is_page = apply_filters( 'is_nmgr_wishlist', $single );
				break;
			case 'account_section':
				$is_page = (($has_archive_shortcode || $has_wishlist_shortcode) &&
					$nmgr_w && $nmgr_action && !in_array( $nmgr_w, $base_actions )) ||
					($is_singular && $nmgr_action) ||
					$ajax_query;
				break;
			default:
				$is_page = false;
				break;
		}
	}

	return apply_filters( 'is_nmgr_wishlist_page', $is_page, [ 'query' => $query ] );
}

/**
 * Check if we are in a specific wishlist template
 * (Still experimental. For internal use only)
 *
 * @global string $nmgr_template The current template being used.
 * @param string $template The path of the template file to check relative to the plugin template directory.
 * @return boolean
 */
function is_nmgr_template( $template ) {
	global $nmgr_template;
	return $nmgr_template === $template;
}

/**
 * Get the current wishlist template being used/
 * (Still experimental. For internal use only)
 *
 * @global string $nmgr_template The current template being used.
 * @return string
 */
function nmgr_get_current_template() {
	global $nmgr_template;
	return $nmgr_template;
}

/**
 * Check if we are in a specific wishlist template
 * (Still experimental. For internal use only)
 *
 * @param string|array $template_name The path (or paths) of the template file(s) to check
 * 																		relative to the plugin template directory.
 * @return boolean|array False if we are not in the template or Array of template arguments if
 * 											 we are in the template
 */
function in_nmgr_template( $template_name ) {
	$names = is_array( $template_name ) ? $template_name : array( $template_name );
	$val = false;

	foreach ( $names as $name ) {
		if ( isset( $GLOBALS[ 'nmgr_templates' ], $GLOBALS[ 'nmgr_templates' ][ $name ] ) ) {
			$val = $GLOBALS[ 'nmgr_templates' ][ $name ];
			break;
		}
	}
	return apply_filters( 'in_nmgr_template', $val, $template_name );
}

/**
 * Returns a html progress bar
 *
 * @param int|float $needed Quantity needed to complete the progress
 * @param int|float $received Quantity of progress made already
 * @param string $title_attribute Title attribute to display over the progress bar. Default none.
 * @param boolean $show_progress Whether to show the progress bar graphic. Default true.
 * @param boolean $show_percent Whether to show the percentage complete. Default true.
 */
function nmgr_progressbar( $needed, $received, $title_attribute = '', $show_progress = true, $show_percent = true ) {
	ob_start();
	$percent = 0 === $received ? '0%' : ceil( ($received / $needed) * 100 ) . '%';
	?>

	<div class="nmgr-progressbar">
		<?php if ( $show_progress ) : ?>
			<div class="progress-wrapper nmgr-tip" title="<?php echo strip_tags( $title_attribute ); ?>">
				<div class="progress" style="width: <?php echo $percent; ?>"></div>
			</div>
		<?php endif; ?>
		<?php if ( $show_percent ) : ?>
			<span class="percent"><?php echo $percent; ?></span>
		<?php endif; ?>
	</div>
	<?php
	return ob_get_clean();
}

/**
 * Get the search term used for searching wishlists
 *
 * The plugin gives priority to wordpress' search term 's' in the wp_query global object.
 * If 's' is not available, the plugin checks for 'nmgr_s' in the wp_query object.
 * If 'nmgr_s' is not available, the plugin checks for 'nmgr_s' in the $_GET array.
 *
 * This function provides a convenient place where all these checks can be made.
 * The search term can be filtered with 'nmgr_get_search_query'
 *
 * @return string Search query term
 */
function nmgr_get_search_query() {
	$s = '';
	if ( get_search_query() ) {
		$s = get_search_query();
	} elseif ( get_query_var( 'nmgr_s' ) ) {
		$s = get_query_var( 'nmgr_s' );
	} elseif ( isset( $_GET[ 'nmgr_s' ] ) ) {
		$s = $_GET[ 'nmgr_s' ];
	}
	return apply_filters( 'nmgr_get_search_query', $s );
}

/**
 * Get the default plugin options used with jquery datatables
 * @return array
 */
function nmgr_get_default_datatable_options() {
	return apply_filters( 'nmgr_default_datatable_options',
		array(
			'order' => array(),
			'paging' => false,
			'searching' => false,
			'info' => false,
			'columnDefs' => array(
				array( 'orderable' => true, 'targets' => array( 'dt-orderable' ) ),
				array( 'orderable' => false, 'targets' => array( '_all' ) ),
			),
			'language' => array(
				'zeroRecords' => nmgr()->is_pro ?
				_x( 'No items', 'Empty wishlist items table', 'nm-gift-registry' ) :
				_x( 'No items', 'Empty wishlist items table', 'nm-gift-registry-lite' ),
			),
		)
	);
}

function nmgr_get_datatables() {
	$default_options = nmgr_get_default_datatable_options();
	$default_col_defs = isset( $default_options[ 'columnDefs' ] ) ? $default_options[ 'columnDefs' ] : array();

	return apply_filters( 'nmgr_datatables', array(
		'items_table' => array(
			'selector' => '.nmgr-items-table',
			'name' => 'items_table',
			'events' => array( 'nmgr_items_reloaded' ),
			'options' => array_merge( $default_options, array(
				'columnDefs' => array_merge(
					array(
						array(
							'orderable' => true,
							'targets' => array(
								'item_title',
								'item_cost',
								'item_purchased_quantity',
								'item_favourite',
								'item_total_cost',
							)
						),
					),
					$default_col_defs
				),
			) ),
		),
		'orders_table' => array(
			'selector' => '.nmgr-orders-table',
			'name' => 'orders_table',
			'events' => array( 'nmgr_orders_reloaded' ),
			'options' => array_merge( $default_options, array(
				'columnDefs' => array_merge(
					array(
						array(
							'orderable' => true,
							'targets' => array(
								'customer',
								'order',
								'order-date',
								'total-spent',
							)
						),
					),
					$default_col_defs
				),
			) ),
		),
		) );
}

/**
 * Get the ids of all orders that contain a particular wishlist
 *
 * @global wpdb $wpdb
 * @param int $wishlist_id The wishlist id
 * @return array Array of order ids
 */
function nmgr_get_wishlist_order_ids( $wishlist_id ) {
	global $wpdb;

	$order_ids = wp_cache_get( 'nmgr_wishlist_order_ids' );

	if ( false === $order_ids ) {
		$order_ids = [];

		$results = $wpdb->get_results( "SELECT post_id, meta_value FROM $wpdb->postmeta AS postmeta LEFT JOIN $wpdb->posts AS posts ON (postmeta.post_id = posts.ID AND posts.post_type = 'shop_order') WHERE postmeta.meta_key = 'nm_gift_registry' " );

		if ( !empty( $results ) ) {
			foreach ( $results as $value ) {
				$unserialized = unserialize( $value->meta_value );
				$wishlist_ids = array_keys( $unserialized );
				if ( in_array( $wishlist_id, $wishlist_ids ) ) {
					$order_ids[] = $value->post_id;
				}
			}

			wp_cache_set( 'nmgr_wishlist_order_ids', $order_ids );
		}
	}

	return $order_ids;
}

/**
 * Get the reference data for all items that have been added to the cart.
 *
 * Typically used when adding to car via ajax if dom elements related
 * to the item being added need to be updated afterwards
 *
 * @return array|null
 */
function nmgr_get_add_to_cart_item_ref_data() {
	return isset( $GLOBALS[ 'nmgr_add_to_cart_ref_data' ] ) ? $GLOBALS[ 'nmgr_add_to_cart_ref_data' ] : '';
}

/**
 * Add reference details of an item that has been added to the cart
 * to the global reference array
 *
 * Typically used when adding to car via ajax if dom elements related
 * to the item being added need to be updated afterwards
 *
 * @param array $data
 */
function nmgr_add_add_to_cart_item_ref_data( $data ) {
	if ( isset( $GLOBALS[ 'nmgr_add_to_cart_ref_data' ] ) ) {
		$GLOBALS[ 'nmgr_add_to_cart_ref_data' ][] = $data;
	} else {
		$GLOBALS[ 'nmgr_add_to_cart_ref_data' ] = array( $data );
	}
}

/**
 * Destroy the reference data stored for all items added to the cart.
 */
function nmgr_destroy_add_to_cart_item_ref_data() {
	if ( isset( $GLOBALS[ 'nmgr_add_to_cart_ref_data' ] ) ) {
		unset( $GLOBALS[ 'nmgr_add_to_cart_ref_data' ] );
	}
}

/**
 * The core action used to add wishlist products to the cart via http or ajax
 *
 * @param array $items_data Array of data of each item to be added to the cart
 * @return array $data Array of data for each item that was supplied to be added to the cart
 */
function nmgr_add_to_cart( $items_data ) {

	/**
	 * Destroy the add to cart item reference data just in case it exists.
	 * We start afresh as we are starting a new add to cart action.
	 */
	if ( wp_doing_ajax() ) {
		nmgr_destroy_add_to_cart_item_ref_data();
	}

	/**
	 * Enable custom add to cart action.
	 * Allow add to cart items data to be modified
	 *
	 * 'do_action_ref_array' is used rather than 'do_action'
	 * because it maintains the state of $items_data
	 * which may be an array of objects.
	 */
	do_action_ref_array( 'nmgr_add_to_cart_action', array( &$items_data ) );

	foreach ( ( array ) $items_data as $item_mixed ) {
		/**
		 * $item_mixed may be an object typically if adding to cart in bulk
		 * so we cast to array
		 */
		$item = ( array ) $item_mixed;

		if ( !isset( $item[ 'nmgr-add-to-cart-product-id' ] ) ) {
			continue;
		}

		$product_id = ( int ) $item[ 'nmgr-add-to-cart-product-id' ];
		$quantity = empty( $item[ 'quantity' ] ) ? 1 : wc_stock_amount( wp_unslash( $item[ 'quantity' ] ) );
		$variation_id = isset( $item[ 'variation_id' ] ) ? ( int ) $item[ 'variation_id' ] : 0;
		$variation = nmgr_get_posted_variations( $variation_id, $item );
		$wishlist_id = ( int ) ($item[ 'nmgr-add-to-cart-wishlist' ] ?? 0);
		$wishlist_item_id = ( int ) ($item[ 'nmgr-add-to-cart-wishlist-item' ] ?? 0);
		$passed_validation = apply_filters( 'woocommerce_add_to_cart_validation', true, $product_id, $quantity );

		if ( $passed_validation && is_a( wc()->cart, 'WC_Cart' ) &&
			false !== WC()->cart->add_to_cart( $product_id, $quantity, $variation_id, $variation ) ) {
			// wc filter (check this filter on updates)

			if ( wp_doing_ajax() ) {
				do_action( 'woocommerce_ajax_added_to_cart', $product_id );
			}

			wc_add_to_cart_message( array( $product_id => $quantity ), true );
		}

		if ( wp_doing_ajax() ) {
			$data = array(
				'product_id' => $product_id,
				'quantity' => $quantity,
				'wishlist_id' => $wishlist_id,
				'wishlist_item_id' => $wishlist_item_id,
			);
			nmgr_add_add_to_cart_item_ref_data( $data );
		}
	}

	return nmgr_get_add_to_cart_item_ref_data();
}

/**
 * Get the checkbox switch used by the plugin
 * @param array $args Arguments supplied to create the checkbox switch:
 * - input_type = {string} The input type, 'checkbox' or 'radio'. Default 'checkbox'
 * - input_id - {string, required} The input id
 * - input_name - {string} The input name
 * - input_value - {mixed} The input value
 * - input_class - {array} The input class
 * - input_attributes = {array} Attributes to go into the input element
 * - label_class - {array} The label class
 * - label_attributes - {array} Attributes to go into the label element
 * - label_text - {string} The label text
 * - label_before - {boolean} Whether the label text should be before the switch. Default false.
 * - checked - {boolean} Whether the checkbox should be checked or not.
 * - show_hidden_input - {mixed} Whether to show a hidden input for the checkbox. Default false.
 * 												If true, show hidden input value is 0 else, value of 'show_hidden_input' is used.
 * @return string
 */
function nmgr_get_checkbox_switch( $args ) {
	return nmgr()->get_toggle_switch( $args );
}

function nmgr_format_attributes( $attributes ) {
	$sanitized_attributes = array();

	if ( !empty( $attributes ) ) {
		foreach ( ( array ) $attributes as $key => $value ) {
			$val = is_array( $value ) ? implode( ' ', array_map( 'esc_attr', $value ) ) : $value;
			$sanitized_attributes[] = esc_attr( $key ) . '="' . $val . '"';
		}
	}
	return implode( ' ', $sanitized_attributes );
}

/**
 * Get the number of days guest wishlist can exist before they are deleted
 * and their cookies expire in the browser
 *
 * The maximum is set at 365 as it probably doesn't makes sense for guest
 * wishlist to last longer than a year. They can register for an account.
 *
 * @return int
 */
function nmgr_get_guest_wishlist_expiry_days() {
	$days = ( int ) apply_filters(
			'nmgr_guest_wishlist_expiry_days',
			nmgr_get_option( 'guest_wishlist_expiry_days', 365 )
	);
	$val = 1 > $days ? 0 : (365 < $days ? 365 : $days);
	return $val;
}

/**
 * Checks if an order has wishlist items
 * @param int|WC_Order $order_id The order id or object
 * @return array|boolean Returns the array of wishlist data for the order or false
 * 				 if the order doesn't contain wishlist items
 */
function nmgr_order_has_wishlist_items( $order_id ) {
	$order = wc_get_order( $order_id );
	$meta = $order->get_meta( 'nm_gift_registry' );
	return !empty( $meta ) ? $meta : false;
}

/**
 * Sync the data for all wishlist items in an order with the data
 * stored for the items themselves.
 *
 * @param int|WC_Order $order_id The order id or object
 */
function nmgr_sync_order_data( $order_id ) {
	nmgr()->order()->update_wishlist_item_purchased_quantity( $order_id );
}

if ( !function_exists( 'nmgr_get_item_purchase_log_template' ) ) {

	/**
	 * Template for showing log of all purchases of a wishlist item
	 *
	 * @param int|NMGR_Wishlist_Item|\NMGR\Sub\Wishlist_Item|array $atts Attributes needed to compose the template.
	 * Currently accepted $atts attributes if array:
	 * - id [int|NMGR_Wishlist_Item] Wishlist item id or instance of NMGR_Wishlist_Item.
	 *
	 * @param boolean $echo Whether to echo the template. Default false.
	 *
	 * @return string Template html
	 */
	function nmgr_get_item_purchase_log_template( $atts = '', $echo = false ) {
		if ( !is_nmgr_enabled() || !is_nmgr_user() || !is_nmgr_admin() ) {
			return;
		}

		$args = shortcode_atts(
			array(
				'id' => is_array( $atts ) ? $atts[ 'id' ] : $atts,
			),
			$atts,
			'nmgr_item_purchase_log'
		);

		$item = nmgr_get_wishlist_item( $args[ 'id' ] );

		$vars = array(
			'item' => $item,
			'template_group' => 'account',
		);

		$template = apply_filters( 'nmgr_item_purchase_log_template', nmgr_get_template( 'account/item-purchase-log.php', $vars ), $vars );

		if ( $echo ) {
			echo $template;
		} else {
			return $template;
		}
	}

}

/**
 * Get the wishlist data in the cart item.
 *
 * @param array|string $cart_item The cart item array value or key
 * @param string $data_type The type of wishlist data for the cart item.
 * Default is 'wishlist_item'. Others are 'crowdfund', 'free_contribution'
 * @return false|array Array of wishlist data if true. False if the cart item doesn't have wishlist data.
 */
function nmgr_get_cart_item_data( $cart_item, $data_type = null ) {
	if ( is_a( wc()->cart, 'WC_Cart' ) ) {
		$cart_item = is_array( $cart_item ) ? $cart_item : wc()->cart->get_cart_item( $cart_item );
		if ( $data_type ) {
			$val = $data_type === ($cart_item[ 'nm_gift_registry' ][ 'type' ] ?? null) ?
				$cart_item[ 'nm_gift_registry' ] : false;
		} else {
			$val = $cart_item[ 'nm_gift_registry' ] ?? false;
		}
		return apply_filters( 'nmgr_get_cart_item_data', $val, $cart_item, $data_type );
	}
}

/**
 * Check if the cart item belongs to a wishlist
 *
 * Alias of nmgr_get_cart_item_data()
 *
 * @param array|string $cart_item The cart item array value or key
 * @param string $data_type The type of wishlist data to check for,
 * Default is 'wishlist_item'. Others are 'crowdfund', 'free_contribution'
 * @return false|array Array of wishlist data if true. False if the cart item doesn't have wishlist data.
 */
function is_nmgr_cart_item( $cart_item, $data_type = null ) {
	return nmgr_get_cart_item_data( $cart_item, $data_type );
}

function nmgr_get_archive_template_args( $atts = [] ) {
	$args = wp_parse_args(
		$atts,
		array(
			'title' => '',
			'show_header' => true,
			'show_title' => true,
			'show_results_count' => true,
			'show_results_if_no_search_query' => true,
			'type' => 'gift-registry',
		)
	);

	return apply_filters( 'nmgr_archive_template_args', $args );
}

if ( !function_exists( 'nmgr_get_archive_template' ) ) {

	/**
	 * Get the template for outputting wishlist archive results
	 *
	 * Works with search results too.
	 *
	 * This function or the shortcode attached to it should be used after wp_loaded hook
	 * as that is when the wp_query global exists.
	 *
	 * @param array $atts Attributes needed to compose the template.
	 * @param bool $echo Whether to echo or return the template.
	 */
	function nmgr_get_archive_template( $atts = array(), $echo = false ) {
		global $wp_query;

		if ( is_nmgr_admin() || !is_a( $wp_query, 'WP_Query' ) ) {
			return;
		}

		if ( !is_nmgr_archive() && !is_nmgr_search() ) {
			return nmgr_get_wishlist_template( $atts, $echo );
		}

		$template_args = nmgr_get_archive_template_args( $atts );

		$query_args = (is_nmgr_archive() && !is_page()) ? $wp_query->query_vars : array();

		if ( isset( $_GET[ 'nmgr_s' ] ) || (!isset( $_GET[ 'nmgr_s' ] ) && $template_args[ 'show_results_if_no_search_query' ]) ) {
			$query_args = array();
			$query_args[ 's' ] = nmgr_get_search_query();
		}

		$query_args[ 'paged' ] = get_query_var( 'paged' ) ? get_query_var( 'paged' ) : '';
		$query_args[ 'post_type' ] = 'nm_gift_registry';

		if ( !empty( $template_args[ 'type' ] ) ) {
			$query_args[ 'tax_query' ] = array( array(
					'taxonomy' => 'nm_gift_registry_type',
					'field' => 'slug',
					'terms' => $template_args[ 'type' ],
					'operator' => 'IN',
				) );
		}
		$query_args[ 'nmgr_query' ] = $template_args; // currently for internal use only

		ob_start();

		$the_query = new WP_Query( $query_args );

		nmgr_archive_loop( $the_query, $template_args );

		$template = ob_get_clean();

		if ( $echo ) {
			echo $template;
		} else {
			return $template;
		}
	}

	function nmgr_get_dropdown() {
		$icon = nmgr_get_svg( [
			'icon' => 'dropdown-toggle',
			'fill' => 'currentColor',
			]
		);
		$dropdown = new Dropdown();
		$dropdown->set_text( $icon );
		$dropdown->hide_arrow();
		$dropdown->toggle_attributes[ 'class' ][] = 'nmgr-dropdown-toggle';
		$dropdown->toggle_attributes[ 'class' ][] = 'nmgr-tip';
		$dropdown->toggle_attributes[ 'data-tiptip-position' ] = 'top';
		$dropdown->toggle_attributes[ 'title' ] = nmgr()->is_pro ?
			__( 'Actions', 'nm-gift-registry' ) :
			__( 'Actions', 'nm-gift-registry-lite' );
		return $dropdown;
	}

}

function nmgr_get_add_items_link() {
	return sprintf( '<a class="button nmgr-add-items-link nmgr-tip" title="%1$s" href="%2$s">%3$s</a>',
		nmgr_get_add_items_text(),
		nmgr_get_add_items_url(),
		( nmgr()->is_pro ?
		__( 'Add item(s)', 'nm-gift-registry' ) :
		__( 'Add item(s)', 'nm-gift-registry-lite' )
		)
	);
}

function nmgr_get_add_to_wishlist_mode( $args = [] ) {
	_deprecated_function( __FUNCTION__, '4.0.0' );
	return apply_filters( 'nmgr_add_to_wishlist_mode', 'simple', $args );
}

function nmgr_get_modal( $args = [] ) {
	$modal = new Modal( $args );
	$modal->add_container_class( [ 'nmgr' ] );
	$modal->set_id( 'nmgr_' . time() );
	$modal->make_vertically_centered();
	return $modal;
}

function nmgr_get_toast( $args = [] ) {
	$toast = new Toast( $args );
	$toast->add_container_class( [ 'nmgr' ] );
	return $toast;
}

function nmgr_get_toast_notice( $content, $notice_type = 'success' ) {
	$comp = nmgr_get_toast();
	$comp->set_id( 'nmgr_toast_' . rand() );
	$comp->set_notice_type( $notice_type );
	$comp->set_content( $content );
	return $comp->get();
}

function nmgr_get_wc_toast_notices() {
	$arr = [];
	foreach ( wc_get_notices() as $type => $notice ) {
		foreach ( $notice as $not ) {
			$arr[] = nmgr_get_toast_notice( $not[ 'notice' ], $type );
		}
	}
	wc_clear_notices();
	return $arr;
}

function nmgr_get_error_toast_notice() {
	return nmgr_get_toast_notice( nmgr_get_error_text(), 'error' );
}

function nmgr_get_success_toast_notice() {
	return nmgr_get_toast_notice( nmgr_get_success_text(), 'success' );
}

function nmgr_show_copy_shipping_address_btn( $address ) {
	?>
	<div class="nmgr-copy-shipping-address-wrapper" style="margin-bottom:15px;">
		<a class="nmgr-copy-shipping-address" href="#"
			 data-nmgr-address="<?php echo esc_attr( htmlspecialchars( json_encode( $address ) ) ); ?>">
				 <?php
				 echo esc_html( nmgr()->is_pro ?
						 __( 'Copy account shipping address', 'nm-gift-registry' ) :
						 __( 'Copy account shipping address', 'nm-gift-registry-lite' )
				 );
				 ?>
		</a>
	</div>
	<?php
}

function nmgr_get_error_text() {
	return nmgr()->is_pro ?
		__( 'Error', 'nm-gift-registry' ) :
		__( 'Error', 'nm-gift-registry-lite' );
}

function nmgr_get_success_text() {
	return nmgr()->is_pro ?
		__( 'Success', 'nm-gift-registry' ) :
		__( 'Success', 'nm-gift-registry-lite' );
}

function nmgr_sort_by_priority( $a, $b ) {
	return nmgr()->sort_by_priority( $a, $b );
}

function is_nmgr_settings_screen( $type = '' ) {
	$is_screen = is_admin() && 'nm_gift_registry' === ($_GET[ 'post_type' ] ?? false);
	$page = $_GET[ 'page' ] ?? false;
	$sections = [
		'wishlist' => 'nmgr-wishlist-settings',
		'gift-registry' => 'nmgr-settings'
	];

	return $type ?
		$is_screen && $page === ($sections[ $type ] ?? null) :
		$is_screen && in_array( $page, $sections );
}

/**
 * Put relevant wishlist data into the global variable 'nmgr'
 *
 * This function is hooked into 'the_post' to make the global variable available on an individual
 * post basis in loops. This can be helpful, for example, in retrieving the right wishlist id for
 * the current post if the current 'post' query is not the main query
 *
 * The function is also hooked into 'add_meta_boxes' to make the global variable available
 * in the admin edit screen for a wishlist.
 *
 * Finally the function is hooked into 'template_redirect' to make the global variable available
 * generally for the frontend but also for enqueued scripts which use 'wp_enqueue_script' hook,
 * as this hook runs before 'the_post' hook
 *
 * The global variable is only available on registered nmgr pages and
 * it would not be available before these hooks are called by wordpress
 *
 * @return Object|\stdClass
 */
function nmgr_set_global_var() {
	$GLOBALS[ 'nmgr' ] = new \stdClass();
	$GLOBALS[ 'nmgr' ]->current_type = nmgr_get_current_type();
	$GLOBALS[ 'nmgr' ]->is_wishlist_page = is_nmgr_wishlist_page();
	$GLOBALS[ 'nmgr' ]->is_wishlist_page_archive = is_nmgr_wishlist_page( 'archive' );
	$GLOBALS[ 'nmgr' ]->is_wishlist_page_single = is_nmgr_wishlist();
	$GLOBALS[ 'nmgr' ]->is_wishlist_page_account_section = is_nmgr_wishlist_page( 'account_section' );
	$GLOBALS[ 'nmgr' ]->is_wishlist_page_new = is_nmgr_wishlist_page( 'new' );
	$GLOBALS[ 'nmgr' ]->is_wishlist_page_home = is_nmgr_wishlist_page( 'home' );
	$GLOBALS[ 'nmgr' ]->is_admin = is_nmgr_admin();
	$GLOBALS[ 'nmgr' ]->is_search = is_nmgr_search();
	$GLOBALS[ 'nmgr' ]->is_modal = false;
	$GLOBALS[ 'nmgr' ]->wishlist_id = nmgr_get_current_wishlist_id();
}

function nmgr_get_global_var() {
	$GLOBALS[ 'nmgr' ] = $GLOBALS[ 'nmgr' ] ?? new \stdClass();
	return $GLOBALS[ 'nmgr' ];
}

function nmgr_get_elements_to_show( $actions ) {
	foreach ( $actions as $key => $args ) {
		if ( (!is_nmgr_settings_screen() && false === ( bool ) ($args[ 'show' ] ?? true)) ||
			( is_nmgr_settings_screen() && false === ( bool ) ($args[ 'show_in_settings' ] ?? true) ) ) {
			unset( $actions[ $key ] );
		}
	}
	return $actions;
}

/**
 * Make sure a user is allowed to perform the action on a wishlist.
 *
 * For ajax usage.
 * This function is similar to check_ajax_referrer in that it kills the script
 * if the wishlist doesn't exist or if the user performing the action is not the
 * wishlist owner or an admin user who can manage the wishlist.
 *
 * @param int|NMGR_Wishlist $wishlist_id The wishlist id or object
 */
function nmgr_check_wishlist_permission( $wishlist_id ) {
	if ( !nmgr_get_wishlist( $wishlist_id ) || !nmgr_user_can_manage_wishlist( $wishlist_id ) ) {
		wp_die();
	}
}

/**
 * Get the wishlist id and wishlist item ids in the $_POST array
 * @return array An array with wishlist_id (int|false) and wishlist_item_ids (array) keys representing
 * the wishlist_id and wishlist_item_ids in $_POST
 */
function nmgr_get_posted_wishlist_and_item_ids() {
	$wishlist_id = wp_unslash( $_POST[ 'wishlist_id' ] ?? false );
	$wishlist_item_ids = array_map( 'absint', wp_unslash( ( array ) ($_POST[ 'wishlist_item_ids' ] ?? []) ) );
	$item_id = ( int ) wp_unslash( $_POST[ 'wishlist_item_id' ] ?? 0 );

	if ( $item_id ) {
		$wishlist_item_ids = [ $item_id ];
	}

	if ( !$wishlist_id && !empty( $wishlist_item_ids ) ) {
		$first_item = nmgr_get_wishlist_item( $wishlist_item_ids[ 0 ] );
		$wishlist_id = $first_item ? $first_item->get_wishlist_id() : false;
	}

	return [
		'wishlist_id' => $wishlist_id ? ( int ) $wishlist_id : false,
		'wishlist_item_id' => !empty( $wishlist_item_ids ) ? $wishlist_item_ids[ 0 ] : false,
		'wishlist_item_ids' => $wishlist_item_ids,
	];
}

function nmgr_get_custom_order_notice() {
	return sprintf(
		/* translators: %s wishlist type title */
		nmgr()->is_pro ? __( 'This order was created manually from the purchase of a %s item.', 'nm-gift-registry' ) : __( 'This order was created manually from the purchase of a %s item.', 'nm-gift-registry-lite' ),
		nmgr_get_type_title()
	);
}

function nmgr_get_purchase_refund_template( $item_id ) {
	ob_start();
	$item = nmgr_get_wishlist_item( $item_id );
	$product = $item->get_product();
	$purchased_quantity = $item->get_purchased_quantity();
	?>

	<style>
		#nmgr-purchase-refund-form {
			display: flex;
			flex-flow: column;
			align-items: center;
		}

		#nmgr-purchase-refund-form .nmgr-title {
			font-size: 1.675em;
		}

		#nmgr-purchase-refund-form input.nmgr-quantity {
			width: 5em;
		}

		#nmgr-purchase-refund-form .nmgr-options {
			border: 1px grey dashed;
			padding: .7em;
			margin-top: 1.5em;
		}
	</style>

	<form id="nmgr-purchase-refund-form">
		<div class="nmgr-title-wrap">
			<span class="nmgr-title"><?php echo esc_html( $product->get_title() ); ?></span>
			<?php
			$desired_qty_title = nmgr()->is_pro ?
				__( 'Desired quantity', 'nm-gift-registry' ) :
				__( 'Desired quantity', 'nm-gift-registry-lite' );
			?>
			<span class="nm badge" style="background-color:#ccc; vertical-align:text-bottom;"
						data-bs-toggle="tooltip"
						data-bs-placement="right"
						title="<?php esc_attr_e( $desired_qty_title ); ?>">
							<?php echo ( int ) $item->get_quantity(); ?>
			</span>
		</div>
		<br>
		<label class="nmgr-new-qty">
			<span>
				<?php
				echo esc_html( nmgr()->is_pro ?
						__( 'Purchased quantity', 'nm-gift-registry' ) :
						__( 'Purchased quantity', 'nm-gift-registry-lite' )
				);
				?>
			</span>
			<input type="number"
						 step="1"
						 placeholder="0"
						 autocomplete="off"
						 size="4"
						 class="nmgr-quantity"
						 value="<?php echo esc_attr( $purchased_quantity ); ?>"
						 data-qty="<?php echo esc_attr( $purchased_quantity ); ?>"
						 name="quantity"
						 min="0"
						 max="<?php echo esc_attr( apply_filters( 'nmgr_quantity_input_max', $product->get_stock_quantity(), $product ) ); ?>"
						 />
		</label>

		<div class="nmgr-options">
			<div>
				<?php
				$args_title = nmgr_get_svg( array(
					'icon' => 'info',
					'class' => 'align-with-text nmgr-tip',
					'style' => 'margin-left:4px;',
					'size' => '.8',
					'title' => nmgr()->is_pro ?
					__( 'Create an order to reflect the purchase of the item.', 'nm-gift-registry' ) :
					__( 'Create an order to reflect the purchase of the item.', 'nm-gift-registry-lite' ),
					'fill' => '#999',
					) );

				$args = [
					'input_id' => 'nmgr_create_order_switch',
					'input_name' => 'create_order',
					'checked' => true,
					'label_text' => (nmgr()->is_pro ?
					__( 'Create order', 'nm-gift-registry' ) :
					__( 'Create order', 'nm-gift-registry-lite' )) . $args_title,
				];
				echo nmgr_get_checkbox_switch( $args );
				?>
			</div>
			<div>
				<?php
				$args2_title = nmgr_get_svg( array(
					'icon' => 'info',
					'class' => 'align-with-text nmgr-tip',
					'style' => 'margin-left:4px;',
					'size' => '.8',
					'title' => nmgr()->is_pro ?
					__( 'Include the price of the item in the order total.', 'nm-gift-registry' ) :
					__( 'Include the price of the item in the order total.', 'nm-gift-registry-lite' ),
					'fill' => '#999',
					) );

				$args2 = [
					'input_id' => 'nmgr_apply_price_switch',
					'input_name' => 'apply_price',
					'checked' => true,
					'label_text' => (nmgr()->is_pro ?
					__( 'Apply price', 'nm-gift-registry' ) :
					__( 'Apply price', 'nm-gift-registry-lite' )) . $args2_title,
				];
				echo nmgr_get_checkbox_switch( $args2 );
				?>
			</div>
		</div>

		<input type="hidden" name="wishlist_item_id" value="<?php echo esc_attr( $item->get_id() ); ?>">
	</form>
	<?php
	return ob_get_clean();
}

/**
 * Wrapper function to update the purchased quantity of an item
 *
 * @param int|NMGR_Wishlist_Item|\NMGR\Sub\Wishlist_Item $item_id Item id or object
 * @param array $args Arguments used to perform the update:
 * - quantity {int} The new purchased quantity.
 * - create_order {boolean} Whether to create an order to reflect the update. Default true.
 * - apply_price - {boolean) Whether to include the price of the item in the created order. Default true.
 * - order_note - {string}. Order note that should be added to the order. Default none.
 * - order_item_meta = {array} Metadata that should be added to the order item if created. Default none.
 */
function nmgr_update_item_purchased_quantity( $item_id, $args = [] ) {
	$item = nmgr_get_wishlist_item( $item_id );

	if ( isset( $args[ 'quantity' ] ) && $item ) {
		$update_qty = ( int ) $args[ 'quantity' ];
		$purchased_qty = $item->get_purchased_quantity();

		if ( $update_qty < $purchased_qty ) {
			$item->set_purchased_quantity( $update_qty );
			$item->save();
			$item->clear_cache();
		} elseif ( $update_qty > $purchased_qty ) {
			if ( false === ( bool ) ( $args[ 'create_order' ] ?? true ) ) {
				$item->set_purchased_quantity( $update_qty );
				$item->save();
				$item->clear_cache();
			} else {
				$order = new \WC_Order();
				$order->set_created_via( 'nmgr_wishlist' );
				$order->set_status( apply_filters( 'nmgr_item_purchased_order_status', 'completed' ) );

				if ( function_exists( 'is_nmgr_shipping_to_wishlist_address' ) &&
					is_nmgr_shipping_to_wishlist_address() ) {
					$wishlist = $item->get_wishlist();
					if ( $wishlist->has_shipping_address() ) {
						foreach ( $wishlist->get_shipping() as $key => $value ) {
							if ( is_callable( array( $order, "set_shipping_{$key}" ) ) ) {
								$order->{"set_shipping_{$key}"}( $value );
							}
						}
					}
				}

				$product = $item->get_product();

				if ( $product ) {
					$order_item = new \WC_Order_Item_Product();
					$quantity = $update_qty;
					$subtotal = wc_get_price_excluding_tax( $product, [ 'qty' => $quantity ] );

					$order_item->set_props( [
						'name' => $product->get_name(),
						'tax_class' => $product->get_tax_class(),
						'product_id' => $product->is_type( 'variation' ) ? $product->get_parent_id() : $product->get_id(),
						'variation_id' => $product->is_type( 'variation' ) ? $product->get_id() : 0,
						'variation' => $product->is_type( 'variation' ) ? $product->get_attributes() : array(),
						'quantity' => $quantity,
						'subtotal' => $subtotal,
						'total' => false === ( bool ) ( $args[ 'apply_price' ] ?? true ) ? 0 : $subtotal,
					] );

					if ( !empty( $args[ 'order_item_meta' ] ) ) {
						foreach ( $args[ 'order_item_meta' ] as $meta_key => $meta_value ) {
							$order_item->add_meta_data( $meta_key, $meta_value );
						}
					}

					$item->add_order_item_meta( $order_item );

					$order->add_item( $order_item );
					$order->calculate_totals();
				}

				nmgr_add_order_meta_data( $order ); // Order is saved here

				$order->add_order_note( nmgr_get_custom_order_notice() );

				if ( !empty( $args[ 'order_note' ] ) ) {
					$order->add_order_note( $args[ 'order_note' ] );
				}
			}
		}
	}
}

/**
 * Add data concerning the wishlists in the order as order meta data
 *
 * This should typically be done once immediately after the order has been created.
 * This information would serve as the main data store from which we can
 * perform gift registry related wishlist actions on the order later
 *
 * @param WC_Order $created_order The order object
 */
function nmgr_add_order_meta_data( $order ) {
	$wishlist_items_data = array();

	if ( !$order || $order->meta_exists( 'nm_gift_registry' ) ) {
		return;
	}

	foreach ( $order->get_items() as $order_item ) {
		$meta = $order_item->get_meta( 'nm_gift_registry' );

		if ( $meta ) {

			$wishlist = nmgr_get_wishlist( $meta[ 'wishlist_id' ], true );
			if ( !$wishlist ) {
				continue;
			}

			$wishlist_item = $wishlist->get_item( $meta[ 'wishlist_item_id' ] );
			if ( !$wishlist_item ) {
				continue;
			}

			$arr = array(
				'wishlist_item_id' => $meta[ 'wishlist_item_id' ], // wishlist_item_id
				'order_item_quantity' => $order_item->get_quantity(), // quantity of item ordered
				'order_item_id' => $order_item->get_id(), // order item id
				'wishlist_id' => $meta[ 'wishlist_id' ], // id of wishlist the item belongs to
			);
			$wishlist_items_data[] = $arr;
		}
	}

	if ( !empty( $wishlist_items_data ) ) {
		$order_meta = array();
		foreach ( $wishlist_items_data as $data ) {
			$order_meta[ $data[ 'wishlist_id' ] ][ 'wishlist_id' ] = $data[ 'wishlist_id' ];
			$order_meta[ $data[ 'wishlist_id' ] ][ 'wishlist_item_ids' ][] = $data[ 'wishlist_item_id' ];
			$order_meta[ $data[ 'wishlist_id' ] ][ 'order_item_ids' ][ $data[ 'wishlist_item_id' ] ] = $data[ 'order_item_id' ];
			$order_meta[ $data[ 'wishlist_id' ] ][ 'sent_customer_purchased_items_email' ] = 'no';

			if ( nmgr()->is_pro && function_exists( 'is_nmgr_cart_shipping_to_wishlist_address' ) ) {
				$shipping_to_wishlist_address = is_nmgr_cart_shipping_to_wishlist_address() ? 'yes' : 'no';
				$order_meta[ $data[ 'wishlist_id' ] ][ 'shipping_to_wishlist_address' ] = $shipping_to_wishlist_address;
			}
		}

		$order->add_meta_data( 'nm_gift_registry', $order_meta );
		$order->save();
		do_action( 'nmgr_created_order', $order->get_id() );
	}
}

function nmgr_get_current_type() {
	global $wp_query;

	$type = null;

	if ( !is_nmgr_admin() ) {
		if ( wp_doing_ajax() && !empty( $GLOBALS[ 'nmgr' ]->current_type ) ) {
			$type = $GLOBALS[ 'nmgr' ]->current_type;
		} elseif ( !wp_doing_ajax() & !empty( $wp_query ) ) {
			$current_id = get_queried_object_id();
			if ( ($current_id && ($current_id === ( int ) nmgr_get_option( 'page_id' ))) ||
				(!$current_id && is_post_type_archive( 'nm_gift_registry' ) ) ) {
				$type = 'gift-registry';
			} elseif ( $current_id && ($current_id === ( int ) nmgr_get_option( 'wishlist_page_id' )) ) {
				$type = 'wishlist';
			}
		}
	}
	return $type;
}

function nmgr_get_base_actions() {
	return [
		'new',
		'home',
	];
}

function nmgr_get_url( $type = 'gift-registry', $action = '' ) {
	$page_id = nmgr_get_type_option( $type, 'page_id' );
	$pre_url = $page_id ? get_permalink( $page_id ) : '';
	$url = $pre_url ? $pre_url : get_post_type_archive_link( 'nm_gift_registry' );

	switch ( $action ) {
		case 'home':
			if ( $page_id && nmgr_post_content_has_shortcode( '[nmgr_wishlist]', $page_id ) ) {
				$url = $url;
			} else {
				$url = trailingslashit( $url ) . $action;
			}
			break;

		default:
			$url = trailingslashit( $url ) . $action;
			break;
	}

	return $url;
}

function nmgr_get_default_type() {
	if ( nmgr_get_type_option( 'gift-registry', 'enable' ) && nmgr_get_type_option( 'wishlist', 'enable' ) ) {
		$type = '';
	} else {
		$type = nmgr_get_type_option( 'wishlist', 'enable' ) ? 'wishlist' : 'gift-registry';
	}
	return apply_filters( 'nmgr_default_type', $type );
}

function nmgr_get_term_id_by_slug( $slug ) {
	$term = get_term_by( 'slug', $slug, 'nm_gift_registry_type' );
	return $term ? $term->term_id : 0;
}

function nmgr_get_add_items_text() {
	return sprintf(
		/* translators: %s: wishlist type title */
		nmgr()->is_pro ? __( 'Go shopping for items to add to your %s.', 'nm-gift-registry' ) : __( 'Go shopping for items to add to your %s.', 'nm-gift-registry-lite' ),
		nmgr_get_type_title()
	);
}

function nmgr_get_relocation_notice() {
	return sprintf(
			/* translators: %s: wishlist type title */
			nmgr()->is_pro ? __( 'You can now manage your %s directly from the main page for viewing it.', 'nm-gift-registry' ) : __( 'You can now manage your %s directly from the main page for viewing it.', 'nm-gift-registry-lite' ),
			nmgr_get_type_title()
		) . ' ' . nmgr_get_click_here_link( nmgr_get_url( 'gift-registry', 'home' ) );
}

function nmgr_post_content_has_shortcode( $shortcode, $post_id = null ) {
	global $wpdb;

	$id = $post_id ?? get_queried_object_id();
	$page = $id ? get_post( $id ) : null;

	if ( $page ) {
		$db_search = strtok( $shortcode, ' ' );
		return $wpdb->get_var( $wpdb->prepare(
					"SELECT ID FROM $wpdb->posts
								WHERE post_type='page'
								AND post_status='publish'
								AND ID=%d
								AND post_content LIKE %s
								LIMIT 1;",
					$page->ID,
					"%{$db_search}%"
			) );
	}
}

function nmgr_get_click_here_link( $url ) {
	$click_here_text = nmgr()->is_pro ?
		__( 'Click here', 'nm-gift-registry' ) :
		__( 'Click here', 'nm-gift-registry-lite' );

	return '<a href="' . $url . '">' . $click_here_text . '</a>';
}

function nmgr_get_nav( $args = [] ) {
	$defaults = [
		'page' => 1,
		'total_pages' => 1,
		'wrapper_attributes' => [
			'class' => [
				'nmgr-navs',
				($args[ 'template' ] ?? ''),
				($args[ 'position' ] ?? ''),
			],
		],
	];
	$pargs = nmgr()->merge_args( $defaults, $args );
	$page = ( int ) $pargs[ 'page' ];
	$total_pages = ( int ) $pargs[ 'total_pages' ];

	if ( 2 > $total_pages ) {
		return;
	}

	$strings = new Strings();

	ob_start();
	?>
	<div <?php echo wp_kses( nmgr()->format_attributes( $pargs[ 'wrapper_attributes' ] ), [] ); ?>>
		<span class="page-info">
			<?php
			printf(
				/* translators: 1: current page, 2: total pages */
				nmgr()->is_pro ? __( 'Page %1$d of %2$d', 'nm-gift-registry' ) : __( 'Page %1$d of %2$d', 'nm-gift-registry-lite' ),
				$page,
				$total_pages
			);
			?>
		</span>
		<?php if ( 1 !== $page ) : ?>
			<a href="#"
				 class="nmgr-tip previous nmgr-nav"
				 data-nav_action="previous"
				 title="<?php echo esc_html( $strings->previous() ); ?>">&#10092;</a>
			 <?php endif; ?>
			 <?php if ( $page !== $total_pages ) : ?>
			<a href="#"
				 class="nmgr-tip next nmgr-nav"
				 data-nav_action="next"
				 title="<?php echo esc_html( $strings->next() ); ?>">&#10093;</a>
			 <?php endif; ?>
	</div>
	<?php
	return ob_get_clean();
}
