<?php

namespace NmeriMedia\V2;

use \NmeriMedia\V2\{
	ReadmeParser,
	Strings
};

defined( 'ABSPATH' ) || exit;

/**
 * Main utils class
 * @class \NmeriMedia\V2\Utils
 */
class Utils {

	/**
	 * @return \NmeriMedia\V2\Strings
	 */
	public static function get_string() {
		$classname = apply_filters( 'nmerimedia_strings_class', Strings::class );
		return new $classname();
	}

	/**
	 * Get the plugin path to the nmerimedia directory
	 * @return string
	 */
	public static function path() {
		return plugin_dir_path( __DIR__ );
	}

	/**
	 * Get the plugin url to the nmerimedia directory
	 * @return string
	 */
	public static function url() {
		return plugin_dir_url( __DIR__ );
	}

	/**
	 * Get the nmerimedia version
	 * @return system
	 */
	public static function version() {
		return __NAMESPACE__;
	}

	/**
	 * Svg tags allowed by the plugin
	 *
	 * @return array
	 */
	public static function allowed_svg_tags() {
		return array(
			'svg' => array(
				'id' => true,
				'role' => true,
				'width' => true,
				'height' => true,
				'class' => true,
				'style' => true,
				'fill' => true,
				'xmlns' => true,
				'viewbox' => true,
				'aria-hidden' => true,
				'focusable' => true,
				'data-notice' => true, // may be deprecated soon. Used temporarily.
			),
			'use' => array(
				'xlink:href' => true
			),
			'title' => array(
				'data-title' => true
			),
			'path' => array(
				'fill' => true,
				'fill-rule' => true,
				'd' => true,
				'transform' => true,
			),
			'polygon' => array(
				'fill' => true,
				'fill-rule' => true,
				'points' => true,
				'transform' => true,
				'focusable' => true,
			),
		);
	}

	/**
	 * Get the allowed post tags
	 * This returns wordpress allowed html tags in post and the plugin's allowed svg tags
	 *
	 * @return array
	 */
	public static function allowed_post_tags() {
		return array_merge( wp_kses_allowed_html( 'post' ), static::allowed_svg_tags() );
	}

	/**
	 * Clean variables using sanitize_text_field. Arrays are cleaned recursively.
	 * Non-scalar values are ignored.
	 *
	 * @param string|array $var Data to sanitize.
	 * @return string|array
	 */
	public static function sanitize( $var ) {
		if ( is_array( $var ) ) {
			return array_map( array( __CLASS__, 'sanitize' ), $var );
		} else {
			return is_scalar( $var ) ? sanitize_text_field( $var ) : $var;
		}
	}

	/**
	 * Compose an svg icon using dynamic arguments provided
	 *
	 * @param string|array $args The name of the svg icon to get or array of icon parameters.
	 *
	 * If $args is an array, registered array keys needed to compose the svg are:
	 * - icon - [required] string The svg icon name. This should correspond to the name of an svg file in the assets/svg directory or the last part of the id of a symbol element in the sprite file - assets/svg/sprite.svg.
	 * - size - [optional] integer|string The svg icon width and height. (uses em unit if no unit is specified). Default 1em (16px).
	 * - class - [optional] array Classes to add to the svg icon.
	 * - sprite - [optional] boolean Whether to use the loaded svg sprite file (Default),
	 *            or to load the single svg icon from another path.
	 * - role - [optional] string Svg role attribute. Default img.
	 * - id - [optional|required] string Svg id attribute. Expected if "sprite" is true.
	 *        If not set, the 'icon' parameter is set as the id by default
	 * - style - [optional] string Svg Inline style.
	 * - title - [optional] string Svg title attribute.
	 * - fill - [optional] string Svg fill attribute.
	 * - path - [optional] Full path to the single svg icon file.
	 *   (Usually used if not using the loaded svg sprite file).
	 *
	 * @return string Svg HTML element
	 */
	public static function get_svg( $args ) {
		// Make sure icon name is given
		if ( !$args || (is_array( $args ) && (false === array_key_exists( 'icon', $args ))) ) {
			// If no arguments are set or the icon key is not in the array, require icon key
			return;
		} elseif ( is_string( $args ) ) {
			//if a string argument is given, assume it is the icon name and set it in the array
			$args = array( 'icon' => $args );
		}

		// Set defaults.
		$defaults = array(
			'id' => $args[ 'icon' ],
			'icon' => $args[ 'icon' ],
			'size' => 1,
			'class' => [
				$args[ 'icon' ],
				'nm-svg',
			],
			'sprite' => true,
			'role' => 'img',
			'path' => '',
		);

		// Parse args.
		$args_1 = apply_filters( 'nmerimedia_svg_args', wp_parse_args( $args, $defaults ), $args );

		if ( true ) {
			$args = $args_1;
		}

		// Make sure default classes are present
		$args[ 'class' ] = ( array ) $args[ 'class' ];
		foreach ( $defaults[ 'class' ] as $class ) {
			if ( !in_array( $class, $args[ 'class' ] ) ) {
				$args[ 'class' ][] = $class;
			}
		}

		// Make sure the size has units (default em)
		$size = is_numeric( $args[ 'size' ] ) ? "{$args[ 'size' ]}em" : $args[ 'size' ];

		// Get extra svg parameters set by user that are not in the default expected arguments
		// e.g 'style', 'title' and 'fill'
		$extra_params = array_diff_key( $args, $defaults );
		$extra_params_string = '';

		// extract the title attribute if it exists so that we can add it separately to the svg
		$title = '';
		if ( isset( $extra_params[ 'title' ] ) && !empty( $extra_params[ 'title' ] ) ) {
			$string = htmlspecialchars( wp_kses_post( $extra_params[ 'title' ] ) );
			$title = sprintf( "<title data-title='%s'>%s</title>", $string, $string );
			unset( $extra_params[ 'title' ] );
		}

		// Create new indexed array from extra svg parameters
		if ( !empty( $extra_params ) ) {
			$arr = array();
			foreach ( $extra_params as $key => $value ) {
				$arr[] = $key . '="' . $value . '"';
			}
			// Compose string from array of extra svg parameters
			$extra_params_string = implode( ' ', $arr );
		}

		// Compose svg with the given attributes
		$composed_svg = sprintf( '<svg role="%s" width="%s" height="%s" class="%s" %s ',
			$args[ 'role' ],
			$size,
			$size,
			implode( ' ', $args[ 'class' ] ),
			$extra_params_string );

		if ( $args[ 'sprite' ] ) {
			/**
			 * we are using a sprite file
			 */
			$svg = sprintf( '%s>%s<use xlink:href="#%s"></use></svg>',
				$composed_svg,
				$title,
				$args[ 'id' ]
			);
		} else {
			/**
			 *  we are using a single svg file
			 */
			// Get the svg file
			$svg = static::get_svg_file( $args[ 'icon' ], $args[ 'path' ] );

			// Remove width and heigh attributes from svg if exists as we are adding our own
			$svg = preg_replace( '/(width|height)="\d*"\s/', '', $svg );

			// Merge composed svg with original svg
			$svg = preg_replace( '/^<svg /', $composed_svg, trim( $svg ) );

			// Add title attribute if it exists
			$svg = str_replace( '</svg>', "$title</svg>", $svg );

			// Remove newlines & tabs.
			$svg = preg_replace( "/([\n\t]+)/", ' ', $svg );

			// Remove white space between SVG tags.
			$svg = preg_replace( '/>\s*</', '><', $svg );
		}

		$svg = apply_filters( 'nmerimedia_svg', $svg, $args );

		return $svg;
	}

	/**
	 * Get a single svg icon file unmodified from the icon directory
	 *
	 * @param string $icon_name The name of the icon e.g. user.
	 * @param string $path The full path to the icon file.
	 *
	 * @return string|false Icon html or false if not found
	 */
	public static function get_svg_file( $icon_name, $path ) {
		$iconfile = trailingslashit( $path ) . "{$icon_name}.svg";
		if ( file_exists( $iconfile ) ) {
			ob_start();
			include $iconfile;
			return ob_get_clean();
		}
		return false;
	}

	/**
	 * Return an array of html attributes as a string to be inserted directly into the html element
	 * @param array $attributes Attributes to flatten
	 * @return string
	 */
	public static function format_attributes( $attributes ) {
		$attr = array();

		if ( !empty( $attributes ) ) {
			foreach ( ( array ) $attributes as $key => $value ) {
				$val = is_array( $value ) ? implode( ' ', $value ) : $value;
				$attr[] = $key . '="' . $val . '"';
			}
		}
		return implode( ' ', $attr );
	}

	/**
	 * Return an array of html attributes as a string to be inserted directly into the html element
	 * (Alias of format_attributes())
	 * @param array $attributes Attributes to flatten
	 * @return string
	 */
	public static function flatten_attributes( $attributes ) {
		return static::format_attributes( $attributes );
	}

	/**
	 * Merge supplied values with default values
	 *
	 * This function doesn't use wp_parse_args because it doesn't merge recursively.
	 *
	 * @param array $defaults Default values
	 * @param array $args Supplied values
	 * @return array
	 */
	public static function merge_args( $defaults, $args ) {
		$merged = $defaults;
		foreach ( ( array ) $args as $key => $val ) {
			if ( isset( $defaults[ $key ] ) ) {
				$merged[ $key ] = is_array( $defaults[ $key ] ) ? array_merge_recursive( $defaults[ $key ], ( array ) $val ) : $val;
			} else {
				$merged[ $key ] = $val;
			}
		}
		return $merged;
	}

	/**
	 * Get a help tip
	 *
	 * @param string $title The text to show as the tip.
	 * @return string
	 */
	public static function help_tip( $title ) {
		return '<span data-bs-toggle="tooltip" class="nm" style="cursor:help;" title="' . $title . '"> &#9432;</span>';
	}

	public static function include_base_sprite_file() {
		$sprite_file = \NmeriMedia\V2\Utils::path() . 'assets/svg/sprite.svg';
		if ( file_exists( $sprite_file ) ) {
			include_once $sprite_file;
		}
	}

	/**
	 * Check if it is an admin request
	 * @return boolean
	 */
	public static function is_admin_request() {
		$current_url = home_url( add_query_arg( null, null ) );
		$admin_url = strtolower( admin_url() );
		$referrer = strtolower( wp_get_referer() );

		// Check if this is a admin request. If true, it
		// could also be a AJAX request from the frontend.
		if ( 0 === strpos( $current_url, $admin_url ) ) {
			// Check if the user comes from a admin page.
			if ( 0 === strpos( $referrer, $admin_url ) ) {
				return true;
			} else {
				return !wp_doing_ajax();
			}
		} else {
			return false;
		}
	}

	/**
	 * Get the current screen id of a page being viewed in the admin area
	 * @return string|null
	 */
	public static function get_current_screen_id() {
		if ( function_exists( 'get_current_screen' ) ) {
			$screen = get_current_screen();
			return $screen ? $screen->id : null;
		}
	}

	/**
	 * Get the checkbox switch used by the plugin
	 * @param array $args Arguments supplied to create the checkbox switch:
	 * - input_type = {string} The input type, 'checkbox' or 'radio'. Default 'checkbox'
	 * - input_id - {string, required} The input id
	 * - input_name - {string} The input name
	 * - input_value - {mixed} The input value
	 * - input_class - {array} The input class
	 * - input_attributes = {array} Attributes to go into the input element
	 * - label_class - {array} The label class
	 * - label_attributes - {array} Attributes to go into the label element
	 * - label_text - {string} The label text
	 * - label_before - {boolean} Whether the label text should be before the switch. Default false.
	 * - checked - {boolean} Whether the checkbox should be checked or not.
	 * - show_hidden_input - {mixed} Whether to show a hidden input for the checkbox. Default false.
	 * 												If true, show hidden input value is 0 else, value of 'show_hidden_input' is used.
	 * @return string
	 */
	public static function get_toggle_switch( $args ) {
		$defaults = array(
			'input_type' => 'checkbox',
			'input_id' => '',
			'input_name' => '',
			'input_value' => 1,
			'input_class' => array(),
			'input_attributes' => array(),
			'label_class' => array(),
			'label_attributes' => array(),
			'label_text' => '',
			'label_before' => false,
			'checked' => '',
			'show_hidden_input' => false,
		);

		$params = wp_parse_args( $args, $defaults );

		if ( isset( $params[ 'input_attributes' ][ 'disabled' ] ) && $params[ 'input_attributes' ][ 'disabled' ] ) {
			$params[ 'label_class' ][] = 'disabled';
		}

		if ( isset( $params[ 'input_attributes' ][ 'readonly' ] ) && $params[ 'input_attributes' ][ 'readonly' ] ) {
			$params[ 'label_class' ][] = 'readonly';
		}

		if ( $params[ 'label_before' ] ) {
			$params[ 'label_class' ][] = 'label-before';
		}

		$label_class = implode( ' ', array_map( 'sanitize_html_class', ( array ) $params[ 'label_class' ] ) );
		$input_class = implode( ' ', array_map( 'sanitize_html_class', ( array ) $params[ 'input_class' ] ) );
		$input_attributes = static::format_attributes( $params[ 'input_attributes' ] );
		$label_attributes = static::format_attributes( $params[ 'label_attributes' ] );
		$checked = $params[ 'checked' ] ? " checked='checked'" : '';
		$hidden = '';

		if ( $params[ 'show_hidden_input' ] ) {
			$hidden_input_val = (true === $params[ 'show_hidden_input' ]) ? 0 : $params[ 'show_hidden_input' ];
			$hidden = '<input type="hidden" value="' . $hidden_input_val . '" name="' . $params[ 'input_name' ] . '">';
		}

		return '<label class="nmgr-checkbox-switch ' . $label_class . '"' . $label_attributes . '>' .
			$hidden .
			'<span>' . $params[ 'label_text' ] . '</span>' .
			'<input type="' . $params[ 'input_type' ] . '" id="' . $params[ 'input_id' ] . '"
                value="' . $params[ 'input_value' ] . '"
								class="' . $input_class . '"
                name="' . $params[ 'input_name' ] . '"' .
			$checked . $input_attributes . '/>
                <label for="' . $params[ 'input_id' ] . '"></label>
            </label>';
	}

	/**
	 * Get the html of an admin notice
	 * @param string $msg The message to display
	 * @param string $notice_type The type of notice (error, info, success)
	 * @param boolean $is_dismissible Whether the notice should be dismissible
	 * @return string
	 */
	public static function get_admin_notice( $msg, $notice_type = 'error', $is_dismissible = true ) {
		$dismiss = $is_dismissible ? 'is-dismissible' : '';
		return "<div class='$dismiss notice notice-{$notice_type}'>$msg</div>";
	}

	/**
	 * Get all the registered Nmeri Media plugins or a specific plugin
	 * @param string $plugin_slug The slug of a specific registered plugin to get
	 * @return array
	 */
	public static function get_plugins( $plugin_slug = null ) {
		$plugins = include \NmeriMedia\V2\Utils::path() . 'templates/plugins.php';
		return $plugin_slug ? ($plugins[ $plugin_slug ] ?? null) : $plugins;
	}

	/**
	 * Add an element to an array
	 * @param array $array The array to add the element to
	 * @param int|string $position The index position in the array to add the element to.
	 * 											 Can be a numeric index in the array or a string representing
	 * 											 a specific array key.
	 * @param mixed $insert The element to insert into the array. Can be a simple array element
	 * 											or an associative array.
	 * @param boolean $after Whether to insert the element after or before $postion. Default is after.
	 * @return array The original array with the element added to it.
	 *
	 */
	public static function array_add( $array, $position, $insert, $after = true ) {
		$pos = is_int( $position ) ? $position : array_search( $position, array_keys( $array ) );
		$ins = is_array( $insert ) ? $insert : array( $insert );
		$length = $after ? $pos + 1 : $pos;
		return array_merge( array_slice( $array, 0, $length ), $ins, array_slice( $array, $length ) );
	}

	/**
	 * Get a plugin's readme.txt file as a string
	 * @param string $filepath Full filepath to the readme.txt file.
	 * @return \NmeriMedia\V2\ReadmeParser
	 */
	public static function get_readme_file( $filepath = null ) {
		return new ReadmeParser( $filepath . 'readme.txt' );
	}

	public static function get_component( $type, $args = [] ) {
		$classname = '\NmeriMedia\\V2\\' . ucfirst( $type );
		if ( class_exists( $classname ) ) {
			$class = new $classname( $args );
			return $class;
		}
	}

	public static function sort_by_priority( $a, $b ) {
		$a[ 'priority' ] = $a[ 'priority' ] ?? 0;
		$b[ 'priority' ] = $b[ 'priority' ] ?? 0;
		return ( $a[ 'priority' ] < $b[ 'priority' ] ) ? -1 : 1;
	}

}
